#!/usr/bin/perl

# $Id: MapWithBlast.pl 633 2011-04-21 13:51:30Z gouzy $

=pod 

=head1 NAME

MapWithBlast.pl - Blast the sequence against the database using parallelization 
                  if the sequence is long. Use Blast+ program.

=head1 SYNOPSIS

    MapWithBlast.pl --sequence sequence.fasta --db database.fasta --output output.blastplus --cfg blast.cfg


=head1 DESCRIPTION

Blast the sequence against the database. If the sequence length is upper than a threshold, 
launch different Blast about subsequences of the reference sequence.
Use paraloop to launch simultaneously the Blast computings.
Use Blast+ program.

=cut


BEGIN
{
    my ($dir,$file) = $0 =~ /(.+)\/(.+)/;
    unshift (@INC,"$dir/../../corelib");
    unshift (@INC,"$dir/../../lib");
}

use strict; 
use warnings;
use ParamParser;
use Carp;
use General;


my $VERBOSE;

MAIN:
{
    my $o_param = New ParamParser('GETOPTLONG', "sequence=s", "db=s", "output=s", "cfg=s", "workingdir=s", "verbose", "noclean", "protein_similarity_search_method=s", "prokaryote");
    $o_param->SetUsage(my $usage=sub { &Usage(); } );
    $o_param->AssertFileExists('cfg');
    $o_param->SetBehaviour('ignore_space');
    $o_param->Update($o_param->Get('cfg'),'A');
    
    $o_param->SetSubstitution('%i', $ENV{'EGNEP'});
    $o_param->SetBehaviour('use_substitution_table');
    
    # Check the parameters
    $o_param->AssertFileExists('sequence');	
    $o_param->AssertDefined('db','output');

    $o_param->AssertInteger('prg_mapwithblast_limit');

    $o_param->AssertDefined('prg_targetedblastsearch');
    $o_param->SetUnlessDefined('prg_targetedblastsearch_param', "");
 	$o_param->AssertAllowedValue('protein_similarity_search_method', "blastx", "ublast_blastx", "diamond_blastx");
	
	my $method = $o_param->Get('protein_similarity_search_method');
	if ($method eq "ublast_blastx")
	{
		$o_param->AssertDefined('prg_usearch', 'prg_usearch_ublast_param');    
	}
	elsif ($method eq "diamond_blastx")
	{
		$o_param->AssertDefined('prg_diamond', 'prg_diamond_param');
	}	
    # check the parameters need by targetblastsearch program
    $o_param->AssertDefined('blast_cmd', 'blast_param');
    $o_param->AssertDefined('prg_blastdbcmd', 'prg_makeblastdb');
    $o_param->AssertDefined('prg_fasta2overlappingwins');
	
	$VERBOSE = ($o_param->IsDefined('verbose')) ? 1 : 0; 
    

    # Get the parameters   
	my $db                  = $o_param->Get('db');   
    my $sequence            = $o_param->Get('sequence');
    my $output              = $o_param->Get('output');
    my $max_seq_length      = $o_param->Get('prg_mapwithblast_limit');
    # complete targetedblastsearch_param according to protein_similarity_search_method
	my $targeted_blast_param = $o_param->Get('prg_targetedblastsearch_param');
	$targeted_blast_param .= " --protein_similarity_search_method $method";
	$targeted_blast_param .= " --prokaryote" if ($o_param->IsDefined('prokaryote'));
	$o_param->Set('prg_targetedblastsearch_param', $targeted_blast_param);
	
	print STDERR "START MapWithBlast SEQ: $sequence OPT: $targeted_blast_param\n" if ($VERBOSE);
		
    # Check the database was indexed 
    if (!-f $db.".nin" && !-f $db.".pin" && !-f $db.".pal" && !-f $db.".nal")  
    {
        croak "The database $db isn't indexed. Use 'makeblastdb' program to format it.\n";
    }

	my ($outdir, $file) = $output =~ /(.+)\/(.+)/; 
    $o_param->SetUnlessDefined('workingdir', $outdir);
    $o_param->AssertDirExists('workingdir');
    my $workingdir = $o_param->Get('workingdir');
    
    # Index database for ublast/diamond for the protein db reduction
    # Done before call targetblastsearch on sliced win to avoid to index database many times
    # and  in the same time!!
	if ($method eq "ublast_blastx")
	{
    	&IndexUBlastDb($o_param, $db);
	}
	elsif ($method eq "diamond_blastx")
	{
		&IndexDiamondDb($o_param, $db);
	}
    
    my $seq_len = &GetSeqLen($sequence);    
    
    if ( $seq_len < $max_seq_length )
    {
	# Blast the sequence against the database $db
	&ComputeOneTargetedBlast($o_param, $sequence, $db, $output);
    }
    else 
    {	
	$o_param->AssertDefined('prg_paraloop');
	$o_param->AssertInteger('prg_mapwithblast_win_size');
	$o_param->AssertInteger('prg_mapwithblast_win_overlap');
	
	# Blast many subsequences of the seq against the db with paraloop
	&ComputeOneLargeTargetedBlast($o_param, $sequence, $db, 
				      $output, $workingdir);
    } 

    if ($VERBOSE)
    {
	print STDERR "STOP MapWithBlast SEQ: $sequence\n";
    }
}



sub IndexUBlastDb
{
    my ($o_param, $db) = @_;
    
    # format for ublast
    if (!-e $db.".udb")
    {
		my $usearch_cmd = $o_param->Get('prg_usearch');
		my $cmd = $usearch_cmd. " -makeudb_ublast $db -output $db.udb";
		$cmd .= " > /dev/null 2>&1" if (!$VERBOSE);
		print STDERR "$cmd\n" if ($VERBOSE);
		system($cmd);
    }

    return;
}


sub IndexDiamondDb
{
    my ($o_param, $db) = @_;
    
    # format for ublast
    if (!-e $db.".dmnd")
    {			
		my $cmd = $o_param->Get('prg_diamond'). " makedb --in $db -d $db";
		$cmd .= " > /dev/null 2>&1" if (!$VERBOSE);
		print STDERR "$cmd\n" if ($VERBOSE);
		system($cmd);
    }

    return;
}

=head2 procedure ComputeOneTargetedBlast

 Title        : ComputeOneTargetedBlast
 Usage        : &ComputeOneTargetedBlast($o_param, $seq, $db, $out_file);
 Prerequisite : $db is a file which was indexed by formatdb, 
                $targeted_blast_cmd program exists.
 Procedure    : Blast $seq against the database $db. The result file is $out_file 
 Returns      : None
 Args         : $o_param
                $seq            Query sequence
                $db             Database
                $out_file       Blast output file
           
 Globals      : none

=cut

sub ComputeOneTargetedBlast
{
    my ($o_param, $seq, $db, $out_file) = @_;
	
    my $cfg                = $o_param->Get('cfg');
    my $targeted_blast_cmd = $o_param->Get('prg_targetedblastsearch');
    my $targeted_blast_param = $o_param->Get('prg_targetedblastsearch_param');
    # Execute Blast
    my $cmd          = "$targeted_blast_cmd $targeted_blast_param --db $db --query $seq --output $out_file --cfg $cfg";	
    $cmd .= " --verbose" if ($VERBOSE);   
    my $system_out   = system($cmd);
    my $blast_output = $system_out >> 8;
    if ($blast_output != 0)
    {
	my $signal_number = $system_out & 127;
    	croak "TargetedBlastSearch execution failed. Exit value: $blast_output, signal number: $signal_number.\n";
    }

    return;
}

=head2 procedure ComputeOneLargeTargetedBlast

 Title        : ComputeOneLargeTargetedBlast
 Usage        : &ComputeOneLargeTargetedBlast($blast_cmd, $blast_param, $seq, $db, 
				$blast_outfile, $paraloop_cmd, 
                                $window_length, $overlap);
 Prerequisite : $db is a file which was indexed by makeblastdb
 Procedure    : Blast $seq against the database $db: to do that, move a window 
                about the sequence. Blast each window against the database. 
                Use Paraloop to parallelize the computing of Blast. 
                The result file is $out_file.
 Returns      : None
 Args         : $o_param
                $seq            Query sequence
                $db             Database
                $out_file       Blast output file (tabular format)
                $workingdir
 Globals      : none

=cut
sub ComputeOneLargeTargetedBlast
{
    my ($o_param, $seq, $db, $out_file, $workingdir) = @_;
    
    my $targeted_blast_cmd = $o_param->Get('prg_targetedblastsearch');    
    my $targeted_blast_param = $o_param->Get('prg_targetedblastsearch_param');
    my $paraloop_cmd       = $o_param->Get('prg_paraloop');
    my $window_length      = $o_param->Get('prg_mapwithblast_win_size');
    my $overlap            = $o_param->Get('prg_mapwithblast_win_overlap');
    my $cfg                = $o_param->Get('cfg');
    chdir($workingdir);
    my $blastxdir = $workingdir."/BlastX.".$$.".".time;
    mkdir($blastxdir);
    chdir($blastxdir);
	
    my $seq_len = &GetSeqLen($seq);
    my ($seqpath, $seqname) = $seq =~ /(.+)\/(.+)/;
    # File of commands
    my $cmd_file = $blastxdir."/".$seqname."_cmd.$$.".time;
    
    my $f_cmd = &General::GetStreamOut($cmd_file);
	
    print STDERR "$cmd_file\n" if ($VERBOSE);
	
    my $begin = 1;
    my $end;
    my $range;
    my $range_blast;
    my $blast_cmd_line;
		
    # Write all the blast commands in the command file
    do
    {
		$end            = $begin + $window_length - 1;
		$end            = ( $end > $seq_len ) ? $seq_len : $end;
		$range          = "$begin...$end";
		$range_blast    = "$begin-$end";
		$end            = $seq_len if ( $end > $seq_len );
		$blast_cmd_line = "export EGNEP=".$ENV{'EGNEP'}.";$targeted_blast_cmd  $targeted_blast_param --query $seq --db $db --output $seq.raw.$$.".time.".$range --query_loc \"$range_blast\" --cfg $cfg";	
		$blast_cmd_line .= " --verbose" if ($VERBOSE);
		# Write the command line in the file
		print $f_cmd "$blast_cmd_line\n";
		$begin          = $begin + $window_length - $overlap - 1;
    } 
    while ( $begin < $seq_len );
	$f_cmd->close;
	
    # Call paraloop about the command file
    my $cde = "$paraloop_cmd --program=Shell --input $cmd_file --output $cmd_file.output ";
    $cde .= " --clean " if (!$o_param->IsDefined('noclean'));
    $cde  = "$paraloop_cmd $cmd_file" if ( $paraloop_cmd =~ /^\w*sh\s*/ );
    print STDERR "$cde\n" if ($VERBOSE);
    `$cde`;
    my ($dir, $filename) = ( $seq =~ /(.+)\/(.+)/ );
    
    # Check there are result files
    $cde = "find $dir -name \"$filename.raw.$$.*...*\" -print";
    print STDERR "$cde\n" if ($VERBOSE);
    my @a_res  = `$cde`;
    my $nb_res = scalar(@a_res);
    if ($nb_res == 0)
    {
	croak "Error during MapWithBlast (ComputeOneLargeTargetedBlast) $seq against $db\n";
    }

    $cde = "find $dir -name \"$filename.raw.$$.*...*\" -exec cat \'{}\' \\; > $out_file";
    print STDERR "$cde\n" if ($VERBOSE);
    system($cde);
   
    if (!$o_param->IsDefined('noclean'))
    {
	$cde = "find $dir -name \"$filename.raw.$$.*...*\" -exec rm \'{}\' \\;";
	print STDERR "$cde\n" if ($VERBOSE);
	system($cde);
    }

    return;
}



=head2 function GetSeqLen

 Title        : GetSeqLen
 Usage        : my $seqlen = &GetSeqLen($seq);
 Prerequisite : -
 Function     : Compute and return the length of the sequence.
 Returns      : The length of the sequence
 Args         : $filename  The sequence.
 Globals      : none

=cut

sub GetSeqLen
{
    my ($filename) = @_;
    
    my $seq    = "";
    my $fh_seq = &General::GetStreamIn($filename); 
    while(my $lign = <$fh_seq>)
    {
		chomp($lign);
		next if ( $lign =~ /^>/ );
		$lign =~ s/\s//g;
		$seq .= $lign;
    }
    $fh_seq->close();
    my $len = length($seq);
    
    return $len;
}

=item Usage

	$o_param->Usage();
	Print the usage of the program

=cut

sub Usage
{
print STDERR<<END
Usage:$0	

[Mandatory]
    --sequence	filename        Sequence file path
    --db        filename        Database path
    --output	filename        Blast output file path
    --cfg       filename        Full path of the configuration file
	--protein_similarity_search_method 	[blastx,ublast_blastx,diamond_blastx]
                               ublast_blastx and diamond_blastx methods perform a protein database reduction step with respectively 
                               ublast or diamond program, then launch blastx on this reduce database
[Optional]
    --prokaryote               Adapt parameters for prokaryotic genomes	
    --workingdir dirname
    --verbose
    --noclean
    --help  			this message
END
}

