#!/usr/bin/perl

BEGIN
{
    my ($dirprg, $nameprg) = $0 =~ /(.+)\/(.+)/;
    $dirprg = "." if ($dirprg eq "");
    push(@INC, "$dirprg/../../lib");
    push(@INC, "$dirprg/../../corelib");
}

use Getopt::Long;
use ParamParser;
use strict;
use General;
use GeneralBioinfo;

my $VERBOSE = 0;
use constant DEFAULT_CONTEXT => 60;
use constant MAX_TP_SITES => 100000;
use constant MAX_TN_SITES => 1000000;

my $MAX_INTRON_LENGTH     = 20000;
my $INTRON_MIN_HIT_LENGTH = 50;

MAIN:
{
    my $o_param = New ParamParser('GETOPTLONG', "sequence=s", "gff3=s", "outprefix=s", "help", "parent_feature=s","child_feature=s","verbose",
								  "context_length=i","maxtpsites=i","maxtnsites=i", "intron_file=s", "intron_min_hit_length=i", "no_alt_site_detection",
                                  "max_intron_length=i", "subsample", "tn_start_sites");

    $o_param->SetUsage(my $usage = sub {&Usage();});
    $o_param->AssertFileExists('sequence','gff3');
    $o_param->AssertDefined('outprefix');
    $o_param->SetUnlessDefined('context_length',&DEFAULT_CONTEXT);
    $o_param->AssertInteger('context_length');
    $o_param->SetUnlessDefined('parent_feature', 'EST_match');
    $o_param->SetUnlessDefined('child_feature',  'match_part');
	
    $o_param->SetUnlessDefined('maxtpsites',&MAX_TP_SITES);
    $o_param->SetUnlessDefined('maxtnsites',&MAX_TN_SITES);
	
	$o_param->SetUnlessDefined('intron_min_hit_length',$INTRON_MIN_HIT_LENGTH);
	$o_param->SetUnlessDefined('max_intron_length',    $MAX_INTRON_LENGTH);

	$o_param->AssertInteger('intron_min_hit_length', 'max_intron_length');

	$VERBOSE = 1 if ( $o_param->Get('verbose') );
	
	# $subsample = 1 to create subsample
	my $subsample = 0;
	$subsample    = 1 if ($o_param->IsDefined('subsample') );
	# $altsplicedsite =1 to detect alternative spliced sites
	my $altsplicedsite = 1;
	$altsplicedsite    = 0 if ($o_param->IsDefined('no_alt_site_detection'));
	
	my %h_seqs ;
	&GeneralBioinfo::FastaToHash($o_param->Get('sequence'),\%h_seqs);
	
	my %h_refseq; # key: ID od a hit, value: genomic sequence
	my %h_parents;
	my %h_exons;

	# everything is loaded to avoid sorting constraints
	&LoadGFF3($o_param->Get('gff3'), $o_param->Get('parent_feature'), $o_param->Get('child_feature'), \%h_parents, \%h_exons, \%h_refseq);


	# the TN are sites on the reverse complement of the Parent region
	&ExtractSpliceSites($o_param->Get('outprefix'), $o_param->Get('context_length'), \%h_parents, \%h_exons, \%h_refseq, \%h_seqs, $o_param->Get('maxtpsites'), $o_param->Get('maxtnsites'), $subsample, $altsplicedsite);
	
	if ($o_param->IsDefined('tn_start_sites'))
	{
	    &ExtractTNStartSites($o_param->Get('outprefix'), $o_param->Get('context_length'), 'TN', \%h_parents,  \%h_refseq, \%h_seqs,$o_param->Get('maxtnsites'), $subsample);
	}
	if ($o_param->IsDefined('intron_file'))
	{
		&ExtractIntrons($o_param->Get('intron_file'),$o_param->Get('intron_min_hit_length'),$o_param->Get('max_intron_length'),\%h_parents,\%h_exons,\%h_refseq, \%h_seqs);
	}
}

sub ExtractSpliceSites
{
	my($outprefix, $contextlen, $rh_parents, $rh_exons, $rh_refseq, $rh_seqs, $maxtpsites, $maxtnsites, $subsample, $altsplicedsite) = @_;

	my $tnfile          = "$outprefix.TN";
	my $tpfile          = "$outprefix.TP";
	
	my $tnfile_donor    = "$tnfile.donor";
	my $tnfile_acceptor = "$tnfile.acceptor";
	my $tpfile_donor    = "$tpfile.donor";
	my $tpfile_acceptor = "$tpfile.acceptor";
	my $altsiteprefix   = "";
	$altsiteprefix      = $outprefix if ($altsplicedsite);
	
	my %h_deja    = ();
	my %h_deja_tn = ();

	printf STDERR "Writing the file of true positives\n" if ( $VERBOSE );
	
	
	my $f_tp_don = &General::GetStreamOut($tpfile_donor);    
	my $f_tp_acc = &General::GetStreamOut($tpfile_acceptor); 
	my $slen = 2;
	
	foreach my $idparent (keys(%$rh_exons))
	{		
		my $seqid   = $$rh_refseq{$idparent};
		my $the_seq = $$rh_seqs{$seqid}{sequence};


		my @a_exons = split(/\n/,$$rh_exons{$idparent});
		for (my $i=0;$i<$#a_exons;$i++)
		{
			my @a_curr = split(/\t/,$a_exons[$i]);
			my @a_next = split(/\t/,$a_exons[$i+1]);
			my $s1_min = $a_curr[1] - $contextlen + 1 ;
			my $s2_min = $a_next[0] - $slen - $contextlen;
			my $strand = $a_curr[3];	
			
	        #my $s1_max = $a_curr[1] + $slen + $contextlen; my $s2_max = $a_next[0] + $contextlen - 1;
			#printf "$strand $a_curr[1] $s1_min-$s1_max delta = %d\n",$s1_max-$s1_min+1;
			#printf "$strand $a_next[0] $s2_min-$s2_max delta = %d\n",$s2_max-$s2_min+1;
			
			my $s1_seq = substr($the_seq, $s1_min-1, 2*$contextlen+$slen);
			my $s2_seq = substr($the_seq, $s2_min-1, 2*$contextlen+$slen);
			if ( $strand eq '-' )
			{
				$s2_seq = &ReverseComplement($s2_seq);
				&DumpSite($f_tp_don, "$seqid:$s2_min:$strand", $s2_seq, "GT", $contextlen, \%h_deja, $altsiteprefix);
				$s1_seq = &ReverseComplement($s1_seq);
				&DumpSite($f_tp_acc, "$seqid:$s1_min:$strand", $s1_seq, "AG", $contextlen, \%h_deja, $altsiteprefix);
			}
			else
			{
				&DumpSite($f_tp_don, "$seqid:$s1_min:$strand", $s1_seq, "GT", $contextlen, \%h_deja, $altsiteprefix);
				&DumpSite($f_tp_acc, "$seqid:$s2_min:$strand", $s2_seq, "AG", $contextlen, \%h_deja, $altsiteprefix);
			}
		}
	}
	$f_tp_don->close;
	$f_tp_acc->close;

	printf STDERR "Writing the file of true negatives\n" if ( $VERBOSE );

	my $f_tn_don = &General::GetStreamOut($tnfile_donor);
	my $f_tn_acc = &General::GetStreamOut($tnfile_acceptor);
	my $cpt_gt=0;
	my $cpt_ag=0;
	my $parent_nb = scalar(keys(%$rh_parents));
	my $gt_dumped_nb = 0;
	my $ag_dumped_nb = 0;
	
	foreach my $idparent (keys(%$rh_parents))
	{
		my $seqid    = $$rh_refseq{$idparent};			
		my $the_seq  = $$rh_seqs{$seqid}{sequence};

		my @a_curr   = split(/\t/, $$rh_parents{$idparent});
		my $strand   = $a_curr[3];
		my $opstrand = ($strand eq '+' ) ? '-' : '+';
		my $start    = $a_curr[0];
		my $end      = $a_curr[1];
		my $len      = abs($end-$start)+1;	
		my $geneseq  = substr($the_seq, $start-1, $len);
		
		if ( $strand eq '+' ) # TN are searched on the opposite strand of the prediction	
		{
			$geneseq = &ReverseComplement($geneseq);
		}
		while($geneseq =~ /(\w{$contextlen}GT\w{$contextlen})/gi )
		{
			$cpt_gt++;
			my $siteseq = $1;
			my $pos     = $start+length($`);
			print STDERR ">DEBUG< Parent $idparent Seq extracted for TN : $seqid $start-$end $siteseq\n" if ($VERBOSE);
			$gt_dumped_nb += &DumpSite($f_tn_don, "$seqid:$pos:$opstrand", $siteseq, "GT", $contextlen, \%h_deja_tn);
		}
		while($geneseq =~ /(\w{$contextlen}AG\w{$contextlen})/gi )
		{
			$cpt_ag++;
			my $siteseq = $1;
			my $pos     = $start+length($`);
			$ag_dumped_nb += &DumpSite($f_tn_acc, "$seqid:$pos:$opstrand", $siteseq, "AG", $contextlen, \%h_deja_tn);
		}
		# on pourrait aussi chercher sur le meme brin sachant que dans h_deja on a les TP mais je suppose qu'il y a a deja assez
	}
	$f_tn_don->close;
	$f_tn_acc->close;

	if ($subsample)
	{
		&SubSample($tpfile_donor,$maxtpsites);
		&SubSample($tnfile_donor,$maxtnsites);
		&SubSample($tpfile_acceptor,$maxtpsites);
		&SubSample($tnfile_acceptor,$maxtnsites);
	}
	if ($VERBOSE)
	{
		print STDERR "Load $parent_nb Parent features.\n";
		print STDERR " Extract $cpt_gt GT regions, Dump $gt_dumped_nb regions\n";
		print STDERR " Extract $cpt_ag AG regions, Dump $ag_dumped_nb regions\n";
	}
}

sub SubSample
{
	my($file,$max) = @_;

	my $cpt=0;
	my $f_in  = &GetStreamIn($file);
	my $f_out = &GetStreamOut("$file.$max");
	while(my $l=<$f_in>)
	{
		if ( $l=~ /^>/ )
		{
			$cpt++;
			last if ($cpt>$max);
		}
		print $f_out $l;
	}
	$f_in->close;
	$f_out->close;

	return;
}


sub ExtractTNStartSites
{
	my($outprefix, $contextlen, $tn, $rh_parents,  $rh_refseq, $rh_seqs,$maxtnsites, $subsample) = @_;

	my $tnfile_start = "$outprefix.$tn.start";
	my $f_tn_start   = &General::GetStreamOut($tnfile_start); 
	my %h_deja_tn    = ();
	my $slen = 3;
	
	printf STDERR "Writing the file of start true negatives\n" if ( $VERBOSE );

	foreach my $idparent (keys(%$rh_parents))
	{
		#my $the_seq = $$rh_seqs{$idparent}{sequence};		
		my $the_seq = $$rh_seqs{$$rh_refseq{$idparent}}{sequence};
		my $seqid =  $$rh_refseq{$idparent};
		my @a_curr = split(/\t/, $$rh_parents{$idparent});
		my $strand = $a_curr[3];
		my $opstrand = ($strand eq '+' ) ? '-' : '+';
		my $start = $a_curr[0];
		my $end   = $a_curr[1];
		my $len   = abs($end-$start)+1;
		my $geneseq = substr( $the_seq, $start-1, $len);
		if ( $strand eq '+' ) # TN are searched on the opposite strand of the prediction	
		{
			$geneseq = &ReverseComplement($geneseq);
		}
		while($geneseq =~ /(\w{$contextlen}ATG\w{$contextlen})/gi )
		{
			my $siteseq = $1;
			my $pos     = $start+length($`);
			&DumpSite($f_tn_start,"$seqid:$pos:$opstrand",$siteseq,"ATG", $contextlen, \%h_deja_tn);
		}
	}
	$f_tn_start->close;
	
	&SubSample($tnfile_start, $maxtnsites) if ($subsample);
}


sub ExtractIntrons
{
	my($outprefix, $min_hit_length, $max_intron_length, $rh_parents, $rh_exons, $rh_refseq, $rh_seqs) = @_;

	my %h_deja = ();

	my $cpt_all = 0;
	my $cpt_ok  = 0;
	my $cpt_nc  = 0;	
	my $cpt_N   = 0;
	my $f_introns = &General::GetStreamOut($outprefix);

	foreach my $idparent (keys(%$rh_exons))
	{

		my $the_seq = $$rh_seqs{$$rh_refseq{$idparent}}{sequence};
		my $seqid = $$rh_refseq{$idparent};
			
		my @a_exons = split(/\n/,$$rh_exons{$idparent});
		for (my $i=0; $i < $#a_exons; $i++)
		{
			$cpt_all++;
			my @a_curr = split(/\t/,$a_exons[$i]);
			my @a_next = split(/\t/,$a_exons[$i+1]);
			my $exon_curr_len = abs($a_curr[1]-$a_curr[0])+1;
			my $exon_next_len = abs($a_next[1]-$a_next[0])+1;
			my $strand        = $a_curr[3];

			my $intron_start = $a_curr[1]+1;
			my $intron_end   = $a_next[0]-1;
			my $intron_len   = $intron_end - $intron_start + 1;

			if ( $exon_curr_len < $min_hit_length || $exon_next_len < $min_hit_length )
			{
				print STDERR "An exon flanking the intron is too short: ";
				print STDERR ">exon_curr_len=$exon_curr_len, exon_next_len=$exon_next_len, intron_len=$intron_len<\n";				
				print STDERR "\n\t$a_exons[$i]\n\t$a_exons[$i+1]\n";
				next;
			}
			if ( $intron_len > $max_intron_length)
			{
				print STDERR "Intron too long (Len = $intron_len)\n";
				print STDERR "\n\t$a_exons[$i]\n\t$a_exons[$i+1]\n";
				next;
			}
	
			
			my $sseq = substr($the_seq, $intron_start-1, $intron_len);
			if ( $strand eq '-' )
			{
				$sseq = &ReverseComplement($sseq);
			}

			if ( $sseq !~ /^(GT|GC)/i || $sseq !~ /AG$/i )
			{
				print STDERR "Warning: removed, non canonical intron (GT/GC...AG) : $sseq\n";
				$cpt_nc++;
				print STDERR "\n\t$a_exons[$i]\n\t$a_exons[$i+1]\n";
				next;
			}
			if ( $sseq =~ /N/i )
			{
				$cpt_N++;
				print STDERR "Warning: removed intron with 'N' : $sseq\n";
				print STDERR "\n\t$a_exons[$i]\n\t$a_exons[$i+1]\n";
				next;
			}
			
			if ( ! defined($h_deja{"$intron_start:$intron_end"}) )
			{
				$cpt_ok++;
				print $f_introns ">$seqid:intron:$intron_start:$intron_end len=$intron_len\n$sseq\n";
				$h_deja{"$intron_start:$intron_end"}=1;
			}
		}
	}
	$f_introns->close;
	
	print STDERR "$cpt_ok / $cpt_all selected non-redundant introns ($cpt_nc non canonical, $cpt_N with 'N') \n";

	return;
}



# If outprefix is defined, extract the non canonical splice sites in supplementar files
# Keep only one copy of each regions to dump
# Dont keep if the seq contains 'N' nucleotides
sub DumpSite
{
	my($flux, $id, $seq, $check, $clen, $rh_deja, $altoutprefix) = @_;

	$seq = "\U$seq";
	
	# keep only one copy of each region
	if ( defined($$rh_deja{$seq}) )
	{
		print STDERR "DEBUGDumpSite\t$seq already read\n" if ($VERBOSE);
		return 0;
	}

	# case noncanonical site
	if ( $seq !~ /^(\w{$clen})($check)(\w{$clen})$/i )
	{
		printf STDERR "Warning: unexpected splice site (expected=$check): $id len=%d $seq\n",length($seq);
		
		if ( defined($altoutprefix) && $altoutprefix ne "" )
		{
			if ( $seq =~ /^(\w{$clen})(\w\w)(\w{$clen})$/ )
			{
				my $altsite = $2;
				if ( $check =~ /AG/i )
				{
					$altsite = "alt-acc.$altsite";
				}
				else
				{
					$altsite = "alt-don.$altsite";
				}
				if ( $altsite =~ /\.[ATGC]+$/ )
				{
					my $f_alt = new IO::File("$altoutprefix.$altsite","a");
					print $f_alt ">$id\n$seq\n";
					$f_alt->close;
					
					$$rh_deja{$seq}     = 1;
					my $revcomp         = &ReverseComplement($seq);
					$$rh_deja{$revcomp} = 1;
				}
			}
		}
		return 0;
	}
	elsif ( $seq !~ /N/i )
	{
		print $flux ">$id\n$seq\n";
		$$rh_deja{$seq}     =1;
		my $revcomp         = &ReverseComplement($seq);
		$$rh_deja{$revcomp} =1;
		return 1;
	}
	else
	{
		print STDERR "DEBUGDumpSite\t$seq contains N\n" if ($VERBOSE);
		return 0;
	}
	
	return 1;
}



sub ReverseComplement
{
	my($seq) = @_;

	$seq =~ tr/ATGCatgc/TACGtacg/;
	return reverse($seq);
}



sub LoadGFF3
{
	my($gfffile, $parent, $child, $rh_parents, $rh_exons, $rh_ref) = @_;
	
	my %h_multiexons = ();

	my $f_gff = &General::GetStreamIn($gfffile);

	if ( $VERBOSE )
	{
		printf STDERR "Loading gff3 file\n";
	}
	while(my $line=<$f_gff>)
	{
		if ( $line =~ /\t$child\t/ && $line =~ /Parent=([^;\s]+)/ )
		{
			my $idparent = $1;
			$line =~ s/^\S+\s+\S+\s+\S+\s+//;	# little compression
			$h_multiexons{$idparent}=1 if ( defined($$rh_exons{$idparent}) );
			$$rh_exons{$idparent} .= $line;
		}
		elsif ( $line =~ /\t$parent\t/ &&  $line =~ /ID=([^;\s]+)/ )
		{
			my $id = $1;
			my ($genome) = $line =~ /^([^\t]+)\t/;
			$$rh_ref{$id} = $genome;
			$line =~ s/^\S+\s+\S+\s+\S+\s+//;
			$$rh_parents{$id} = $line;
		}
	}
	$f_gff->close;
	
	if ( $VERBOSE )
	{
		printf STDERR "\t%d $parent loaded\n",keys(%$rh_parents)+1;
		printf STDERR "Removing single exon hits\n";
	}
	
	foreach my $idparent (keys(%$rh_parents))
	{
		if ( !defined($h_multiexons{$idparent}) )
		{
			delete($$rh_parents{$idparent});
			delete($$rh_exons{$idparent});
		}
	}
	if ( $VERBOSE )
	{
		printf STDERR "\t%d $parent remaining\n",keys(%$rh_parents)+1;
	}

}



sub Usage
{
    print STDERR<<END
$0
    --help
	--sequence   file		fastafile
	--gff3       file		gff3 formatted file
	--outprefix  string		prefix for output files

[Optional]
    --parent_feature        string  [EST_match]  defines the Parent elements
    --child_feature         string  [match_part] defines the child elements
    --context_length        integer 
    --intron_file           file    If fill, extract intronic sequence 
    --max_intron_length     integer [$MAX_INTRON_LENGTH]
    --intron_min_hit_length integer [$INTRON_MIN_HIT_LENGTH]
    --tn_start_sites        flag    Create TN start sites
    --no_alt_site_detection flag 	Do not detect uncanonical splicing sites
		
[Subsampling]           
    --subsample	            flag    Create subsamples
    --maxtpsites            integer [30000]
    --maxtnsites            integer [300000]
END
      ;
}
