#!/usr/bin/perl

BEGIN
{
    $ENV{LANG} = "C";
}

use FindBin;
use lib "$FindBin::Bin/../../lib/";
use lib "$FindBin::Bin/../../corelib/"; 


use strict;
use warnings;
use General;
use GeneralBioinfo;
use ParamParser;

my $DEBUG = 1;

my $DEF_CONTEXT_LENGTH=60;
my $DEF_ORDER="";


MAIN:
{
	my $o_param = New ParamParser('GETOPTLONG', \&Usage, "outdir=s", "cfg=s", "pcs=i", "pci=i", 
                                   "context_length=i", "order=i", "training_data_dir=s", "force");
    $o_param->Update('ENV', "A");	 
	$o_param->AssertFileExists('cfg');
    $o_param->AssertFullPath('cfg');
	# assert EGNEP is defined
	$o_param->AssertDirExists('EGNEP');
	$o_param->Update($o_param->Get('cfg'), 'A');
	$o_param->SetBehaviour('use_substitution_table');	
    $o_param->SetSubstitution('%i', $o_param->Get('EGNEP'));
    $o_param->SetSubstitution('%e', $o_param->Get('EUGENEDIR'));
	$o_param->Update($o_param->Get('cfg'), 'U');
 
	$o_param->AssertDefined('wam_species', 'outdir');
	$o_param->AssertFullPath('outdir');
	$o_param->AssertDefined('prg_gmapbuild', 'prg_get_est_gmap', 'prg_gmap','prg_eugene_wambuilder');
	$o_param->AssertDefined('prg_gmap_param', 'prg_gmap_filter_param','prg_build_training_dataset', 'prg_nrdb');
	&AssertData($o_param);
	
	$o_param->SetUnlessDefined('pcs', 90);
	$o_param->SetUnlessDefined('pci', 99);
	$o_param->SetUnlessDefined('context_length', $DEF_CONTEXT_LENGTH);	
	$o_param->SetUnlessDefined('order', $DEF_ORDER);

	my $species = $o_param->Get('wam_species');	
	my $outdir  = $o_param->Get('outdir');
	system("mkdir $outdir") if (! -d $outdir);
	
	my $training_data_dir; # directory containing the acceptor/donor TP|TN spliced site region files

	
	if ($o_param->IsDefined('training_data_dir'))
	{
		$o_param->AssertDirExists('training_data_dir');
		$training_data_dir = $o_param->Get('training_data_dir');
	}
	else
	{	
		# For each dataset, map and extract region around splice site
		my @a_fasta_root;
		my $list = $o_param->Get('wam_dataset_list');    # "2 5 6"
    	my @a_nb = split(/\s+/, $list);
	
	
		foreach my $nb (@a_nb)
    	{
			# create directory for this dataset
			my $datasetdir = "$outdir/dataset_$nb";
			rmdir ($datasetdir) if (-d $datasetdir && $o_param->IsDefined('force'));
			mkdir ($datasetdir) if (!-d $datasetdir);
		
       		my $genome        = $o_param->Get("wam_dataset_$nb\_genome");
			my $gmap_res      = "$datasetdir/gmap";
		
			if (! -e "$datasetdir/dataset$nb.TN.acceptor.success")
			{
				my $ra_transcriptome = $o_param->GetListOfFiles("wam_dataset_$nb\_transcriptome");
				my $cpt              = 0;
				unlink("$gmap_res.gff3") if (-e "$gmap_res.gff3");
				
				foreach my $transcriptome (@$ra_transcriptome)
				{
					$cpt++;
					&Mapping($o_param, $genome, $transcriptome, "$gmap_res.$cpt");
					system("cat $gmap_res.$cpt.gff3 >> $gmap_res.gff3");
                    #unlink ("gmap_res.$cpt.gff3");
				}

				&BuildSplicedSites($o_param, $genome, $gmap_res.".gff3", $datasetdir, "dataset$nb");
				`touch $datasetdir/dataset$nb.TN.acceptor.success`;
			}
			else
			{
				print "Dataset $nb already exists in $datasetdir\n";
			}
			push(@a_fasta_root, "$datasetdir/dataset$nb");
		}

		my $datasetpooldir = "$outdir/dataset_$species";
		system("mkdir $datasetpooldir") if (! -d $datasetpooldir);
		
		&Nrdb($o_param, \@a_fasta_root, "TN.acceptor", "AG.TN", $datasetpooldir);	
		&Nrdb($o_param, \@a_fasta_root, "TP.acceptor", "AG.TP", $datasetpooldir);
		&Nrdb($o_param, \@a_fasta_root, "TN.donor",    "GT.TN", $datasetpooldir);
		&Nrdb($o_param, \@a_fasta_root, "TP.donor",    "GT.TP", $datasetpooldir);
		
		$training_data_dir = $datasetpooldir;
	}

	# Build wam models and create a tar.gz
	&BuildWam($o_param, $training_data_dir, $species, $outdir, $o_param->Get('order'));
}


sub Nrdb
{
	my ($o_param, $ra_prefix, $suffix, $new_suffix, $outdir) = @_;
	
	my $species        = $o_param->Get('wam_species');
	my $context_length = $o_param->Get('context_length');
	my $resfile        = sprintf ("$outdir/$species.uc%03d.dc%03d.%s", $context_length, $context_length, $new_suffix);

	my $cmd = $o_param->Get('prg_nrdb') ." -o $resfile ";
	foreach my $f (@$ra_prefix)
	{
		$cmd .= " $f.$suffix ";
	}
	
	system($cmd);
	print "$cmd\n" if ($DEBUG);
	
	return;
	
}

sub AssertData
{
	my ($o_param) =@_;
		
	$o_param->AssertDefined('wam_dataset_list');

    my $list = $o_param->Get('wam_dataset_list');    # "2 5 6"
    my @a_db_nb = split(/\s+/, $list);

    foreach my $nb (@a_db_nb)
    {
        $o_param->AssertFileExists("wam_dataset_$nb\_genome");
        $o_param->AssertDefined("wam_dataset_$nb\_transcriptome");	
	}
	return;
}


sub Mapping
{
	my ($o_param, $sequence, $transcripts, $outfile) = @_;

	my $pcs = $o_param->Get('pcs');
	my $pci = $o_param->Get('pci');
	my $cfg = $o_param->Get('cfg');
	
	if (-e "$outfile.gff3.success")
	{
		print STDERR "$outfile.gff3.success exists. Skip mapping\n";
		return;
	}
	
	my $cmd = "export EGNPP=" . $o_param->Get('EGNEP') . ";";
    $cmd .= "export EGNEP=" . $o_param->Get('EGNEP') . ";";
    $cmd .= $o_param->Get('prg_get_est_gmap') . " --sequence $sequence --mrnadb $transcripts --pcs $pcs --pci $pci --cfg $cfg";
    $cmd .= " --outfile $outfile " if ($outfile ne "");
    $cmd .= " --min_length_short_unspliced 100000 ";
	$cmd .= " > /dev/null 2>&1 " if (!$DEBUG);
	
	print STDERR $cmd."\n" if ($DEBUG);
	system($cmd);
	
	`touch $outfile.gff3.success` if (-e "$outfile.gff3");
	
	return;

}


sub BuildSplicedSites
{
	my ($o_param, $sequence, $gff3, $outdir, $name) = @_;

	my $context_length = $o_param->Get('context_length');

	chdir ($outdir);
	my $cmd = $o_param->Get('prg_build_training_dataset')." --sequence $sequence --gff3 $gff3 --no_alt_site_detection --parent_feature EST_match --child_feature match_part --outprefix $name ";
	$cmd .= " --context_length $context_length ";
	$cmd .= $o_param->Get('prg_build_training_dataset_param') if $o_param->IsDefined('prg_build_training_dataset_param');
	$cmd .= "2> $outdir/build_training_dataset.log";
	
	print STDERR "$cmd\n" if ($DEBUG);
	system ($cmd);
	
	return;
}



sub BuildWam
{
	my ($o_param, $datasetdir, $species, $outdir, $order) = @_;

	# 1 Create a directory where save the wam models
	my $wamdir = "$outdir/WAM";
	mkdir($wamdir) if (! -d $wamdir);
	
	chdir($datasetdir);
	my $context_length = $o_param->Get('context_length');

	if ($order eq "")
	{
		$order = &SelectWamOrder($species, $context_length);
	}
	
	foreach my $site ('AG','GT')
	{
		foreach my $type ('TP','TN')
		{
			my $file_name = sprintf ("%s.uc%03d.dc%03d.%s.%s ", $species, $context_length, $context_length, $site, $type);
			my $site_name = ($site eq "AG") ? "acceptor" : "donor";
			my %h_seq;
			my $nbseq = &GeneralBioinfo::FastaToHash($file_name,\%h_seq, 1);
			print "Load $nbseq $type $site_name sequences for training WAM...\n";
			my $cmd = $o_param->Get('prg_eugene_wambuilder')." $file_name $species $site_name $site $type $context_length $context_length $order $wamdir 2> $datasetdir/eugene_wambuilder.$species.$context_length.$order.$site.$type.log\n";
			print $cmd;
			system ($cmd);
		}
	}

	chdir($wamdir);
	
	system("rm $outdir/$species.tar.gz") if (-e "$outdir/$species.tar.gz");
	
	my $cmd = "tar cvzf $outdir/$species.tar.gz $species/ "; 	
	$cmd .= " > /dev/null 2>&1 " if (!$DEBUG);
	system($cmd);
	print STDERR "$cmd\n"  if ($DEBUG);
		
	chdir($outdir);	
	system("rm -rf $wamdir");
		
	print "Wam models for $species created: $species.tar.gz\n";
	
	return;
}



# TODO: improve code!!
# return maximum order to compyte wam according to number of seq in our dataset
sub SelectWamOrder
{
	my ($species, $context_length) = @_;
	
	my $min_seq_nb = 0;
	# In the 4 training datasets, search the one with the ninimum nb of sequences
	foreach my $site ('AG','GT')
	{
		foreach my $type ('TP','TN')
		{
			my $file_name = sprintf ("%s.uc%03d.dc%03d.%s.%s ", $species, $context_length, $context_length, $site, $type);
			my %h_seq;
			my $nbseq = &GeneralBioinfo::FastaToHash($file_name,\%h_seq, 1);
			$min_seq_nb = $nbseq if ($min_seq_nb == 0 || $nbseq < $min_seq_nb);
		}
	}

	my $threshold = 400;
	my $order     = -1;
	while ($min_seq_nb >= $threshold)
	{
		$order++;
		$threshold = $threshold*4;
	}

	if ($order < 0)
	{
		print STDERR "Not enough sequences to build wam ($min_seq_nb < $threshold)\n";
		exit;
		return -1;
	}
	# we know that (threshold/4) < $min_seq_nb < $threshold. 
	# order  = $order + 1 if $min_seq_nb is closest to $threshold than ($threshold/4)
	$order++ if ( ($threshold - $min_seq_nb) <=  ($min_seq_nb-($threshold/4)) );

	print "Computed WAM order=$order. [Optimal value with a dataset of $min_seq_nb sequences]\n";
	
	return $order;
}

	
sub Usage
{
    print STDERR<<END
$0
Build wam models for a species. 
	--help
	--cfg                  file	
    --outdir               string
[Optional]		
    --order                integer  [Default choose according to training sequence number] wam order 		
    --context_length       integer  [Default \$DEF_CONTEXT_LENGTH] 
    --pcs                  integer  [Default 90] Percentage of the aligned mrna
    --pci                  integer  [Default 99] Percentage of identity
    --force                         Relaunch all compute 		
    --training_data_dir    dir      Directory contaning the 4 multifasta files of TP|TN acceptor and donor sites.
	                                Example: speciesX/speciesX.uc060.dc060.[AG|GT].[TP|TN]
		                            => Skip mapping step and only build WAM models.
END
      ;
}
