#!/usr/bin/perl
#
# $Id
#

# Copyright INRA

# Jerome.Gouzy@toulouse.inra.fr
# Erika.Sallet@toulouse.inra.fr
# Thomas.Schiex@toulouse.inra.fr

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=pod

=head1 NAME

	get_Est_Gmap.pl

=head1 SYNOPSIS

 $EGNP/bin/int/get_Est_Gmap.pl --sequence fasta_file [--mrnadb mrnadb  --pcs 80 --pci 100] [--split /results/]

=head1 DESCRIPTION

get_Est_Gmap.pl - Execute gmap, filter about pcs and pci and save in GFF3 format EST_match/match_part

If --split:
 Split by sequence and save the results in the sequence subdirectory. 
 Create empty files if no result for some sequences.



 --sequence  file   input fasta file 

Optional qualifiers:
 --split    dir                      Root directory where save the results splitted by sequence.
 --mrnadb   file                     Multifasta file of transcripts
                                     [Default values are keys 'prg_getest_db*' read in the cfg file]
 --outfile  file   [{sequence}.est]  Output file. The suffix ".gff3" is added to the result files
 --pcs      float  [$pcs]            percent of est overlapping 
 --pci      float  [$pci]            identity percentage
 --split    fof

=cut

BEGIN
{
    $ENV{LANG} = "C";
}

use FindBin;
use lib "$FindBin::Bin/../../corelib/";
use lib "$FindBin::Bin/../../lib/"; 


use strict;
use ParamParser;
use Global;
use General;
use GeneralBioinfo;

my $DEBUG;
my $DEFAULT_OUT_EXT = "est";
my $DEFAULT_PCS = 80;
my $DEFAULT_PCI = 90;

my ( $O_CMD);	

MAIN:
{
	my $o_param = New ParamParser('GETOPTLONG',  \&Usage, "sequence=s", 
				      "mrnadb=s", "outfile=s", "pci=i", "pcs=i", "cfg=s", "split=s", 
				      "noclean", "outext=s", "gmapbuild_dir=s", "min_length_short_unspliced=i", 'debug');

	$o_param->SetBehaviour('ignore_space');
	$o_param->Update($o_param->Get('cfg'),'A');
	$o_param->SetSubstitution('%i', $ENV{'EGNEP'});
	$o_param->SetBehaviour('use_substitution_table');


	$DEBUG = 0;
	$DEBUG = 1 if ($o_param->IsDefined('debug'));

	$o_param->AssertFileExists('sequence');
	$o_param->AssertFullPath('sequence');
	$o_param->AssertFileExists('split') if ($o_param->IsDefined('split'));
	if ($o_param->IsDefined('split') && $o_param->IsDefined('outfile'))
	{
		print STDERR "Warning: 'outfile' parameter ignored in split mode. Only 'outext' parameter can be defined.\n";
	}
	if (!$o_param->IsDefined('split') && $o_param->IsDefined('outext') && $o_param->IsDefined('outfile'))
	{
		print STDERR "Warning: 'outext' parameter is ignored when 'outfile' is defined.\n";
	}

	my $seqfasta = $o_param->Get('sequence');
	
	$o_param->SetUnlessDefined('outext', $DEFAULT_OUT_EXT);
	my $outext = $o_param->Get('outext');
	$o_param->SetUnlessDefined('outfile', "$seqfasta.$outext");
	$o_param->SetUnlessDefined('pcs', $DEFAULT_PCS);
	$o_param->SetUnlessDefined('pci', $DEFAULT_PCI);
	$o_param->AssertDefined('prg_gmapbuild', 'prg_gmap', 'prg_gmap_param', 'prg_gmap_filter');
	$o_param->AssertFullPath('outfile');
	$o_param->AssertInteger('min_length_short_unspliced') if ($o_param->IsDefined('min_length_short_unspliced'));

	my $outfile            = $o_param->Get('outfile');
	my ($outdir, $outname) = $outfile =~ /(.+)\/(.+)/;
	my @a_lib              = sort $o_param->GetKeys('prg_getest_db');
	if ( $o_param->IsDefined('mrnadb'))
	{
		$o_param->AssertFileExists('mrnadb');
		@a_lib = ();
		$a_lib[0] = 'prg_getest_db1';
		$o_param->Set('prg_getest_db1',$o_param->Get('mrnadb'));
	}

	# Read the paths of the different genomic sequences
	my %h_seq_path;
	if ($o_param->IsDefined('split'))
	{
		my $f_fof = $o_param->GetStreamIn('split');
		while (my $line = <$f_fof>)
		{
			chomp($line);
			my ($seq) = $line =~ /.+\/([^\/]+)$/;
			$h_seq_path{$seq} = $line;
		}
		$f_fof->close;
	}
	
	my $gmapbuild_dir ;
	if ($o_param->IsDefined('gmapbuild_dir'))
	{
		$o_param->AssertDirExists('gmapbuild_dir');
		$gmapbuild_dir = $o_param->Get('gmapbuild_dir');
	}
	else
	{
		$gmapbuild_dir = $outdir."/GMAP_BUILD_$$";
		&GmapBuild($o_param, $seqfasta, $gmapbuild_dir);
	}
	
	my @a_ext ; # Array of file suffixes to compute

	foreach my $libkey (@a_lib)
    {
		my ($i) = $libkey =~ /(\d+)$/;
        $o_param->AssertDefined($libkey);

        my $db               = $o_param->Get($libkey);
		my $extlib           = ( $i == 1 ) ? "" : "$i";		
		my $outputgmap       = $o_param->Get('outfile')."$extlib.gmap_raw.gff3";
		my $outputgmapfilter = $o_param->Get('outfile')."$extlib.gmap_filter.gff3";		
		my $output           = $o_param->Get('outfile')."$extlib.gff3";
		my $output_unsorted = $output.".unsorted.gff3";
		
		&GMap($o_param, $seqfasta, $db, $gmapbuild_dir, $outputgmap);

		&GMapFilter($o_param, $outputgmap, $outputgmapfilter, $seqfasta);

		&ChangeSourceName($outputgmapfilter, $output_unsorted, "gmap$extlib");

		&SortFileByItem($output_unsorted, $output, $o_param);
		
		# Remove tmp gmap and gmapfilter file
		unlink($outputgmapfilter, $output_unsorted) if (!$o_param->IsDefined('noclean'));
		
		# Zip raw gmap file
		system("gzip -f $outputgmap");

		if ($o_param->IsDefined('split'))
		{
			&Split($o_param, $output, \%h_seq_path, "$outext$extlib.gff3");
			push(@a_ext, "$outext$extlib.gff3");
			
			# Remove global result file
			unlink($output) if (!$o_param->IsDefined('noclean'));
		}

		if ($o_param->IsDefined('min_length_short_unspliced'))
		{
			my $minlen_su = $o_param->Get('min_length_short_unspliced');
			# file exists
			my $unspliced = $o_param->Get('outfile')."$extlib.gmap_filter.rm_unspliced_${minlen_su}.gff3";
			my $o         = $o_param->Get('outfile')."$extlib.rm_unspliced_${minlen_su}.gff3";
			&ChangeSourceName($unspliced, $unspliced.".unsorted.gff3", "gmap$extlib");
			&SortFileByItem($unspliced.".unsorted.gff3", $o, $o_param);
            # Remove tmp  file
		    unlink($unspliced, $unspliced.".unsorted.gff3") if (!$o_param->IsDefined('noclean'));
			
			if ($o_param->IsDefined('split'))
			{
				&Split($o_param, $o, \%h_seq_path, "$outext$extlib.rm_unspliced_${minlen_su}.gff3");	
				push(@a_ext, "$outext$extlib.rm_unspliced_${minlen_su}.gff3");	
                # Remove global result file
				unlink($o) if (!$o_param->IsDefined('noclean'));
			}

		}
	}
	
	if ($o_param->IsDefined('split'))
	{
		&CompleteFileTree(\%h_seq_path, \@a_ext);
	}


	# Remove gmap index directory
	if (!$o_param->IsDefined('noclean') && ! $o_param->IsDefined('gmapbuild_dir'))
	{
		`rm -rf $gmapbuild_dir`;
	}
	
}



=head2 procedure CompleteFileTree

    Title        : CompleteFileTree
    Usage        : CompleteFileTree ($rh_seq_f, \@ra_ext)
    Prerequisite : -
    Procedure    : 
    Args         : $seq_fof         ref to an hash. Key=seqid, value=seq file
                   $ra_ext          ref to an array of file extension
    Globals      : none
=cut
sub CompleteFileTree
{
	my ($rh_seq, $ra_ext) = @_;
	
	foreach my $seqid (keys %$rh_seq)
	{
		my $seq = $$rh_seq{$seqid};
		foreach my $ext (@$ra_ext)
		{
			if (! -e "$seq.$ext")
			{
				print STDERR "Create $seq.$ext\n" if ($DEBUG);
				my $f_out = &GetStreamOut("$seq.$ext");
				$f_out->close;
			}
		}
	}
	
	return;
}

=head2 procedure GmapBuild

    Title        : GmapBuild
    Usage        : &GmapBuild($o_param, $seqfasta, $dir);
    Prerequisite : none
    Procedure    : Create gmap indexation directory (if not exists) 
                   and index genome sequence for gmap
    Args         : $o_param
                   $genome          Fasta sequence
                   $outdir          Directory of gmap indexation
    Globals      : none
=cut
sub GmapBuild
{
	my ($o_param, $genome, $outdir) = @_;

	my ($genomedir, $genomename) = $genome =~ /(.+)\/(.+)/; 
	
	mkdir($outdir) if (! -d  $outdir);

	# -d STRING  genome name
	# -D STRING  destination directory for installation
	my $cmd = $o_param->Get("prg_gmapbuild") . " -d $genomename -D $outdir $genome";
	print STDERR $cmd ."\n"; 
	system($cmd);

	return;
}




=head2 procedure Gmap

    Title        : Gmap
    Usage        : &GMap($o_param, $seqfasta, $db, $gmapbuilddir, $outputgmap );
    Prerequisite : prg_gmap, prg_gmap_param are keys of $o_param. 
                   $gmapbuild_dir exists.
    Procedure    : Launch gmap. Output is in GFF3
    Args         : $o_param         ParamPareser object
                   $genome          Fasta sequence
                   $db              Multifasta of transcripts
                   $gmapbuild_dir   Directory of gmap indexation
                   $outfile         Gmap results file
    Globals      : none
=cut
sub GMap
{
	my ($o_param, $genome, $db, $gmapbuild_dir, $outfile) = @_;
	
	my ($genomedir, $genomename) = $genome =~ /(.+)\/(.+)/; 

	my $cmd = $o_param->Get('prg_gmap') . " -D $gmapbuild_dir -d $genomename -f gff3_gene ";
	$cmd .= " " . $o_param->Get('prg_gmap_param');
	$cmd .= " $db > $outfile 2> /dev/null";

	print STDERR $cmd."\n"; 
	system($cmd);

	return;
}



=head2 procedure GmapFilter

    Title        : GmapFilter
    Usage        : &GMapFilter($o_param, $outputgmap, $output, $seqfasta);
    Prerequisite : prg_gmap_filter, pcs and pci  are existing keys of $o_param. 
    Procedure    : Filter gmap results according to pcs and pci. Save results in GFF3 (EST_match/match_part)
    Args         : $o_param         ParamParser object
                   $in              GFF3 gmap result file
                   $out             Filtering result GFF3 file
                   $seq             Fasta sequence
    Globals      : none
=cut
sub GMapFilter
{
	my ($o_param, $in, $out, $seq) = @_;
	
	my $pcs = $o_param->Get('pcs');
	my $pci = $o_param->Get('pci');
	
	my $cmd = $o_param->Get('prg_gmap_filter') . " --in $in --out $out --pci $pci --pcs $pcs --format est --seqin $seq";
	$cmd .= " " . $o_param->Get('prg_gmap_filter_param') if ($o_param->IsDefined('prg_gmap_filter_param'));
	$cmd .= " --min_length_short_unspliced ". $o_param->Get('min_length_short_unspliced') if ($o_param->IsDefined('min_length_short_unspliced'));
	print STDERR $cmd."\n"; 
	system($cmd);

	return;
}


sub ChangeSourceName
{
	my ($ingff3, $outgff3, $source) = @_;

	
	my $f_in  = &General::GetStreamIn($ingff3);
	my $f_out = &General::GetStreamOut($outgff3);
	while (my $line = <$f_in>)
	{
		chomp($line);
		my @a_elt = split(/\t/, $line);
		$a_elt[1] = $source;
		print $f_out join("\t", @a_elt) . "\n";
	}
	
	$f_in->close;
	$f_out->close;
	
	return;
}

=head2 procedure Split

    Title        : Split
    Usage        : &Split($o_param, $file, $fof, $ext)
    Prerequisite : -
    Procedure    : split GFF3 $file per sequence name (column 1) and 
                   save each result file in the same directory than the sequence 
                   (sequence list in $fof)
    Args         : $o_param
                   $file        GFF3 file
                   $rh_seq_file Ref to a hash. Key=seqid, value=sequence file
                   $ext         String Extension of the result files 
   Prerequisite  : GFF3 file is sorted according to the sequence name
    Globals      : none
=cut
sub Split
{
	my ($o_param, $file, $rh_seq_file, $ext) = @_;

	my $seqid = '';
	my $f_out = undef;

	my $f_in  = &GetStreamIn($file);
	while(my $l = <$f_in>)
	{
		chomp($l);
		next if ( $l =~ /^#/ );
		my @a_= split(/\t/, $l);
		
		if ( ! defined($f_out) || $a_[0] ne $seqid )
		{
			$f_out->close if defined ($f_out);
			$seqid = $a_[0];
			if (!defined $$rh_seq_file{$seqid})
			{
				print STDERR "GMAP Warning during split result: The file of files does not contain the $seqid sequence. $seqid results are ignored\n";
				next;
			}
			my $seq = $$rh_seq_file{$seqid};
			$f_out = &GetStreamOut("$seq.$ext");
		}
		print $f_out "$l\n";
	}
	$f_in->close;
	$f_out->close if defined ($f_out);

	
	return;
}



sub SortFileByItem
{
	my($in, $out, $o_param) = @_;
	
	my $f_in;
	my $f_out;
	
	$f_in  = &GetStreamIn($in);
	$f_out = &GetStreamOut("$out.Parent.gff3");
	my %h_gff3;
	
	# 1. je fais la liste des parents
	while (my $line = <$f_in>)
	{
		&ParseGFF3Line(\$line, 1, \%h_gff3);
		if ($h_gff3{type} eq "EST_match")
		{		
			print $f_out $line;
		}
	}
	$f_in->close;
	$f_out->close;
    
	#  2 Tri des parents par seq puis par position de debut
	my $cmd = "sort -s -k4,4n $out.Parent.gff3 | sort -s -k1,1 > $out.Parent.gff3.sorted";
	system($cmd);
	
	
	# 3 sauvegarde de l'ordre des parents
	my %h_parent_sort;
	my $cpt = 0;
	my $f_parent = &GetStreamIn("$out.Parent.gff3.sorted");
	while(my $l = <$f_parent>)
	{
		chomp($l);
		my ($id) = $l =~ /ID=([^;]+);?/;
		if (defined($h_parent_sort{$id}))
		{
			print STDERR "ERROR Get_Est_gmap: duplicated id in gmap results.\n";
			exit;
		}
		$cpt++;
		$h_parent_sort{$id} = $cpt;
	}
	$f_parent->close;

	# 4 creation d'un fichier avec le numero des parents
	$f_in     = &GetStreamIn($in);
	my $f_num = &GetStreamOut("$out.Numbered.gff3");
	while(my $l = <$f_in>)
	{
		my $id;
		chomp($l);
		&ParseGFF3Line(\$l, 1, \%h_gff3);
		if ($h_gff3{type} eq "EST_match")
		{
			$id = $h_gff3{ID};
		}
		else
		{
			$id = $h_gff3{Parent};
		}
		print $f_num $h_parent_sort{$id}."\t$l\n";

	}
	$f_in->close;
	$f_num->close;

	# 5 tri en fonction du numero des parents (premiere colonne)
	$cmd = "sort -k1,1n -s $out.Numbered.gff3 > $out.Numbered.gff3.sorted";
	system($cmd);

	$f_in     = &GetStreamIn("$out.Numbered.gff3.sorted");
	$f_out = &GetStreamOut("$out");
	my $current = 0;	
	my @a_ok    = ();
	while(my $l = <$f_in>)
	{
		my ($nb, $line) = $l =~ /^(\d+)\t(.+)$/;
		$line .= "\n";
		if ($current != $nb && scalar(@a_ok) > 0 )
		{
			print $f_out sort SortByItem @a_ok;
			@a_ok = ();
		}
		$current = $nb;
		push(@a_ok, $line);
	}
	
	print $f_out sort SortByItem @a_ok;
	
	$f_in->close;
	$f_out->close;

	if (!$o_param->IsDefined('noclean'))
	{
		unlink("$out.Parent.gff3");
		unlink("$out.Parent.gff3.sorted");
		unlink("$out.Numbered.gff3");
		unlink("$out.Numbered.gff3.sorted");
	}

	return;
}


sub SortByItem
{
	my %h_a=();
	&ParseGFF3Line(\$a,0,\%h_a);
	my %h_b=();
	&ParseGFF3Line(\$b,0,\%h_b);

	if ( $h_a{type} ne $h_b{type} )
	{
		return $h_a{type} cmp $h_b{type};
	}
	else
	{
		return $h_a{start} <=> $h_b{start};
	}
}


=head2 Function Usage

 Title        : Usage
 Usage        : &Usage()

=cut

sub Usage
{

    print <<END
Usage: $0

PREREQUISITE: 'sort' commands in the path

Mandatory qualifiers:
 --sequence  file   input fasta file 
 --cfg       file   configuration file

Optional qualifiers:

 --help                              this message
 --mrnadb   file                     Multifasta file of transcripts
                                     [Default values are keys 'prg_getest_db*' read in the cfg file]
 --outfile  file   [{sequence}.$DEFAULT_OUT_EXT]  Output file. The suffix ".gff3" is added to the result files
 --outext   string [$DEFAULT_OUT_EXT]             Extension of the output file[s]. The suffix ".gff3" is added to the result files
 --pcs      float  [$DEFAULT_PCS]              Percent of est overlapping 
 --pci      float  [$DEFAULT_PCI]              Identity percentage
 --min_length_short_unspliced integer [0] remove short unspliced matches
 --split    fof                      File of sequence files. If defined, split by sequence and save result files 
                                     in the sequence directories.
                                     
 --noclean
 --debug

Advanced qualifiers:
  --gmapbuild_dir dir              Path of the gmap index directory

END
;
}




