#!/usr/bin/perl

#
# $Id$
#

# Copyright INRA

# Erika.Sallet@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Sebastien.Carrere@toulouse.inra.fr
# Ludovic.Cottret@toulouse.inra.fr
# Ludovic.Legrand@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based system for detecting, clustering and annotating non protein
# coding RNA genes.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1

=head1 NAME

    lipm_annot2match.pl

=head1 SYNOPSIS

   ./lipm_annot2match.pl --infile genome.annotation.gff3 --outprefix genome.match

=head1

=cut


BEGIN
{
	$ENV{LANG} = "C";
	my ($dirprg, $nameprg) = $0 =~ /(.+)\/(.+)/;
	$dirprg = "." if ($dirprg eq "");
	unshift(@INC, "$dirprg/../../../corelib");
	unshift(@INC, "$dirprg/../../../lib");
}

use strict;
use warnings;
use ParamParser;
use General;
use GeneralBioinfo;

my $ALLOWED_FEATURES = "mRNA;rRNA;tRNA;ncRNA;five_prime_UTR;three_prime_UTR;exon;CDS;";
my $PARENT_FEATURE = "EST_match";
my $DEBUG = 0;
my $COMMENT = 0;

MAIN:
{
    my $o_param = New ParamParser;
    $o_param->SetUsage(my $usage = sub {&Usage();});
    $o_param->Update( "GETOPTLONG", "I", ( "infile=s", "outprefix=s", "eugene", "parent_feature=s"));
	
    $o_param->SetBehaviour('assert_empty_file_allowed') ;

    $o_param->AssertFileExists('infile');
	$o_param->AssertDefined('outprefix');
	$o_param->SetUnlessDefined('parent_feature', $PARENT_FEATURE);
	my $parent_feature = $o_param->Get('parent_feature');
	my $f_in    = $o_param->GetStreamIn('infile');
	my $outfile = $o_param->Get('outprefix');
	my $f_out_mrna  = &General::GetStreamOut("$outfile.mRNA.gff3");
	my $f_out_ncrna = &General::GetStreamOut("$outfile.ncRNA.gff3");

	if ($o_param->IsDefined('eugene'))
	{
		&ParseEuGeneFile($f_in, $f_out_mrna, $f_out_ncrna, $parent_feature);
	}
	else
	{
		&ParseFile($f_in, $f_out_mrna, $f_out_ncrna, $parent_feature);
	}
	
   
	$f_in->close;
	$f_out_ncrna->close;
	$f_out_mrna->close;
}


# Parseur qui gere les varaitn d'epissage
# Un exon peut avoir plusieurs parents
# Je cree un hash : key=parent_id, value=lignes qui ont comme parent parent_ig
sub ParseFile
{	
	my ($f_in, $f_out_mrna, $f_out_ncrna, $parent_feature) = @_;
	
	my %h_res;
	my %h_parent;
	
	while (my $line = <$f_in>)
	{
	   	next if ($line =~ /^#/);
		
		&GeneralBioinfo::ParseGFF3Line(\$line, 1, \%h_res);
			
		if ($h_res{type} eq "gene")
		{
			# je suis sur un nouveau gene: j'affiche ce qui a ete precedemment suavegarde dans le hash
			# et je reinitialise le hash
			foreach my $p (keys %h_parent)
			{
				&PrintIfOk($h_parent{$p}, $f_out_mrna, $f_out_ncrna,1, $parent_feature);
			}
			%h_parent = ();
		}
		elsif ($h_res{type} eq "mRNA" || $h_res{type} eq "ncRNA")
		{
			$h_parent{$h_res{ID}} = $line;
		}
		elsif ($ALLOWED_FEATURES =~ /$h_res{type}/)
		{
			my @a_parents = split(/,/, $h_res{Parent});
			my $cpt = 1;
			# Pour chaque parent, je sauvegarde cette ligne dans le hash
			# avant cela, je met a jour le champ Parent= et j'ajoute un increment 
            # à lidentifiant pour m'assurer que les ID sont uniques
			foreach my $p (@a_parents)
			{
				my $okline = $line;
				chomp($okline);
				$okline =~ s/Parent=[^;]+/Parent=$p/;
				$okline =~ s/ID=([^;]+)/ID=$1\.$cpt/;
				$h_parent{$p} .= $okline."\n";
				$cpt++;
			}
		}
	}
	
	foreach my $p (keys %h_parent)
	{
		&PrintIfOk($h_parent{$p}, $f_out_mrna, $f_out_ncrna,1, $parent_feature);
	}
		
	return;
	
}



sub ParseEuGeneFile
{
	my ($f_in, $f_out_mrna, $f_out_ncrna, $parent_feature) = @_;
	
	my %h_res;
	my $match_lines = "";
	
	while (my $line = <$f_in>)
	{
	   	next if ($line =~ /^#/);
		
		&GeneralBioinfo::ParseGFF3Line(\$line, 1, \%h_res); #seqid, source, type, start, end, len, score, strand, phase, attributes
		if ($h_res{type} eq "mRNA" || $h_res{type} eq "ncRNA")
		{
			&PrintIfOk($match_lines, $f_out_mrna, $f_out_ncrna, 0, $parent_feature);
			$match_lines = $line;
		}
		elsif ($ALLOWED_FEATURES =~ /$h_res{type}/)
		{
			$match_lines .= $line;
		}
	}
	&PrintIfOk($match_lines, $f_out_mrna, $f_out_ncrna, 0, $parent_feature);

	return;
}



sub PrintIfOk
{
	my ($lines, $f_mrna, $f_ncrna, $sort, $parent_feature) = @_;

	return if ($lines eq "");

	print "In PrintIfOK\n" if ($DEBUG);

	&SortByPos(\$lines) if ($sort);

	# Cas d un mRNA
	if ($lines =~ /\tmRNA\t/ && $lines =~ /\tfive_prime_UTR\t/ && $lines =~ /\tthree_prime_UTR\t/)
	{
		&PrintExonFeature($lines, "mRNA", $f_mrna, $parent_feature);
	}
	# Cas d'un ncRNA
	elsif ($lines =~ /\tncRNA\t/)
	{
		# Cas d'un ncRNA contenant des introns
		if ($lines =~ /\texon\t/)
		{
			&PrintExonFeature($lines, "ncRNA", $f_ncrna, $parent_feature);
		}
		else
		{
			my @a_lines = split(/\n/, $lines);
		    foreach my $line (@a_lines)
			{
				if ($line =~ /\tncRNA\t/)
				{
					my $match_part_line = $line;
					$line =~ s/\tncRNA\t/\t$parent_feature\t/;
					$line =~ s/;Parent=[^;]+//;
					my ($parent) = $line =~ /ID=([^;]+);?/;
					$match_part_line =~ s/\tncRNA\t/\tmatch_part\t/;
					$match_part_line =~ s/Parent=[^;]+/Parent=$parent/;
					$match_part_line =~ s/ID=([^;]+)/ID=$1.ncrna0/;
					print $f_ncrna "##\n" if ($COMMENT);
					print $f_ncrna "$line\n$match_part_line\n";
					print $f_ncrna "##\n" if ($COMMENT);				
				}
			}
		}
	}
	
	return;
}


sub SortByPos
{
	my ($r_lines) = @_;

	if ($DEBUG)
	{
		print ">>AVANT>>\n";
		print "$$r_lines\n";
	}
	
	my @a_lines = split(/\n/, $$r_lines);
	my $first = shift(@a_lines);
	my @a_sort = sort gff3sort @a_lines;
	$$r_lines = $first."\n".join ("\n", @a_sort);

	if ($DEBUG)
	{
		print "<<APRES<<\n";
		print $$r_lines;
		print "########\n";
	}
	
	return;
}


sub gff3sort
{
	my %A; 
	my %B; 
	&GeneralBioinfo::ParseGFF3Line(\$a, 1, \%A);
	&GeneralBioinfo::ParseGFF3Line(\$b, 1, \%B);
	
   
	if ($A{seqid} eq $B{seqid}) # If same chr
	{
		if ($A{start} == $B{start}) # same genomic start position
		{
			return $A{end} <=> $B{end}; # sort by end
		}
		else
		{
			return $A{start} <=> $B{start}; # sort by start
		}
	}
	else
	{
		return $A{seqid} cmp $B{seqid}; # sort by protein name 
	}

	return;	
}


sub PrintExonFeature
{
	my ($lines, $parent, $f_out, $parent_feature) = @_;

	print $f_out "##\n" if ($COMMENT);
	my @a_lines  = split(/\n/, $lines);
	foreach my $line (@a_lines)
	{
		if ($line =~ /\t$parent\t/)
		{
			$line =~ s/\t$parent\t/\t$parent_feature\t/;
			$line =~ s/;Parent=[^;]+//;
			print $f_out "$line\n";
		}
		elsif ($line =~ /\texon\t/)
		{
			$line =~ s/\texon\t/\tmatch_part\t/;
			print $f_out "$line\n";
		}
	}
	print $f_out "##\n" if ($COMMENT);
	
	return;
}

		
=item Usage

	$o_param->Usage();
	Print the usage of the program

=cut

sub Usage
{
    print STDERR<<END
$0     
Convert a valid GFF3 structural annotation (gene/mRNA/ncRNA/exon) in EST_match/match_part. 
Keep only the genes with 5'UTR and 3'UTR  and ncRNA genes.

MANDATORY

 --infile        GFF3         Structural Annotation (gene/mRNA/ncRNA ..)
 --outprefix     file prefix  Prefix of the two result files: outprefix.mRNA.gff3 and outprefix.ncRNA.gff3.

OPTIONAL
		--eugene                   If the infile is generated by EuGene. It avoids to check the presence of variants/to check that subfeatures are sorted, etc.
		--parent_feature    [Default=EST_match]
  --help this message
END
;

}
