#
#       Sebastien.Letort@toulouse.inra.fr
#       Created: October 08, 2007
#       $Id: Appli.pm 363 2008-07-01 09:48:58Z sletort $
#

#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Sebastien.Carrere@toulouse.inra.fr
# Erika.Sallet@toulouse.inra.fr
# Ludovic.Cottret@toulouse.inra.fr
# Ludovic.Legrand@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package Appli;

=pod

=head1 NAME

	Appli, a class to describe application

=head1 QUICKSTART

	#A small program to copy files
	use ParamParser;
	use Appli;
	
	my $o_appli = &SetAppli();
	my $o_param = New ParamParser('APPLI', $o_appli);
	$o_param->AssertFileExists('SOURCE');
	$o_param->AssertDefined('DEST');
	my ($source_file,$dest_file)   = $o_param->Get('SOURCE','DEST');
	my $force_option  = ($o_param->IsDefined('force')) ? ' --force' : "";
	my $suffix_option = "--suffix=" . $o_param->Get('suffix');
	system("cp $suffix_option $force_option $source_file $dest_file");
	
	sub SetAppli
	{
	    my %h_inputs = ('SOURCE' => {descr => 'source file', type => 'String'});
	    my %h_outputs = ('DEST' => {descr => 'destination file',type  => 'String'});
	    my %h_parameters = (
	        'force' => {
	            descr => 'if an existing destination file cannot be opened, remove it and try again ',
	            type  => 'Boolean'
	          },
	        'suffix' => {
				descr => 'override the usual backup suffix', 
				type => 'String', 
				default => 'backup'}
	        );
	    my $o_appli = New Appli(-inputs => \%h_inputs, -outputs => \%h_outputs, -params => \%h_parameters);
	    return $o_appli;
	}

=head1 SYNOPSIS

	my %h_general = (
			name      =>	'NarcisseGetAnnotationsFromLocus_tags', # the only one mandatory
			descr     =>	"Given a locus tag or a file containing a list of loci tags, it will return the annotation of the sequence.",
			authors   =>	'sebastien.letort@toulouse.inra.fr',
			version   =>	'0.1',
			cmd       =>	abs_path( $0 ) # default value is the prog name
			category  =>	'Usefull prog.'
			reference =>	[],
			use_biomoby_ontology => true or false (default is false to speed up usage display)
		);

	# ouputs
	my %h_outputs = (
			'Annotation' => {
				label       =>  "optional pretty parameter name (used by Galaxy)",
				descr       =>	"a list of ids with their annotation.",
				namespace_biomoby   =>	'Narcisse',
				type_biomoby=>	'text-formatted',
				type_galaxy=>	'tabular',
				cmd         =>	'--outfile=$value',
				card        =>	'1,n',
				filenames   =>	'*.out'
			}
		);

	# inputs
	my %h_inputs = (
			'NarcisseLocus_tags_list'	=>	{
				label       =>  "optional pretty parameter name (used by Galaxy)",
				descr       =>	"a locus tag or a file containing a list of locus tags.",
				namespace_biomoby   =>	'Narcisse',
				type_biomoby=>	'List_Text',
				type_galaxy=>	'text',
				cmd         =>	'--locus=$value',
				card        =>	'1,1'
			}
		);

	# parameters
	my %h_parameters = (
			'db'	=>	{
				type    =>	'Choice',
				type_biomoby=>	'String',
				enum        =>	{
									'proteic' => 'proteic search',
									'nucleic' => 'nucleic search'
								},
				default     =>	'proteic',
				descr       =>	"which DB you want to search into.",
				label       =>  "optional pretty parameter name (used by Galaxy)"
			}
		);

	my $appli = New Appli( -general => \%h_general );
	$appli->AddOutput( %h_output );
	$appli->SetParameters( %h_parameters );
	my $usage = $appli->Usage();
	print $usage;
	$appli = undef;

#	-----------------------------------

	$appli = New Appli(
			-general	=> \%h_general,
			-inputs 	=> \%h_inputs,
			-outputs	=> \%h_output,
			-params 	=> \%h_parameters
		);


=head1 DESCRIPTION

	This class aims to describe programs simply.
	It can generate a mobyle formatted file describing the application.

	I will describe here the keys of hashes i use into the methods.
	key marked with * are optional, default value is between parenthesis if not computed.

	'name' : string - should be only one line, name of the thing.
		the name in general hash, will also be used as title of the mobyle file.
	'descr'* : string - description of the thing. ('no description.')
	'cmd'* : string -
		in general = the command that launch the program default is the program name.
		in input/output and params = see playmoby notes.

	what's uniquely into general hash :
	'authors' : string - authors names coma-separated
	'version'* : string - version of the prog ('0.0')
	'doclink'* : ref onto a list of strings - each string is an url of online documentation
	'reference'* : ref onto a list of strings - the bibliographics references
	'category'* : string - to classified the programs in group (ex. Service, DNA analysis, phylogeny)
		(Service)
		cf playmoby notes
	'use_biomoby_ontoloy': boolean - use ontoloy server to find best mapping between biomoby datatypes and mobyle ones (cf. Appli::OntologyMap)

	what's common to outputs, inputs and params hashs
	'type' : string - type of the data, Sequence , Structure , Matrix , Choice
		please refer to the latest mobyle's dtd. As I writing, there's no more explanation.
		but using int, bool, float, string seems acceptable.
	'type_biomoby'* : string - this is a biomoby type
		please consider if you're describing a parameter (secondary article),
		an input or an output (primary article) and have a look to
		http://lipm-bioinfo.toulouse.inra.fr/registry/cgi/registry.cgi?registry=mobycentral&form=choice&what=objects
		if you use type int, float bool string, the biomoby type can generally be omitted.
	'type_galaxy'* : string - this is a galaxy data type
	'mandatory'* : 0 or 1
		by default, it's 1 for inputs and outputs

	what is uniquely into input/output hash :
	'card': cardinality '1,1' or '1,n'
		by default it's '1,1' (single input or output)
		'1,n' means 1 or many input/output from the same type (homogene collection in BioMoby world)
	'filenames': a string describing the name of input/ouput files
		this is very usefules when your program produces many files (number unknown): in that case
		you could produce this files into a directory with the same extension ".foo" and use
		the filenames attribute "*.foo" to use Appli

	what uniquely into params hash :
	'enum'* : a hash ref or a href (hypertext link) to a vlist XML file.
		keys are value of the param, and values are labels.
	'default'* : any type, the default value for the parameter
	'position'* : not yet implemented
	'min'* : numeric, the minimum value for the param.
	'max'* : numeric, the maximum value for the param.

=head2 Notes for playmoby user

	'cmd' will be evaluated _don't be mean please.
		Because sometime cmd are complicated like
		"($value != 0) ? '-bad joker' : 'batman'"
		it is important to us to have a variable that will contain the value of
			the parameter when the service will be executed.
		As mobyle responsables use $value in their code, we choose to use the same.
		So your command should have the string '$value' inside.
		Here are some common example of cmd string for parameters.
		*with GetOptLong* => '--param=$value'
			=> Note that $value will not be evaluated here.
		*if the param is just a value in the command string* => '$value'
		*if your program wrap another complicated one you could have something like*
		 => 'if(0==$value){ return "zero"; }elsif(5==$value){ return "anything"; }else{ return $value; }'
	'category' we use this as the ServiceType Biomoby.
		You can find a list of possible values here : 		http://lipm-bioinfo.toulouse.inra.fr/registry/cgi/registry.cgi?registry=mobycentral&form=choice&what=servicetypes&extended=1

	'namespace_biomoby' is another key of the input/output hashes. It is the biomoby namespace.
		you can have a look of the possible vlues here :
		http://lipm-bioinfo.toulouse.inra.fr/registry/cgi/registry.cgi?registry=mobycentral&form=choice&what=namespaces

=head2 Notes for developpers

 Appli->{
	name => 'nemo',	# prog name
	descr => 'no description.', # string that describe the prog
	cmd => undef, # name of the command
	authors => 'no author', # authors names coma-separated
	version => '0.0', # version of the prog
	category => 'Service', # category of the prog
	reference=> [], # prog. references
	doclink => [] # urls of online doc.
		# no default value set yet : TODO

	inputs => {
		'in_name1' => {
			type => '',
			type_biomoby => '',
			descr => '',
			mandatory => 0/1,
			cmd => '',
			namespace_biomoby => '',
			card => '',
			filenames => ''
		},
		'another_input_name' => { ... } # there can be many inputs
	}, # inputs key

	outputs => {
		'out_name1' => {
			...	# this is exactely the same keys as inputs
		},
		'another_outputs_name' => { ... } # there can be many outputs
	}, # outputs key

	params => {
		'par_name1' => {
			type => '',
			type_biomoby => '',
			descr => '',
			mandatory => 0/1,
			cmd => '',
			default => '',
			position => i,	# integer
			min => num, # any numerical value
			max => num,
			enum => { 'value1' => 'label for this value', num => 'another label' } | 'http://mobyle-dev.toulouse.inra.fr/core/Local/Services/Programs/Entities/nucdbs.xml'
		},
		'another_param_name' => { ... } # there can be many params
	} # params key
} # Appli

=head1 TODO

	It would be greater to separate mobyle/playmoby output functions
	to make it easier to maintain.
	
	Integrate EDAM Ontology

=head1 SUBROUTINES

=cut

use base qw( LipmObject );
use strict;
use warnings;

use XML::Twig;
use IO::File;
use JSON;

use Data::Dumper;

## de lipmutils
use LipmError::ParameterException;
use LipmError::IOException;

# use Enum;
use General;    #IsNotNull SetUnlessDefined

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 363 $ =~ /\d+/g); $r[0]};
    my $abs_path_module = $INC{'Appli.pm'};
    my ($dir_module) = ($abs_path_module =~ /(.+)\/.+/);
    unshift(@INC, $dir_module . '/Appli');
}

# fermeture
{
    my @__a_biomoby_primitives = qw( String Boolean Integer Float DateTime );

    sub GetTypeBiomoby
    {
        my ($type) = @_;

        # /!\ ATTENTION datetime renverra Datetime qui n'est pas DateTime /!\ #
        foreach my $biomoby_primitive (@__a_biomoby_primitives)
        {
            return $biomoby_primitive if ($biomoby_primitive =~ /^$type/i);
        }

        #return ucfirst($type) if (map(/^$type$/i, @__a_biomoby_primitives));
        return 'Object';
    }

    sub GetTypeMobyle
    {
        my ($type, $central_alive) = @_;

        eval 'use OntologyMap';
        return 0 if $@;

        return OntologyMap::BiomobyToMobyle($type, $central_alive);

    }

    # keys attr common to inputs, outputs and params
    my @__a_attrs = qw( descr mandatory type type_biomoby type_galaxy cmd label );

    # keys attr common to inputs and outputs
    my @__a_xputs_attrs = (@__a_attrs, qw( namespace_biomoby card filenames ));

    sub GetXputsKeys {return @__a_xputs_attrs;}

    # keys attr for params
    my @__a_params_attrs = (@__a_attrs, qw( default position enum min max ));

    sub GetParamsKeys {return @__a_params_attrs;}
}    #fermeture

=head2 Function New

	Title      :	New
	Usage      :	my $o_appli = New Appli( -general => \%h_gen,
	            							-inputs => \%h_in,
	            							-outputs => \%h_out,
	            							-params => \%h_params
	            						);
	Prerequiste:	none
	constructor:	an Appli object should be used to described simply application
	Args       :	a hash with at least the -general key with a ref to another hash as value
	            		this hash have descr, author, and version for keys
	            		cf SetOutputs, SetInputs and Set param for others
	Error      :    LipmError::ParameterException
	Globals    :    none

=cut

sub _Init
{
    my $self = shift;
    my %h_attrs = (
        -general => {},
        -inputs  => {},
        -outputs => {},
        -params  => {},

        #-central_alive => &__IsCentralAlive(),
        -central_alive => &FALSE,
        @_
        );

    # general data
    my $rh_general = $h_attrs{-general};

    use Cwd;
    my $path_bin = Cwd::abs_path($0);
    my ($dir, $file) = ($path_bin =~ /(.+)\/(.+)/);

    $rh_general->{version} = $main::VERSION if ((!defined($rh_general->{version})) && defined($main::VERSION));

    # the default attributes values
    my %h_default_general = (
                             name                 => $file,        # prog name
                             descr                => $file,        # string that describe the prog
                             #SC: on n'ecrit pas le full path du programme (pas portable)
                             #cmd                  => $path_bin,    # name of the command
                             cmd                  => $file,    # name of the command
                             authors              => 'unknown',    # authors names coma-separated
                             version              => 'unknown',    # version of the prog
                             category             => 'Service',    # category of the prog
                             reference            => [],           # prog. references
                             doclink              => [],           # urls of online doc.
                             use_biomoby_ontology => &FALSE
                             );

    # I want only keys of %h_def_general as attributes
    foreach my $key (keys %h_default_general)
    {
        my $user_val    = $$rh_general{$key};
        my $default_val = $h_default_general{$key};
        $self->{$key} = (defined($user_val)) ? $user_val : $default_val;
    }

    if ($self->{'use_biomoby_ontology'} == &TRUE)
    {
        $h_attrs{-central_alive} = &__IsCentralAlive();
    }

    # other affectation of attributes
    # set $self->{ inputs }
    $self->__Set('inputs', $h_attrs{-inputs}, &TRUE);

    # set $self->{ outputs }
    $self->__Set('outputs', $h_attrs{-outputs}, &TRUE);

    # set $self->{ params }
    $self->__Set('params', $h_attrs{-params}, &TRUE);
    

    return $self;
}

=head2 Function __IsCentralAlive

	Title      :	__IsCentralAlive
	Usage      :	$bool = __IsCentralAlive()
	Prerequiste:	none
	Procedure  :	ping against Biomoby central url http://moby.toulouse.inra.fr
	                with a timeout of 3 seconds
	Args       :	NONE
	Returns    :    &TRUE or &FALSE
	Globals    :    none

=cut

sub __IsCentralAlive
{

    my $url     = 'http://moby.toulouse.inra.fr';
    my $timeout = 3;
    use LWP::Simple qw($ua);
    $LWP::Simple::ua->timeout($timeout);
    my $cgi_ok = LWP::Simple::get($url);
    my $alive  = &FALSE;
    $alive = &TRUE if (defined $cgi_ok);
    return $alive;
}

=head2 Procedure __Set

	Title      :	__Set
	Usage      :	$o_appli->__Set( $key, $r_hash, $do_erase );
	Prerequiste:	none
	Procedure  :	set the outputs/inputs/params for the current application
	Args       :	$key, a string which should be 'outputs', 'inputs' or 'params'
	            	$r_hash, a ref to an hash
	            		which keys are the name of the in/out/params
	            		and values are ref to hash
	            				which keys should be in/out attrs or params attrs
	            	$do_erase, a boolean, if true will erase previous data.
	Error      :    LipmError::ParameterException
	Globals    :    none

	Note       :	you can have more than one object but in that case, the type should be Collection

=cut

sub __Set
{
    my $self = shift;
    my ($key, $rh_attr, $do_erase) = @_;

    $self->{$key} = {} if ($do_erase);

    # mandatory attributes
    my @a_attrs = ();
    @a_attrs = GetXputsKeys()  if ($key eq 'inputs');
    @a_attrs = GetXputsKeys()  if ($key eq 'outputs');
    @a_attrs = GetParamsKeys() if ($key eq 'params');

    #ENUM => HREF VERS VLIST XML MOBYLE ?
    while (my ($name, $rh_data) = each(%$rh_attr))
    {

        Carp::croak(
                  "ERROR: >$name< invalid parameter name - only alphanumeric characters are allowed in parameter names")
          if ($name =~ /\W/);

        # all the $$rh_data{ $key } will be copied into $h_params{ $key }
        # where $key is an element of the @a_attrs list.
        my %h_params;
        @h_params{@a_attrs} = @$rh_data{@a_attrs};

        foreach my $key (keys %{$rh_data})
        {
            Carp::croak("ERROR: >$key< invalid parameter name (valid:@a_attrs)") unless (grep /^$key$/, @a_attrs);
        }

        # 1- mandatories
        if (!defined($$rh_data{type_biomoby}) && !defined($$rh_data{type}))
        {
            $h_params{type} = 'Boolean';
            $$rh_data{type} = $h_params{type};
        }

        if ((!defined $$rh_data{type_biomoby}) && (defined($$rh_data{type})))
        {

            $h_params{type_biomoby} = ucfirst($$rh_data{type});
            $$rh_data{type_biomoby} = $h_params{type_biomoby};
        }

        if ((!defined $$rh_data{type}) && (defined $$rh_data{type_biomoby}))
        {

            my $alternate_mobyle_type = GetTypeMobyle($$rh_data{type_biomoby}, $self->{central_alive});
            if ($alternate_mobyle_type != 0)
            {
                $$rh_data{type} = $alternate_mobyle_type;
                SetUnlessDefined(\$h_params{type}, $alternate_mobyle_type);
            }
        }

        #        if (defined $$rh_data{enum} && $$rh_data{enum} =~ /^http:\/\//)
        #        {
        #				$h_params{enum} = &__VlistXmlURLToHashRef( $$rh_data{enum} );
        #		}

        LipmError::ParameterException::Assert(
            'name' => $name    #,
                               #'datatype' => $$rh_data{ datatype }
              );

        #all

        my $alternate_biomoby_type = GetTypeBiomoby($$rh_data{type}) if (!defined $$rh_data{type_biomoby});
        SetUnlessDefined(\$h_params{type_biomoby}, $alternate_biomoby_type);

        if ($h_params{type_biomoby} eq 'Boolean')
        {
            SetUnlessDefined(\$h_params{cmd}, " --$name");
        }
        else
        {

            #ajout " pour les listes de fichiers (collection)

            SetUnlessDefined(\$h_params{cmd}, " --$name=\$value");
        }

        if ($key eq 'inputs' or $key eq 'outputs')
        {
            SetUnlessDefined(\$h_params{card},      '1,1');
            SetUnlessDefined(\$h_params{mandatory}, '1');     # utile ou futile ?

            # namespace remplace par namespace_biomoby
            #		SetUnlessDefined( \$h_params{ namespace }, $self->{ def_namespace } );
            # verif de filenames si key == outputs ?
            # => besoin d'avoir une valeur par defaut
        }
        $self->{$key}{$name} = \%h_params;
    }

    return;
}    # __Set

=head2 Procedure SetInputs

	Title      :	SetInputs
	Usage      :	$o_appli->SetInputs( %hash );
	Prerequiste:	none
	Procedure  :	this is an encapsulation of the __Set private methods
	Args       :	%hash cf __Set
	Error      :    none
	Globals    :    none

	Note       :	you can have more than one object but in that case, the type should be Collection

=cut

sub SetInputs
{
    my $self = shift;
    my (%h_inputs) = @_;

    return $self->__Set('inputs', \%h_inputs);
}

=head2 Procedure SetOutputs

	Title      :	SetOutputs
	Usage      :	$o_appli->SetOutputs( %hash );
	Prerequiste:	none
	Procedure  :	this is an encapsulation of the __Set private methods
	Args       :	%hash cf __Set
	Error      :    none
	Globals    :    none

	Note       :	you can have more than one object but in that case, the type should be Collection

=cut

sub SetOutputs
{
    my $self = shift;
    my (%h_outputs) = @_;

    return $self->__Set('outputs', \%h_outputs);
}

=head2 Procedure SetParams

	Title      :	SetParams
	Usage      :	$o_appli->SetParams( %hash );
	Prerequiste:	none
	Procedure  :	Set params to the current application.
	Args       :	a hash which keys are the name of the parameters
	            	values are ref to hash like
	            		type	=>	'string',
	            		mandatory	=>	0,	a boolean
	            		descr   	=>	"part of the long name or begining of the short name."
	            		position	=>	1,	an integer
	            		cmd     	=>	"--string="
	            		default 	=>	''
	            		type_biomoby=>	'String'
	Error      :    Error::ParameterException
	Globals    :    none

=cut

sub SetParams
{
    my $self = shift;
    my (%h_params) = @_;

    return $self->__Set('params', \%h_params);
}

=head2 Procedure AddParam

	Title      : AddParam
	Usage      :	$o_appli->AddParam( %hash );
	Prerequiste:	none
	Procedure  : alias of SetParams
	Args       :	a hash which keys are
	            		name	=> "ppp",
	            		type    	=>	'string',
	            		mandatory	=>	0,	a boolean
	            		descr   	=>	"part of the long name or begining of the short name."
	            		position	=>	1,	an integer
	            		cmd     	=>	"--string="
	            		default 	=>	''
	            		type_biomoby=>	'String'
	            		min     	=> 1,	an integer
	            		max     	=> 100, an integer
	Error      :    Error::ParameterException
	Globals    :    none

=cut

sub AddParam
{
    my $self   = shift;
    my (%hash) = @_;
    my $name   = delete $hash{name};

    return $self->SetParams($name => \%hash);
}

=head2 Procedure AddOutput

	Title      :	AddOutput
	Usage      :	$o_appli->AddOutput( %hash);
	Prerequiste:	none
	Procedure  :	Add one output to the application.
	Args       :	a hash with the following keys
	            	name descr namespace_biomoby type_biomoby mandatory
			type (Mobyle) is taken from dictionnary via OntologyMap.pm
	Error      :    none
	Globals    :    none

=cut

sub AddOutput
{
    my $self = shift;
    my (%hash) = @_;

    my $name = delete $hash{name};

    return $self->SetOutputs($name => \%hash);
}

=head2 Procedure AddInput

	Title      :	AddInput
	Usage      :	$o_appli->AddInput( %hash);
	Prerequiste:	exactly the same as AddOutput
	Procedure  :	Add one input to the application.
	Args       :	a hash with the following keys
	            	name descr namespace_biomoby type_biomoby mandatory card
			type (Mobyle) is taken from dictionnary via OntologyMap.pm
	Error      :    none
	Globals    :    none

=cut

sub AddInput
{
    my $self = shift;
    my (%hash) = @_;

    my $name = delete $hash{name};

    return $self->SetInputs($name => \%hash);
}

=head2 Procedure AddDoclink

	Title      :	AddDoclink
	Usage      :	$o_appli->AddDoclink( @a_urls );
	Prerequiste:	none
	Procedure  :	Add all strings of @a_urls as mobyle doclink.
	Args       :	@a_urls, a list of string representing the urls of the online doc.
	Error      :    none

=cut

sub AddDoclink
{
    my $self = shift;
    my (@a_urls) = @_;
    

    my $ra_doclink = $self->{doclink};

    push(@$ra_doclink, @a_urls);

    return;
}

=head2 Function GetUsage

	Title      :	GetUsage
	Usage      :	my $page = $o_appli->GetUsage();
	Prerequiste:	none
	Function   :	build a standardized usage string which explain how to use the application
	Returns    :	a string
	Args       :	none
	Error      :    none
	Globals    :    none

=cut

sub GetUsage
{
    my $self = shift;
    my $page;
    $page .= "Name:\t\t" . $self->{name} . "\n";
    $page .= "Program:\t" . $self->{cmd} . " (v." . $self->{version} . ') by '. $self->{authors} ."\n";
    $page .= "Description:\t" . $self->{descr} . "\n";
    $page .= "Documentation:\t" . join(', ', @{$self->{doclink}}) . "\n" if (scalar @{$self->{doclink}} > 0);

    my @a_inout = ('inputs', 'outputs');
    foreach my $interface (@a_inout)
    {
        $page .= "\n" . ucfirst($interface) . ":\n";
        my $rh_inputs = $self->{$interface};
        while (my ($name, $rh_attr) = each(%$rh_inputs))
        {
            my $cmd     = $rh_attr->{cmd};
            my $biomoby = $rh_attr->{type_biomoby};

            my $collection = '';
            $collection = 'Collection of ' if ($rh_attr->{card} =~ /n/i);
            $cmd =~ s/\$value/<$collection$biomoby>/;
            my $default = '';
            $default = '[' . $rh_attr->{filenames} . "]\t" if (defined $rh_attr->{filenames});
            $page .= "\t" . $cmd . "\t" . $default . "$name: " . $rh_attr->{descr} . "\n";
        }
    }

    if (defined $self->{params} && ref($self->{params}) =~ /hash/i && length(keys(%{$self->{params}}) > 0))
    {
        $page .= "\nOptionnal:\n";
        my $rh_params = $self->{params};
        while (my ($name, $rh_attr) = each(%$rh_params))
        {
            my $cmd     = $rh_attr->{cmd};
            my $biomoby = $rh_attr->{type_biomoby};
            my $default = '';
            $default = $rh_attr->{default} if (defined($rh_attr->{default}));
            $default = 'false' if ($biomoby eq 'Boolean' && $default eq '');

            $cmd =~ s/\$value/<$biomoby>/;
            my $min = '';
            $min = " (Min:" . $rh_attr->{min} . ")" if (defined $rh_attr->{min});
            my $max = '';
            $max = " (Max:" . $rh_attr->{max} . ")" if (defined $rh_attr->{max});
            
            $page .= "\t" . $cmd . "\t" . "[" . $default . "]$min$max" . "\t" . "$name: " . $rh_attr->{descr} . "\n";
            if (defined $rh_attr->{enum})
            {
                my $rh_enum = {};
                if ($rh_attr->{enum} =~ /^http:\/\//)
                {
                    $rh_enum = &__VlistXmlURLToHashRef($rh_attr->{enum});

                }
                else
                {
                    $rh_enum = $rh_attr->{enum};
                }
                $page .= "\t\tValues:\n";
                while (my ($enum_key, $enum_value) = (each(%{$rh_enum})))
                {
                    $page .= "\t\t\t$enum_key : $enum_value \n";
                }
            }
            

        }
    }
    $page .= "\nAppli related:\n";
    $page .= "\t" . '--help' . "\t\tPrint this usage\n";
    #$page .= "\t" . '--mobyle' . "\t\tPrint Mobyle XML data on STDOUT (Appli::GetMobyleXml must be implemented)\n";
    #$page .= "\t" . '--pmbtest' . "\t\tPrint PlayMOBY test XML data on STDOUT (Appli::GetPlaymobyTestXml must be implemented)\n";
    #$page .= "\t" . '--mobyle_dir'  . "=<output directory>\tRedirect STDOUT for the --mobyle and --pmbtest parameters to <output_directory>/" . $self->{name} . ".xml|.pmb.t\n";
    $page .= "\t" . '--galaxy' . "\t\tPrint Galaxy XML data on STDOUT (Appli::GetGalaxyXml must be implemented)\n";
    $page .= "\t" . '--htmldesc' . "\t\tPrint HTML formatted help on STDOUT (Appli::GetHtmlDesc must be implemented)\n";
    $page .= "\t" . '--jsondesc' . "\t\tPrint JSON formatted help on STDOUT (Appli::GetJSONDesc must be implemented)\n";
    #$page .= "\t" . '--pipelinesubs' . "\t\tPrint a perl string with all necessary sub routines for Pipelines CGI interface STDOUT (Appli::GetPipelinesPerlSubs must be implemented)\n";

    

    return $page;
}

=head2 Function GetPipelinesPerlSubs

	Title      :	GetPipelinesPerlSubs
	Usage      :	my $page = $o_appli->GetPipelinesPerlSubs();
	Prerequiste:	none
	Function   :	build a perl string with all necessary sub routines for Pipelines CGI interface 
	Returns    :	a string
	Args       :	none
	Error      :    none
	Globals    :    none

=cut

sub GetPipelinesPerlSubs
{
    my $self = shift;

    my $page = '
sub PipelineDescription
{
    my $o_webbuilder = shift;
    my $o_param = $o_webbuilder->GetObjectParam;
    my $pipeline_id = $o_param->Get(\'pipeline_id\');
    my $pipeline_bin =  $o_param->Get(\'pipeline_\' . $pipeline_id . \'_bin\');
    my $cde_pipeline_description = $pipeline_bin . \' --htmldesc \';
    my $pipeline_description =  `$cde_pipeline_description`;
    print $pipeline_description;
    return;
    
}
';

    my @a_inout = ('inputs', 'outputs');
    foreach my $interface (@a_inout)
    {
        $page .= 'sub Pipeline' . ucfirst($interface) . "\n";
        $page .= '{' . "\n";

        $page .= "\t" . 'my %h_files = (' . "\n";

        my $rh_inputs = $self->{$interface};
        while (my ($name, $rh_attr) = each(%$rh_inputs))
        {
            my $type     = $rh_attr->{type};
            my $biomoby  = $rh_attr->{type_biomoby};
            my $desc     = $rh_attr->{descr};
            my $filename = $name . '.input';
            $filename = $name . '.' . $type    if (defined $type                 && ($type                 ne ''));
            $filename = $name . '.' . $biomoby if (defined $biomoby              && ($biomoby              ne ''));
            $filename = $rh_attr->{filenames}  if (defined $rh_attr->{filenames} && ($rh_attr->{filenames} ne ''));
            $page .= "\t" x 4 . "'$name' => { file => '$filename' , desc => '$desc', mandatory => 1} , " . "\n";
        }
        $page =~ s/,\n$/\n/;
        $page .= "\t" x 4 . ');' . "\n";
        $page .= "\t" . 'return \%h_files;' . "\n";
        $page .= '}' . "\n\n";

    }

    $page .= 'sub PipelineParameters' . "\n";
    $page .= '{' . "\n";
    $page .= "\t" . 'my %h_parameters = (' . "\n";
    if (defined $self->{params} && ref($self->{params}) =~ /hash/i && length(keys(%{$self->{params}}) > 0))
    {
        my $rh_params = $self->{params};
        while (my ($name, $rh_attr) = each(%$rh_params))
        {

            my $biomoby = $rh_attr->{type_biomoby};
            my $default = '';
            $default = $rh_attr->{default} if (defined($rh_attr->{default}));
            $default = 'false' if ($biomoby eq 'Boolean' && $default eq '');
            my $desc = $rh_attr->{descr};

            my $enum = '{}';
            if (defined $rh_attr->{enum})
            {
                my $rh_enum = {};
                if ($rh_attr->{enum} =~ /^http:\/\//)
                {
                    $rh_enum = &__VlistXmlURLToHashRef($rh_attr->{enum});

                }
                else
                {
                    $rh_enum = $rh_attr->{enum};
                }

                $enum = '{';
                while (my ($enum_key, $enum_value) = (each(%{$rh_enum})))
                {
                    $enum .= "'$enum_key' => '$enum_value',";
                }
                $enum =~ s/,$//;
                $enum .= '}';
            }
            $page .= "\t" x 4 . "'$name' => { default => '$default' , desc => '$desc', enum => $enum} ," . "\n";

        }
    }

    $page =~ s/,\n$/\n/;

    $page .= "\t" x 4 . ');' . "\n";
    $page .= "\t" . 'return \%h_parameters;' . "\n";
    $page .= '}' . "\n\n";

    return $page;

}

=head2 Function GetHtmlDesc

	Title      :	GetHtmlDesc
	Usage      :	my $page = $o_appli->GetHtmlDesc();
	Prerequiste:	none
	Function   :	build a HTML formatted table string which explain how to use the application
	Returns    :	a string
	Args       :	none
	Error      :    none
	Globals    :    none

=cut

sub GetHtmlDesc
{
    my $self = shift;
    my $page = "\n<div class='htmldesc'>\n<table>";
    $page .= "\n<tr><td class='htmldesc_item'>Program</td><td>" . $self->{cmd} . "</td></tr>";
    $page .= "\n<tr><td class='htmldesc_item'>Name</td><td>" . $self->{name} . "</td></tr>";
    $page .= "\n<tr><td class='htmldesc_item'>Description</td><td>" . $self->{descr} . "</td></tr>";
    $page .= "\n<tr><td class='htmldesc_item'>Authors</td><td>" . $self->{authors} . "</td></tr>";
    $page .= "\n<tr><td class='htmldesc_item'>Version</td><td>" . $self->{version} . "</td></tr>";
    $page .= "\ntr><td class='htmldesc_item'>Documentation</td><td>" . join(', ', @{$self->{doclink}}) . "</td></tr>"
      if (defined($self->{doclink}) && (scalar @{$self->{doclink}} > 0));

    my @a_inout = ('inputs', 'outputs');
    foreach my $interface (@a_inout)
    {
        $page .= "\n<tr><td class='htmldesc_item' colspan='2'>" . ucfirst($interface) . "</td></tr>";
        my $rh_inputs = $self->{$interface};
        while (my ($name, $rh_attr) = each(%$rh_inputs))
        {
            my $cmd     = $rh_attr->{cmd};
            my $biomoby = $rh_attr->{type_biomoby};

            my $collection = '';
            $collection = 'Collection of ' if ($rh_attr->{card} =~ /n/i);
            $cmd =~ s/\$value/<$collection$biomoby>/;
            my $default = '';
            $default = '[' . $rh_attr->{filenames} . "]\t" if (defined $rh_attr->{filenames});
            $page .= "\n<tr><td class='htmldesc_item'>$name</td><td>" . $rh_attr->{descr} . " $default </td></tr>";
        }
    }

    if (defined $self->{params} && ref($self->{params}) =~ /hash/i && length(keys(%{$self->{params}}) > 0))
    {
        $page .= "\n<tr><td class='htmldesc_item' colspan='2'>Parameters</td></tr>";
        my $rh_params = $self->{params};
        while (my ($name, $rh_attr) = each(%$rh_params))
        {
            my $cmd     = $rh_attr->{cmd};
            my $biomoby = $rh_attr->{type_biomoby};
            my $default = '';
            $default = '[' . $rh_attr->{default} . ']' if (defined($rh_attr->{default}));
            $default = 'false' if ($biomoby eq 'Boolean' && $default eq '');

            $cmd =~ s/\$value/<$biomoby>/;
            $page .= "\n<tr><td class='htmldesc_item'>$name</td><td>" . $rh_attr->{descr} . " $default";
            if (defined $rh_attr->{enum})
            {
                my $rh_enum = {};
                if ($rh_attr->{enum} =~ /^http:\/\//)
                {
                    $rh_enum = &__VlistXmlURLToHashRef($rh_attr->{enum});

                }
                else
                {
                    $rh_enum = $rh_attr->{enum};
                }
                $page .= "\n<br><i>Values:</i>";
                while (my ($enum_key, $enum_value) = (each(%{$rh_enum})))
                {
                    $page .= "\n<br />&nbsp;&nbsp;&nbsp;$enum_key : $enum_value";
                }
            }
            $page .= "\n<br /><i>Min</i>:" . $rh_attr->{min} if (defined $rh_attr->{min});
            $page .= "\n<br /><i>Max</i>:" . $rh_attr->{max} if (defined $rh_attr->{max});

            $page .= "</td></tr>\n";
        }
    }

    $page .= "</table>\n</div>\n";

    return $page;

}

sub GetJSONDesc
{
    my $self   = shift;
    my %h_desc = ();

    $h_desc{cmd}     = $self->{cmd};
    $h_desc{name}    = $self->{name};
    $h_desc{descr}   = $self->{descr};
    $h_desc{authors} = $self->{authors};
    $h_desc{version} = $self->{version};
    $h_desc{doclink} = $self->{doclink};
    my @a_inout = ('inputs', 'outputs');
    foreach my $interface (@a_inout)
    {
        $h_desc{$interface} = {};
        my $rh_inputs = $self->{$interface};

        while (my ($name, $rh_attr) = each(%$rh_inputs))
        {
            $h_desc{$interface}->{$name}                 = {};
            $h_desc{$interface}->{$name}->{cmd}          = $rh_attr->{cmd};
            $h_desc{$interface}->{$name}->{type_biomoby} = $rh_attr->{type_biomoby};
            $h_desc{$interface}->{$name}->{card}         = $rh_attr->{card};
            $h_desc{$interface}->{$name}->{filenames}    = $rh_attr->{filenames};
            $h_desc{$interface}->{$name}->{descr}        = $rh_attr->{descr};
        }
    }

    if (defined $self->{params} && ref($self->{params}) =~ /hash/i && length(keys(%{$self->{params}}) > 0))
    {
        $h_desc{params} = {};
        my $rh_params = $self->{params};
        while (my ($name, $rh_attr) = each(%$rh_params))
        {
            my $default = '';
            $default = $rh_attr->{default} if (defined($rh_attr->{default}));
            $default = 'false'
              if (defined $rh_attr->{type_biomoby} && $rh_attr->{type_biomoby} eq 'Boolean' && $default eq '');

            $h_desc{params}->{$name}                 = {};
            $h_desc{params}->{$name}->{cmd}          = $rh_attr->{cmd};
            $h_desc{params}->{$name}->{type_biomoby} = $rh_attr->{type_biomoby};
            $h_desc{params}->{$name}->{default}      = $rh_attr->{card};
            $h_desc{params}->{$name}->{descr}        = $rh_attr->{descr};

            if (defined $rh_attr->{enum})
            {
                my $rh_enum = {};
                if ($rh_attr->{enum} =~ /^http:\/\//)
                {
                    $rh_enum = &__VlistXmlURLToHashRef($rh_attr->{enum});
                }
                else
                {
                    $rh_enum = $rh_attr->{enum};
                }
                $h_desc{params}->{$name}->{enum} = $rh_enum;

            }
            $h_desc{params}->{$name}->{min} = $rh_attr->{min} if (defined $rh_attr->{min});
            $h_desc{params}->{$name}->{max} = $rh_attr->{max} if (defined $rh_attr->{max});

        }
    }

    return to_json(\%h_desc, {pretty => 1, ascii => 1});
}

=head2 Function GetOptLongArray

	Title      :	GetOptLongArray
	Usage      :	my @a_getopt_params = $o_appli->GetOptLongArray(  );
	Prerequiste:	none
	Function   :	build Getopt::Long parameters array from Appli structure
	Returns    :	an array (param=type)
	Args       :	none
	Error      :    none
	Globals    :    none

=cut

sub GetOptLongArray
{
    my $self                = shift;
    my %h_getoptlong_types  = ('Integer' => '=i', 'Boolean' => '', 'Float' => '=f');
    my @a_getoptlong_params = ();

    my @a_inout = ('inputs', 'outputs', 'params');
    foreach my $interface (@a_inout)
    {
        next if (!defined($self->{$interface}));
        my $rh_inputs = $self->{$interface};
        while (my ($name, $rh_attr) = each(%$rh_inputs))
        {
            my $biomoby = $rh_attr->{type_biomoby};
            my $str = (defined($h_getoptlong_types{$biomoby}) ? $h_getoptlong_types{$biomoby} : "=s");
            push(@a_getoptlong_params, $name . $str);
        }
    }

    push(@a_getoptlong_params, 'mobyle', 'pmbtest', 'galaxy', 'htmldesc', 'jsondesc', 'pipelinesubs', 'help');
    return @a_getoptlong_params;
}

sub GetDefaultValues
{
    my $self             = shift;
    my %h_default_values = ();
    return {} if (!defined($self->{params}));

    my $rh_params = $self->{params};
    while (my ($name, $rh_attr) = each(%$rh_params))
    {
        $h_default_values{$name} = $rh_attr->{default} if (defined($rh_attr->{default}));
    }
    return \%h_default_values;
}

=head2 Function GetGalaxyXml

	Title      :	GetGalaxyXml
	Usage      :	my $xml = $o_appli->GetGalaxyXml( $outdir );
	Prerequiste:	none
	Function   :	write an xml based on the Galaxy dtd which describe the application
					http://wiki.g2.bx.psu.edu/Admin/Tools/Tool%20Config%20Syntax
	Returns    :	undef if $outdir is defined and not empty
	            	a string containing the xml otherwise.
	Args       :	$outdir, facultative string, the output directory
	Error      :    none
	Globals    :    none

=cut

sub GetGalaxyXml
{
    my $self = shift;
    my ($outdir) = @_;

    eval 'use XML::LibXML';
    if ($@)
    {
        print "ERROR - XML::LibXML perl module not found; Please fix this problem before\n";
        exit -1;
    }

    my $description = $self->{'descr'};
    my $id          = $self->{'name'};
    my $version     = $self->{'version'};
    my $authors     = $self->{'authors'};
    my $cmd         = $self->{'cmd'};

    my %h_interface_xml_string = ();

    my @a_inout = ('inputs', 'outputs');
    foreach my $interface (@a_inout)
    {
        my $attribute = 'param';
        $attribute = 'data' if ($interface eq 'outputs');
        my $rh_inputs = $self->{$interface};
        while (my ($name, $rh_attr) = each(%$rh_inputs))
        {

            my $param_cmd = $rh_attr->{'cmd'};

            my $type_galaxy = 'text';
            $type_galaxy = $rh_attr->{'type_galaxy'} if (defined $rh_attr->{'type_galaxy'} && $rh_attr->{'type_galaxy'} ne '');
            
            my $label = ( defined $rh_attr->{'label'} && $rh_attr->{'label'} ne '' ) ? $rh_attr->{'label'} : $name;
            
            my $description = $rh_attr->{descr};
            $h_interface_xml_string{$interface} .=
              "<$attribute optional=\"false\" format=\"$type_galaxy\" name=\"$name\" label=\"$label\" type=\"data\" help=\"$description\"/>\n";

            if (($interface eq 'inputs') && $rh_attr->{'card'} =~ /n/i)
            {
                $param_cmd =~ s/\$value/"\$$name/g;

                $h_interface_xml_string{$interface} .=
                  "<repeat name=\"collection_$name\" title=\"Additional $description\"><param format=\"$type_galaxy\" name=\"additional_$name\" label=\"$label\" type=\"data\" help=\"$description\"/></repeat>";
                $param_cmd .= "\n"
                  . '## Inputs.' . "\n"
                  . '#for $input_file in $collection_'
                  . $name . ':' . "\n"
                  . ',${input_file.additional_'
                  . $name . '}' . "\n"
                  . '#end for' . "\n" . '"';
            }
            else
            {
                $param_cmd =~ s/\$value/\$$name/g;
            }

            $cmd .= ' ' . $param_cmd;
            my $default = '';
            $default = '[' . $rh_attr->{'filenames'} . "]\t" if (defined $rh_attr->{'filenames'});

        }

    }

    if (defined $self->{'params'} && ref($self->{'params'}) =~ /hash/i && length(keys(%{$self->{'params'}}) > 0))
    {
        my $rh_params = $self->{'params'};
        while (my ($name, $rh_attr) = each(%$rh_params))
        {

            my $param_cmd = $rh_attr->{'cmd'};
            $param_cmd =~ s/\$value/\$$name/g;

			if ($rh_attr->{'type'} =~ /boolean/i)
			{
				$param_cmd = "\$$name";
			}
			elsif (! defined $rh_attr->{'default'} or ($rh_attr->{'default'} eq ''))
			{
				$param_cmd = "\n#if (\$$name):\n--$name \$$name\n#end if\n";
			}
			

            $cmd .= ' ' . $param_cmd;

            my $type_galaxy = lc ($rh_attr->{'type'});
            $type_galaxy = 'text' if ($rh_attr->{'type'} =~ /string/i);
            $type_galaxy = $rh_attr->{'type_galaxy'} if (defined $rh_attr->{'type_galaxy'} && $rh_attr->{'type_galaxy'} ne '');
           
            my $description = $rh_attr->{'descr'};

            my $default = '';
            $default = ' value="' . $rh_attr->{'default'} . '" ' if (defined $rh_attr->{'default'});

            my $min = '';
            $min = ' min="' . $rh_attr->{'min'} . '" ' if (defined $rh_attr->{'min'});

            my $max = '';
            $max = ' max="' . $rh_attr->{'max'} . '" ' if (defined $rh_attr->{'max'});
            
            my $label = ( defined $rh_attr->{'label'} && $rh_attr->{'label'} ne '' ) ? $rh_attr->{'label'} : $name;

            if (defined $rh_attr->{'enum'})
            {
                my $rh_enum = $rh_attr->{enum};

                $h_interface_xml_string{'inputs'} .=
                  "<param optional=\"true\" name=\"$name\" type=\"select\" help=\"$description\" label=\"$label\" $default $min $max >\n";

                while (my ($enum_key, $enum_value) = (each(%{$rh_enum})))
                {
                    my $selected = '';
                    if ($enum_key eq $rh_attr->{'default'})
                    {
                        $selected = ' selected="true" ';
                    }
                    $h_interface_xml_string{'inputs'} .= "<option value=\"$enum_key\" $selected>$enum_value</option>\n";
                }
                $h_interface_xml_string{'inputs'} .= "</param>\n";
            }
            elsif ($type_galaxy =~ /boolean/i)
            {
				$default = 'checked="false"';
				$default = 'checked="true"' if (defined $rh_attr->{'default'} && $rh_attr->{'default'} == &TRUE);
				$h_interface_xml_string{'inputs'} .=  "<param optional=\"true\" name=\"$name\" type=\"boolean\" truevalue=\"--$name\" falsevalue=\"\" help=\"$description\" label=\"$label\" $default/>\n";
        
			}
            else
            {
                $h_interface_xml_string{'inputs'} .=
                  "<param optional=\"true\" name=\"$name\" type=\"$type_galaxy\" help=\"$description\" label=\"$label\" $default $min $max />\n";
            }

        }
    }

    my $help = $self->__GetGalaxyHelp;

    my $galaxy_xml = '';

    $galaxy_xml .= <<END;
<tool id="$id" name="$id" version="$version">
	<requirements>
    	<requirement type="binary">$self->{'cmd'}</requirement>
    </requirements>
	<description>$description</description>
	<command interpreter="perl">$cmd</command>
END
    foreach my $interface (@a_inout)
    {
        $galaxy_xml .= "<$interface>\n" . $h_interface_xml_string{$interface} . "</$interface>";
    }

	$galaxy_xml .= '<stdio><exit_code range="1:" level="fatal" /></stdio>';

    $galaxy_xml .= "<help>$help</help>";
    $galaxy_xml .= "</tool>";
    my $o_xml = new XML::Twig(pretty_print => 'indented', empty_tags => 'html');
    $o_xml->parse($galaxy_xml);
    if (defined $outdir && ($outdir ne ''))
    {
        my $file       = '>' . $outdir . '/' . $self->{name} . '.xml';
        my $outfile_fh = new IO::File($file)
          or throw LipmError::IOException($file, 'f');
        $o_xml->print($outfile_fh);
        $outfile_fh->close;

        return;
    }

    my $page = $o_xml->sprint();
    return $page;
}

=head2 Function __GetGalaxyHelp

	Title      :	GetHtmlDesc
	Usage      :	my $galaxy_wiki_help = $self->__GetGalaxyHelp();
	Prerequiste:	none
	Function   :	build a Wiki reStructuredText formatted help for galaxy xml dump
	Returns    :	a string
	Args       :	none
	Error      :    none
	Globals    :    none

=cut

sub __GetGalaxyHelp
{
    my $self = shift;
    
    my $warning = "\n\n.. class:: warningmark\n\n";
    
    my $page = '';
    $page .= "\n**Program**: " . $self->{'cmd'} . "\n";
    $page .= "\n**Name**: " . $self->{'name'} . "\n";
    $page .= "\n**Description**: " . $self->{'descr'} . "\n";
    $page .= "\n**Authors**: " . $self->{'authors'} . "\n";
    $page .= "\n**Version**: " . $self->{'version'} . "\n";
    $page .= "\n**Documentation**: " . join(', ', @{$self->{'doclink'}}) . "\n"  if (defined($self->{'doclink'}) && (scalar @{$self->{'doclink'}} > 0));

    $page .= "\n-----\n";

    my @a_inout = ('inputs', 'outputs');
    foreach my $interface (@a_inout)
    {
        $page .= "\n**" . ucfirst($interface) . "**\n";
        my $rh_inputs = $self->{$interface};
        while (my ($name, $rh_attr) = each(%$rh_inputs))
        {
            my $cmd           = $rh_attr->{'cmd'};
            my $galaxy_format = $rh_attr->{'type_galaxy'};

            my $collection = '';
            $collection = 'Collection of ' if ($rh_attr->{'card'} =~ /n/i);
            $cmd =~ s/\$value/<$collection$galaxy_format>/;
            my $default = '';
            $default = '[' . $rh_attr->{'filenames'} . "]\t" if (defined $rh_attr->{'filenames'});
            $page .= "\n" . "- **$name** " . $rh_attr->{'descr'} . " $default ($galaxy_format format)\n";
        }
        $page .= "\n-----\n";
    }

    if (defined $self->{'params'} && ref($self->{'params'}) =~ /hash/i && length(keys(%{$self->{'params'}}) > 0))
    {
        $page .= "\n**Parameters**\n";
        my $rh_params = $self->{'params'};
        while (my ($name, $rh_attr) = each(%$rh_params))
        {
            my $cmd     = $rh_attr->{'cmd'};
            my $biomoby = $rh_attr->{'type_biomoby'};
            my $default = '';
            $default = '[' . $rh_attr->{'default'} . ']' if (defined($rh_attr->{'default'}));
            $default = 'false' if ($biomoby eq 'Boolean' && $default eq '');

			my $label = ( defined $rh_attr->{'label'} && $rh_attr->{'label'} ne '' ) ? $rh_attr->{'label'} : $name;

            $cmd =~ s/\$value/<$biomoby>/;
            $page .= "\n\n" . "**$label**";
            
            $page .= " [$rh_attr->{'type'}]"       if (defined $rh_attr->{'type'} && !defined $rh_attr->{'enum'});
            $page .= " [" . join(',', keys(%{$rh_attr->{enum}}) ) . "]"       if (defined $rh_attr->{'enum'});
            $page .= " *default:* " . $rh_attr->{'default'}  if (defined $rh_attr->{'default'});
            $page .= " *min:* " . $rh_attr->{'min'}         if (defined $rh_attr->{'min'});
            $page .= " *max:* " . $rh_attr->{'max'}         if (defined $rh_attr->{'max'});
            
            my $description = $rh_attr->{'descr'};
            $description =~ s/warning:?\s*/$warning/i;
            $page .= "\n\n$description\n";
        }
    }

    return $page;

}

=head2 Function GetMobyleXml

	Title      :	GetMobyleXml
	Usage      :	my $xml = $o_appli->GetMobyleXml( $outdir );
	Prerequiste:	none
	Function   :	write an xml based on the Mobyle dtd which describe the application
	Returns    :	undef if $outdir is defined and not empty
	            	a string containing the xml otherwise.
	Args       :	$outdir, facultative string, the output directory
	Error      :    none
	Globals    :    none

=cut

sub GetMobyleXml
{
    my $self = shift;
    my ($outdir) = @_;

    eval 'use XML::LibXML';
    if ($@)
    {
        print "ERROR - XML::LibXML perl module not found; Please fix this problem before\n";
        exit -1;
    }

    my $o_xml = new XML::Twig(pretty_print => 'indented',
                              empty_tags   => 'html');
    $o_xml->parse("<program></program>");

    # $o_xml->print( *STDOUT );

    # head element

    my $head = $self->__GetMobyleXmlHead();
    $head->paste(last_child => $o_xml->root());

    # parameters element
    my $parameters = $self->__GetMobyleXmlParameters();
    $parameters->paste(last_child => $o_xml->root());

    #L'element superieur mobyle n'existe plus
    #$o_xml->root()->wrap_in('mobyle');

    $o_xml->set_encoding();
    $o_xml->set_output_encoding('utf-8');

    #plutot mettre l'adresse de la dtd de pasteur
    #$o_xml->set_doctype( 'mobyle', "" );
    if (defined $outdir && ($outdir ne ''))
    {
        my $file       = '>' . $outdir . '/' . $self->{name} . '.xml';
        my $outfile_fh = new IO::File($file)
          or throw LipmError::IOException($file, 'f');
        $o_xml->print($outfile_fh);
        $outfile_fh->close;

        return;
    }

    my $page = $o_xml->sprint();
    return $page;
}

#	------------------------------------------------------------------	#
#	----------------------- MOBYLE -----------------------------------	#
#	------------------------------------------------------------------	#

# tester l'acces depuis l'exterieur :
# my $test = Appli::__BuildTwigElement
# doit merder pour etre bon.
# <-- ne fonctionne pas, seules les variables on une portee limitee dans une fermeture
{

    # $head est la balise racine renvoyee
    # $ra_child est une ref sur une liste de ref de hash
    #	tq	<head><key1>value1</key1><key2>value2</key2><head>
    sub __BuildTwigElement
    {
        my ($head, $ra_child) = @_;

        my $head_elt = new XML::Twig::Elt($head);
        foreach my $rh (@$ra_child)
        {
            my ($tag, $value) = each(%$rh);

            my $elt = new XML::Twig::Elt($tag => $value);
            $elt->paste(last_child => $head_elt);
        }

        return $head_elt;
    }

    # add each val of the $ra_val list after the $dad elt with the tag $tag
    sub __AddLastToElt
    {
        my ($tag, $ra_val, $dad) = @_;

        foreach my $val (@$ra_val)
        {
            my $elt = new XML::Twig::Elt($tag => $val);
            $elt->paste(last_child => $dad);
        }

        return;
    }

    #<head>
    #	<name>$self->{ name }</name>
    #	<version>$self->{ version }</version>
    #	<doc>
    #		<title>$self->{ name }</title>
    #		<description><text></text></description>
    #		<authors>$self->{ authors }</authors>
    #		<reference>$self->{ reference }</reference>
    #		<doclink>$self->{ doclink }</doclink>
    #		<help><text>$self->GetUsage()</text></help>
    #	</doc>
    #	<category>$self->{ category }</category>
    #</head>
    sub __GetMobyleXmlHead
    {
        my $self = shift;

        my @a_head =
          ({name => $self->{name}}, {version => $self->{version}}, {category => $self->{category}}, {doc => ''});
        my $head = __BuildTwigElement('head', \@a_head);

        # doc element, part of head element
        my @a_doc = (
            {title   => $self->{name}},
            {authors => $self->{authors}}    #,

            # 				{ reference => undef },
            # 				{ doclink => undef }
            );

        my $doc = __BuildTwigElement('doc', \@a_doc);

        # description elt, part of doc element, follows 'title'
        my $descr = __BuildTwigElement('description', []);

        my $descr_text = new XML::Twig::Elt('text' => $self->{descr});
        $descr_text->set_att('lang', 'en');
        $descr_text->paste($descr);

        $descr->paste(after => $doc->first_child('title'));

        # ref elt, part of doc element
        my $tag = 'reference';
        __AddLastToElt($tag, $self->{$tag}, $doc);

        # doclink elt, part of doc element
        $tag = 'doclink';
        __AddLastToElt($tag, $self->{$tag}, $doc);

        # help elt, part of doc element
        #SEB 26.05.2008 DISPARU DERNIERE DTD
        #my $help = new XML::Twig::Elt('text' => $self->GetUsage())->wrap_in('help');
        #$help->paste(last_child => $doc);

        $doc->replace($head->last_child());
        ## fin doc elt

        # env elt, part of head
        # besoin d'analyser le xml contenant les variable d'environnement.

        return $head;
    }

    # return a valid parameters Elt
    sub __GetMobyleXmlParameters
    {
        my $self = shift;

        my $params_elt = new XML::Twig::Elt('parameters');
        $params_elt->set_att('xmlns:xi' => 'http://www.w3.org/2001/XInclude');

        # la commande : le nom du programme est celui du service
        #	par defaut la commande a executer est le nom du service

        my $param_elt = $self->__GetMobyleXmlXputs(
                                                   'command',
                                                   $self->{name},
                                                   {
                                                    descr => 'command',
                                                    type  => 'String',
                                                    cmd   => $self->{cmd}
                                                   }
                                                   );
        $param_elt->paste(last_child => $params_elt);

        #plus tard gerer peut-etre les paragraphes
        #inputs
        my $rh_inputs = $self->{inputs};
        while (my ($name, $rh_attr) = each(%$rh_inputs))
        {
            $param_elt = $self->__GetMobyleXmlXputs('inputs', $name, $rh_attr);

            $param_elt->paste(last_child => $params_elt);
        }

        #outputs
        my $rh_outputs = $self->{outputs};
        while (my ($name, $rh_attr) = each(%$rh_outputs))
        {
            my ($param_a_elt, $param_b_elt) = $self->__GetMobyleXmlOutputs('outputs', $name, $rh_attr);
            $param_b_elt->paste(last_child => $params_elt);
            $param_a_elt->paste(last_child => $params_elt);
        }

        #params
        my $rh_param = $self->{params};
        while (my ($name, $rh_attr) = each(%$rh_param))
        {
            my $param_elt = $self->__GetMobyleXmlParams($name, $rh_attr);
            $param_elt->paste(last_child => $params_elt);
        }

        return $params_elt;
    }

    # renvoie
    #<parameter (attr) >
    #	<name>$name</name>
    #	<prompt>$$rh_attr{ descr }</prompt>
    #	<type>cf __GetMobyleXmlType</type>
    #	<></>
    #</parameter>
    sub __GetMobyleXmlXputs
    {
        my $self = shift;
        my ($key, $name, $rh_attr) = @_;

        $$rh_attr{descr} = $name if (!defined $$rh_attr{descr} || ($$rh_attr{descr} eq ''));

        my @a_param = ({name => $name}, {prompt => $$rh_attr{descr}});
        my $parameter = __BuildTwigElement('parameter', \@a_param);

        #attributs
        my $rh_in_out = {ismaininput => '1'} if ($key eq 'inputs');

        $rh_in_out = {iscommand => '1', ishidden => '1'} if ($key eq 'command');

        $parameter->set_atts($rh_in_out);

        my $type_elt =
          __GetMobyleXmlType($$rh_attr{type}, $$rh_attr{type_biomoby}, $$rh_attr{namespace_biomoby}, $$rh_attr{card});
        $type_elt->paste(last_child => $parameter);

        if (defined($$rh_attr{cmd}))
        {

            #ICI pour passer le vrai nom de l'executable
            #my $format_elt = new XML::Twig::Elt( code => { proglang => 'perl' },
            #									"\" $$rh_attr{ cmd }\"" )
            #					->wrap_in('format');

            #$format_elt->paste(last_child => $parameter);
            my $format_elt = new XML::Twig::Elt('format');

            my $perl_cmd = $$rh_attr{cmd};
            my $code_perl_elt = new XML::Twig::Elt(code => {proglang => 'perl'}, "\" $perl_cmd\"");
            $code_perl_elt->paste(last_child => $format_elt);
            my $python_cmd = $perl_cmd;

            #$python_cmd =~ s/"/\\"/g;
            $python_cmd =~ s/\\"//g;
            $python_cmd =~ s/["']//g;

            if ($python_cmd =~ /\$/)
            {
                $python_cmd =~ s/\$(\w+)/%s" % $1/;
            }
            else
            {
                $python_cmd .= '"';
            }
            my $code_python_elt = new XML::Twig::Elt(code => {proglang => 'python'}, "\" $python_cmd");
            $code_python_elt->paste(last_child => $format_elt);

            $format_elt->paste(last_child => $parameter);
        }

        if (defined($$rh_attr{filenames}))
        {

            my $filenames_elt = new XML::Twig::Elt(
                                                   code => {proglang => 'perl'},
                                                   "\"$$rh_attr{ filenames }\""
                                                   )->wrap_in('filenames');
            $filenames_elt->paste(last_child => $parameter);
        }

        return $parameter;
    }

    sub __GetMobyleXmlOutputs
    {
        my $self = shift;
        my ($key, $name, $rh_attr) = @_;
        $$rh_attr{descr} = $name if (!defined $$rh_attr{descr} || ($$rh_attr{descr} eq ''));
        my $no_dash_name = $name;
        $no_dash_name =~ s/-/_/g;

        my @a_param = ({name => $name}, {prompt => $$rh_attr{descr}});
        my $parameter_a = __BuildTwigElement('parameter', \@a_param);

        #attributs
        my $rh_in_out = {isout => '1'} if ($key eq 'outputs');
        $parameter_a->set_atts($rh_in_out);

        my $type_elt =
          __GetMobyleXmlType($$rh_attr{type}, $$rh_attr{type_biomoby}, $$rh_attr{namespace_biomoby}, $$rh_attr{card});
        $type_elt->paste(last_child => $parameter_a);

        #<code proglang="perl">$RmesReport_name</code>
        #<code proglang="python">RmesReport_name</code>

        if (defined($$rh_attr{filenames}))
        {
            my $filenames       = $$rh_attr{filenames};
            my $format_elt      = new XML::Twig::Elt('filenames');
            my $code_perl_elt   = new XML::Twig::Elt(code => {proglang => 'perl'}, "\"$filenames\"");
            my $code_python_elt = new XML::Twig::Elt(code => {proglang => 'python'}, "\"$filenames\"");
            my $filenames_elt   = new XML::Twig::Elt('filenames');
            $code_perl_elt->paste(last_child => $filenames_elt);
            $code_python_elt->paste(last_child => $filenames_elt);
            $filenames_elt->paste(last_child => $parameter_a);
        }
        else
        {
            my $code_perl_elt   = new XML::Twig::Elt(code => {proglang => 'perl'},   '$' . $no_dash_name . '_name');
            my $code_python_elt = new XML::Twig::Elt(code => {proglang => 'python'}, $no_dash_name . '_name');
            my $filenames_elt = new XML::Twig::Elt('filenames');
            $code_perl_elt->paste(last_child => $filenames_elt);
            $code_python_elt->paste(last_child => $filenames_elt);
            $filenames_elt->paste(last_child => $parameter_a);
        }

        my @a_param_b = ({name => $no_dash_name . '_name'}, {prompt => $name . ' output file name'});

        my $parameter_b = __BuildTwigElement('parameter', \@a_param_b);
        $parameter_b->set_atts({ishidden => 1});

        my $datatype_elt = __BuildTwigElement('datatype', [{class => 'Filename'}]);

        $type_elt = new XML::Twig::Elt('type');
        $datatype_elt->paste(last_child => $type_elt);
        $type_elt->paste(last_child => $parameter_b);

        my $extension = 'out';
        if (ref($$rh_attr{type}) =~ /XML::Twig::Elt/i)
        {

            if (defined $$rh_attr{type}->{att}->{fileextension}
                && ($$rh_attr{type}->{att}->{fileextension} ne ''))
            {
                $extension = $$rh_attr{type}->{att}->{fileextension};
            }
            elsif (defined $$rh_attr{type}->{att}->{dataformat} && ($$rh_attr{type}->{att}->{dataformat} ne ''))
            {
                $extension = $$rh_attr{type}->{att}->{dataformat};
            }
        }
        my $vdef_elt = __BuildTwigElement('vdef', [{value => $no_dash_name . '.' . $extension}]);
        $vdef_elt->paste(last_child => $parameter_b);

        if (defined($$rh_attr{cmd}))
        {

            #ICI pour passer le vrai nom de l'executable
            my $format_elt = new XML::Twig::Elt('format');

            my $perl_cmd = $$rh_attr{cmd};
            my $code_perl_elt = new XML::Twig::Elt(code => {proglang => 'perl'}, "\" $perl_cmd\"");
            $code_perl_elt->paste(last_child => $format_elt);

            #on va essayer de traduire du perl vers le python
            #pour les cmd simples !

            #$python_cmd =~ s/"/\\"/g;
            my $python_cmd = $perl_cmd;
            $python_cmd =~ s/\\"//g;
            $python_cmd =~ s/["']//g;

            $python_cmd =~ s/\$(\w+)/%s" % $1/;

            my $code_python_elt = new XML::Twig::Elt(code => {proglang => 'python'}, "\" $python_cmd");
            $code_python_elt->paste(last_child => $format_elt);

            $format_elt->paste(last_child => $parameter_b);
        }

        return ($parameter_a, $parameter_b);
    }

    sub __GetMobyleXmlParams
    {
        my $self = shift;
        my ($name, $rh_attr) = @_;

        $$rh_attr{descr} = $name if (!defined $$rh_attr{descr} || ($$rh_attr{descr} eq ''));
        my @a_param = (
            {name   => $name},
            {prompt => $$rh_attr{descr}},
            {type   => ''},

            # 				# la commande est -$name=xxx
            # 				{ cmd	=>	"-$name=" }
            );
        my $parameter = __BuildTwigElement('parameter', \@a_param);

        #attributs
        if (defined $$rh_attr{mandatory}) {$parameter->set_atts({ismandatory => $$rh_attr{mandatory}});}

        $self->__CompleteParameterElement($parameter, $rh_attr);

        return $parameter;
    }

    sub __CompleteParameterElement
    {
        my $self = shift;
        my ($param_elt, $rh_attr) = @_;

        my $type_elt = __GetMobyleXmlType($$rh_attr{type}, $$rh_attr{type_biomoby}, '', '', $$rh_attr{enum});
        $type_elt->replace($param_elt->first_child('type'));

        if (defined $$rh_attr{default})
        {

            # la dtd considere que plusieurs valeurs par defaut sont possibles
            #	pas de ÃÂ§a chez nous !
            if ($$rh_attr{default} ne '')
            {
                my $vdef = new XML::Twig::Elt('value' => $$rh_attr{default})->wrap_in('vdef');
                $vdef->paste(last_child => $param_elt);
            }
        }

        # enum <=> vlist
        #print STDERR "TOTO\n\n";
        #print STDERR $$rh_data{enum};
        if (defined $$rh_attr{enum} && $$rh_attr{enum} =~ /^http:\/\//)
        {
            my $xinclude_string = "
<xi:include href=\"$$rh_attr{enum}\">
    <xi:fallback>
       <vlist>
          <velem undef=\"1\">
             <value>null</value>
             <label>Choose a database</label>
          </velem>
       </vlist>
    </xi:fallback>
</xi:include>";

            my $o_twig_vlist = new XML::Twig();
            $o_twig_vlist->parse($xinclude_string);
            my $vlist = $o_twig_vlist->root;
            $vlist->paste(last_child => $param_elt);
        }
        elsif (defined $$rh_attr{enum})
        {
            my $vlist = new XML::Twig::Elt('vlist');

            my $rh_enum = $$rh_attr{enum};
            foreach my $key (sort keys %$rh_enum)
            {
                my $label   = $$rh_enum{$key};
                my @a_velem = ({value => $key}, {label => $label});
                my $velem   = __BuildTwigElement('velem', \@a_velem);
                $velem->paste(last_child => $vlist);
            }

            #Tri des ENUM - demande SLETORT 20091105
            #while (my ($instance, $label) = each(%$rh_enum))
            #{
            #   my @a_velem = ({value => $instance}, {label => $label});
            #    my $velem = __BuildTwigElement('velem', \@a_velem);
            #    $velem->paste(last_child => $vlist);
            #}

            $vlist->paste(last_child => $param_elt);
        }    #fin Enum <=> vlist

        # cmd <=> format

        #        my $format_elt = new XML::Twig::Elt(
        #                                            code => {proglang => 'perl'},
        #                                            "\"$$rh_attr{ cmd }\""
        #                                            )->wrap_in('format');
        #        $format_elt->paste(last_child => $param_elt);

        my $format_elt = new XML::Twig::Elt('format');

        my $raw_cmd  = $$rh_attr{cmd};
        my $perl_cmd = $raw_cmd . "\"";
        if ($$rh_attr{type_biomoby} eq 'Boolean')
        {
            $perl_cmd .= ' if ( defined $value && ($value ne \'\') &&($value != 0) && $value !~ /false/i))';
        }
        elsif (!defined $$rh_attr{enum})
        {
            $perl_cmd .= ' if ($value ne \'\')';
        }

        my $code_perl_elt = new XML::Twig::Elt(code => {proglang => 'perl'}, "\" $perl_cmd");
        $code_perl_elt->paste(last_child => $format_elt);

        #on va essayer de traduire du perl vers le python
        #pour les cmd simples !
        my $python_cmd = $raw_cmd;

        #$python_cmd =~ s/"/\\"/g;
        $python_cmd =~ s/\\"//g;
        $python_cmd =~ s/["']//g;

        $python_cmd =~ s/\$(\w+)/%s" % $1/;

        if ($$rh_attr{type_biomoby} eq 'Boolean')
        {
            $python_cmd = '(""," ' . $raw_cmd . '")[ bool(value) ]';
        }
        elsif (!defined $$rh_attr{enum})
        {
            $python_cmd = '(""," ' . $python_cmd . ')[' . $1 . ' is not None]';
        }
        else
        {
            $python_cmd = '" ' . $python_cmd;
        }
        my $code_python_elt = new XML::Twig::Elt(code => {proglang => 'python'}, "$python_cmd");
        $code_python_elt->paste(last_child => $format_elt);

        $format_elt->paste(last_child => $param_elt);

        #fin cmd <=> format

        # min-max <=> scale
        if (defined $$rh_attr{min} or defined $$rh_attr{max})
        {
            if (   (defined $$rh_attr{min} && !defined $$rh_attr{max})
                or (defined $$rh_attr{max} && !defined $$rh_attr{min}))
            {
                my $dump =
                    "\n--------------------------------------------------"
                  . "\n\tdescription : "
                  . $$rh_attr{descr} . "\n\t"
                  . "cmd : "
                  . $$rh_attr{cmd} . "\n\t"
                  . "min : "
                  . $$rh_attr{min} . "\n\t"
                  . "max : "
                  . $$rh_attr{max}
                  . "\n--------------------------------------------------\n";

                Carp::croak("ERROR - min AND max: once you declared one, you must declare the other one $dump");
            }

            my $scale_elt = new XML::Twig::Elt('scale');
            my $min_elt = new XML::Twig::Elt('value' => $$rh_attr{min})->wrap_in('min');
            $min_elt->paste(last_child => $scale_elt);
            my $max_elt = new XML::Twig::Elt('value' => $$rh_attr{max})->wrap_in('max');
            $max_elt->paste(last_child => $scale_elt);

            $scale_elt->paste(last_child => $param_elt);
        }    #fin min-max <=> scale

        return;
    }

    # return
    #<type>
    #	<datatype><class>$type</class></datatype>
    #	<card>$card</card>
    #	<biomoby>$type_biomoby</biomoby>
    #	<biomoby_namespace>$namespace_biomoby</biomoby_namespace>
    #</type>

    sub __GetMobyleXmlType
    {
        my ($type, $type_biomoby, $namespace_biomoby, $card, $enum) = @_;

        Carp::croak("__GetMobyleXmlType:type for $type_biomoby undef") unless defined $type;

        # type element part of parameter element
        my $type_elt = new XML::Twig::Elt('type');

        if ((defined $enum) && ($enum ne ''))
        {
            $type = 'Choice';
            my $datatype = new XML::Twig::Elt(class => ucfirst($type))->wrap_in('datatype');
            $datatype->paste(last_child => $type_elt);
        }
        elsif (ref($type) !~ /xml::twig::elt/i)
        {

            my $datatype = new XML::Twig::Elt(class => ucfirst($type))->wrap_in('datatype');
            $datatype->paste(last_child => $type_elt);
        }
        else
        {
            my $datatype = new XML::Twig::Elt('datatype');

            if ($type->{att}->{class} ne '')
            {
                my $class = new XML::Twig::Elt(class => $type->{att}->{class});
                $class->paste(last_child => $datatype);
            }
            if ($type->{att}->{superclass})
            {
                my $superclass = new XML::Twig::Elt(superclass => $type->{att}->{superclass});
                $superclass->paste(last_child => $datatype);
            }

            $datatype->paste(last_child => $type_elt);

            if ($type->{att}->{biotype} ne '')
            {
                my $biotype = new XML::Twig::Elt(biotype => $type->{att}->{biotype});
                $biotype->paste(last_child => $type_elt);
            }

            if ($type->{att}->{dataformat} ne '')
            {
                my $dataformat = new XML::Twig::Elt(dataFormat => $type->{att}->{dataformat});

                #SC 20101227: Mobyle 0.97 -> Mobyle 1.0
                #my $accepteddataformats = new XML::Twig::Elt('acceptedDataFormats');
                #$dataformat->paste(last_child => $accepteddataformats);
                #$accepteddataformats->paste(last_child => $type_elt);
                $dataformat->paste(last_child => $type_elt);
            }
        }

        if (IsNotNull($card))
        {
            my $card_elt = new XML::Twig::Elt('card' => $card);
            $card_elt->paste(last_child => $type_elt);
        }

        #<biomoby><datatype>FASTA</datatype><namespace>NCBI</namespace></biomoby>
        if (IsNotNull($type_biomoby) || IsNotNull($namespace_biomoby))
        {
            my $biomoby = new XML::Twig::Elt('biomoby');

            if (IsNotNull($type_biomoby))
            {
                my $biomoby_datatype = new XML::Twig::Elt('datatype' => $type_biomoby);
                $biomoby_datatype->paste(last_child => $biomoby);
            }

            if (IsNotNull($namespace_biomoby))
            {
                my $biomoby_namespace = new XML::Twig::Elt('namespace' => $namespace_biomoby);
                $biomoby_namespace->paste(last_child => $biomoby);
            }
            $biomoby->paste(last_child => $type_elt);
        }
        return $type_elt;
    }

}

#	------------------------------------------------------------------	#
#	----------------------- PLAYMOBY --------------------------      #
#	------------------------------------------------------------------	#

=head2 Function GetPlaymobyTestXml

	Title      :	GetPlaymobyTestXml
	Usage      :	my $xml = $o_appli->GetPlaymobyTestXml( $outdir );
	Prerequiste:	none
	Function   :	write an xml to launch Playmoby tests
	Returns    :	undef if $outdir is defined and not empty
	            	a string containing the xml otherwise.
	Args       :	$outdir, facultative string, the output directory
	Error      :    none
	Globals    :    none

=cut

sub GetPlaymobyTestXml
{
    my $self = shift;
    my ($outdir) = @_;

    my $servicename = $self->{name};

    my $xml_for_test = "<playmoby:test>";
    $xml_for_test .= "<playmoby:service name=\"$servicename\" >";

    foreach my $output_name (keys %{$self->{outputs}})
    {
        $xml_for_test .= "<playmoby:article name=\"$output_name\" >";
        $xml_for_test .= "<playmoby:format>";
        $xml_for_test .= <<END;
<![CDATA[

#these are default tests for format consistancy
#you can/should write your own test 
#your test MUST return a boolean and a free text

return (0, "empty article") if (\$_ eq '');
return (1, "ok");
]]>
END
        $xml_for_test .= "</playmoby:format>";
        $xml_for_test .= "<playmoby:content>";
        $xml_for_test .= <<END;
<![CDATA[

#these are default tests for content analysis
#you can/should write your own test 
#your test MUST return a boolean and a free text

return (0, "empty article") if (\$_ eq '');
return (1, "ok");
]]>
END
        $xml_for_test .= "</playmoby:content>";
        $xml_for_test .= "</playmoby:article>";

    }

    $xml_for_test .= "</playmoby:service>";
    $xml_for_test .= "</playmoby:test>";

    my $o_xml = new XML::Twig(pretty_print => 'indented',
                              empty_tags   => 'html');
    $o_xml->parse($xml_for_test);
    $o_xml->set_encoding();
    $o_xml->set_output_encoding('utf-8');

    if (defined $outdir && ($outdir ne ''))
    {
        my $outfile_fh = new IO::File('>' . $outdir . '/' . $self->{name} . '.pmb.t');
        $o_xml->print($outfile_fh);
        $outfile_fh->close;

        return;
    }
    else
    {
        my $page;
        $page = $o_xml->sprint();
        return $page;
    }

}

=pod
<vlist>
	<velem undef="1">
		<value>null</value>
		<label>Choose a database</label>
	</velem>
	<velem>
		<value>embl</value>
		<label>embl: Embl last release + updates</label>
	</velem>
</vlist
=cut

sub __VlistXmlURLToHashRef
{

    my $url    = shift;
    my %h_enum = ();

    my $o_twig         = XML::Twig->nparse($url);
    my @a_o_velem_elts = $o_twig->get_xpath('/vlist/velem/');
    foreach my $o_velem_elt (@a_o_velem_elts)
    {
        my $value = $o_velem_elt->first_child('value')->text;
        my $label = $o_velem_elt->first_child('label')->text;
        $h_enum{$value} = $label;
    }
    return \%h_enum;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
