#
# $Id: AccWarn.pm 835 2012-02-23 19:25:49Z manu $
#

package AccWarn;

use strict;
use warnings;

use base qw( LipmObject );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

AccWarn - Accumulate warnings, then display some of them

=head1 SYNOPSIS
 my $o_warn = New AccWarn('^WARNING \d+ ');
 ... A lot of warnings generated by the program, most of them match with the regex above
 # Display the accumulated warnings:
 $o_warn->DisplayWarnings();

=head1 DESCRIPTION

 AccWarn is a "Warning Accumulator": If not already configured, it takes the __WARN__ handle to intercept all the warnings
 (a warning is a string passed to the warn function)
 
 A first regexp match (against $IMPORTANT_WARN_REGEX), then a second (against $WARN_REGEX): If one of both succeeds, 
 the warning is counted by the object in a hash, whose key is the matched substring.
 
 If the message matched against $IMPORTANT_WARN_REGEX, all warnings of this type are kept in some array.
 If the message matched only against $WARN_REGEX, only the N (default = 10) first warnings are conserved
 If both the matches fail, the warning is blocked or not (configurable).
 
 When the method DisplayWarnings is called, all the accumulated warnings are printed in a synthetized way.
 
 This object is useful to avoid printing too many warnings, which does not make sense for the users as they do not read, 
 but also keeping track of those warnings. 
 
 If you want to display some line to stderr without being trapped by AccWarn, please use print STDERR instead of warn. 
 This is useful for debug-related messages
 
 WARNING - Please note that as soon as AccWarn is initialized, ALL warn are trapped, even those emitted by other perl modules
          (ParamParser for instance)
          This can be useful (a lot of messages emitted by bioperl are trapped), but this can be embarrassing, as those messages may be useful
          The solution is to let the other warnings pass, OR to defer the construction of AccWarn.
 
=cut

use General;
use Error qw(:try);
use LipmError::ParameterException;
use LipmError::ObjectException;
use English qw( -no_match_vars ) ;

use File::Temp qw/ :seekable /;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 835 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title   : Constructor
 Usage   : my $o_warnings = New AccWarn('^(WARNING )',5,'^(ERROR )');
 Function: Initialize the object, redefining the warn handler if not already defined. The object is thus a defacto singleton.
           We DO NOT use $self to store data, we use call variables (our construct).
 Args    : $regex                      The $nb first messages matching with this regex will be trapped and accumulated
           $nb (optional, default=10)  see above
           $important_regex (optional) ALL the messages matching with this regex will be trapped and accumulated

=cut

# Because AccWarn is a singleton (see under), we can use module variables as private fields, 
# this should be faster than using $self !
our $WARN_REGEX           = "";
our $IMPORTANT_WARN_REGEX = "";
our %H_WARNINGS           = ();
our %H_IMPORTANT_WARNINGS = ();
our $BLOCK_WARNINGS       = 0;
our $WLIMIT               = 0;

#
# The %H_WARNINGS hash:
#     - key = The matched substring ( see ${^MATCH} in the perl documentation - NOTE THIS CODE WORKS ONLY WITH perl 5.10 AND ABOVE)
#     - val = A rh with following keys:
#             'occurences' The number of occurrences of this warning
#             'warnings'   A ra with the N first warnings 
#

sub _Init
{
	my $self = shift;
	my ($regex,$wlimit,$important_regex) = @_;
	$wlimit = 10 unless ( defined $wlimit );
	
	$self->SUPER::_Init(@_);

	$WARN_REGEX = $regex;
	$IMPORTANT_WARN_REGEX = $important_regex if (defined $important_regex);
	
	# If something else is already using the warn handler, we throw an exception - AccWarn is thus a singleton !
	if ( defined ( $SIG { __WARN__ } ) && $SIG { __WARN__ } == \&WarnHandler )
	{
		throw LipmError::ObjectException( 'AccWarn', '!build', -text => "An AccWarn object ALREADY EXISTS in your program !\n" );
	}
	else
	{
		$SIG { __WARN__ } = \&WarnHandler;
	}
	
	# some default values
	$BLOCK_WARNINGS = 1;
	$WLIMIT         = $wlimit;
}

=head2 procedure WarnHandler

 Title    : WarnHandler
 Usage    : Called by the system when warn is called by any function.
 Procedure: Try to match the message with the regex, then accumulated the message if matched.
            The matched substring is the key inside %H_WARNINGS
            If it does not match, ignore the message or call warn to display it (when called from its handler warn works in the default way)
 Args     : $msg  The warn parameter

=cut
sub WarnHandler
{
	my ($msg) = @_;
	if ( $IMPORTANT_WARN_REGEX ne "" && $msg =~ /$IMPORTANT_WARN_REGEX/p )
	{
		my $key = ${^MATCH};
		if ( ! exists $H_IMPORTANT_WARNINGS { $key } )
		{
			$H_IMPORTANT_WARNINGS{$key} = {};
			$H_IMPORTANT_WARNINGS{$key} -> { warnings }   = [];
		};
		push ( @{$H_IMPORTANT_WARNINGS{$key} -> { warnings }}, ${^POSTMATCH} );
	}
	elsif ( $msg =~ /$WARN_REGEX/p ) 
	{
		my $key = ${^MATCH};
		if ( ! exists $H_WARNINGS { $key } )
		{
			$H_WARNINGS{$key} = {};
			$H_WARNINGS{$key} -> { occurences } = 0;
			$H_WARNINGS{$key} -> { warnings }   = [];
		};
		if ( ++$H_WARNINGS{$key} -> { occurences } <= $WLIMIT )
		{
			push ( @{$H_WARNINGS{$key} -> { warnings }}, ${^POSTMATCH} );
		}
	}
	else
	{
		if ( $BLOCK_WARNINGS == 0 )
		{
			warn $msg;
		}
	}
}
			
=head2 procedure DisplayWarnings

 Title    : DisplayWarnings
 Usage    : $o_warnings->DisplayWarnings();
 Procedure: Display the accumulated warnings, call this at the end of the program
 Args     : none

=cut
sub DisplayWarnings
{
	my @a_keys = sort keys %H_IMPORTANT_WARNINGS;
	if ( @a_keys > 0 )
	{
		print STDERR ("=" x 40) . "\n";
		print STDERR "IMPORTANT WARNINGS\n";
		foreach my $key (@a_keys)
		{
			print STDERR ("=" x 40) . "\n";
			print STDERR "$key ==> " . @{$H_IMPORTANT_WARNINGS{$key} -> { warnings }} . " occurences\n";
			print STDERR ("=" x 40) . "\n";
			foreach my $msg ( @{$H_IMPORTANT_WARNINGS{$key} -> { warnings }} )
			{
				print STDERR "$msg\n";
			}
		}
	}
	
	@a_keys = sort keys %H_WARNINGS;
	if ( @a_keys > 0 )
	{
		print STDERR ("=" x 40) . "\n";
		print STDERR "OTHER WARNINGS\n";
		foreach my $key (@a_keys)
		{
			print STDERR ("=" x 40) . "\n";
			print STDERR "$key ==> " . $H_WARNINGS{$key} -> { occurences } . " occurences\n";
			print STDERR ("=" x 40) . "\n";
			
			foreach my $msg ( @{$H_WARNINGS{$key} -> { warnings }} )
			{
				print STDERR "$msg\n";
			}
			if ( $H_WARNINGS{$key} -> { occurences } > $WLIMIT )
			{
				print STDERR "...\n";
			}
		}
		print STDERR ("=" x 40) . "\n";
	}
}

=head2 procedures BlockOtherWarnings, PassOtherWarnings
 Self explanatory, isn't it ?
=cut
sub BlockOtherWarnings { $BLOCK_WARNINGS = 1; };
sub PassOtherWarnings  { $BLOCK_WARNINGS = 0; };

=head2 function GetNbWarnings

 Title    : GetNbWarnings
 Usage    : my $nb_warnings = $o_warnings->GetNbWarnings();
 Function : Return the total number of accumulated warnings.
            The other warnings, passed or blocked, are NOT counted
 Args     : none
 Return   : the number of accumulated warnings

=cut
sub GetNbWarnings
{
	my $nb = 0;
	my @a_keys = keys %H_WARNINGS;
	foreach my $key (@a_keys)
	{
		$nb += $H_WARNINGS{$key} -> { occurences };
	}
	@a_keys = keys %H_IMPORTANT_WARNINGS;
	foreach my $key (@a_keys)
	{
		$nb += @{$H_IMPORTANT_WARNINGS{$key} -> { warnings }};
	}
	
	return $nb;
}
	
=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
