#
# $Id: AnnotatedSeq.pm 1375 2014-03-17 10:13:47Z manu $
#

package AnnotatedSeq;

use strict;
use warnings;

use base qw( LipmObject );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

AnnotatedSeq - A class to model a genomic sequence with annotations (ie a chromosome, a contig, etc)

=head1 SYNOPSIS

An AnnotatedSeq contains 4 arrays of objects derived from the Locus class:
      - An array of Gene objects
      - An array of Ncrnas objects
      - An array of Repeat objects
      - An array of MiscRegion objects
      
An AnnotatedSeq MAY contain also:
      - An AnnotatedSeqFileReader to be initalized from some file
      - An AnnotatedSeqFileWriter to dump data to some file using some format
      - A ChromsomeBuilder to build a chromosome from a lot of contigs and an agp file

An annotatedSeq has a lot of mutators and accessors to manipulate the internal data structures.

=head1 DESCRIPTION

A general and very simple converter program should look like this:
      my $o_ann_seq = New AnnotatedSeq ( -in_ann_file  => "input.gff3",
                                         -in_ann_fmt   => "gff3",
                                         -in_seq_file  => "input.fasta",
                                         -in_seq_fmt   => "fasta",
                                         -out_ann_file => "output.gbk",
                                         -out_seq_fmt  => "genbank" );
      $o_ann_seq -> Load();  # Reading input file
      $o_ann_seq -> Dump();  # Dumping output file
      $o_ann_seq = undef;    # Destroying the object

But another use case could be to set the biological objects from another program:
      my $o_ann_seq = New AnnotatedSeq ();
      my $o_gene = New Locus::Gene ( ...);
      $o_ann_seq -> AddGene ( $o_gene );
      etc.
       
It is even possible to mix both approaches:

      # Add programmatically some biological objects
      $o_ann_seq -> AddGene ( $o_gene );
      etc.
      
      # Load sequence and other annotations from gff3 + fasta files
      $o_ann_seq -> SetInput ( -in_ann_file  => "annotations.gff3", -in_ann_fmt => "gff3", -in_seq_file => "sequence.fasta" );
      $o_ann_seq -> SetOutput( -out_ann_file => "output.gbk", -out_ann_fmt => "genbank" );
      $o_ann_seq -> Load();
      $o_ann_seq -> Dump();
      
=cut

use Carp;
use Error qw(:try);
use Error::Simple;
use LipmError::ParameterException;
use LipmError::ObjectException;

use General;

use FileFormats;
use Locus;
use Locus::Gene;
use Locus::Mrna;
use Locus::Ncrna;
use Locus::Cds;
use Locus::CdsRegion;
use Locus::ExonSet;
use Locus::Exon;
use Locus::Utr;
use Locus::UtrRegion;
use Locus::Repeat;
use Locus::MiscRegion;

use Data::Dumper;
use Splitter::Fasta;

# a mettre dans un dumper xml
use FileHandle;
use XML::Twig;

# This is used only with the switch --renumer-locus_tags
# cf. the sub __RenumberLocusTags
our $LOCUS_TAG_NB = -1;


BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1375 $ =~ /\d+/g); $r[0]};
}

our $H_IN_ANN_FMT = { 
	GFF3    => 1,
	GTF     => 1,
	GENBANK => 1,
	EMBL    => 1,
	ENSEMBL => 1
};
our $H_OUT_ANN_FMT = {
	GFF3     => 1,
	GENBANK  => 1,
	EMBL     => 1,
	FASTA    => 1,
	NARCISSE => 1,
	TBL      => 1,
	TSV      => 1,
	EUGENE   => 1
	};
our $H_IN_SEQ_FMT = {
	FASTA   => 1
	};
our $H_OUT_SEQ_FMT = {
	FASTA   => 1
	};
	
our %H_KINGDOM = (
	bacteria => 1,
	eukaryote => 1
	);

use constant UNKNOWN => 'Unknown';

sub GetAllowedFormats
{
	my $self = shift;
	my $blabla;
	$blabla .= "ALLOWED INPUT FORMATS FOR ANNOTATION FILES : ";
	$blabla .= join ( ',',keys %$AnnotatedSeq::H_IN_ANN_FMT ) . "\n";
	$blabla .= "ALLOWED INPUT FORMATS FOR SEQUENCE FILES   : ";
	$blabla .= join ( ',',keys %$AnnotatedSeq::H_IN_SEQ_FMT ) . "\n";
	$blabla .= "ALLOWED OUTPUT FORMATS FOR ANNOTATION FILES: ";
	$blabla .= join ( ',',keys %$AnnotatedSeq::H_OUT_ANN_FMT ) . "\n";
	$blabla .= "ALLOWED OUTPUT FORMATS FOR SEQUENCE FILES  : ";
	$blabla .= join ( ',',keys %$AnnotatedSeq::H_OUT_SEQ_FMT ) . "\n";
	
	return $blabla;
}

=head2 procedure _Init

 Title        : Constructor
 Usage        : my $o_annot_seq = New AnnotatedSeq( ... );
 Args         : Named arguments, optional start with a '-':
                kingdom            The kingdom - Must be declared in %H_KINGDOM
                species            The species - generally short name
                -contig            The contig label - Free text, no space (?????) - If not specified, we try to automatically
                                   detect it from the file (this will be possible only when SetInput will be called)
                -chromosome        The chromosome label - If not specified, same as contig label.
                                   If -o_assembly is specified, the chromosome is specified by this object, and this parameter is not used
                -o_assembly        The Assembly object, if defined and used
                
                -in_ann_file       The input annotation file name
                -in_ann_fmt[GFF3]  The input annotation file format - Must be a key of %H_INFORMAT
                -in_seq_file       The input sequence file name, if any
                -in_seq_fmt        The input sequence file format (currently only FASTA)
                
                -out_ann_file      The output annotation file name
                -out_ann_fmt[GFF3] The output annotation file format - Must be a key of %H_outFORMAT
                -out_seq_file      The output sequence file name, if any
                -out_seq_fmt       The output sequence file format (currently only FASTA)
                
                -tmpdir[/tmp]      The temporary directory
                 
                -create_cds        &TRUE/&FALSE
                -desc              &TRUE/&FALSE

                -consistency_limit  The ??????????????????????????
                -codon_usage_table  The Codon Usage Table, chosen by bioperl for eukaryotes (for CDS translation) ?????
                -cds_overlap_allowed If True, allow overlaps between CDS (see AddGeneCDSorExon), but this probably means that the annotation is poor
                -debug[&FALSE]       More messages if specified
 Globals      : none

=cut

sub _Init
{
	my $self = shift;

	my %h_args = (
		kingdom           => undef,
		species           => undef,
		-contig           => undef,
		-chromosome       => undef,
		-o_assembly       => undef,
		-o_reassemble_assembly => undef,
		-renumber_lt      => undef,
		-filter_small_utr => &FALSE,
		-with_sequence_embedded => &FALSE,
		-with_partial           => &FALSE,
		-filter_bad_proteins => undef,
		-create_utr       => &FALSE,
		-tmpdir           => "/tmp",
		-debug            => &FALSE,
		-new_rep_origin      => undef,
		-consistency_limit   => 0,
		-codon_usage_table   => undef,
		-cds_overlap_allowed => &FALSE,
		-create_cds          => &FALSE,
		-desc_file           => "",
		-header_file         => "",
		-submit              => 0,
		-compute_gaps        => &FALSE,
		-keep_gap_size       => &FALSE,
		-check_gaps          => &FALSE,
		-build_assembly      => &FALSE,
		-disassemble_assembly => &FALSE,
		-output_utr          => &FALSE,
		-output_intron       => &FALSE,
		-noclean_fasta_header => &FALSE,
		@_);

	# Validate required fields
	foreach my $key ( qw( kingdom species ) )
	{
		unless ( defined $h_args { $key } )
		{
			throw LipmError::ParameterException ($key,'missing', -text => "required parameter: $key\n");
		}
	}

	# Validate value for some fields
	unless ( exists $H_KINGDOM { $h_args { kingdom } } )
	{
		throw LipmError::ParameterException ('kingdom','value', -text => "kingdom " . $h_args { kingdom } . " NOT SUPPORTED (only " . join(' ',keys(%H_KINGDOM)) . " SUPPORTED)\n");
	}
	unless ( -d $h_args { -tmpdir } )
	{
		throw LipmError::ParameterException ('-tmpdir','value', -text => "Directory " . $h_args { -tmpdir } . " DOES NOT EXIST");
	}
	unless ( $h_args { -consistency_limit } =~ /^\d+$/ )
	{
		throw LipmError::ParameterException ('-consistency_limit','value', -text => "Consistency limit must be a number ! Not " . $h_args { -consistency_limit });
	}

	unless ( !defined($h_args { -codon_usage_table }) || $h_args { -codon_usage_table } =~ /^\d+$/ )
	{
		throw LipmError::ParameterException ('-codon_usage_table','value', -text => "Codon usage table must be a number ! Not " . $h_args { -codon_usage_table });
	}

	# Init the private fields
	foreach my $key ( qw( kingdom species -with_sequence_embedded -with_partial -contig -chromosome -o_assembly -o_reassemble_assembly -build_assembly -disassemble_assembly -renumber_lt -filter_small_utr -filter_bad_proteins -new_rep_origin -create_utr -tmpdir -debug -create_cds -consistency_limit -codon_usage_table -cds_overlap_allowed  -desc_file -header_file -submit -compute_gaps -keep_gap_size -check_gaps -output_utr -output_intron -noclean_fasta_header ) )
	{
		my $k = $key;
		$k =~ s/-//;
		$k = '__' . $k;
		$self -> { $k } = $h_args{$key} if ( defined($h_args{$key}) &&  $h_args{$key} ne '' );
	}
	
    # An array of Locus::Gene objects and an index to those objects
	$self -> { __a_genes } = [];
	
	# An index to the Locus::Gene objects
	# key = locus tag
	# val = Index of the FIRST gene found with this locus tag in the array of genes
	$self -> { __idx_genes } = {};
	
	# An index of mrnas objects:
	# key = locus tag
	# val = (ind_g,ind_r)(ind_g,ind_r)...
	#       where ind_g is the index of the gene parent of the mrna, ind_r is the index of the mrna found in the gene
	#       There may be several mrnas with the same locus tag
	$self -> { __idx_mrnas } = {};

	# An array of Locus::Ncrna objects
	$self -> { __a_ncrnas } = [];
	
	# An array of Locus:Repeat objects
	$self -> { __a_repeats }  = [];

	# An array of Locus:MiscRegion objects
	$self -> { __a_regions }  = [];

	# The genomic sequence, see LoadGenomicSequence and GetGenomicSequence
	$self -> { __genomic_sequence } = undef;

	# Call setinput/SetOutput, validation input-output parameters
	$self->SetInput(@_);
	$self->SetOutput(@_);

	# Try to automatically detect the contig, unless passed by parameter
	if ( ! defined $self -> { __contig } && defined $self -> { __o_reader } )
	{
		$self -> { __o_reader } -> DetectContig();
	}
	
	# The __narcisse_id is an internal id, specific to each object - The narcisse id here is the species+contig id
	#$self -> { '__narcisse_id'} = $self->GetSpecies().'_s'.$self->GetContigLabel();
	my $species = $self->GetSpecies();
	if ( $species ne &AnnotatedSeq::UNKNOWN )
	{
		$self -> { '__narcisse_id'} = $self->GetSpecies().'_s'.$self->GetChromosomeLabel();
	}
	else
	{
		$self -> { '__narcisse_id'} = 's'.$self->GetChromosomeLabel();
	}
   
	# The sequence_region start and end
	# Should be updated by gff3 sequence-region, gbk source, etc. 
	# Is ALSO updated by Locus::SetParent
	# NOTE - __sequence_region_start is initialized to 1, so that ALL SEQUENCE REGIONS SHOULD START AT 1
	$self -> { __sequence_region_start } = 1;
	$self -> { __sequence_region_end   } = 0;
	
	# Read the headers from the cfg file, if specified
	if ( defined $self->{__header_file} )
	{
		my $o_headers = New ParamParser( $self->{__header_file} );
		$self->{__o_header} = $o_headers;
	}
	
	# keep headers in a specific hash
	$self -> {__rh_headers} = {};
	
	return $self;
}

sub GetCreateCds { my $self = shift; return $self -> { __create_cds }; }
sub GetDescFile  { my $self = shift; return (defined $self -> { __desc_file }) ? $self -> { __desc_file }: ""; }
sub GetHeaders   { my $self = shift; return $self -> { __rh_headers }; };

sub SetHeaderData
{
	my $self = shift;
	my ( $key,$value) = @_;
	$self -> {__rh_headers} -> { $key } = $value;
}

# Dummy procedure, see Locus for a really useful procedure
sub AdjustBoundaries {};

=head2 SetInput
 Title     : SetInput
 Usage     : $o_annot->SetInput ( -in_ann_file => 'data.gff3' );
 Procedure : Validate the parameters and reset the input parameters (4 optional parameters), calling __AnnotatedSeqFileReaderFactory
=cut
sub SetInput
{
	my $self = shift;
	my %h_args = (	
		-in_ann_file => "",
		-in_ann_fmt  => "",
		-in_seq_file => "",
		-in_seq_fmt  => "",
		-contig      => undef,
		@_);

	unless ( exists $self -> { __o_reader } )
	{
		$self -> { __o_reader } = undef;
	}
	###return if ( $h_args{ -in_ann_file } eq "" );
	
	# Guess the file formats if not specified
	if ($h_args{-in_ann_fmt} eq "" || $h_args{-in_seq_fmt} eq "")
	{
		my $o_fmt = New FileFormats;
		$h_args{-in_ann_fmt}=$o_fmt -> GuessFormat($h_args{-in_ann_file}) if ($h_args{-in_ann_file} ne "" && $h_args{-in_ann_fmt} eq "");
		$h_args{-in_seq_fmt}=$o_fmt -> GuessFormat($h_args{-in_seq_file}) if ($h_args{-in_seq_file} ne "" && $h_args{-in_seq_fmt} eq "");
	}
	
	# Validate value for some fields
	if (  $h_args { -in_ann_file ne "" } && !exists $H_IN_ANN_FMT -> { $h_args { -in_ann_fmt } } )
	{
		throw LipmError::ParameterException ('-in_ann_fmt','value', -text => "Annotation Input format " . $h_args { -in_ann_fmt } . " NOT SUPPORTED (only " . join(',',keys(%$H_IN_ANN_FMT)) . ") SUPPORTED\n");
	}
	if (  $h_args { -in_seq_file ne "" } && !exists $H_IN_SEQ_FMT -> { $h_args { -in_seq_fmt } } )
	{
		throw LipmError::ParameterException ('-in_seq_fmt','value', -text => "Sequence Input format " . $h_args { -in_seq_fmt } . " NOT SUPPORTED (only " . join(',',keys(%$H_IN_ANN_FMT)) . ") SUPPORTED\n");
	}
	
	foreach my $key ( qw( -in_ann_file -in_ann_fmt -in_seq_file -in_seq_fmt -contig ) )
	{
		my $k = $key;
		$k =~ s/^-/__/;
		$self -> { $k } = $h_args { $key } if ( defined $h_args { $key } );
	}
	
	$self -> __AnnotatedSeqFileReaderFactory();
}

=head2 SetOutput
 Title     : SetOutput
 Usage     : $o_annot->SetOutput ( -out_ann_file => 'data.gff3' );
 Procedure : Validate the parameters and reset the output parameters (4 optional parameters), calloutg __AnnotatedSeqFileWriterFactory
 Args      : -out_ann_file       The output annotation file name
             -out_ann_fmt[GFF3]  The output annotation file format - Must be a key of %H_OUT_ANN_FMT
             -out_seq_file       The output sequence file name, if any
             -out_seq_fmt        The output sequence file format (currently only FASTA)
=cut
sub SetOutput
{
	my $self = shift;
	my %h_args = (	
		-out_ann_file  => "",
		-out_ann_fmt  => "",
		-out_seq_file  => "",
		-out_seq_fmt  => "",
		@_);
	
	unless ( exists $self -> { __o_writer } )
	{
		$self -> { __o_writer } = undef;
	}
	#return if ( $h_args{ -out_ann_file } eq "" );

	foreach my $key ( qw( -out_ann_file -out_ann_fmt -out_seq_file -out_seq_fmt ) )
	{
		my $k = $key;
		$k    =~ s/^-/__/;
		$self -> { $k } = $h_args { $key };
	}
	
	# Guess the file formats if not specified
	if ( $h_args{-out_ann_file} ne "")
	{
		if ($h_args{-out_ann_fmt} eq "")
		{
			my $o_fmt = New FileFormats;
			$self -> {__out_ann_fmt}=$o_fmt -> GuessFormat($self->{__out_ann_file});
			#$h_args{-out_seq_fmt}=$o_fmt -> GuessFormat($h_args{-out_seq_file}) if ($h_args{-out_seq_fmt} eq "");
		}

		# Validate value 
		if ( !exists $H_OUT_ANN_FMT->{$self->{__out_ann_fmt}})
		{
			throw LipmError::ParameterException ('-out_ann_fmt','value', -text => "Output format " . $self->{ __out_ann_fmt } . " NOT SUPPORTED (only " . join(',',keys(%$H_OUT_ANN_FMT)) . ") SUPPORTED\n");
		}

		$self -> __AnnotatedSeqFileWriterFactory();

	}
	
	# resetting the list of dumped files
	delete $self -> { __rh_dumped }; 
	
	# call SetOutput for every embedded contigs, if any (see the switch --disassemble-assembly)
	#if ( $self -> { __disassemble_assembly } == &TRUE )
	#{
	#	$self -> __CreateDisassembledContigs();
	#	my $ra_ann = $self -> { __ra_disassembled_contigs };
	#	for my $o_ann (@$ra_ann)
	#	{
	#		$o_ann -> SetOutput ( %h_args );
	#	}
	#}
}

=head2 __AnnotatedSeqFileWriterFactory

 Title    : __AnnotatedSeqFileWriterFactory
 Usage    : my $o_writer = $this -> &__AnnotatedSeqFileWriterFactory();
 Procedure: Build a correct AnnotatedSeqFileWriter
 Args     : None
=cut
sub __AnnotatedSeqFileWriterFactory
{
	my $self = shift;
	my $o_format = New FileFormats;

	my $out_ann_fmt = $self -> { __out_ann_fmt };
	
	my $o_writer_class = "AnnotatedSeqFileWriter::" . $o_format->GetClassName($out_ann_fmt);
	eval "use $o_writer_class";
	throw Error::Simple( $@ ) if ( $@ ne "" );
	my $o_writer       = eval "New $o_writer_class( \$self )";
	throw Error::Simple( $@ ) if ( $@ ne "" );
	$self -> { __o_writer } = $o_writer;
}

=head2 __AnnotatedSeqFileReaderFactory

 Title    : __AnnotatedSeqFileReaderFactory
 Usage    : my $o_reader = $this -> &__AnnotatedSeqFileReaderFactory();
 Procedure: Build a correct AnnotatedSeqFileReader
 Args     : None
=cut
sub __AnnotatedSeqFileReaderFactory
{
	my $self = shift;
	my $o_format   = New FileFormats;
	my $in_ann_fmt = $self -> { __in_ann_fmt };
	if ( $in_ann_fmt ne '' )
	{
		my $o_reader_class = "AnnotatedSeqFileReader::" . $o_format -> GetClassName($in_ann_fmt);
		eval "use $o_reader_class";
		throw Error::Simple( $@ ) if ( $@ ne "" );
		my $o_reader       = eval "New $o_reader_class( \$self )";
		throw Error::Simple( $@ ) if ( $@ ne "" );
		$self -> { __o_reader } = $o_reader;
	}
}

=head2 Some accessors
=cut
sub GetFormat()
{
	throw LipmError::ObjectException( 'AnnotatedSeqFile', 'abstract method' );
}
sub GetDebug
{
	my $self = shift;
	return $self -> { __debug };
}
sub GetWithSequenceEmbedded
{
	my $self = shift;
	return $self -> { __with_sequence_embedded };
}
sub GetWithPartial
{
	my $self = shift;
	return $self -> { __with_partial };
}
sub GetKingdom
{
	my $self = shift;
	return $self -> { __kingdom } if (defined $self -> { __kingdom });
	throw LipmError::ObjectException('AnnotatedSeqFile', '!attribute', -text => '__kingdom not initialized');
}
sub GetSpecies
{
	my $self = shift;
	return $self -> { __species } if (defined $self -> { __species });
	throw LipmError::ObjectException('AnnotatedSeqFile', '!attribute', -text => '__species not initialized');
}

sub GetContigLabel
{
	my $self = shift;
	return (defined $self -> { __contig }) ? $self -> { __contig } : '';
#	throw LipmError::ObjectException('AnnotatedSeqFile', '!attribute', -text => '__contig not initialized');
}

sub GetSubmit
{
	my $self = shift;
	return $self -> { __submit };
}
sub GetOutputIntron
{
	my $self = shift;
	return $self -> { __output_intron };
}
sub GetOutputUtr
{
	my $self = shift;
	return $self -> { __output_utr };
}

sub GetNoCleanFastaHeader
{
	my $self = shift;
	return $self -> { __noclean_fasta_header };
}

sub GetChromosomeLabel
{
	my $self = shift;
	my $chromosome_label = '';
	if ( defined $self -> { __o_assembly } )
	{
		$chromosome_label = $self -> {__o_assembly} -> GetChromosomeLabel();
	}
	if ( $chromosome_label eq '' && defined ($self -> { __chromosome }) )
	{
		$chromosome_label = $self -> { __chromosome };
	}
	if ( $chromosome_label eq '' )
	{
		$chromosome_label = $self -> GetContigLabel();
	}
	return $chromosome_label;
}

sub GetCodonTable
{
	my $self = shift;
	if ( defined($self -> { __codon_usage_table }))
	{
		return $self -> { __codon_usage_table };
	}
	else
	{
		return ( $self -> GetKingdom() eq 'bacteria' ) ? 11 : 1;
	}
}

sub GetNarcisseId
{
	my $self = shift;
	return $self -> { __narcisse_id };
}

=head2 SetDebug/UnsetDebug
=cut
sub SetDebug
{
	my $self = shift;
	$self -> { __debug } = &TRUE;
}
sub UnsetDebug
{
	my $self = shift;
	$self -> { __debug } = &FALSE;
}

=head2 function AddGene

 Title        : AddGene
 Usage        : $o_annot -> AddGene($o_gene)
 Function     : Add a new protein coding gene to this object:
                    - update the index
                    - update the gene parent and the gene narcisseId
 Returns      : the added gene
 Args         : $o_gene: a Locus::Gene object

=cut

sub AddGene
{
	my($self,$o_gene) = @_;
	throw LipmError::ParameterException( 'o_gene', 'type', -text => 'should be a Locus::Gene object instead of a ' . ref($o_gene) ) unless ref $o_gene eq 'Locus::Gene';

	# Put the gene inside the __a_genes array, and update the index	
	my $ra_genes = $self -> { __a_genes };
	push @$ra_genes, $o_gene;
	
	my $locus_tag = $o_gene->GetLocusTag();

	my $rh_index = $self -> { __idx_genes };
	if ( !defined $rh_index -> { $locus_tag } )
	{
		$rh_index -> { $locus_tag } = [];
	}
	push ( @{$rh_index -> { $locus_tag }}, $o_gene );

	# Set the gene parent and compute the Narcisse Id
	$o_gene -> SetParent ( $self );

	return $o_gene;
}

=head2 function AddGeneMRNA

 Title        : AddGeneMRNA
 Usage        : &AddGeneMRNA($o_mrna,$locus_tag)
 Prerequisite : none
 Function     : Based on the locus_tag and the coordinates, search the parent gene in this object.
                If found, add a new transcript to the gene and keep this transcript to the mrnas index
 Returns      : $o_locus if OK
 Error        : Throwing an error if no gene found
 Args         : $o_mrna   : A Locus::Mrna object
                $locus_tag: the locus_tag value of its parent (gene)
 Globals      : none

=cut

sub AddGeneMRNA
{
	my($self,$o_mrna,$locus_tag) = @_;
	
	my $o_gene = $self -> SearchGene( $o_mrna, $locus_tag );
	if ( ! defined $o_gene )
	{
		my ($s, $e) = $o_mrna -> GetEnds();
		throw Error::Simple "ERROR - Trying to add the mRNA $s..$e to a gene: the gene $locus_tag was not found\n";
	}
	
	$o_gene -> AddmRNA( $o_mrna );
	
	my $rh_index       = $self -> { __idx_mrnas };
	my $mrna_locus_tag = $o_mrna -> GetLocusTag();
	if ( !defined $rh_index -> { $mrna_locus_tag } )
	{
		$rh_index -> { $mrna_locus_tag } = [];
	}
	push ( @{$rh_index -> { $mrna_locus_tag }}, $o_mrna );
	return $o_mrna;
}

=head2 function AddGeneCDSorExon

 Title        : AddGeneCDSorExon
 Usage        : $o_annot -> AddGeneCDSorExon($o_region,$locus_tag)
 Prerequisite : The "parent" of the CDS/ExonSet must be the mRNA or the gene identified by $locus_tag (as in Genbank formatted files)
 Function     : look for a Mrna surrounding the CDS/ExonSet/Utr and call AddCDS/AddExon/AddUtr on it
                If not found, try to create a new Mrna, then call AddGeneMRNA and call AddCDS/AddExon/AddUtr to the new Mrna
 Returns      : $o_rmna    : The found or added mrna
 Error        : Throwing an error if no gene found
 Args         : $o_region  : a Locus::CdsRegion/Locus::Exon/Locus::UtrRegion object
                $locus_tag : the locus_tag value of the gene parent
 Globals      : none

=cut
sub AddGeneCDSorExon
{
	my($self,$o_region,$locus_tag) = @_;
	my $region_type;
	$region_type = 0 if (ref ( $o_region ) =~ /Locus::Cds/ );			# Cds or CdsRegion
	$region_type = 1 if (ref ( $o_region ) =~ /Locus::Exon/);			# Exon or ExonSet
	$region_type = 2 if (ref ( $o_region ) =~ /Locus::Utr/);			# Utr or UtrRegion

	my $utr_type = '';
	$utr_type = $o_region->GetSubType() if ( $region_type == 2);
	
	my $o_mrna = undef;
	my $i = 0;
	while ( $i++>=0 )   # Infinite loop, incrementing $i at each iteration
	{
		$o_mrna = $self -> SearchMRNA($o_region,$locus_tag,$i);
		
		# $o_mrna found
		if (defined $o_mrna)
		{
			# Adding Cds
			if ($region_type==0)
			{
				my $o_cds = $o_mrna -> GetCDS();
				
				# New region to add to this mrna
				if ( !defined($o_cds) || $o_cds->IsRegionOverlap($o_region)==&FALSE )
				{
					$o_mrna -> AddCDS($o_region);
					return $o_mrna;
				}
				
				# try again with next mrna, same locus_tag
				else
				{
					next;
				}
			}
			
			# Adding an exon
			elsif ($region_type==1)
			{
				my $o_exon_set = $o_mrna -> GetExonSet();

				# New region to add to this mrna
				if ( !defined($o_exon_set) || $o_exon_set->IsRegionOverlap($o_region)==&FALSE )
				{
					$o_mrna -> AddExon($o_region);
					return $o_mrna;
				}
				
				# try again with next mrna, same locus_tag
				else
				{
					next;
				}
			}
			
			# Adding an Utr
			elsif ($region_type==2)
			{
				my $o_utr = $o_mrna -> GetUtr($utr_type);

				# New region to add to this mrna
				if ( !defined($o_utr) || $o_utr->IsRegionOverlap($o_region)==&FALSE )
				{
					$o_mrna -> AddUtr($o_region);
					return $o_mrna;
				}
				
				# try again with next mrna, same locus_tag
				else
				{
					next;
				}
			}
		}
		
		# $o_mrna not found, exit from loop
		else
		{
			last;
		}
	}

	# Coming here: no mnra found, looking for a gene
	my $o_gene = $self -> SearchGene ( $o_region, $locus_tag );
	
	# gene not found, no more hope -> throw an error !
	if ( !defined $o_gene )
	{
		my ($s, $e) = $o_region->GetEnds();
		throw Error::Simple ( "ERROR - Could not insert this " . ref ( $o_region ) . " $locus_tag  [$s..$e] to an mRNA/gene - May be there is some overlab between regions ?" );
	}
	#return undef unless defined $o_gene;

	# gene found, we give him a new mrna
	my %h_param;
	$h_param { locus_tag } = $locus_tag;
	$o_mrna = New Locus::Mrna ( %h_param, $o_gene );
	my $o_added_mrna = $self -> AddGeneMRNA ( $o_mrna, $locus_tag );
	
	$o_added_mrna -> AddCDS  ( $o_region ) if ($region_type==0); 
	$o_added_mrna -> AddExon ( $o_region ) if ($region_type==1); 
	$o_added_mrna -> AddUtr  ( $o_region ) if ($region_type==2); 

	warn $self -> GetWarningHeader() . $o_region->GetpkExtended() . " A new mrna was created from a cds, an exon or an utr\n" if ($self -> GetDebug());
	
	return $o_added_mrna;
}

=head2 Procedure AddNcrna

 Title        : AddNcrna
 Usage        : $o_annot -> AddGene($o_ncnra)
 Procedure    : Add a new ncRNA_gene
                If an ncRNA_gene is found in the index, call the gene's AddNcrna method
                Else, set the rncna's parent to $self and add the ncRNA to the index
 Args         : $o_ncrna: a Locus::Ncrna object
                $locus_tag (optional): the locus_tag value of its parent

=cut

sub AddNcrna
{
	my $self = shift;
	my ($o_ncrna,$locus_tag) = @_;
	throw LipmError::ParameterException( 'o_ncrna', 'type', -text => 'should by a Locus::Ncrna object instead of a ' . ref($o_ncrna) ) unless ref $o_ncrna eq 'Locus::Ncrna';

	my $ra_ncrnas = $self -> { __a_ncrnas };

	# Set the parent and compute the Narcisse Id
	if (defined $locus_tag)
	{
		my $o_gene = $self -> SearchGene( $o_ncrna, $locus_tag );
		if ( defined ($o_gene) )
		{
			$o_gene -> AddNcrna( $o_ncrna );
		}
		else
		{
			warn $self -> GetWarningHeader() ."the ncRNA_gene $locus_tag was not found, storing the Ncrna region without a ncRNA_gene (you can ignore this warning)\n";
			push @$ra_ncrnas, $o_ncrna;
			$o_ncrna -> SetParent($self);
		}
	}
	else
	{	
		push @$ra_ncrnas, $o_ncrna;
		$o_ncrna -> SetParent ( $self );
	}
}

=head2 functions AddRepeat,AddMiscRegion

 Title        : AddRepeat/AddMiscRegion
 Usage        : $o_annot -> AddRepeat($o_repeat)
 Function     : Add a new Repeat/MiscRegion
 Returns      : the number of Repeat/MiscRegion objects already stored
 Args         : a Locus::Repeat/Locus::MiscRegion object

=cut

sub AddMiscRegion { return __AddLocus(@_);};
sub AddRepeat     { return __AddLocus(@_);};

=head2 function __AddLocus

 Title        : __AddLocus
 Usage        : Called by AddRepeat/AddMiscRegion
 Function     : Add a new Repeat/MiscRegion
 Returns      : the number of Repeat/MiscRegion objects already stored
 Args         : $o_locus   : the object to add

=cut
sub __AddLocus
{
	my($self,$o_locus) = @_;
	my $locus_type = ref($o_locus);
	throw LipmError::ParameterException( $locus_type, 'value', -text => "should be Locus::Repeat or Locus::MiscRegion" ) unless ($locus_type eq 'Locus::MiscRegion' || $locus_type eq 'Locus::Repeat');

	my $ra_loci = ($locus_type eq 'Locus::Repeat') ? $self -> { __a_repeats } : $self -> { __a_regions };
	push @$ra_loci, $o_locus;

	# Set the parent and compute the Narcisse Id
	$o_locus -> SetParent ( $self );
	
	return $#$ra_loci;
}

=head2 function GetNbRepeats

 Title        : GetNbRepeats
 Usage        : my $nb_rep = &GetNbRepeats()
 Prerequisite : none
 Function     : returns the number of repeats loaded
 Returns      : the number of repeats loaded
 Args         : none
 Globals      : none

=cut

sub GetNbRepeats
{
	my ($self) = @_;
	my $ra_repeats = $self -> { __a_repeats };
	return @$ra_repeats;
}

sub GetNbMiscRegions
{
	my ($self) = @_;
	my $ra_regions = $self -> { __a_regions };
	return @$ra_regions;
}
=head2 function GetNbNcrnas

 Title        : GetNbNcrnas
 Usage        : my $nb_ncrnas = &GetNbNcrnas()
 Function     : returns the number of non protein coding Ncrnas loaded
 Returns      : the number of non protein coding Ncrnas
 Args         : none
 Globals      : none

=cut

sub GetNbNcrnas
{
	my ($self) = @_;
	my $ra_ncrnas = $self -> { __a_ncrnas };
	return scalar( @$ra_ncrnas );
}

=head2 function GetNbGenes

 Title        : GetNbGenes
 Usage        : (1) my $nb_genes   = $o_annot -> GetNbGenes()
                (2) my @a_nb_genes = $o_annot -> GetNbGenes()
 Function     : returns the number of genes
 Returns      : (1) return the total number of genes
                (2) return the number of gene, protein_codign gene, ncRNA_gene

=cut

sub GetNbGenes
{
	my ($self) = @_;
	my $ra_genes = $self -> { __a_genes };
	if ( ! wantarray )
	{
		return scalar( @$ra_genes );
	}
	else
	{
		my ($nb_o_genes,$nb_p_genes,$nb_r_genes) = (0,0,0);
		for my $o_gene (@$ra_genes)
		{
			next unless defined $o_gene;
			$nb_o_genes++ if ( $o_gene->GetSubType() eq 'gene' );
			$nb_p_genes++ if ( $o_gene->GetSubType() eq 'protein_coding_gene' );
			$nb_r_genes++ if ( $o_gene->GetSubType() eq 'ncRNA_gene' );
		}
		return ($nb_o_genes,$nb_p_genes,$nb_r_genes)
	}
}

=head2 function GetNbProteinSequences

 Title        : GetNbProteinSequences
 Usage        : my $nb_prot = $o_annot -> GetNbProteinSequences();
 Function     : returns the number of proteins translated from the protein-coding genes
                Warning: Some genes are NOT translated because of consistency problems
 Returns      : the number of protein sequences
 Args         : none
 Globals      : none

=cut

sub GetNbProteinSequences
{
	my ($self) = @_;
	my $o_annot_writer = $self -> { __o_writer };
	return $o_annot_writer -> GetFastaNbSequences ( 'prot' );
}

=head2 function GetNbGeneSequences

 Title        : GetNbGeneSequences
 Usage        : my $nb_genes = $o_annot -> GetNbGeneSequences();
 Function     : returns the number of gene sequences kept in the database
 Returns      : the number of gene sequences
 Args         : none
 Globals      : none

=cut

sub GetNbGeneSequences
{
	my ($self) = @_;
	my $o_annot_writer = $self -> { __o_writer };
	return $o_annot_writer -> GetFastaNbSequences ( 'gene' );
}

=head2 function GetNbCdsSequences

 Title        : GetNbCdsSequences
 Usage        : my $nb_cds = $o_annot -> GetNbCdsSequences();
 Function     : returns the number of cds sequences kept in the database
 Returns      : the number of cds sequences
 Args         : none
 Globals      : none

=cut

sub GetNbCdsSequences
{
	my ($self) = @_;
	my $o_annot_writer = $self -> { __o_writer };
	return $o_annot_writer -> GetFastaNbSequences ( 'cds' );

}

=head2 function GetNbTranscriptSequences

 Title        : GetNbTranscriptSequences
 Usage        : my $nb_prot = $o_annot -> GetNbTranscriptSequences();
 Function     : returns the number of transcripts kept in the database
 Returns      : the number of transcripts
 Args         : none
 Globals      : none

=cut

sub GetNbTranscriptSequences
{
	my ($self) = @_;
	my $o_annot_writer = $self -> { __o_writer };
	return $o_annot_writer -> GetFastaNbSequences ( 'mrna' );
}

=head2 function GetNbTranscripts

 Title        : GetNbTranscripts
 Usage        : my $nb_prot = &GetNbTranscripts()
 Function     : returns the number of transcripts 
 Returns      : the number of transcripts
 Args         : none
 Globals      : none

=cut

sub GetNbTranscripts
{
	my ($self) = @_;
	my $nb_mrnas = 0;
	my $ra_genes = $self -> GetGenes();
	foreach my $o_gene (@$ra_genes)
	{
		if ( $o_gene->GetSubType() eq 'protein_coding_gene' )
		{
			my $ra_mrnas = $o_gene -> GetMrnas();
			$nb_mrnas += @$ra_mrnas;
		}
	}
	return $nb_mrnas;
}

=head2 function GetNbExons

 Title        : GetNbExons
 Usage        : my $nb_prot = &GetNbExons()
 Function     : returns the number of exons found - This is more than the number of proteins
 Returns      : ($nb_cds_regions,$nb_exons) the number of CDS regions AND the number of exons
 Args         : none
 Globals      : none

=cut

sub GetNbExons
{
	my ($self) = @_;

	my $nb_cds_regions  = 0;
	my $nb_exons = 0;
	my $ra_genes = $self -> GetGenes();
	foreach my $o_gene (@$ra_genes)
	{
		if ( $o_gene->GetSubType() eq 'protein_coding_gene' )
		{
			my $ra_mrnas = $o_gene -> GetMrnas();
			foreach my $o_mrna (@$ra_mrnas)
			{
				my $o_cds = $o_mrna -> GetCDS();
				if ( defined $o_cds )
				{
					my @a_cds_regions = $o_cds -> GetRegions();
					$nb_cds_regions += @a_cds_regions;
				}
				my $o_exonset= $o_mrna -> GetExonSet();
				if ( defined $o_exonset )
				{
					my @a_exons = $o_exonset -> GetRegions();
					$nb_exons += @a_exons;
				}
			}
		}
	}
	return ($nb_cds_regions,$nb_exons);
}

=head2 GetGenes,GetRepeats,GetNcrnas

 Title    : GetGenes,GetRepeats,GetNcrnas
 Function : Accessor to the array of genes/repeats/nc_rnas
 Usage    : my $ra_genes = $o_annot -> GetGenes();
 Arg      : None
 Return   : A ref to an array of Gene/Repeat/Ncrna objects
 
=cut
sub GetGenes
{
	my $self = shift;
	return $self -> { __a_genes };
}
sub GetRepeats
{
	my $self = shift;
	return $self -> { __a_repeats };
}
sub GetMiscRegions
{
	my $self = shift;
	return $self -> { __a_regions };
}
sub GetNcrnas
{
	my $self = shift;
	return $self -> { __a_ncrnas };
}

=head2 function GetSortedLoci

 Title        : GetSortedLoci
 Usage        : my $ra_objects = $o_annot->GetSortedObjects();
 Function     : Return loci (genes, ncrnas, repeat_regions, misc regions), sorted in start positions
 Returns      : A ref to an array of Locus:: objects
 Args         : $with_exons (default &FALSE): If &TRUE, returns also the exons

=cut
sub GetSortedLoci
{
	my $self = shift;
	my ( $with_exons ) = @_;
	$with_exons = &FALSE unless ( defined $with_exons );
	
	my @a_objects;
	push @a_objects, @{$self -> { __a_genes }};
	push @a_objects, @{$self -> { __a_ncrnas }};
	push @a_objects, @{$self -> { __a_repeats }};
	push @a_objects, @{$self -> { __a_regions }};
	if ($with_exons == &TRUE )
	{
		foreach my $o_gene ( @{$self -> { __a_genes }} )
		{
			next if ($o_gene -> GetSubType() ne 'protein_coding_gene');
			foreach my $o_mrna ( @{$o_gene -> GetMrnas()} )
			{
				my $o_exon_set = $o_mrna -> GetExonSet();
				push (@a_objects, $o_exon_set -> GetRegions()) if ( defined $o_exon_set );
			}
		}
	}
	my @a_sorted_objects = sort { $a->GetStart() <=> $b->GetStart() } @a_objects;
	return \@a_sorted_objects;
}

=head2 function GetAllLoci

 Title        : GetAllLoci
 Usage        : my $ra_objects = $o_annot->GetAllLoci();
 Function     : Call GetSortedLoci, but return also the children of the top-level loci
 Returns      : A ref to an array of Locus:: objects
 Args         : none

=cut
sub GetAllLoci
{
	my $self = shift;
	
	my $ra_top_loci = $self -> GetSortedLoci();
	my @a_all_loci;
	foreach my $o_locus ( @$ra_top_loci )
	{
		push @a_all_loci, $o_locus, @{$o_locus -> GetAllChildren()};
	}
	return \@a_all_loci;
}

=head2 function GetGene

 Title        : GetGene
 Usage        : my $o_gene = $o_annot->GetGene( $index );
 Function     : Return the gene found at the index $index
 Returns      : The gene
 Args         : The index

=cut

sub GetGene
{
	my $self = shift;
	my ($index) = @_;
	return $self -> { __a_genes } -> [ $index ];
}
sub GetRepeat
{
	my $self = shift;
	my ($index) = @_;
	return $self -> { __a_repeats } -> [ $index ];
}
sub GetMiscRegion
{
	my $self = shift;
	my ($index) = @_;
	return $self -> { __a_regions } -> [ $index ];
}
sub GetNcrna
{
	my $self = shift;
	my ($index) = @_;
	return $self -> { __a_ncrnas } -> [ $index ];
}

=head2 procedure Print

 Title        : Print
 Usage        : &Print($level);
 Prerequisite : none
 Function     : Print the data structure
 Returns      : the number of protein coding genes
 Args         : $level: if 'full' then a full report of the datastructure is printed else only a short summary
 Globals      : none

=cut

sub Print
{
	my $self = shift;
	my ($level) = @_;

	$level = '' if ( !defined($level));
	print STDOUT "contig   :\t".$self->{'__contig'}."\n";
	print STDOUT "infile   :\t".(defined ($self->{'__infile'}) ? $self->{'__infile'} : "NO ANNOTATION FILE DEFINED")."\n";
	print STDOUT "informat :\t".$self->GetFormat()."\n";
	print STDOUT "tmpdir   :\t".$self->{'__tmpdir'}."\n";
	print STDOUT "outformat:\t".$self->{'__outfmt'}."\n";

	my $nb_genes = $self->GetNbGenes();
	print STDOUT $nb_genes." protein coding genes.......\n";
	for(my $i=0; $level eq 'full' && $i<$nb_genes; $i++)
	{
		$$self{'__a_genes'}[$i]->Print();
	}
	
	my $nb_ncrnas = $self->GetNbNcrnas();
	print STDOUT $nb_ncrnas." non protein coding genes.......\n";
	for(my $i=0; $level eq 'full' && $i<$nb_ncrnas; $i++)
    {
		$$self{'__a_ncrnas'}[$i]->Print();
    }
    
    my $nb_repeats = $self-> GetNbRepeats();
	print STDOUT $nb_repeats." repeat regions.......\n";
	for(my $i=0; $level eq 'full' && $i<$nb_repeats; $i++)
    {
		$$self{'__a_repeats'}[$i]->Print();
    }
}

=head2 procedure LoadHeaders

 Title        : LoadHeaders
 Function     : Read the headers from $o_header, if defined
 Usage        : $self->LoadHeaders();
                Must be called AFTER $self->Load(), because Load clears the headers
 Args         : none

=cut

sub LoadHeaders
{
	my $self = shift;
	if ( defined $self -> { __o_header } )
	{
		my $o_header  = $self -> { __o_header };
		my $rh_headers= $self -> { __rh_headers };
		foreach my $key ($o_header->GetKeys )
		{
			$rh_headers -> { $key } = $o_header -> Get ( $key ) unless defined ($rh_headers -> { $key });
		}
	}
}

=head2 procedure __ComputeAssemblyGaps

 Title        : __ComputeAssemblyGaps
 Function     : If an Assembly object is alive, add MiscRegion objects, 1 for each assembly gap
 Usage        : $self->__ComputeAssemblyGaps()
 Args         : none

=cut
sub __ComputeAssemblyGaps
{
	my $self = shift;
	if ( defined $self -> { __o_assembly } )
	{
		my $o_assembly = $self -> { __o_assembly };

		# If the chromosome is not defined at this stage, consider it is the same as the contig label
		my $chrom = $o_assembly -> GetChromosomeLabel();
		if ( !defined $chrom || $chrom eq '')
		{
			$o_assembly -> SetCurrentChromosome($self->GetContigLabel());
		}
		
		# Look for MiscRegions with sub_type assembly_gap and deselect them, avoiding duplication
		my $ra_regions = $self -> GetMiscRegions();
		for my $o_region (@$ra_regions)
		{
			$o_region -> Deselect() if ( $o_region -> GetSubType() eq 'assembly_gap' );
		}
		
		# We clear the locus assembly, because the assembly gap positions is counted from the chromosome origin.
		# First, keep track of the current assembly object ($o_assembly OR undef)
		my $o_locus_assembly = &Locus::GetAssembly();
		&Locus::ClearAssembly();
		
		my $ra_assembly_gaps = $o_assembly -> GetAssemblyGaps();
		foreach my $rh_assembly_gap (@$ra_assembly_gaps)
		{
			my $s = $rh_assembly_gap -> { object_beg };
			my $e = $rh_assembly_gap -> { object_end };
			my $gt= $rh_assembly_gap -> { gap_type };
			my $le= $rh_assembly_gap -> { linkage_evidence };
			my $lg= ($rh_assembly_gap -> { component_type } eq 'U' ? 'unknown' : $rh_assembly_gap -> { gap_length });
			my %h_param = ( start => $s, end => $e, strand => 1, sub_type => 'assembly_gap', gap_type => $gt, estimated_length=> $lg);
			if ( defined ($le) )
			{
				$h_param{ linkage_evidence } = $le;
			}
			#SC: assembly_gap pas accepte dans annotation --> gap en mode submit
			if ($self -> {__submit} == &TRUE)
			{
				%h_param = ( start => $s, end => $e, strand => 1, sub_type => 'gap', estimated_length=> $rh_assembly_gap -> { gap_length });
			}
			my $o_region = New Locus::MiscRegion( %h_param );
			$self -> AddMiscRegion( $o_region);
		}
		
		if ( defined($o_locus_assembly) )
		{
			&Locus::InitAssembly($o_locus_assembly);
		}
	}
}

=head2 procedure Load

 Title        : Load
 Function     : Call the reader's Load method
 Args         : none

=cut

sub Load
{
	my ($self) = @_;

	# Clear the disassembled contigs, if any
	delete $self -> { __ra_disassembled_contigs };
	
	# If building assembly, configure the Locus assembly object
	if ( $self -> { __build_assembly } == &TRUE )
	{
		my $o_assembly = $self -> { __o_assembly };
		&Locus::InitAssembly($o_assembly);
	}
	else
	{
		&Locus::ClearAssembly();
	}
	
	if (defined $self -> { __o_reader })
	{
		# clear the headers
		$self -> { __rh_headers } = {};
		$self -> { __o_reader } -> Load();
		
		if ( defined ($self -> { __fatal_error }) )
		{
			throw Error::Simple ( "FATAL ERROR OCCURRED WHILE LOADING DATA - PLEASE HAVE A LOOK TO THE IMPORTANT WARNINGS" );
		}
	}
}

=head2 procedure PostLoadingTasks

 Title        : PostLoadingTasks
 Prerequisite : Load should already have been called
 Function     : Should be called AFTER Load (but may be after several calls to Load), executes some post loding tasks
                Some tasks are always called, while other tasks are called only if the corresponding switch is on
 Args         : none

=cut

sub PostLoadingTasks
{
	my ($self) = @_;

	# Load Headers from cfg file, may be replacing the headers already read from the data file
	$self -> LoadHeaders();

	# Load assembly_gaps now from agp file if building the assembly
	if ($self -> { __build_assembly } == &TRUE)
	{
		$self -> __ComputeAssemblyGaps();
	}
	
	# If wanted, relocate the replication origin
	if ( defined $self-> { __new_rep_origin } )
	{
		my $rep_origin = $self->{ __new_rep_origin  };
		$self -> __NewRepOrigin($rep_origin);
	}
	
	# If wanted, renumber the locus tags
	if ( defined $self->{ __renumber_lt } )
	{
		my $lt_templ = $self->{ __renumber_lt };
		$self -> __RenumberLocusTags($lt_templ);
	}

	# All mrnas do not have an ExonSet defined, here we check and create a trivial exon when necessary
	$self->__CreateTrivialExon();

	# Unless switch --create-cds specified, remove exons for genes without cds 
	unless ( $self->{ __create_cds } == &TRUE )
	{
		$self->__FilterMrnasWithoutCds();
	}
	
	# Translate every CDS, may be modifying the consistency status of the CDS
	# Call SetConsistencyFromRegions for the CDS and the mrnas
	# Then call SetConsistencyFromParent for the first and last exon
	$self->__TranslateCDS();

	# If wanted, translate CDS then filter proteins whose protein sequence contains 'X'
	# The filter is done with setting the CDS as Inconsistent
	if ( defined $self->{__filter_bad_proteins} )
	{
		my $threshold = $self -> {__filter_bad_proteins};
		$self -> __FilterBadProteins ($threshold);
	}

	# Scan all the genes having only ONE mRNA
	# If the corresponding CDS is tagged as a fragment, try to propagate this information to:
	#  - the mRNA 
	#  - the gene
	# BUT propagate the information ONLY if the corresponding boundary is the same
	$self->__PropagateFragmentInformation();
	
	# If wanted, add utr when necessary
	if ($self->{__create_utr}==1 || $self->{__output_utr}==1 || $self->{__output_intron}==1 )
	{
		$self -> __CreateUtr ();
	}

	# If wanted, remove the UTR with legnth 1pb (this could be added by the previous sub)
	if ( $self->{__filter_small_utr}==1 || $self->{__create_utr}==1 )
	{
		$self -> __FilterSmallUtr ();
	}
	
	# If wanted, clear then compute the gaps
	if ( $self->{ __compute_gaps } == &TRUE )
	{
		my $keep=$self->{__keep_gap_size};
		$self -> __ComputeGaps($keep);
	}
	
	# If wanted, check the gaps
	if ( $self->{ __check_gaps } == &TRUE )
	{
		$self -> __CheckGaps();
	}
}

# post-treatments, those functions are called from PostLoadingTasks
=head __CreateTrivialExon

 Title     : __CreateTrivialExon
 Procedure : Create a single Exon if not already defined in each mran
 Args      : None
    
=cut
sub __CreateTrivialExon
{
	my $self = shift;
	
	my $ra_genes = $self -> GetGenes();
	foreach my $o_gene ( @$ra_genes )
	{
		next unless ( $o_gene -> GetSubType() eq 'protein_coding_gene' );
		my $ra_mrnas = $o_gene -> GetMrnas();
		foreach my $o_mrna (@$ra_mrnas)
		{
			my $o_exon_set = $o_mrna -> GetExonSet();
			if ( ! defined $o_exon_set )
			{
				my $o_exon = New Locus::Exon ( $o_mrna );
				$o_mrna -> AddExon ( $o_exon );
			}
		}
	}
}

=head __TranslateCDS

 Title     : __TranslateCDS
 Procedure : Translate every CDS, may be modifying the consistency status of the CDS
             Call SetConsistencyFromRegions for the CDS and the mrnas
             Then call SetConsistencyFromParent for the first and last exon
 Args      : None
    
=cut
sub __TranslateCDS
{
	my $self = shift;
	
	my $create_cds = $self -> GetCreateCds();
	my $ra_genes = $self -> GetGenes();
	foreach my $o_gene ( @$ra_genes )
	{
		next unless ( $o_gene -> GetSubType() eq 'protein_coding_gene' );
		my $ra_mrnas = $o_gene -> GetMrnas();
		foreach my $o_mrna ( @$ra_mrnas )
		{
			my $o_cds  = $o_mrna -> GetCDS($create_cds);
			if ( defined ($o_cds) )
			{
				$o_cds->GetTranslation();
				$o_cds->SyncConsistency();
			}
		}
	}
}

=head __PropagateFragmentInformation

 Title     : __PropagateFragmentInformation
 Procedure : Scan all the genes having only ONE mRNA
             If the corresponding CDS is tagged as a fragment, try to propagate this information to:
                - the mRNA 
                - the gene
             BUT propagate the information ONLY if the corresponding boundary is the same

 Args      : None
    
=cut
sub __PropagateFragmentInformation
{
	my $self = shift;

	my $ra_genes = $self -> GetGenes();
	foreach my $o_gene ( @$ra_genes )
	{
		next unless ( $o_gene -> GetSubType() eq 'protein_coding_gene' );
		my $ra_mrnas = $o_gene -> GetMrnas();
		
		# $alt_flg = &TRUE if several mrnas for this gene
		my $alt_flg = ( @$ra_mrnas > 1 );

		foreach my $o_mrna ( @$ra_mrnas )
		{
			my $o_cds  = $o_mrna -> GetCDS();
			next if ( !defined ($o_cds) || $o_cds -> GetConsistency() eq 'C' );
			
			my ( $g_start,$g_end,$r_start,$r_end,$c_start,$c_end);
			if ( $o_gene -> GetStrand() == -1 )
			{
				($g_end,$g_start) = $o_gene -> GetEnds();
				($r_end,$r_start) = $o_mrna -> GetEnds();
				($c_end,$c_start) = $o_cds  -> GetEnds();
			}
			else
			{
				($g_start,$g_end) = $o_gene -> GetEnds();
				($r_start,$r_end) = $o_mrna -> GetEnds();
				($c_start,$c_end) = $o_cds  -> GetEnds();
			}
			if ( $o_cds -> GetConsistency() =~ /FS/ )
			{
				if ($r_start == $c_start)
				{
					$o_mrna -> SetFragment('S');
					$o_mrna -> GetExonSet() -> SetFragment('S');
					$o_gene -> SetFragment('S') if ($alt_flg == &FALSE && $g_start == $c_start);
				}
			}
			if ( $o_cds -> GetConsistency() =~ /FE/ )
			{
				if ($r_end == $c_end)
				{
					$o_mrna -> SetFragment('E');
					$o_mrna -> GetExonSet() -> SetFragment('E');
					$o_gene -> SetFragment('E') if ($alt_flg == &FALSE && $g_end == $c_end);
				}
			}
			$o_mrna -> GetExonSet() -> SyncConsistency();
		}
	}
}

=head __ComputeGaps

 Title     : __ComputeGaps
 Usage     : $self->__ComputeGaps()
 Procedure : Deselect all MiscRegions with sub_type gap, and compute new 'gap' MiscRegions, reading the gaps from the sequence
 Args      : $keep: If $keep is &TRUE, keep the REAL gap size in estimated_length
    
=cut
sub __ComputeGaps
{
	my $self = shift;
	my ($keep) = @_;
	
	my $seq_length = $self -> GetLength();
	if ( $seq_length == 0 )
	{
		throw Error::Simple( "There is no sequence loaded, cannot compute gaps");
	}
	
	print "computing the gaps annotation from the sequence\n";
	
	# Deselect all gaps
	my $ra_loci = $self -> GetAllLoci();
	foreach my $o_locus (@$ra_loci)
	{
		if ( $o_locus->isa('Locus::MiscRegion') && $o_locus->GetSubType() eq 'gap')
		{
			$o_locus->Deselect();
		}
	}

	# Look for the gaps in the genomic sequence and create the gaps
	my $r_gseq = $self -> GetGenomicSequence();
	my $offset=0;
	my $pos=0;
	while(1)
	{
		#print "$pos\n";
		my $seq = substr $$r_gseq, $pos;
		if ( $seq =~ /[nN]+/p )
		{
			my $s = $offset + length(${^PREMATCH})+1;
			my $e = $s       + length(${^MATCH})-1;
			my $sz= ($keep)?$e-$s+1:'unknown';
			#print $s.'..'.$e."\n";
			my $o_region = New Locus::MiscRegion( start => $s, end => $e, strand => 1, sub_type => 'gap', estimated_length=> $sz );
			$self -> AddMiscRegion( $o_region);
			$pos = $e + 1;
			$offset = $pos;
		}
		else
		{
			last;
		}
	}
}

=head __CheckGaps

 Title     : __CheckGaps
 Usage     : $self->__CheckGaps()
 Procedure : Check the MiscRegions with sub_type 'gap' and if they are not in sync with the sequence, write to stderr the difference and throw an Error::Simple
 Args      : None
    
=cut
sub __CheckGaps
{
	my $self = shift;
	
	my $seq_length = $self -> GetLength();
	if ( $seq_length == 0 )
	{
		throw Error::Simple( "There is no sequence loaded, cannot check gap annotations");
	}
	
	print "Checking gap annotations against the sequence\n";
	
	# Find all annotated gaps and store their pk in a hash
	my %h_ann_gaps;
	my $ra_loci = $self -> GetAllLoci();
	foreach my $o_locus (@$ra_loci)
	{
		if ( $o_locus->isa('Locus::MiscRegion') && $o_locus->GetSubType() eq 'gap')
		{
			my $pk = $o_locus->Getpk();
			my $pke= $o_locus->GetpkExtended();
			$h_ann_gaps{$pk} = $pke;
		}
	}
	
	# Look for the gaps in the genomic sequence and populate %h_seq_gaps
	my %h_seq_gaps;
	my $r_gseq = $self -> GetGenomicSequence();
	my $offset=0;
	my $pos=0;
	while(1)
	{
		my $seq = substr $$r_gseq, $pos;
		if ( $seq =~ /[nN]+/p )
		{
			my $s = $offset + length(${^PREMATCH})+1;
			my $e = $s       + length(${^MATCH})-1;
			$h_seq_gaps{"+$s..$e"} = 1;
			#print "$s..$e\n";
			$pos = $e + 1;
			$offset = $pos;
		}
		else
		{
			last;
		}
	}
	
	# Compare both hashes
	my (@a_not_in_ann,@a_not_in_seq);
	my @a_k_ann = sort keys %h_ann_gaps;
	my @a_k_seq = sort keys %h_seq_gaps;
	while ( my $ka = shift @a_k_ann)
	{
		my $kb;
		while ( @a_k_seq > 0 )
		{
			$kb = shift @a_k_seq ;
			if ($kb lt $ka)
			{
				push @a_not_in_ann, $kb;
			}
			else
			{
				last;
			}
		}
		if ( $ka lt $kb || ($ka gt $kb && @a_k_seq == 0))
		{
			push @a_not_in_seq, $ka;
		}
	}
	
	my $error_flg=&FALSE;
	if ( @a_not_in_ann>0)
	{
		print "\n";
		print "------------------------------------------------------------------\n";
		print "SOME GAPS WERE DETECTED IN THE SEQUENCE BUT NOT IN THE ANNOTATIONS\n";
		print "------------------------------------------------------------------\n";
		foreach my $k (@a_not_in_ann)
		{
			print $k."\n";
		}
		$error_flg=&TRUE;
	}
	if ( @a_not_in_seq>0)
	{
		print "\n";
		print "------------------------------------------------------------------\n";
		print "SOME GAPS WERE DETECTED IN THE ANNOTATIONS BUT NOT IN THE SEQUENCE\n";
		print "------------------------------------------------------------------\n";
		foreach my $k (@a_not_in_seq)
		{
			print $h_ann_gaps{$k}."\n";
		}
		$error_flg=&TRUE;
	}
	if ( $error_flg == &FALSE )
	{
		print "The gap annotations are consistent with the gaps found in the sequence\n";
	}
}


=head __NewRepOrigin

 Title     : __NewRepOrigin
 Usage     : $self->__NewRepOrigin ($rep_origin)
 Procedure : Relocate the replication origin of every annotated object AND change the sequecne accordingly
             A negative number means: "change the strand"
 Args      : $rep_origin, the new replication origin
    
=cut

sub __NewRepOrigin
{
	my $self = shift;
	my ($rep_origin) = @_;
	
	# Change the strand if necessary
	my $seq_length = $self -> GetLength();
	if ( $rep_origin < 0 )
	{
		$self->RevComp();
		$rep_origin = - $rep_origin;
		$rep_origin = $seq_length - $rep_origin + 1;
	}
	
	# Change the replication origin in the annotations
	my $ra_loci = $self -> GetAllLoci();
	foreach my $o_locus (@$ra_loci)
	{
		$o_locus -> NewOrigin($rep_origin,$seq_length);
	}
	
	# Change also the sequence
	my $r_gseq = $self -> GetGenomicSequence();
	my $part_1= substr($$r_gseq,0,$rep_origin-1,'');
	$$r_gseq .= $part_1;
}

=head RevComp

 Title     : RevComp
 Usage     : $self->RevComp();
 Procedure : Reverse complement the genomic sequence, which should be loaded, and the coordinates of ALL biological objects
             Mark the sequence and the annotations as locekd, to keep the synchro    
=cut

sub RevComp
{
	my $self = shift;
	
	my $ra_loci = $self -> GetAllLoci();
	my $seq_length = $self -> GetLength();
	
	# Reverse complement all the biological objects
	foreach my $o_locus (@$ra_loci)
	{
		$o_locus -> RevComp($seq_length);
	}
	
	# Reverse complement the genomic sequence
	my $r_gseq = $self->GetGenomicSequence();
	#$$r_gseq = reverse($$r_gseq);
	my $tmp = reverse($$r_gseq);
	$$r_gseq = $tmp;
	$$r_gseq =~ tr/ATGCatgc/TACGtacg/;
}

=head __RenumberLocusTags

 Title     : __RenumberLocusTags
 Usage     : $self->__RenumberLocusTags( $ra_genes, $lt_templ_cpt)
 Procedure : renumber the locus tags, starting from $cpt or from $AnnotatedSeq::LOCUS_TAG_NB
 Args      : $lt_templ_cpt, see the BFC Usage
    
=cut

sub __RenumberLocusTags
{
	my $self = shift;
	my ( $lt_templ_cpt ) = @_;
	
	my ($lt_templ,$cpt,$step) = split ',',$lt_templ_cpt;
	
	# If step not defined, step = 1
	if ( ! defined $step )
	{
		$step = 1;
	}
	
	# if not defined, use the static variables
	if ( ! defined $cpt )
	{
		if ( $AnnotatedSeq::LOCUS_TAG_NB==-1 )
		{
			$AnnotatedSeq::LOCUS_TAG_NB=1;
		}
		$cpt = $AnnotatedSeq::LOCUS_TAG_NB;
	}
	
	# if defined and at start, use the defined value
	elsif ( $AnnotatedSeq::LOCUS_TAG_NB==-1 )
	{
		$AnnotatedSeq::LOCUS_TAG_NB = $cpt;
	}
	
	# if defined but not the first contig, use the static variable
	else
	{
		$cpt = $AnnotatedSeq::LOCUS_TAG_NB;
	}
	
	my $ra_genes = $self -> GetGenes();
	my ($tmpl)  = ($lt_templ =~ /(\.+)/);
	my $fmt_len = length $tmpl;
	$tmpl =~ s/\./\\./g;
	foreach my $o_gene ( @$ra_genes )
	{
		my $cpt_lt = '0' x ($fmt_len-length($cpt)) . $cpt;
		#print "KOUKOU " . length($cpt) . ' ' . $cpt_lt. "\n";

		my $locus_tag = $lt_templ;
		#print "LOCUS_TAG $locus_tag\n";
		$locus_tag =~ s/$tmpl/$cpt_lt/;
		#print "LOCUS_TAG $locus_tag\n";
		$o_gene -> SetLocusTag($locus_tag);
		my $ra_children = $o_gene -> GetAllChildren();
		foreach my $o_locus ( @$ra_children )
		{
			$o_locus -> SetLocusTag($locus_tag);
		}
		$cpt += $step;
	}
	$AnnotatedSeq::LOCUS_TAG_NB = $cpt;
}

=head __FilterSmallUtr

 Title     : __FilterSmallUtr
 Usage     : $self->__FilterSmallUtr()
 Procedure : Remove the utr which have only 1 bp length, computing new positions for the corresponding exons, mrna, gene.
    
=cut
sub __FilterSmallUtr
{
	my $self = shift;

	my $ra_genes = $self -> GetGenes();
	foreach my $o_gene ( @$ra_genes )
	{
		next unless ( $o_gene->IsProteinCoding() );
		my $ra_mrnas = $o_gene -> GetMrnas();
		foreach my $o_mrna ( @$ra_mrnas )
		{
			foreach my $type ( qw/ three_prime_UTR five_prime_UTR / )
			{
				my $o_utr = $o_mrna -> GetUtr ( $type );
				if ( defined($o_utr) && $o_utr->GetLength()==1 )
				{
					warn ( $self -> GetWarningHeader() . "Removed utr " . $o_utr->GetpkExtended() ."\n" );
					$o_mrna -> RemoveUtr( $type );
					my ($u_start,$u_end) = $o_utr  -> GetEnds();
					my ($r_start,$r_end) = $o_mrna -> GetEnds();
					
					# The locus and the whole parent chain must shrink from the left
					if ( $u_start == $r_start )
					{
						my $o_locus = $o_mrna->GetExonSet()->GetRegion0();
						do
						{
							$o_locus->SetStart($u_start+1);
							$o_locus = $o_locus -> GetParent();
						} 
						while ( $o_locus -> can ( 'GetParent' ) );
					}

					# The locus and the whole parent chain must shrink from the right
					elsif ( $u_end == $r_end )
					{
						my $o_locus = $o_mrna->GetExonSet()->GetRegionLast();
						do
						{
							$o_locus->SetEnd($u_end-1);
							$o_locus = $o_locus -> GetParent();
						} 
						while ( $o_locus -> can ( 'GetParent' ) );
					}
					
					else
					{
						warn ( $self -> GetWarningHeader() . "utr " . $o_utr->GetpkExtended() . " does not match with the mrna\n" );
					}
				}
			}
		}
		
		# Adjust again the gene boundaries, this is useful if there are several mRNAS
		foreach my $o_mrna ( @$ra_mrnas )
		{
			$o_gene->AdjustBoundaries($o_mrna);
		}
	}
}

=head __FilterBadProteins

 Title     : __FilterBadProteins
 Usage     : $self->__FilterBadProteins()
 Procedure : When a cds contains at least one 'N', it is tagged as Inconsistent
    
=cut
sub __FilterBadProteins
{
	my $self = shift;
	my ($threshold) = @_;

	my $seq_length = $self -> GetLength();
	if ( $seq_length == 0 )
	{
		throw Error::Simple( "There is no sequence loaded, cannot filter proteins");
	}

	my $ra_genes = $self -> GetGenes();
	foreach my $o_gene ( @$ra_genes )
	{
		next unless ( $o_gene->IsProteinCoding() );
		my $ra_mrnas = $o_gene -> GetMrnas();
		my $mrna_nb  = scalar(@$ra_mrnas);
		foreach my $o_mrna ( @$ra_mrnas )
		{
			my $o_cds = $o_mrna -> GetCDS();
			next if (!defined($o_cds));
			my $seq = $o_cds -> GetTranslation();
				
			if ( $seq =~ /X/ )
			{
				my $seq1= $seq;
				$seq1 =~ s/X//g;
				my $ratio = int(100*(length($seq)-length($seq1)) / length($seq));
				if ( $ratio >= $threshold )
				{
					if ($mrna_nb == 1)
					{
						$o_gene -> Deselect();
						warn ( $self -> GetWarningHeader() . "gene " . $o_gene->GetpkExtended() ." removed because there were $ratio%, ie more than $threshold% X in the proteic sequence\n" );
					}
					else
					{
						$o_mrna -> Deselect();
						warn ( $self -> GetWarningHeader() . "mrna " . $o_mrna->GetpkExtended() ." removed because there were $ratio%, ie more than $threshold% X in the proteic sequence\n" );
					}
				}
				else
				{
					# Put the cds inconsistent if not already
					next if ($o_cds->GetConsistency() ne 'C');
					$o_cds -> SetInconsistent();
					$o_cds -> SyncConsistency();
					$o_mrna -> AddValue ( 'note', 'This protein was removed because some X are embedded in its sequence' );
					warn ( $self -> GetWarningHeader() . "cds " . $o_cds->GetpkExtended() ." declared inconsistent because some X were embedded in its sequence\n" );
				}
			}
		}
	}
}

=head __CreateUtr

 Title     : __CreateUtr
 Usage     : $self->__CreateUtr()
 Procedure : Create an UTR when the Exon and CDS positions are different and if the UTR does not already exist
             NOTE - The sub __FilterSmallUtr is called also, so that if small UTR are created by this sub, they will be removed just after
    
=cut
sub __CreateUtr
{
	my $self = shift;
	
	my $ra_genes = $self -> GetGenes();
	foreach my $o_gene ( @$ra_genes )
	{
		next unless ( $o_gene->IsProteinCoding() );
		my $ra_mrnas = $o_gene -> GetMrnas();
		foreach my $o_mrna ( @$ra_mrnas )
		{
			$o_mrna -> CreateUtr();
		}
	}
}

=head __FilterMrnasWithoutCds

 Title     : __FilterMrnasWithoutCds
 Usage     : $self->__FilterMrnasWithoutCds()
 Procedure : If in some mRNA there are exons but NO CDS, we remove the Exons and put a note to the gene. 
             If the is NO CORRECT mRNA, replace the gene line with a region line
    
=cut
sub __FilterMrnasWithoutCds
{
	my $self = shift;

	my $ra_genes = $self -> GetGenes();
	foreach my $o_gene ( @$ra_genes )
	{
		next unless ( $o_gene->IsProteinCoding() );
		my $ra_mrnas = $o_gene -> GetMrnas();
		
		# copy the array of mrnas, because we are using the gene method ShiftMrna
		my @a_removed_mrnas;  
		foreach my $o_mrna ( @$ra_mrnas )
		{
			my $o_cds = $o_mrna->GetCDS();
			next if defined($o_cds);
			
			push @a_removed_mrnas,$o_mrna->GetpkExtended();
			warn ( $self -> GetWarningHeader() . "mrna " . $o_mrna->GetpkExtended() ." removed because no CDS were declared\n" );
			$o_mrna -> Deselect();
		}
		
		if ( @a_removed_mrnas != 0 )
		{
			my $note = "Some mrnas were removed because no CDS was declared for them: ".join(' ',@a_removed_mrnas);
			$o_gene->AddValue('note',$note);
		}
	}
}

=head SetFatalError

 Title     : SetFatalError
 Usage     : $o_annot->SetFatalError()
 Procedure : Set a flag '__FatalError'. When set, some methods (see Log()) will generate an exception at END of processing.
             Thus, you may accumulated a lot of error messages, generate an exception and display all those messages at the end
    
=cut
sub SetFatalError
{
	my ($self) = @_;
	$self -> { __fatal_error } = &TRUE;
}

=head2 procedure LoadGenomicSequence

 Title        : LoadGenomicSequence
 Function     : If the member __in_seq_file is defined, load the sequence from this file
                Else if the reader is defined, call the loader's LoadSequence
                Else throw an exception
                The sequence is CONCATENATED to the current genomic sequence
                
 Args         : none

=cut

sub LoadGenomicSequence
{
	my ($self) = @_;
	
	my $r_new_gseq = "";
	my $o_assembly = $self -> { __o_assembly };
	my $assembly_flg = $self -> { __build_assembly };

	if ( ! defined ( $self -> { __genomic_sequence } ) )
	{
		my $genomic_sequence = "";
		$self -> { __genomic_sequence } = \$genomic_sequence;
	}

	# During the assembly, it is wise calling LoadGenomicSequence at start, just to include the first gap
	if ( $assembly_flg && $o_assembly -> GetCurrentContigName() eq 'start' )
	{
		$r_new_gseq = $self->__ReadStartContig();
	}
	
	# No assembly, or not at start of assembly: load a contig
	# Note - We consider here that if no assembly, the whole sequence is an unique contig
	else
	{
		$r_new_gseq = $self -> __ReadContig();
	}
	
	# If the sequence is empty, we just return !
	if ( $$r_new_gseq eq '' )
	{
		return;
	}
	
	# During assembly, but not at start: more processing to do
	if ( $assembly_flg && $o_assembly -> GetCurrentContigName() ne 'start')
	{

		# Select the part of contig to add
		# If assembly, we extract only from feature_start to feature_end
		my $f_start  = $o_assembly -> GetFeatureStart();
		my $f_end    = $o_assembly -> GetFeatureEnd();
		my $f_len    = $f_end - $f_start + 1;
		
		# Call substr
		$$r_new_gseq = substr ( $$r_new_gseq, $f_start - 1, $f_len );
		
		# Reverse complement if we live on the '-' strand
		if ( $o_assembly -> GetStrand() eq '-' )
		{
			$$r_new_gseq = reverse($$r_new_gseq);
			$$r_new_gseq =~ tr/ATGCatgc/TACGtacg/;
		}
	
		# Add sequence buffer after the contig
		my $current_end   = $o_assembly -> GetEnd();
		$o_assembly -> NextContig();
		my $next_start  = $o_assembly -> GetStart();
		$o_assembly -> PrevContig();

		my $gap_len  =  $next_start - $current_end - 1;
		$$r_new_gseq .= 'N' x $gap_len;
	}
	
	# Concatenate the new sequence + buffer with the existing genomic sequence
	my $r_genomic_sequence = $self -> { __genomic_sequence };
	$$r_genomic_sequence .= $$r_new_gseq if ( defined $$r_new_gseq );
	
	# Update the sequence limits
	$self -> UpdateSeqRegion(1,length($$r_genomic_sequence));
	
	return;
}

sub __ReadStartContig
{
	my $self = shift;
	my $r_new_gseq = "";
	my $o_assembly = $self -> { __o_assembly };

	my $current_end = $o_assembly -> GetEnd();
	$o_assembly -> NextContig();
	my $next_start  = $o_assembly -> GetStart();
	$o_assembly -> PrevContig();
	
	my $gap_len     = $next_start -  $current_end;
	my $genomic_sequence = 'N' x $gap_len;
	$r_new_gseq = \$genomic_sequence;
	return $r_new_gseq;
}

=head2 function __Readcontig

 Title        : __ReadContig
 Function     : If the member __in_seq_mem is defined, the sequence is already somewhere in memory (see __ReassembleContig)
                If the member __in_seq_file is defined, load the contig from this file
                If __contig is defined (see DetectContig), get this contig
                Else Read the ONLY contig in from the fasta file (and check it is not a multifasta)
                Return a ref to the sequence and set __contig
 Args         : A ref to a scalar

=cut
sub __ReadContig{
	my $self = shift;
	my $r_new_gseq = "";
	
	# __in_seq_mem defined, the sequence is already in the memory !
	if (defined $self -> { __in_seq_mem } && $self -> { __in_seq_mem } ne '')
	{
		$r_new_gseq = $self -> { __in_seq_mem };
	}
	
	# __in_seq_file specified, read the sequence from this file
	elsif (defined $self -> { __in_seq_file } && $self -> { __in_seq_file } ne '')
	{
		# TODO - faire un objet separe si on veut traiter plusieurs types de fichiers de sequence
		# TODO - en fait on utilise un fasta avec une seule sequence donc Splitter n'est pas obligatoirement judicieux ici
		#        A moins de dire a Splitter qu'il s'agit d'un fasta et pas d'un multifasta ?
		my $f_fasta = $self -> { __in_seq_file };
		my $o_fasta = New Splitter::Fasta ( $f_fasta );
		if ( defined ( $self -> { __contig } ))
		{
			my $contig_id = $self -> { __contig };
			$r_new_gseq = $o_fasta -> GetContig( $contig_id );
			if ( $$r_new_gseq eq "" )
			{
				throw Error::Simple( "Cannot load the genomic sequence, contig $contig_id unknown");
			}
		}
		else
		{
			my $ra_contigs_id = $o_fasta -> GetContigNames();
			if ( @$ra_contigs_id == 0 )
			{
				throw Error::Simple( "Sorry, $f_fasta is empty, or is not a fasta file" );
			}
			if ( @$ra_contigs_id != 1 )
			{
				throw Error::Simple( "Sorry, multifasta file is not supported if the contig id is not specified" );
			}
			$r_new_gseq = $o_fasta -> GetContig( $ra_contigs_id -> [0] );
			$self -> { __contig } = $ra_contigs_id -> [0];
		}
		$$r_new_gseq =~ s/^.+\n//m;	# Remove the header
		$$r_new_gseq =~ s/\n//gm;   # Remove all EOL characters
	}

	# If no __in_seq_file specified, but __o_load specified, the sequence may be embedded in the annotation file ?
	elsif (defined $self -> { __o_reader })
	{
		$r_new_gseq = $self -> { __o_reader } -> LoadGenomicSequence();
	}
	
	# Nowhere to load from, throw an exception
	else
	{
		throw Error::Simple( "Cannot load the genomic sequence ");
	}
	return $r_new_gseq;
}

=head2 procedure __DumpGenomicSequence

 Title    : __DumpGenomicSequence
 Prerequisite : If the genomic sequence is not loaded, or __out_seq_file not defined, throw an exception
 Function : Called by Dump, dump the genomic sequence to the ouput sequence file
            If not in append mode, open the file with GetStreamOut and write the fasta header
            If in append mode, open the file in 'a'
            If $Locus::O_ASSEMBLY is defined, we are assembling the genome, thus:
               1/ If append, do not write the header (only one sequence will be written at the end)
               2/ Write some 'N' after the sequence if there is a gap between this contig and the next one
            
 Args     : -append => 0/1
 Access   : private

=cut
sub __DumpGenomicSequence
{
	my $self = shift;
	my (%h_params) = (-append => 0, @_);
	
	if ( !defined ($self -> { __genomic_sequence }) || !defined ($self -> { __out_seq_file } ) )
	{
		throw Error::Simple( "No genomic sequence loaded or no output sequence file defined" );
	}
	
	my $r_sequence = $self -> { __genomic_sequence };
	my $f_fasta    = $self -> { __out_seq_file };
	my $fh_fasta;
	my $append     = $h_params {-append};

	my $new_file = &TRUE;
	if ( $append )
	{
		# Append mode: Are we writing to a new file ?
		$new_file = &FALSE if ( -f $f_fasta );

		# Open file, append or write mode	
		$fh_fasta = new IO::File( $f_fasta, 'a' ) || throw LipmError::IOException( $f_fasta, 'f' );
	}
	else
	{
		$fh_fasta = &GetStreamOut( $f_fasta );
	}
	
	# print header
	print $fh_fasta '>' . $self -> GetChromosomeLabel() . "\n";
	
	# Print sequence	
	my $seq_length = length $$r_sequence;
	my $offset     = 0;
	while ($offset < $seq_length)
	{
		print $fh_fasta substr ( $$r_sequence, $offset, 60 ) . "\n" ;
		$offset += 60;
	}
}

=head2 Accessor GetLength

 Title        : GetLength
 Usage        : my $len = $o_obj -> GetLength()
 Prerequisite : LoadGenomicSequence() must have been called, else the length 0 is returned
 Args         : nothing
 Return       : The contig length if the genomic sequence is loaded, 0 if not
 
=cut

sub GetLength
{
	my $self  = shift;
	
	if ( ! defined $self -> { __length } )
	{
		try
		{
			my $r_seq = $self -> GetGenomicSequence();
			$self -> { __length } = length $$r_seq;
		}
		catch Error with 
		{
			$self -> { __length } = 0;
		}
	}
	return $self -> { __length };
}
 
=head2 procedure Dump

 Title        : Dump
 Usage        : $o_obj -> Dump()
 Procedure    : If the output annotation file is specified, call the Dump procedure from the writer object
                If the output sequence file is specified, call the DumpGenomicSequence procedure
 Args         : -append => 0/1

=cut
sub Dump
{
	my $self = shift;
	if ( defined $self -> { __o_reassemble_assembly } )
	{
		$self -> __ReassembleAndDump(@_);
	}
	elsif ( $self -> { __disassemble_assembly } == &TRUE )
	{
		$self -> __DisassembleAndDump(@_);
	}
	else
	{
		$self -> __DumpNormal(@_);
	}
	#if ( defined ($self -> { __fatal_error }) )
	#{
	#	throw Error::Simple ( "FATAL ERROR OCCURRED WHILE DUMPING DATA - PLEASE HAVE A LOOK TO THE IMPORTANT WARNINGS" );
	#}

}

sub __ReassembleAndDump
{
	my $self = shift;
	$self -> __ReassembleContig();
	$self -> __DumpNormal(@_);
}

sub __DisassembleAndDump
{
	my $self = shift;
	my $o_writer = $self -> { __o_writer };
	my $out_ann_file = $self -> { '__out_ann_file' };
	my $out_ann_fmt  = $self -> { '__out_ann_fmt'  };
	my $out_seq_file = $self -> { '__out_seq_file' };
	my $out_seq_fmt  = $self -> { '__out_seq_fmt'  };

	$self -> __CreateDisassembledContigs();
	my $ra_disassembled_contigs = $self -> { __ra_disassembled_contigs };

	foreach my $o_annot ( @$ra_disassembled_contigs )
	{
		$o_annot -> SetOutput ( -out_ann_file=>$out_ann_file, -out_ann_fmt=>$out_ann_fmt, -out_seq_file=>$out_seq_file,-out_seq_fmt=>$out_seq_fmt );
		$o_annot -> __DumpNormal( -append => &TRUE );
	}
}
 
sub __DumpNormal
{
	my $self = shift;
	my $o_writer = $self -> { __o_writer };
	
	my $out_ann_file = $self -> { __out_ann_file };
	my $out_ann_fmt  = $self -> { __out_ann_fmt };
	my $out_seq_file = $self -> { __out_seq_file };
	if ( $out_ann_file ne "" )
	{
		if ( defined $o_writer )
		{
			$o_writer -> Dump (@_);
		}
		else
		{
			my $msg = "Writer object not initialized - File = $out_ann_file ($out_ann_fmt)";
			throw LipmError::ObjectException('AnnotatedSeqFile', '!attribute', -text => $msg);
		}
	}
	if ( $out_seq_file ne "" )
	{
		$self -> __DumpGenomicSequence(@_);
		$self -> AddDumped($out_seq_file);
	}
}

=head2 procedure __ReassembleContig

 Title    : __ReassembleContig
 Usage    : $self -> __ReassembleContig();
 Prerequisite : An annotated sequence must have been loaded, and $self -> GetChromosomeLabel() must return something
                The __o_assembly propety must be in the form: file1.agp,file2.agp (, is the separator between both names)
 Function : Call __CreateDisassembledContigs (complaining if already done), using file1.agp
            there is not more locus in $self
            THEN InitAssembly to file2.agp and re-populate $self, emptying the contigs, correcting again the coordinates
 Args     : None
 NOTE     : After a call to this function, there will be NO MORE LOCUS inside the assembled contigs

=cut
sub __ReassembleContig
{
	my $self = shift;
	
	# Return if already disasembled and reassembled
	return if ( defined $self -> { __ra_disassembled_contigs } );

	my $chr  = $self -> GetContigLabel();
	return if ( $chr eq '');
	
	# Disassemble things...
	my $o_assembly = $self -> { __o_reassemble_assembly };
	$o_assembly -> SetCurrentChromosome( $chr );
	$self -> __CreateDisassembledContigs($o_assembly);

	# ...then reassemble them !
	$o_assembly = $self -> { __o_assembly };
	$o_assembly -> SetCurrentChromosome( $chr );
	Locus::InitAssembly($o_assembly);
	$self -> { __sequence_region_start } = 1;
	$self -> { __sequence_region_end   } = 0;

	# Reset the genomic sequence
	my $genomic_sequence = "";
	$self -> { __genomic_sequence } = \$genomic_sequence;
	
	# set the __build_assembly flag, it is used by LoadGenomicSequence
	$self -> { __build_assembly } = &TRUE;
	
	# The key is the contig names, we should find all contig names in the agp
	my $rh_annot = $self -> { __rh_disassembled_contigs };
	for ( $o_assembly -> SetFirstContig(); $o_assembly->GetCurrentContigName() ne 'end'; $o_assembly -> NextContig() )
	{
		my $contig = $o_assembly->GetCurrentContigName();
		next if ( $contig eq 'start' );
		if ( ! defined $rh_annot -> { $contig } )
		{
			warn ( $self -> GetWarningHeader() . "contig $contig NOT FOUND in the assembly agp ! Bad agp file ?\n" );
			next;
		}
		
		my $o_annot = $rh_annot -> { $contig };
		my $ra_loci = $o_annot -> GetSortedLoci();
		$self->UpdateSeqRegion($o_assembly -> GetFeatureStart(),$o_assembly -> GetFeatureEnd());
		
		foreach my $o_locus ( @$ra_loci )
		{
			$o_locus -> AbsCoords();
			$o_locus -> ResetParent($self);
		}
		
		$o_annot -> { __a_genes } = [];
		$o_annot -> { __a_ncrnas } = [];
		$o_annot -> { __a_repeats } = [];
		$o_annot -> { __a_regions } = [];
		
		# Load the genomic sequence from the sequence already in the contig (see the __ReadContig function for __in_seq_mem)
		$self -> { __in_seq_mem } = $o_annot -> GetGenomicSequence();
		$self -> LoadGenomicSequence(); 
	}

	$self -> __ComputeAssemblyGaps();# if ($self -> { __submit} == &FALSE);
}

=head2 procedure __CreateDisassembledContigs

 Title    : __CreateDisassembledContigs
 Usage    : $self -> __CreateDisassembleContigs();
 Prerequisite : An annotated contig must have been loaded, and $self -> GetContigLabel() must return something
 Function : If not already done, create one AnnotatedSeq object / contig found in the agp
            Populate them with the Locus objects found in $self, following the information of the agp
 Args     : $o_assembly (optional, if undefined we use the __o_assembly private member)
 NOTE     : After a call to this function, there will be NO MORE LOCUS inside $self !

=cut
sub __CreateDisassembledContigs
{
	my $self = shift;
	my ( $o_assembly) = @_;
	$o_assembly = $self -> { __o_assembly } unless defined ( $o_assembly );
	
	return if ( defined $self -> { __ra_disassembled_contigs } );
	
	# The disassembled chromosome label is the loaded contig label
	my $chr  = $self -> GetContigLabel();
	return if ( $chr eq '');

	$o_assembly -> SetCurrentChromosome( $chr );
	Locus::InitAssembly($o_assembly);

	my %h_annot;
	my @a_annot;

	my $ra_genes   = $self -> { __a_genes };
	my $ra_ncrnas  = $self -> { __a_ncrnas };
	my $ra_repeats = $self -> { __a_repeats };
	my $ra_regions = $self -> { __a_regions };

	# Dispatch all the loci between the AnnotatedSeq objects
	my $i=0;
	for ( $o_assembly -> SetFirstContig(); $o_assembly->GetCurrentContigName() ne 'end'; $o_assembly -> NextContig() )
	{
		next if ($o_assembly->GetCurrentContigName() eq 'start');
		
		my $contig = $o_assembly->GetCurrentContigName();
		if ( !exists $h_annot { $contig } )
		{
			$h_annot { $contig } = New AnnotatedSeq (
			kingdom       => $self->{__kingdom},
			species       => $self->{__species},
			-contig       => $o_assembly->GetCurrentContigName(),
			-chromosome   => $o_assembly->GetCurrentContigName(),
			-tmpdir       => $self->{__tmpdir},
			-debug        => $self->{__debug},
			-out_ann_file => $self->{__out_ann_file},
			-out_ann_fmt  => $self->{__out_ann_fmt},
			-out_seq_file => $self->{__out_seq_file},
			-out_seq_fmt  => $self->{__out_seq_fmt}
			);
			push @a_annot,$h_annot { $contig };
		}

		my $o_annot = $h_annot { $contig  };
		my $c_start = $o_assembly -> GetStart();
		my $c_end   = $o_assembly -> GetEnd();
		$o_annot->UpdateSeqRegion($o_assembly -> GetFeatureStart(),$o_assembly -> GetFeatureEnd());
		if ( defined $self -> { __genomic_sequence } )
		{
			$o_annot->__SetSequence($self -> { __genomic_sequence }, $o_assembly);
		}
		foreach my $ra_loci ( $ra_genes,$ra_ncrnas,$ra_repeats,$ra_regions )
		{
			for (my $j=0; $j<@$ra_loci;$j++)
			{
				my $o_locus = $ra_loci->[$j];
				next unless defined $o_locus;
				if ( $o_locus->GetStart() >= $c_start && $o_locus->GetEnd() <= $c_end )
				{
					$o_locus -> RelCoords();
					$o_locus->ResetParent( $o_annot);
					$ra_loci->[$j]=undef;
				}
				elsif ( $o_locus->GetStart() < $c_start && $o_locus->GetEnd() >= $c_start )
				{
					my ($s,$e) = $o_locus->GetEnds();
					warn ( "ERROR locus is NOT in the contig $contig: Loc strt=$s Cont strt=$c_start Loc end=$e " . $o_locus->GetpkExtended() ." - locus skipped\n" );
					$self->SetFatalError();
				}
				elsif ( $o_locus->GetStart() <= $c_end && $o_locus->GetEnd() > $c_end )
				{
					my ($s,$e) = $o_locus->GetEnds();
					warn ( "ERROR locus is NOT in the contig $contig: Loc strt=$s Cont end=$c_end Loc end=$e " . $o_locus->GetpkExtended() ." - locus skipped\n" );
					$self->SetFatalError();
				}
			}
		}
	}

	$self -> { __a_genes } = [];
	$self -> { __a_ncrnas } = [];
	$self -> { __a_repeats } = [];
	$self -> { __a_regions } = [];

	$self -> { __ra_disassembled_contigs } = \@a_annot;
	$self -> { __rh_disassembled_contigs } = \%h_annot;
}

=head2 procedure __SetSequence

 Title : __SeqSequence
 Usage : $o_annot -> __SetSequence($r_src_seq,$start,$end)
         Called by __CreateDisassembledContigs to set the sequence of a contig
         NOTE - $start, $end are 1-BASED
 Args  : $r_src_seq : A ref to the source sequence
         $o_assembly: Compute the sequence from the current contig
 
=cut
sub __SetSequence
{
	my $self = shift;
	my ($r_src,$o_assembly) = @_;
	
	my $src_start = $o_assembly -> GetStart();
	my $src_end   = $o_assembly -> GetEnd();
	my $dst_start = $o_assembly -> GetFeatureStart();
	my $dst_end   = $o_assembly -> GetFeatureEnd();
	my $strand    = $o_assembly -> GetStrand();
	
	if (!defined $self -> { __genomic_sequence })
	{
		my $dst = "";
		$self -> { __genomic_sequence } = \$dst;
	}
	my $r_dst = $self -> { __genomic_sequence };
	my $dst_len = length ($$r_dst);
	
	my $add_len = $dst_end - $dst_start + 1;
	
	# $dst_start > dst_len => complete the string from the left
	# $dst_start  = 6
	# $dst_end    = 10
	# $$r_dst = GAGA   ==> GAGANNNNNNN ==> GAGANNTCTCT
	# OR
	# $$r_dst = GAGAGA ==> GAGAGATCTCT
	# 
	if ( $dst_start >= $dst_len )
	{
		my $diff = $dst_start - $dst_len - 1;
		$$r_dst .= 'N' x $diff;
		my $new = substr ( $$r_src, $src_start-1, $add_len );
		if ( $strand eq '-' )
		{
			$new = reverse($new);
			$new =~ tr/ATGCatgc/TACGtacg/;
		}
		$$r_dst .= $new;
	}
	# $dst_end <= $dst_len => replace some sharacters of $$r_dst
	# $dst_start  = 6
	# $dst_end    = 10
	# $$r_dst = GAGAGAGAGAGA ==> GAGAGTCTCTGA
	# OR
	# $$r_dst = GAGAGAGAGA   ==> GAGAGTCTCT
	elsif ( $dst_end <= $dst_len )
	{
		my $old = substr ( $$r_dst, $dst_start-1, $add_len );
		my $new = substr ( $$r_src, $src_start-1, $add_len );
		if ( $strand eq '-' )
		{
			$new = reverse($new);
			$new =~ tr/ATGCatgc/TACGtacg/;
		}
		$$r_dst =~ s/$old/$new/;
		#substr $$r_dst, $dst_start-1, $add_len, $new;
	}
}
		
=head2

 Title    : __CreateDisassembleContigs
 Usage    : $self -> __CreateDisassembleContigs();
 Prerequisite : An annotated sequence must have been loaded, and $self -> GetChromosomeLabel() must return something
 Function : If not already done, create one AnnotatedSeq object / contig found in the agp
            Populate them with the Locus objects found in $self, following the information of the agp
 Args     : None
 NOTE     : After a call to this function, there will be NO MORE LOCUS inside $self !

=cut


=head2 procedure AddDumped

 Title        : AddDumped
 Usage        : $o_annot -> AddDumped( $file )
 Procedure    : Called by Dump, adds $file to the list of dumped files - Note that some objects do not create 1 file
                when dumping data, several (cf Fasta.pm) objects write in fact several files
                The files are maintained as keys in a hash, so you can dump several times the same file, it will be counted once
                WARNING - This list is RESET by Setoutput()
 Args         : $file: The file to add to the list
 Globals      : none

=cut
sub AddDumped
{
	my $self = shift;
	my ($dumped) = @_;
	
	if ( !defined $self -> { __rh_dumped } )
	{
		$self -> { __rh_dumped } = {};
	}
	
	$self->{ __rh_dumped }->{$dumped} = 1;
}

=head2 procedure GetDumped

 Title    : GetDumped
 Usage    : my $ra_o_comp = $o_annot -> GetDumped()
 Function : Returns the list of FileCompressedAppend objects created when calling Dump
            The calling program may keep a ref to those objects, delaying the compression phase, which occurs at object destruction
 Args     : none
 Returns  : A ref to an array, containing the list of FileCompressedAppend objects (the order is irrelevant).

=cut
sub GetDumped
{
	my $self = shift;
	my @a_rvl = keys %{$self -> { __rh_dumped }};
	return \@a_rvl;
}
	
=head2 procedure AssertConsistencyLimit

 Title     : AssertConsistencyLimit
 Usage     : $o_obj -> AssertConsistencyLimit()
 Procedure : Croak if the nb of inconsistent cds is under the consistency limit
 Args      : none
 Globals   : none
 TODO      : UN check SERAIT UN PEU MIEUX !!!

=cut
sub AssertConsistencyLimit
{
	my $self = shift;
	
	my $ctg = $self -> GetContigLabel();
	
	# Consistency limit	
	my $consistency_limit=$$self{'__consistency_limit'};
	if ( defined($consistency_limit) && $consistency_limit =~ /^\d+$/ )
	{
		my $nb_protein    = $self -> GetNbProteinSequences();
		my $nb_transcript = $self -> GetNbTranscriptSequences();
		my $nb_gene       = $self -> GetNbGenes();
		if ( $nb_transcript != 0 )
		{
			my $consistency   = sprintf("%04.1f",100.0 * $nb_protein / $nb_transcript);
			
			if ($nb_protein < $nb_transcript * $consistency_limit/100.0)
			{
	            my $msg = "Fatal error: contig $ctg: too many inconsistencies during protein extraction:\n";
	            $msg .= "nb of genes       = $nb_gene;\nnb of transcripts = $nb_transcript\nnb of proteins    = $nb_protein\nconsistency       = $consistency% (consistency_limit = $consistency_limit\%)\n";
	            $msg .= "You should: 1) check the translation table (parameter TABLE_$ctg in your species.cfg\n";
	            $msg .= "            2) check the annotation file format (is it REALLY correctly formatted ?\n";
	            $msg .= "            3) At a last resort, increase the parameter CONSISTENCY_LIMIT_$ctg in your species.cfg file\n";
				Carp::croak($msg);
			}
			else
			{
				print "Consistency check OK: $nb_protein proteins for $nb_transcript transcripts ($nb_gene genes) \nconsistency = $consistency% (consistency limit = $consistency_limit%)\n\n";
			}
		}
	}
	else
	{
		print "Warning: consistency checks skipped, consistency_limit is not initialized or is not an integer\%)\n";
	}
}

=head1 PRIVATE METHODS

=cut

=head2 function SearchMRNA

 Title        : SearchMRNA
 Usage        : $self->SearchMRNA ($o_locus,$locus_tag)
 Prerequisite : none
 Function     : Search a mRNA with a given transcript id and matching the coordinates of the given locus
                Matches means that $o_locus is INCLUDED inside the retrieved mrna
 Returns      : $o_mrna: the mrna retrieved, undef if not found
 Args         : $o_locus: a Locus::Mrna object
                $locus_tag: the locus_tag value of the transcript to search
                $cnt [optional, default = 1] If specified, looks for the nth mRNA matching the given condition
 Globals      : none

=cut

sub SearchMRNA
{
	my($self,$o_locus,$locus_tag,$cnt) = @_;
	$cnt = 1 unless defined $cnt;
	
	if ( !defined $locus_tag )
	{
		warn ( $self -> GetWarningHeader() . "locus_tag NOT specified " . $o_locus->GetpkExtended() ."\n" );
		return undef;
	}

	my $rh_index = $self -> { __idx_mrnas };
	if (!defined ( $rh_index -> { $locus_tag } ) )
	{
		print STDERR "Warning: unable to find the mRNA $locus_tag\n" if $self->GetDebug();
		return undef;
	}
	else
	{
		# Looping among the mrnas with same locus_tag, we return the nth mrna
		# surrounding $o_locus
		my $ra_mrnas = $rh_index -> { $locus_tag };
		my ($start,$end) = $o_locus -> GetEnds();
		my $i = 1;
		foreach my $o_mrna (@$ra_mrnas)
		{
			my ($mrna_start,$mrna_end) = $o_mrna -> GetEnds();
			if ( ($start >= $mrna_start) && ($start <= $mrna_end) && ($end >=$mrna_start) && ($end <= $mrna_end) )
			{
				#print STDOUT "Trouve: $locus, $i, ($start >= $gene_start) && ($start <= $gene_end) && ($end >=$gene_start) && ($end <= $gene_end)\n";
				if ($i == $cnt)
				{
					return $o_mrna;
				}
				else
				{
					$i++;
				}
			}
		}
		print STDERR "Warning: unable to find the mRNA $locus_tag\n" if $self->GetDebug();
		return undef;
	}
}	

=head2 function SearchGene

 Title        : SearchGene
 Usage        : $self -> SearchGene($o_locus,$locus_tag)
 Prerequisite : none
 Function     : Search a gene with a given locus_tag which matches the coordinates of a given locus
                Matches means that $o_locus is INCLUDED inside the retrieved gene
                ===> NOOOOOOOOOO means that $o_locus and the gene OVERLAP !
 Returns      : The Locus::Gene found, undef if none found
 Args         : $o_locus  : A locus object
                $locus_tag: A locus tag
 Globals      : none

=cut

sub SearchGene
{
	my($self,$o_locus,$locus_tag) = @_;
	$locus_tag ="" unless (defined $locus_tag);

	my $rh_index = $self -> { __idx_genes };
	if (!defined ( $rh_index -> { $locus_tag } ) )
	{
		return undef;
	}
	else
	{
		# Looping among the genes with same locus_tag, we return the first gene
		# surrounding $o_locus
		my $ra_genes = $rh_index -> { $locus_tag };
		my ($start,$end)         = $o_locus -> GetEnds();
		foreach my $o_gene (@$ra_genes)
		{
			my ($gene_start,$gene_end) = ($o_gene -> GetEnds());
			if ( ($start>=$gene_start && $start<=$gene_end) || ($end>=$gene_start && $end <= $gene_end) )
			{
				#print STDOUT "Trouve: $locus, $i, ($start >= $gene_start) && ($start <= $gene_end) && ($end >=$gene_start) && ($end <= $gene_end)\n";
				return $o_gene;
			}
		}
		return undef;
	}
}	

=head2 function GetSequenceFile, GetAnnotationFile

 Title        : GetSequenceFile, GetAnnotationFile
 Usage        : my $seq_file   = $self -> GetSequenceFile();
                my $annot_file = $self -> GetAnnotationFile();
 Function     : Accessors
 Returns      : Sequence or annotation file names
 Args         : None
 
=cut
sub GetSequenceFile
{
	my $self = shift;
	return $self -> { __in_seq_file };
}
sub GetAnnotationFile
{
	my $self = shift;
	return $self -> { __in_ann_file };
}
sub GetOutputFile
{
	my $self = shift;
	return $self -> { __out_ann_file };
}
=head2 function GetGenomicSequence

 Title        : GetGenomicSequence
 Usage        : my $r_seq = $self -> GetGenomicSequence();
 Function     : Get a ref to the genomic sequence
                If the genomic sequence is not yet loaded, call LoadGenomicSequence (lazy loading)
 Returns      : The ref to the genomic sequence (can be huge data)
 Args         : None
 Globals      : none
 
=cut
sub GetGenomicSequence
{
	my $self = shift;
	if (!defined $self -> { __genomic_sequence })
	{
		$self -> LoadGenomicSequence();
	}
	return $self -> { __genomic_sequence };
}

=head2 function GetSeqRegionStart, GetSeqRegionEnd

 Title        : GetSeqRegionStart, GetSeqRegionEnd

=cut
sub GetSeqRegionStart   { my $self = shift; return $self -> { __sequence_region_start } };
sub GetSeqRegionEnd   { my $self = shift; return $self -> { __sequence_region_end } };

=head2 procedure UpdateSeqRegion

 Title        : UpdateSeqRegion
 Usage        : $self -> UpdateSeqRegion( $start, $end );
 Function     : Update SeqRegion if necessary, ie only if enlarged
 Args         : $start, $end New start/end positions to store, ONLY if not already inside the current region
 Globals      : none
 
=cut

sub UpdateSeqRegion
{
	my $self = shift;
	my ( $start, $end ) = @_;
	
	$self -> { __sequence_region_start } = $start if ( $start < $self -> { __sequence_region_start } );
	$self -> { __sequence_region_end }   = $end   if ( $end    > $self -> { __sequence_region_end } );
}

=head2 Validation

 Title    : Validation
 Usage    : $o_annot -> Validation()
 Procedure: FOR DEBUG ONLY - Execute some sanity checks on our objects
 NOTE     : NE MARCHE PAS ACTUELLEMENT !!!
 
=cut
sub Validation
{
	my $self = shift;
	print "======================== Starting validation =============================\n";
	
	# Verifying that EACH gene is properly indexed
	$self -> __GeneIndexationValidate();
	
	# Verifying that EACH mrna is properly indexed
	$self -> __MrnaIndexationValidate();
	
	# Verifying that starting from each cds we can go to $self, using GetParent()
	$self -> __ParentValidate();
	
	# Verifying that starting from each cds, the cds is inside the mrna who is inside the gene
	# Verifying also that the regions are inside the locus
	$self -> __BoundaryValidate();
	
	# Verifying that each object has a Narcisse Id, and that there are NO two equal NarcisseIds 
	$self -> __NarcisseIdValidate();
	print "======================== End of validation ========================\n";
}

sub __GeneIndexationValidate
{
	my $self = shift;
	
	print "__GeneIndexationValidate\n";

	my $ra_genes     = $self -> { __a_genes };
	my $rh_idx_genes = 	$self -> { __idx_genes };
	foreach my $o_gene (@$ra_genes)
	{
		my $locus_tag = $o_gene -> GetLocusTag();
		if ( defined $rh_idx_genes -> { $locus_tag } )
		{
			my $ra_indexed_genes = $rh_idx_genes -> { $locus_tag };
			my $fnd = 0;
			foreach my $o_igene (@$ra_indexed_genes)
			{
				if ($o_gene == $o_igene) 
				{
					$fnd = 1;
					last;
				}
			}
			if ($fnd==0)
			{
				print $self -> GetWarningHeader() . "The gene " . $o_gene->GetpkExtended() . " is NOT indexed, but the locus tag is indexed\n";
			}
		}
		else
		{
			print $self -> GetWarningHeader() . "The gene " . $o_gene->GetpkExtended() . " is NOT indexed, neither its locus tag\n";
		}
	}
}


sub __MrnaIndexationValidate
{
	my $self = shift;
	
	print "__MrnaIndexationValidate";

	my $ra_mrnas     = $self -> { __a_mrnas };
	my $rh_idx_mrnas = 	$self -> { __idx_mrnas };
	foreach my $o_mrna (@$ra_mrnas)
	{
		my $locus_tag = $o_mrna -> GetLocusTag();
		if ( defined $rh_idx_mrnas -> { $locus_tag } )
		{
			my $ra_indexed_mrnas = $rh_idx_mrnas -> { $locus_tag };
			my $fnd = 0;
			foreach my $o_imrna (@$ra_indexed_mrnas)
			{
				if ($o_mrna == $o_imrna) 
				{
					$fnd = 1;
					last;
				}
			}
			if ($fnd==0)
			{
				print $self -> GetWarningHeader() . "The mrna " . $o_mrna->GetpkExtended() . " is NOT indexed, but the locus tag is indexed\n";
			}
		}
		else
		{
			print $self -> GetWarningHeader() . "The mrna " . $o_mrna->GetpkExtended() . " is NOT indexed, neither its locus tag\n";
		}
	}
}					

sub __ParentValidate
{
	my $self = shift;
	
	print "__ParentValidate";

	my $ra_genes = $self -> { __a_genes };
	foreach my $o_gene (@$ra_genes)
	{
		if ($o_gene -> GetParent() != $self)
		{
			print $self -> GetWarningHeader() . "The gene " . $o_gene->GetpkExtended() . " should have as parent the annotatedSeqfile object\n";
		};
		for (my $i=0; $i<$o_gene->GetNbMrnas(); $i++)
		{
			my $o_mrna = $o_gene -> GetMrna( $i );
			my $o_cds  = $o_mrna -> GetCDS();
			if ($o_mrna -> GetParent() != $o_gene)
			{
				print $self -> GetWarningHeader() . "The mrna " . $o_mrna->GetpkExtended() . " should have as parent the GENE " . $o_gene->GetLocusTag()."\n";
			}
			if ($o_cds -> GetParent() != $o_mrna)
			{
				print $self -> GetWarningHeader() . "The cds " . $o_cds->GetpkExtended() . " should have as parent the MRNA " . $o_mrna->GetLocusTag()."\n";
			}
		}
	}
}

sub __BoundaryValidate
{
	my $self = shift;
	
	print "__BoundaryValidate\n";

	my $ra_genes = $self -> { __a_genes };
	foreach my $o_gene (@$ra_genes)
	{
		my ($gene_start,$gene_end) = $o_gene -> GetEnds();
		if ($gene_start < 1)
		{
			print $self -> GetWarningHeader() . "AnnotatedSeqFile::__BoundaryValidate - The gene " . $o_gene->GetLocusTag() . " starts at " . $gene_start . "\n";
		};

		# I CANNOT CHECK THE END, BECAUSE I DO NOT KNOW THE CHR LENGTH
		for (my $i=0; $i<$o_gene->GetNbMrnas(); $i++)
		{
			my $o_mrna = $o_gene -> GetMrna( $i );
			my ($mrna_start, $mrna_end) = $o_mrna -> GetEnds();
			if ($mrna_start < $gene_start)
			{
				print $self -> GetWarningHeader() . "The mrna " . $o_mrna->GetLocusTag() . " starts at " . $mrna_start . " gene start = $gene_start\n";
			};
			if ($mrna_end > $gene_end)
			{
				print $self -> GetWarningHeader() . "The mrna " . $o_mrna->GetLocusTag() . " ends at " . $mrna_end . " gene end  = $gene_end\n";
			};
			my $mrna_regions = $o_mrna -> GetRegions();
			if ($mrna_regions eq '')
			{
				print $self -> GetWarningHeader() . "The mrna " . $o_mrna->GetLocusTag() . " has NO region defined\n";
			}
			else
			{
				my ($mrna_start_region,$mrna_end_region) = ($mrna_regions=~/^(\d+).+(\d+)$/);
				if ($mrna_start_region < $mrna_start)
				{
					print $self -> GetWarningHeader() . "The mrna " . $o_mrna->GetLocusTag() . " has a region starting at " . $mrna_start_region . " mrna start = $mrna_start\n";
				};
				if ($mrna_end_region > $mrna_end)
				{
					print $self -> GetWarningHeader() . "The mrna " . $o_mrna->GetLocusTag() . " has a region ending    at " . $mrna_end_region . " mrna end = $mrna_end\n";
				};
			}

			my $o_cds  = $o_mrna -> GetCDS();
			my ($cds_start, $cds_end) = $o_cds -> GetEnds();
			if ($cds_start < $mrna_start)
			{
				print $self -> GetWarningHeader() . "The CDS " . $o_cds->GetLocusTag() . " starts at " . $cds_start . " mrna start = $mrna_start\n";
			};
			if ($cds_end > $mrna_end)
			{
				print $self -> GetWarningHeader() . "The CDS " . $o_mrna->GetLocusTag() . " ends at " . $cds_end . " mrna end  = $mrna_end\n";
			};
			my $cds_regions = $o_cds -> GetRegions();
			if ($cds_regions eq '')
			{
				print $self -> GetWarningHeader() . "The CDS " . $o_cds->GetLocusTag() . " has NO region defined\n";
			}
			else
			{
				my ($cds_start_region,$cds_end_region) = ($cds_regions=~/^(\d+).+(\d+)$/);
				if ($cds_start_region < $cds_start)
				{
					print $self -> GetWarningHeader() . "The cds " . $o_cds->GetLocusTag() . " has a region starting at " . $cds_start_region . " cds start = $cds_start\n";
				};
				if ($cds_end_region > $cds_end)
				{
					print $self -> GetWarningHeader() . "The cds " . $o_cds->GetLocusTag() . " has a region ending    at " . $cds_end_region . " cds end = $cds_end\n";
				};
			}
		}
	}
}
			
sub __NarcisseIdValidate
{
	my $self = shift;
	
	print "__NarcisseIdValidate\n";

	my %h_narcisse_id;

	my $ra_genes = $self -> { __a_genes };
	foreach my $o_gene (@$ra_genes)
	{
		my $narcisse_id = $o_gene -> GetNarcisseId();
		if ( ! defined $narcisse_id )
		{
			print $self -> GetWarningHeader() . "The gene " . $o_gene->GetLocusTag() . " should have a Narcisse Id\n";
		}
		else
		{
			if (exists $h_narcisse_id { $narcisse_id })
			{
				print $self -> GetWarningHeader() . "gene " . $o_gene->GetLocusTag() . " its narcisse Id $narcisse_id is already used\n";
			}
			$h_narcisse_id { $narcisse_id } = 1;
		}
		for (my $i=0; $i<$o_gene->GetNbMrnas(); $i++)
		{
			my $o_mrna = $o_gene -> GetMrna( $i );
			my $narcisse_id = $o_mrna -> GetNarcisseId();
			if ( ! defined $narcisse_id )
			{
				print $self -> GetWarningHeader() . "The mrna " . $o_mrna->GetLocusTag() . " should have a Narcisse Id\n";
			}
			else
			{
				if (exists $h_narcisse_id { $narcisse_id })
				{
					print $self -> GetWarningHeader() . "mrna " . $o_mrna->GetLocusTag() . " its narcisse Id $narcisse_id is already used\n";
				}
				$h_narcisse_id { $narcisse_id } = 1;
			}
			
			my $o_cds = $o_mrna -> GetCDS();
			$narcisse_id = $o_cds -> GetNarcisseId();
			if ( ! defined $narcisse_id )
			{
				print $self -> GetWarningHeader() . "The cds " . $o_cds->GetLocusTag() . " should have a Narcisse Id\n";
			}
			else
			{
				if (exists $h_narcisse_id { $narcisse_id })
				{
					print $self -> GetWarningHeader() . "cds " . $o_cds->GetLocusTag() . " its narcisse Id $narcisse_id is already used\n";
				}
				$h_narcisse_id { $narcisse_id } = 1;
			}
		}
	}
	
	my $ra_ncrnas = $self -> { __a_ncrnas };
	foreach my $o_ncrna (@$ra_ncrnas)
	{
		my $narcisse_id = $o_ncrna -> GetNarcisseId();
		if ( ! defined $narcisse_id )
		{
			print $self -> GetWarningHeader() ."The ncrna " . $o_ncrna->GetLocusTag() . " should have a Narcisse Id\n";
		}
		else
		{
			if (exists $h_narcisse_id { $narcisse_id })
			{
				print $self -> GetWarningHeader() . "The ncrna " . $o_ncrna->GetLocusTag() . " its narcisse Id $narcisse_id is already used\n";
			}
			$h_narcisse_id { $narcisse_id } = 1;
		}
	}
	
	my $ra_repeats = $self -> { __a_repeats };
	foreach my $o_repeat (@$ra_repeats)
	{
		my $narcisse_id = $o_repeat -> GetNarcisseId();
		if ( ! defined $narcisse_id )
		{
			print $self -> GetWarningHeader() . "The repeat " . $o_repeat->GetLocusTag() . " should have a Narcisse Id\n";
		}
		else
		{
			if (exists $h_narcisse_id { $narcisse_id })
			{
				print $self -> GetWarningHeader() . "The repeat " . $o_repeat->GetLocusTag() . " its narcisse Id $narcisse_id is already used\n";
			}
			$h_narcisse_id { $narcisse_id } = 1;
		}
	}
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut



1;
