#
# $Id: GFF3.pm 1346 2014-01-08 13:15:54Z manu $
#

package AnnotatedSeqFileWriter::Tsv;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

AnnotatedSeqFileWriter::tsv - Dumping AnnotatedSeqFile objects in tsv
                              This format is simply derived from gff3 files, except that the field nb 9 is exploded is as many fieds as necessary
                              The separator is TAB

=head1 SYNOPSIS

 my $o_dumper = New AnnotatedSeqFileWriter::tsv( o_annot => $o_annot );
 $o_dumper -> Dump();

=head1 DESCRIPTION

 this class extends the AnnotatedSeqFileWriter::GFF3, it dumps data from AnnotatedSeq to tsv files

=cut

use strict;
use warnings;
use Carp;
use URI::Escape;

use Error qw(:try);
#use Bio::Seq;
use Data::Dumper;
use General;
use Scalar::Util qw(looks_like_number);

use Locus;
use SeqOntology;
use base qw( AnnotatedSeqFileWriter::GFF3 );
use FileCompressedAppend;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1346 $ =~ /\d+/g); $r[0]};
}

# The headers corresponding to columns 1..8
our @A_HEADERS    = qw( seqid source type start end score strand phase );

# The headers corresponding the the GFF3 column 9
# The properties NOT in this header will be ignored
our @A_F9_HEADERS = qw( ID Name synonym Alias locus_tag Parent function product ec_number Dbxref Ontology_term Is_circular Note Derives_from Target Gap );

sub _Init
{
	my $self = shift;
	$self->SUPER::_Init(@_);
	# See GFF3::__DumpF45678
	$self -> { __is_gff3 } = &FALSE;
}

=head2 procedure _DumpHeader

 Title     : _DumpHeader
 Usage     : $o_annot_writer -> _DumpHeader( $fh_out, $o_annot );
 Procedure : Dump header data, called by Dump
             If any, dump the data found in the header hash maintained by $o_annot
             Dump lines as (key,value) pairs

=cut

sub _DumpHeader
{
	my $self = shift;
	my ( $fh_out, $o_annot ) = @_;
	
	my $l = join "\t", @AnnotatedSeqFileWriter::Tsv::A_HEADERS, @AnnotatedSeqFileWriter::Tsv::A_F9_HEADERS;
	print $fh_out "$l\n";
}

=head2 procedure _DumpF9

 Title     : __DumpF9
 Usage     : $o_annot_writer -> __DumpF9 ($fh_out, $o_locus)
 Procedure : Dump the Field 9 of gff3, exploding it in several columns (see @A_F9_HEADERS)
             This is called by subs declared in the parent class
             We call _ComputeF9 from GFF3, the we simply parse the line produced
 Args      : $fh_out  An output opened stream
             $o_locus The locus to dump
 Access    : PROTECTED

=cut

sub _DumpF9
{
	my $self = shift;
	my ($fh_out,$o_locus)  = @_;
	my $field_9 = $self -> _ComputeF9($o_locus);
	
	my $rh_sub_fields = $self -> __Line2Hash($field_9);
	my $line = '';
	my @a_f;
	foreach my $k ( @AnnotatedSeqFileWriter::Tsv::A_F9_HEADERS )
	{
		push @a_f,((defined $rh_sub_fields->{$k})?$rh_sub_fields->{$k}:'');
	}
	
	print $fh_out join ("\t",@a_f)."\n";
}

sub __Line2Hash
{
	my $self = shift;
	my ($line) = @_;
	my @a_line = split ';', $line;
	my %h_line;
	foreach my $f (@a_line)
	{
		my ($k,$v) = split '=',$f;
		$h_line{$k}=$v;
	}
	return \%h_line;
}


1;


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;


