#
# $Id: Cds.pm 1336 2013-12-04 13:35:33Z manu $
#

package Locus::Cds;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

Locus::Cds - describes a Cds

=head1 SYNOPSIS

 my $o_gene = New Locus::Cds('start'=>integer,'end'=>integer,'strand'=>'-1|1|+|-','locus_tag'=>string,parent=>$o_gene);

=head1 DESCRIPTION

 A Cds extends the Locus class. A cds is a Coding Dna Sequence, it is an array of regions
 the parent MUST be a Locus::Mrna object
 A Cds should not have a locus_tag nor gene_name attribute, but a Cds has a NarcisseId
 It should NOT be printed, only the cdsregions will be printed
 Thus, a cds is a container for cdsregions

=cut


use strict;
use warnings;
use General;
use Carp;
use Bio::Seq;
use Bio::Tools::CodonTable;
use Data::Dumper;
use Locus::CdsRegion;

use base qw( Locus );
sub GetDefaultSubType { return 'CDS' };

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1336 $ =~ /\d+/g); $r[0]};
}

=head2 procedures Select,Deselect

 See Locus.pm for the generic documentation
 we do not change the status of the children (see CdsRegion::IsSelected for the details)

=cut
sub Select
{
	my $self = shift;
	delete $self -> { __deselected };
}
sub Deselect
{
	my $self   = shift;
	$self -> { __deselected } = &TRUE;
}
=head2 function GetProteinId

 Title : GetProteinId
 Usage : my $id = $o_cds -> GetProteinId();
=cut
sub GetProteinId
{
	my $self = shift;
	my @a_regions = $self -> GetRegions();
	return ($a_regions[0]->isa('Locus::CdsRegion')) ? $a_regions[0]->GetProteinId(@_) : '';
}

=head2 function GetProduct

 Title        : GetProduct
 Function     : Return the product (= annotation) if found
 Usage        : my $gene_name = $o_mcds->GetProduct();
 Returns      : The product, or the product of the first region

=cut
sub GetProduct
{
	my $self = shift;
	
	my $product = $self -> Get ( 'product' );
	if ( $product eq '' )
	{
		my @a_regions = $self -> GetRegions();
		return ($a_regions[0]->isa('Locus::CdsRegion')) ? $a_regions[0]->GetProduct() : '';
	}
	else
	{
		return $product;
	}
}
=head2 GetLocusTag, GetGeneNameSimple

 return _Get0, ie may be the locus_tag/gene of the first region

=cut
sub GetLocusTag       { my $self = shift; return $self -> _Get0 ( 'locus_tag'); }
sub GetGeneNameSimple { my $self = shift; return $self -> _Get0 ( 'gene'); };

=head2 procedure SetParent

 Title        : SetParent
 Usage        : $o_cds->setParent( $o_mrna );
 Procedure    : Set the parent
 Prerequisite : The parent should not already set, $o_mrna should be a Locus::Mrna object
 
 Args         : $o_rmna : The parent mRNA
 Globals      : none

=cut
sub SetParent
{
	my $self = shift;
	my ($o_mrna) = @_;
	
	if ( ref($o_mrna) ne 'Locus::Mrna' )
	{
		throw LipmError::ParameterException( 'o_parent', 'type', -text => 'should be a Locus::Mrna object' );
	}
	# caeel_sIg1t1 ==> caeel_sIg1c1
	my $nc_id = $o_mrna -> GetNarcisseId();
	$nc_id    =~ s/t(\d+)$/c$1/;
	$self -> { __narcisse_id } = $nc_id;

	$self->SUPER::SetParent( $o_mrna );
}

=head2 Function GetProteinSequenceLength

 Title        : GetProteinSequenceLength
 Function     : If consistent or fragment: Computes the length of the cds, using GetRegionsLength, then / 3 and substract 1
				If inconsistent: returns 0
 Usage        : my $seq = $o_locus->GetProteinSequenceLength();
 Returns      : The proteinsequence length
 Args         : None
 Globals      : none
=cut
sub GetProteinSequenceLength
{
	my $self = shift;
	if ($self -> GetConsistency eq 'I')
	{
		return 0;
	}
	else
	{
		# Return the nucleic sequence length / 3 and take into account the start codon
		my $nucl_len = $self -> GetRegionsLength();
		return int ($nucl_len / 3)  - 1
	}
}

=head2 procedure GetTranslation

 Title     : GetTranslation
 Usage     : my $prot_seq = $o_cds->GetTranslation();
 Function  : Compute and return the protein translation of the CDS, using bioperl
             MAY SET THE CONSISTENCY FLAG OF THE OBJECT (thus the object is MODIFIED by this function)
             Keep the result of the translation inside the object because GetTranslation may be called several times

 Args      : $verbose if &TRUE, print more warnings to STDERR
 Return    : The translated protein, "" if nothing to translate or if inconsistency found

=cut
sub GetTranslation
{
	my $self = shift;
	my ($verbose) = @_;
	$verbose = &FALSE unless defined($verbose);
	
	#print "KOUKOU ".$self->GetGeneName."\n";
	
	if ( !defined $self -> { __seq_prot } )
	{
		# TODO - A REVOIR - Ensembl NOT SUPPORTED YET
		my $data_source_ensembl = &FALSE;					
		# JG changement suite au merge API my ($len,$sseq) = $self->GetSequence($r_seq,$h_param{strand},$exoncds);
		my ($len,$sseq);
		if ( $data_source_ensembl )
		{
			#($len,$sseq) = $self->GetParent()->GetSequence($r_seq);
		}
		else
		{
			$sseq = $self->GetSequence();
			$len  = $self->GetSequenceLength();
		}
		return "" if ( $len == 0 );
		
		my $o_mrna = $self -> GetParent();
		if ( $data_source_ensembl )
		{
			# On a deja fait un GetSequence pourquoi on le corrige ? (MANU)
			$sseq = substr($sseq,$o_mrna->Get('cds_start')-1,$o_mrna->Get('cds_end')-$o_mrna->Get('cds_start')+1);
			$len = length $sseq;
			#printf STDERR "$sseq#%d\n",$r_mrna->Get('cds_end')-$r_mrna->Get('cds_start')+1;
		}
					
		# Useful for warnings or error messages
		my $mrna_gene_name = $o_mrna -> GetGeneName();
		my $mrna_locus_tag = $o_mrna -> GetLocusTag();
		my $transcript_id  = $o_mrna -> GetTranscriptId();
		my $protein_id     = $self   -> GetProteinId();
	
		# Detecting fragment or inconsistent proteins:
		#
		# 1/ If START or STOP codon is missing, the protein is a FRAGMENT
		# 2/ If there is a STOP codon inside the CDS, the protein is a FRAGMENT and we try another phase
		# 3/ If there is still a STOP codon inside the CDS, AND we are in phase 2, we give up and the protein is INCONSISTENT
	
		my $codon_table=$self -> GetAncestor() -> GetCodonTable();
		my $o_codon_table = Bio::Tools::CodonTable -> new ();
		$o_codon_table -> id ( $codon_table );
		
		my $display_id = "prot:" . $self -> GetProteinId();
		my $gene_obj     = Bio::Seq->new( -display_id => "$display_id" );
		$gene_obj -> seq ( $sseq, 'dna' );

		my $seq_prot;

		my $phase0;
		foreach my $phase (0,1,2)
		{
			$phase0 = $phase;
			
			my $prot_obj = $gene_obj->translate(undef,undef,$phase,$codon_table,1);
	
			$seq_prot = $prot_obj->seq;
	
			my $start_codon = substr $sseq,$phase,3;
			if ( $o_codon_table -> is_start_codon($start_codon) == 0 )
			{
				$self -> SetFragment('S');
				print STDERR $self -> GetWarningHeader('S') . " - $mrna_locus_tag ($transcript_id $protein_id gene $mrna_gene_name) does not start with M -> it is a fragment\n" if ($verbose);
			}
	
			# If we have stop codons in the middle, it is at least a fragment, but we check another phase
			if ( $seq_prot =~ /\*/ )
			{
				$self -> SetFragment('E');
				print STDERR $self -> GetWarningHeader('E') . " - $mrna_locus_tag ($transcript_id $protein_id gene $mrna_gene_name) has a stop codon in the middle (trying phase $phase)\n" if ($verbose);
	
				#... and if in phase 2, we mark the protein as inconsistent before giving up
				if ($phase == 2)
				{
					$self -> SetInconsistent();
					print STDERR $self -> GetWarningHeader('I') . " - giving up - $mrna_locus_tag ($transcript_id $protein_id gene $mrna_gene_name) is inconsistent (skipped)\n" if ($verbose);
				}
				next;
			}
		
			# OK - No terminator codon inside CDS
			# If we have a stop codon at end it's OK
			# NOTE - We have to check the end codon at the nucleic level because the terminator is chopped by the translate function
			my ($lastcodon) = ($sseq =~ /(\w{3})$/);
			#print "\nLASTCODON $lastcodon\n";
			if  ( $lastcodon =~ /(TAA|TGA|TAG)$/i)
			{
				last;
			}
			
			# if there is no stop codon, or if stop codon contains 'N', as in EuGene, it is a fragment
			else
			{
			  $self -> SetFragment('E');
			  print STDERR $self -> GetWarningHeader('T') . " - $mrna_locus_tag ($transcript_id $protein_id gene $mrna_gene_name) has not correct stop codon -> it is a fragment (but kept)\n" if ($verbose);
			  last;
			}
		}
		
		if ( $self-> GetConsistency() eq 'C' )
		{
			if ($len % 3 != 0)
			{
				$self -> SetFragment('F');
				print STDERR $self -> GetWarningHeader('F') . " - $mrna_locus_tag ($transcript_id) is not divisible / 3 -> it is a fragment\n" if ($verbose);
			}
		}
	

		# warn about the consistency status if necessary
		if ($self -> GetConsistency() eq 'I')
		{
			warn $self ->GetWarningHeader('I') . "$mrna_locus_tag ($transcript_id $protein_id gene $mrna_gene_name) - protein is inconsistent (removed)\n";
			$seq_prot = "";
		}
		elsif ( $self -> GetConsistency() =~ /F/ )
		{
			warn $self ->GetWarningHeader('F') . $self -> GetConsistency() ." "."$mrna_locus_tag ($transcript_id $protein_id gene $mrna_gene_name) - protein is a fragment (kept)\n";
			
		}
		
		$self -> { __seq_prot } = $seq_prot;
		$self -> { __phase0 }   = $phase0;
	}
	return $self -> { __seq_prot };
} 

=head2 procedure AddRegion

 Title        : AddRegion
 Prerequisite : GetTranslation should NOT have benn called on this CDS
 Procedure    : If __seq_prot already defined, throw an exception, else call Locus::AddRegion

=cut

sub AddRegion
{
	my $self = shift;
	
	if ( defined $self -> { __seq_prot } )
	{
		my $msg = "ERROR - AddRegion CANNOT BE CALLED ANY MORE, because GetTranslation was already called on this CDS: ";
		$msg .= $self->GetpkExtended();
		throw Error::Simple $msg;
	}
	else
	{
		$self->SUPER::AddRegion(@_);
	}
}

=head2 procedure ComputePhase

 Title     : ComputePhase
 Procedure : If not already done, compute the phase of each CdsRegion
 Note      : This should be done AFTER calling GetTranslation, as GetTranslation may detect some framshift in the first cds region
             If GetTranslation was not called before, we consider that the phase of the first cds region is 0
 Usage     : $o_cds -> ComputePhase();
=cut

sub ComputePhase
{
	my $self = shift;
	
	my @a_regions = $self -> GetRegions();
	return if (@a_regions==0);
	
	# Return if the work was already done
	return if ($a_regions[0]->GetPhase() ne '.');

	if ( $self -> GetStrand() == 1 )
	{
		my $o_current  = $a_regions[0];
		$o_current -> { _d_rh_core } -> { phase } = (exists $self -> { __phase0 }) ? $self -> { __phase0 } : 0;
		
		# For each region except the first one, the phase is computed as followes: 
		#       1/ p = (Lp - Pp) % 3
		#       2/ The phase is: (3 - p) % 3, or simpler said: 0->0 1->2 2->1
		#
		# where Lp = Length of the previous CdsRegion
		#       Pp = Phase of the previous CdsRegion
		my $o_previous = $o_current;
		for (my $i=1; $i<@a_regions; $i++)
		{
			$o_current = $a_regions[$i];
			my $p      = ( $o_previous -> GetLength() - $o_previous -> GetPhase() ) % 3;
			$o_current -> { _d_rh_core } -> { phase } = (3 - $p ) % 3;
			$o_previous = $o_current;
		}
	}
	
	# same thing but we start from the LAST exon
	else
	{
		my $o_current  = $a_regions[-1];
		$o_current -> { _d_rh_core } -> { phase } = (exists $self -> { __phase0 }) ? $self -> { __phase0 } : 0;
		my $o_previous = $o_current;
		for (my $i=-2; $i>=-@a_regions; $i--)
		{
			$o_current = $a_regions[$i];
			my $p      = ( $o_previous -> GetLength() - $o_previous -> GetPhase() ) % 3;
			$o_current -> { _d_rh_core } -> { phase } = (3 - $p ) % 3;
			$o_previous = $o_current;
		}
	}
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
