#
# $Id: SeqOntology.pm 1301 2013-11-08 12:47:24Z manu $
#

package SeqOntology;

use strict;
use warnings;
use 5.10.0;			# For the state variable

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

SeqOntology - The gateway between sequence ontology terms and the objects derived from Locus
              This module is used by AnnotatedSeqFileReader and AnnotatedSeqFileWriter

=head1 SYNOPSIS
 my $o_so = New SeqOntology();
 if ( $o_so->IsTypeInSeqOntology( $o_locus ) ...
 my $locus_class = GetLocusClass( $term);

=head1 DESCRIPTION

 This object may help to create a Locus:: object from a Sequence Ontology term, (col 3 of the gff3 files).
 It may also tell you if the type of some locus belongs to the sequence ontology, so if you may directly write its type to the col3 of the gff3 file
 See http://www.ebi.ac.uk/ena/WebFeat/ for the genbank tags 
 See http://www.sequenceontology.org/ for the sequence ontology terms
 
 THIS OBJECT IS A SINGLETON
 
=cut

use General;
use Error qw(:try);
use base qw( LipmSingleton );
use LipmError::ParameterException;
use LipmError::ObjectException;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1301 $ =~ /\d+/g); $r[0]};
}

# The data are stored inside the SeqontologyData.pl file, this file is evaled
# The following global variables will be filled:
use constant MODULE_DATA_FILE => 'SeqOntologyData.pl';

# keys   = The gene ontology terms
# values = The corresponding classes
# This hash will be computed from %H_SEQ_ONTOLOGY_TYPE by the constructor
# (see SeqOntologyData.pl)

our %H_SEQ_ONTOLOGY_TYPE;
our %H_SEQ_ONTOLOGY_PARENT;
our %H_SEQ_ONTOLOGY_PREFERRED_CHILD;

our $RH_CLASSES;


sub _Init
{
	my $self = shift;
	eval $self->_InitModuleData(&MODULE_DATA_FILE);
	if ($@) 
	{
		die $@;
	}

	if ( ! defined $SeqOntology::RH_CLASSES )
	{
		my %h_classes;
		foreach my $locus_class (keys %SeqOntology::H_SEQ_ONTOLOGY_TYPE)
		{
			my $rh_so = $SeqOntology::H_SEQ_ONTOLOGY_TYPE{$locus_class};
			foreach my $so_term ( keys %$rh_so )
			{
				if (exists $h_classes{$so_term})
				{
					throw LipmError::ParameterException('%H_SEQ_ONTOLOGY_TYPE','value', -text => "Duplicate value in %H_SEQ_ONTOLOGY_TYPE: $so_term");
				}
				$h_classes{ $so_term } = $locus_class;
			}
		}
		$SeqOntology::RH_CLASSES = \%h_classes;
	}
}

=head2 function IsTypeInSeqOntology

 Title    : IsTypeInSeqOntology
 Usage    : if ( $o_so -> IsTypeInSeqOntology( $o_locus ) ...
 Function : returns &TRUE if the $o_locus passed by parameters is not undef AND if its type returns a term
            found in the Sequence Ontology
            returns &FALSE in any other case
 Args     : $o_locus Locus::XXX object
 Return   : &TRUE/&FALSE
 
 Global   : %H_SEQ_ONTOLOGY_TYPE
 
=cut
sub IsTypeInSeqOntology
{
	my $self = shift;
	my ($o_locus) = @_;
	if ( $o_locus -> isa ( 'Locus' ) && exists ($SeqOntology::H_SEQ_ONTOLOGY_TYPE{ref($o_locus)} -> {$o_locus -> GetSubType()}) )
	{
		return &TRUE;
	}
	else
	{
		return &FALSE;
	}
}

=head2 function GetLocusClass

 Title    : GetLocusClass
 Usage    : my $class = $o_so -> GetLocusClass( $so_term );
 Function : return the type of Locus object corresponding to the sequence ontology term passed by parameter
 Args     : $so_term A string, should be a sequence ontology term
 Return   : The Locus::XXX class, or "" if none
 
 Global   : $RH_CLASSES; 

=cut
sub GetLocusClass
{
	my $self = shift;
	my ($so_term) = @_;
	if ( exists ( $SeqOntology::RH_CLASSES -> {$so_term} ) )
	{
		return $SeqOntology::RH_CLASSES -> {$so_term};
	}
	else
	{
		return "";
	}
}

=head2 function IsADescendantOf

 Title    : IsADescendantOf
 Usage    : if ( $o_so -> IsADescendantOf($child,$ancestor) )...
 Function : return &TRUE if the first name ($child) is a child, or little-child etc., of 2nd name ($ancestor)
 Args     : $child, $ancestro
 Return   : 0/1
 TODO - REMPLIR AUTOMATIQUEMENT H_ONTOLOGY_PARENT A PARTIR DU FICHIER FOURNI PAR LA GENE_ONTOLOGY ?
 Global   : $H_ONTOLOGY_PARENT

=cut
sub IsADescendantOf
{
	my $self = shift;
	my ($child,$ancestor) = @_;
	if ( exists $SeqOntology::H_SEQ_ONTOLOGY_PARENT{ $child } )
	{
		if ( $SeqOntology::H_SEQ_ONTOLOGY_PARENT{ $child } eq $ancestor )
		{
			return &TRUE;
		}
		else
		{
			return $self->IsADescendantOf($SeqOntology::H_SEQ_ONTOLOGY_PARENT{ $child }, $ancestor);
		}
	}
	else
	{
		return &FALSE;
	}
}

=head2 function GetPreferredChild

 Title    : GetPreferredChild
 Usage    : my $child = $o_sp->GetPreferredChild($parent);
 Function : return a name which is surely a "part_of" the oarameter
 Args     : $parent
 Return   : $child
 TODO - REMPLIR AUTOMATIQUEMENT $H_SEQ_ONTOLOGY_PREFERRED_CHILD A PARTIR DU FICHIER FOURNI PAR LA GENE_ONTOLOGY ?
 Global   : $H_SEQ_ONTOLOGY_PREFERRED_CHILD

=cut
sub GetPreferredChild
{
	my $self = shift;
	my ($parent) = @_;
	
	return $SeqOntology::H_SEQ_ONTOLOGY_PREFERRED_CHILD{$parent};
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
