#!/usr/bin/perl

# ajoute par manu pour trouver les bibs
use FindBin;
use lib "$FindBin::Bin/../../../corelib";
use lib "$FindBin::Bin/../..";

use strict;
use General;
use ParamParser;
use GeneralBioinfo;
use Data::Dumper;
use Cwd;


=pod

=head1 UC: soumission d'un genome complet a GenBank

http://www.ncbi.nlm.nih.gov/genbank/genomesubmit

=head2 FORMAT TBL POUR SOUMISSION


>Feature Seq1
<1    >1050    gene
                        gene          ATH1
<1    1009    CDS
                        product       acid trehalase
                        product       Athlp
                        codon_start   2
<1    >1050    mRNA
                        product       acid trehalase

>Feature Seq2
2626  2590    tRNA
2570  2535
                        product       tRNA-Phe

>Feature Seq3
1080  1210  CDS
1275  1315
                        product       actin
                        note          alternatively spliced
1055  1210  mRNA
1275  1340
                        product       actin
1055  1340  gene
                        gene          ACT
1055  1079  5'UTR
1316  1340  3'UTR

=head2 COMMANDE POUR VALIDATION ET CREATION DU FICHIER ASN

	Il faut au prealable:
	- avoir cree le BioProject (https://submit.ncbi.nlm.nih.gov/subs/bioproject/)
	- genere le fichier Template.sbt (http://www.ncbi.nlm.nih.gov/WebSub/template.cgi)

=head3 Ex melilo

	tbl2asn -i GMI11495-Rm2011G.20130218.submit.fna -t GMI11495-Rm2011G.20130218.submit.sbt -a s -f GMI11495-Rm2011G.20130218.submit.tbl -V vb -k m -H '05/01/2013' -j '[gcode=11]'
	Doc:http://www.ncbi.nlm.nih.gov/genbank/tbl2asn2/
	Sources qualifiers (-j) : http://www.ncbi.nlm.nih.gov/Sequin/modifiers.html


=cut

my @A_START_CODONS = ('atg', 'ttg', 'ctg', 'gtg');
my @A_STOP_CODONS = ('tag', 'tga', 'taa');

MAIN:
{
    my $o_param = New ParamParser('GETOPTLONG', 'gff3=s', 'tbl=s', 'sequence=s');

    $o_param->AssertFileExists('gff3');
    $o_param->AssertFileExists('sequence');
    $o_param->AssertDefined('tbl');

    my $sequence_file = Cwd::abs_path($o_param->Get('sequence'));
    my $rh_sequence   = {};
    &FastaToHash($sequence_file, $rh_sequence);

    my %h_locus_tags      = ();
    my %h_locus_tags_type = ();
    my %h_tbl             = ();
    my %h_types           = ();
    my %h_cds             = ();
    my %h_utr             = ();

    my $fh_gff3 = $o_param->GetStreamIn('gff3');
    while (my $line = <$fh_gff3>)
    {
        chomp $line;
        next if ($line =~ /^#/);
        my $rh_gff3 = {};
        &ParseGFF3Line(\$line, 1, $rh_gff3);
        my $seqid       = $rh_gff3->{'seqid'};
        my $start       = $rh_gff3->{'start'};
        my $end         = $rh_gff3->{'end'};
        my $strand      = $rh_gff3->{'strand'};
        my $rh_fragment = &GetFragment($seqid, $start, $end, $strand, $rh_sequence);

        $h_tbl{$seqid} = '' if (!defined $h_tbl{$seqid});

        if ($strand eq '-')
        {
            my $tmp = $end;
            $end   = $start;
            $start = $tmp;
        }

        my $type   = $rh_gff3->{'type'};
        my @a_note = ();
        if (($type eq 'three_prime_UTR') && ($end - $start == 0))
        {
            #SC: suppression des UTR 3' de 1nt.
            #SC: normalement cela n'est plus necessaire, c'est fait au niveau de EuGene
            #print STDERR "SKIP 3'UTR 1nt $end - $start\n";
            next;

        }
        if ($type =~ /_prime_UTR$/)
        {
#SC: dans le cas de melilo, le RNAseq avait permis de definir les UTR
#SC: chez les bacteries, si on a des UTR c'est surement lie au fait que l'on avait du RNAseq oriente pour les detecter
            push(@a_note, 'based on oriented RNAseq data');
        }

        #tag SO -> Feature EMBL/GBK
        $type =~ s/five_prime_UTR/5'UTR/;
        $type =~ s/three_prime_UTR/3'UTR/;
        $type =~ s/tRNA_primary_transcript/precursor_RNA/;
        $type =~ s/insertion_sequence/mobile_element/;

        my $gene_synonym = $rh_gff3->{'Name'};

#MESSAGE NCBI: Gene features should include only one 'gene' name; other/alternate gene names should be provided as a 'gene_synonym' or as a 'note'

#SC: dans le GFF3 de melilo, le champ product etait dans le champ Note
#SC: il faudrait voir avec Erika qui avait fait cela, il me semble que le champs "product=xxxx" de la colonne 9 correspondrait plus.
        my $product = $rh_gff3->{'Note'};

        #SC: le locus_tag (/locus_tag de EMBL) est dans la colonne 9, champs "locus_tag=xxxX"
        #SC: il etait absent du GFF3 original, nous l'avons rajoute via un script
        my $locus_tag = $rh_gff3->{'locus_tag'};

        #SC: le nom de gene (/gene de EMBL) est dans la colonne 9, champs "gn=xxxX"
        my $gene_name = $rh_gff3->{'gn'};

        #
        # GESTION DES TRANSCRITS ALTERNATIFS (demarrage de transcription differents en fonction des conditions)
        #
        #SC: Les genes avec des debuts de transcription alternatifs ont le même nom et le suffixe _TRCondition
        #SC: donc je recupere le nom de gene et je rajoute l'info=suffixe dans un champ note
        #SC: voir avec Erika si c'est le fonctionnement standard de Eugene
        if ($gene_name =~ /(\w+)(_TR\S*)/)
        {
            $gene_name = $1;
            my $note = $2;
            $note =~ s/_/ /g;
            $note =~ s/^\s+//g;
            $note =~ s/\s+$//g;
            push(@a_note, $note);
        }

        #SC: On ne garde qu'une CDS (on ne reporte pas X fois la meme)
        #SC: pour l'identifier on se sert de sa position et on garde le locus_tag de la premiere rencontree
        if ($type eq 'CDS')
        {
            if (defined $h_cds{$seqid . $start . $end})
            {
                print STDERR "REDONDANT_CDS\t$locus_tag\t$h_cds{$seqid.$start.$end}\n";
                next;
            }
            else
            {
                $h_cds{$seqid . $start . $end} = $locus_tag;
            }
        }

        #SC: idem pour les 3'UTR
        if ($type eq "3'UTR")
        {
            if (defined $h_utr{$seqid . $start . $end})
            {
                next;
            }
            else
            {
                $h_utr{$seqid . $start . $end} = 1;
            }

        }

        #SC: pour la soumission de genomes bacteriens, on ne soumet pas les EXONS et MRNA
        next if ($type eq 'exon');
        next if ($type eq 'mRNA');

        if ($type eq 'region')
        {
            next if ($gene_synonym !~ /^SM[abc]\d+$/);
            if   ($gene_synonym =~ /^oriC$/) {$type = 'rep_origin'}
            else                             {$type = 'misc_feature'}

            if ($gene_name eq 'rrn')
            {
                my $note = 'Ribosomal RNA operon';
                push(@a_note, $note);
            }
        }

#SC: les numeros EC ont ete ajoute au GFF3 dans un champs "ec_number=list,of,ec" via un script QD tout comme les locus_tags
        my @a_ecnums = ();
        @a_ecnums = split(',', $rh_gff3->{'ec_number'}) if (defined $rh_gff3->{'ec_number'});

        #SC: juste pour le comptage final pour verif
        $h_types{$type} = 0 if (!defined $h_types{$type});
        $h_types{$type}++;

        #SC: le champ product des tRNA est contraint pour la soumission
        #SC: tRNA-Aam
        #SC: cf. avec Erika si le format anti_codon=xxx_xxx est stable en sortie de Eugene

        if ($type eq 'tRNA')
        {
            $product = 'tRNA-' . $rh_gff3->{'anti_codon'};
            $product =~ s/_\w{3}$//;
        }

        #SC: cas particulier d'un tRNA de RFAM qui ne passe pas la validation
        #SC: dans ce cas le product par defaut est tRNA-Xxx
        if ($locus_tag eq 'SM2011_a0022')
        {
            $product = 'tRNA-Xxx';
            my $note = 'tRNA-SeC(p)';
            push(@a_note, $note);
        }

        #SC: les precursor_RNA ne peuvent pas avoir de champ product
        #SC:  donc on reverse dans une note
        if ($type eq 'precursor_RNA')
        {
            push(@a_note, $product);
            $product = undef;
        }

        #SC: les CDS ont obligatoirement un champ product
        #SC: valeur par defaut:hypothetical protein
        $product = 'hypothetical protein' if ((!defined $product) && ($type eq 'CDS'));

        #SC: dans le GFF3 Eugne, certains genes sont annotes via RFAM
        #SC: cela passe pas les validations
        #SC: donc je recupere cette info que je reverse dans un champ note et je nettoie le product
        if ($product =~ /(\s+RFAM:(RF\d+))/)
        {
            my $rf_link = $1;
            my $note    = $2;
            $product =~ s/$rf_link//;
            push(@a_note, $note);
        }

        #SC: les mobile_element|repeat_region|rep_origin ne peuvent pas avoir de champ product
        #SC: les mobile_element|repeat_region|rep_origin ne peuvent pas avoir de champ gene (nom de gene)
        #SC:  donc on reverse dans des notes
        if ($type =~ /mobile_element|repeat_region|rep_origin/)
        {

            push(@a_note, $product)   if (defined $product);
            push(@a_note, $gene_name) if (defined $gene_name);
            $product   = undef;
            $gene_name = undef;
        }

        #SC: Gene features should not include 'product' qualifiers.
        $product = undef if ($type eq 'gene');

        #SC: il faut supprimer les ponctuations finales
        $product =~ s/\.$//;

        #SC: Fuzzy sequences
        #SC: je faisais le controle en interne (sub GetFragment) et a partir de la je plaçais les '<' ou '>'
        if ($type eq 'CDS')
        {
            if ($rh_fragment->{start} == 0)
            {
                $start = '<' . $start;
            }
            if ($rh_fragment->{stop} == 0)
            {
                $end = '>' . $end;
            }
            if ($rh_fragment->{modulo} == 0)
            {
                die "ERROR - CDS $locus_tag not modulo 3\n";
            }

            my @a_split  = split(',', $product);
            my $productA = shift(@a_split);
            my $productB = join(',', @a_split);

            if ($productB =~ /partial|fragment|term/)
            {
                $product = $productA;
                push(@a_note, $productB);

            }
        }

        #
        #SC: on commence a ecrire le contenu du fichier tbl pour l'objet en question
        #SC: on le met dans un hach dont la cle est l'id de la pseudomolecule/contig
        #

        $h_tbl{$seqid} .= "$start\t$end\t$type\n";

        if ($type !~ /UTR|mRNA|exon/)
        {

            #SC: les gene names ne peuvent pas etre des accessions/locus_tags
            #SC: gestion particuliere des codes SMabc que l'on reverse en note
            if ($gene_name =~ /^SM[abc]\d+$/i)
            {
                $h_tbl{$seqid} .= "\t" x 3 . 'note' . "\tSynonym:" . $gene_name . "\n";
            }
            else
            {
                $h_tbl{$seqid} .= "\t" x 3 . 'gene' . "\t" . $gene_name . "\n" if (defined $gene_name);
            }


            #SC: ajout du locus_tag
            $h_tbl{$seqid} .= "\t" x 3 . 'locus_tag' . "\t" . $locus_tag . "\n" if (defined $locus_tag);

#SC: les genes synonymes ne peuvent pas etre mis ailleurs que dans des notes
#SC: nous avons fait le choix de structure ce champ note avec le mot cle "Synonym:"
#SC: les synonymes sont par exemple des locus_tag de genes de reference sur lequel on a recupere l'annotation par transfert
#SC: ex melilo: locus_tag=SM2011_a0001, synonym=SMa0001

            $h_tbl{$seqid} .= "\t" x 3 . 'note' . "\tSynonym:" . $gene_synonym . "\n" if (defined $gene_synonym);

            #SC: les pseudogenes ne peuvent pas avoir de champ product
            #SC: on reverse dans note
            if (!defined $rh_gff3->{'pseudogene'})
            {
                $h_tbl{$seqid} .= "\t" x 3 . 'product' . "\t" . $product . "\n" if (defined $product);
            }
            else
            {
                $h_tbl{$seqid} .= "\t" x 3 . 'note' . "\t" . $product . "\n" if (defined $product);
            }

        }

        #SC: pour les CDS qui ne sont pas des pseudogenes
        if (($type eq 'CDS') && (!defined $rh_gff3->{'pseudogene'}))
        {
            #SC: le protein_id est obligatoire
            #SC: nous avons opte pour le prefixe gnl|LIPM| comme namespace suivit du locus_tag
            #SC: ce protein_id devant etre unique, cela colle
            $h_tbl{$seqid} .= "\t" x 3 . 'protein_id' . "\tgnl|LIPM|$locus_tag\n";

            #SC: ajout des numeros EC
            if (scalar @a_ecnums > 0)
            {
                foreach my $ecnum (@a_ecnums)
                {
                    $h_tbl{$seqid} .= "\t" x 3 . 'EC_number' . "\t$ecnum\n";
                }
            }
        }

        #SC: ajout du tag pseudo pour les pseudogenes
        if (defined $rh_gff3->{'pseudogene'})
        {
            $h_tbl{$seqid} .= "\t" x 3 . 'pseudo' . "\n";
        }

#SC: dans notre cas tous les precursor_RNA etaient des precurseurs de tRNA, je ne pense pas que ce soit le cas general
        if ($type eq 'precursor_RNA')
        {
            $h_tbl{$seqid} .= "\t" x 3 . 'note' . "\ttRNA primary transcript\n";
        }

#SC: dans notre cas tous les mobile_element etaient des "insertion sequence", je ne pense pas que ce soit le cas general
        if ($type eq 'mobile_element')
        {
            $h_tbl{$seqid} .= "\t" x 3 . 'mobile_element_type' . "\tinsertion sequence\n";
        }

        #SC: pour les ncRNA, le champ product est obligatoire ainsi que la classe
        if ($type eq 'ncRNA')
        {
            $h_tbl{$seqid} .= "\t" x 3 . 'ncRNA_class' . "\tother\n";
            $h_tbl{$seqid} .= "\t" x 3 . 'note' . "\tbased on oriented RNAseq data\n";
            $h_tbl{$seqid} .= "\t" x 3 . 'product' . "\t" . 'novel ncRNA, function unknown' . "\n"
              if (!defined $product);
        }

        #SC: ajout des notes
        if (scalar @a_note > 0)
        {
            foreach my $note (@a_note)
            {
                $h_tbl{$seqid} .= "\t" x 3 . 'note' . "\t$note\n";
            }
        }

    }

    $fh_gff3->close;

    my $fh_tbl = $o_param->GetStreamOut('tbl');
    foreach my $seqid (keys %h_tbl)
    {
        print $fh_tbl ">Feature $seqid\n";
        print $fh_tbl $h_tbl{$seqid};
    }
    $fh_tbl->close;

    #SC: rapport
    print "#LOCUS TAGS NON UNIQUES\n";
    foreach my $locus_tag (keys %h_locus_tags)
    {
        if ($h_locus_tags{$locus_tag} > 1)
        {
            print "$locus_tag\t$h_locus_tags{$locus_tag}\t$h_locus_tags_type{$locus_tag}\n";
        }
    }

    print "#TYPES\n";
    foreach my $type (keys %h_types)
    {
        print "$type\t$h_types{$type}\n";
    }

}

sub GetFragment
{
    my ($chr, $begin, $end, $strand, $rh_sequence) = @_;

    my $seq = $rh_sequence->{$chr}->{sequence};

    die "ERROR - Strand '$strand' unknown [+/-]\n" if (($strand ne '+') && ($strand ne '-'));

    my $cds_seq = &ExtractSeq(\$seq, $begin, $end, $strand);

    #print STDERR "INFO - $cds_seq\n";
    my %h_fragment = (start => 1, stop => 1, modulo => 1);

    my $start_codon = substr($cds_seq, 0,  3);
    my $stop_codon  = substr($cds_seq, -3, 3);
    my $modulo      = length($cds_seq) % 3;

    #print STDERR length($cds_seq) . "-- $modulo \n";
    $h_fragment{start} = 0 if (!grep (/^$start_codon$/i, @A_START_CODONS));
    $h_fragment{stop}  = 0 if (!grep (/^$stop_codon$/i,  @A_STOP_CODONS));
    $h_fragment{modulo} = 0 if ($modulo != 0);

    return \%h_fragment;

}

