#
# $Id$
#

package EnzymeValidation;
use base qw( LipmObject );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

EnzymeValidation - Download the files enzyme.dat and enzclass.dat from expasy
			Corrects the EC number during the Locis initialization
           Corrects the positions using the assembly
           Create a single fasta file, using a multifasta as entry

=head1 SYNOPSIS

 my $o_env_val = New EnzymeValidation();

=head1 DESCRIPTION

 
=cut

use strict;
use warnings;
use Data::Dumper;
use Error::Simple;
use Carp;
use LipmObject;
use General;
use base qw( LipmObject );
use Tree::Simple;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1009 $ =~ /\d+/g); $r[0]};
}

use constant FULLEC_DB_URL => "ftp://ftp.expasy.org/databases/enzyme/enzyme.dat";
use constant CLASS_DB_URL => "ftp://ftp.expasy.org/databases/enzyme/enzclass.txt";

=head2 procedure _Init

 Title        : Constructor
 Usage        : my $o_enz_val = New EnzymeValidation($o_param);
 Function     : Initialize the object, downloading two files: enzyme.dat, enzclass.txt
                this function is copied from lipm_valid_ec_numbers.pl
 Args         : $debug (0/1/2, optional, default 0)
                The default is to download enzyme and class files to a temporary file
                If debug != 0, the enzyme and class files are downloaded to some file in the current directory

=cut

sub _Init
{
	my $self = shift;
	my ($debug) = @_;
	$debug = 0 unless defined $debug;
	$self->{ __no_warnings } = &FALSE;
	$self->{ __timestamp } = $self->__ComputeTimestamp();
	
	my %h_refcodes;
	$self -> { __rh_refcodes }    = \%h_refcodes;

	$self->__GetClassesInfo($debug);
	$self->__GetEnzymeInfo($debug);

	$self -> { __rh_validated }   = {};
	$self -> { __rh_transferred } = {};
	$self -> { __rh_deleted }     = {};

}

# NoWarnings - When NoWarnings has been called, no warnings are issued when an ec is deleted or transferred
sub NoWarnings     { my $self = shift; $self -> { __no_warnings } = &TRUE; }
sub GetValidated   { my $self=shift; return $self->{__rh_validated}; }
sub GetTransferred { my $self=shift; return $self->{__rh_transferred}; }
sub GetDeleted     { my $self=shift; return $self->{__rh_deleted}; }

sub __ComputeTimestamp
{
	my $self = shift;
	my @a_time = localtime;
	$a_time[5] += 1900;
	$a_time[4] += 1;
	return $a_time[5].'-'.$a_time[4].'-'.$a_time[3];
}

=head2 function Check

 Title        : Check
 Usage        : my $ec_number = $o_enz_val->Check(2.6.1.73,2.6.1.74)
 Function     : Check the ec_number, and return The not changed ec, OR the transferred ec, OR nothing (if ec deleted)
 Args         : One or several EC_numbers to check
 Return       : A ref to an array of EC_numbers

=cut
sub Check {
	my $self = shift;
	my @a_ec_numbers = @_;
	my $no_warnings = $self -> { __no_warnings };

	my $rh_refcodes    = $self -> { __rh_refcodes };
	my $rh_validated   = $self -> { __rh_validated };
	my $rh_transferred = $self -> { __rh_transferred };
	my $rh_deleted      = $self -> { __rh_deleted };
	
	#my @a_ec_numbers = split ',',$ec_numbers;
	my @a_t_ec_numbers;
	foreach my $ecnum (@a_ec_numbers)
	{
		$ecnum =~ s/.*([0-6]\.[0-9-]+\.[0-9-]+\.n?[0-9-]+).*/$1/;
		if (exists $rh_refcodes->{$ecnum} )
		{
			my $new_ecnum = $rh_refcodes->{$ecnum};
			 if ( $ecnum eq $new_ecnum )
			 {
				 $rh_validated ->{$ecnum} = 1;
			 }
			 else
			 {
				 $rh_transferred -> {$ecnum} = $new_ecnum;
				warn $self -> GetWarningHeader() . "$ecnum transferred to $new_ecnum\n" unless ( $no_warnings == &TRUE );

			 }
			 push @a_t_ec_numbers, $new_ecnum;
		 }
		 else
		 {
			 $rh_deleted -> {$ecnum} = 1;
			warn $self -> GetWarningHeader() . "$ecnum deleted\n" unless ( $no_warnings == &TRUE );
		 }
	}
	#return join (',', @a_t_ec_numbers);
	return \@a_t_ec_numbers;
}

# PRIVATE METHODS

# Get class information from the enzyme database, then reformat it to the hash
# Arg: $debug
# return: nothing
# modifies: $self -> { __rh_refcodes };
sub __GetClassesInfo
{
	my $self = shift;
	my ($debug) = @_;
	
	my $rh_refcodes = $self -> { __rh_refcodes };
	my @a_ref = $self->__GetClassTxt();

	if ( $debug > 0 )
	{
		# Saving downloaded file for later reference
		my $f_enzclass = "enzclass-".$self->{__timestamp}.".txt";
		print STDERR "INFO: Now saving data downloaded from " . &CLASS_DB_URL . "to file $f_enzclass \n" if ($debug >= 1);
		my $fh_enzclass= &GetStreamOut($f_enzclass);
		print $fh_enzclass @a_ref;
		$fh_enzclass->close();
	}
	chomp @a_ref;
	
	foreach my $line (@a_ref)
	{
		# 1.1.1.-
		chomp $line;
		if ($line =~ /^([^\.]+\.[^\.]+\.[^\.]+\.\S+)/)
		{
			my $code = $1;
			$code =~ s/\s+//g;
			$rh_refcodes->{$code} = $code;
		}
	}
}

# Get code information from the enzyme database, then reformat it to the hash
# Arg: $debug
# return: nothing
# modifies: $self -> { __rh_refcodes };
sub __GetEnzymeInfo
{
	my $self = shift;
	my ($debug) = @_;

	my $rh_refcodes = $self -> { __rh_refcodes };
	my @a_ref = $self->__GetEnzymeDat();

	if ($debug > 0 )
	{
		# Saving downloaded file for later reference
		my $f_enzyme = "enzyme-".$self->{__timestamp}.".dat";
		print STDERR "INFO: Now saving data downloaded from " . &FULLEC_DB_URL . " to file $f_enzyme \n" if ($debug >= 1);
		my $fh_enzyme= &GetStreamOut($f_enzyme);
		print $fh_enzyme @a_ref;
		$fh_enzyme->close();
	}

	my $last_id = undef;
    foreach my $line (@a_ref)
    {
        chomp $line;
        if ($line =~ /^ID\s+([^\.]+\.[^\.]+\.[^\.]+\.[^\.]+)/)
        {
            my $code = $1;
            $code =~ s/\s+//g;
            $rh_refcodes->{$code} = $code;
            $last_id = $code;
        }
        elsif ($line =~ /^DE\s+Transferred\s+entry:\s+([^\.]+\.[^\.]+\.[^\.]+\.\w+)\.$/)
        {
			my $newcode = $1;
            $newcode =~ s/\s+//g;
            if ($newcode =~ /n/)
            {
				
            	print STDERR "$last_id should be transferred to preliminary code $newcode ==> SKIPPED, will be DELETED\n"  if ($debug >= 2);
            	delete $rh_refcodes->{$last_id};
			}
			else
			{
				print STDERR "$last_id will be transferred to $newcode\n"  if ($debug >= 2);
				$rh_refcodes->{$last_id} = $newcode;
			}
			
		}
		
		# DE   Transferred entry: 1.1.1.198, 1.1.1.227 and 1.1.1.228.
		elsif ($line =~ /^DE\s+Transferred\s+entry:\s+(.+)$/)
        {
			my $newcodes = $1;
            if ($newcodes =~ /,|and/)
            {
				# 1.1.1.198, 1.1.1.227 and 1.1.1.228. => an array with the three codes
				$newcodes =~ s/and|,//g;
				$newcodes =~ s/ +/ /g;
				$newcodes =~ s/\.$//;
				my (@a_newcodes) = split ' ',$newcodes;
				my $class = $self->__Codes2Class(@a_newcodes);
				if ( exists $rh_refcodes->{$class } )
				{
					$rh_refcodes->{$last_id} = $class;
					print STDERR "$last_id could be transferred to many codes $newcodes ==> will be transferred to the class $class\n"  if ($debug >= 2);
				}
				else
				{
					print STDERR "$last_id could be transferred to many codes $newcodes ==> No suitable class identified ($class) ==> SKIPPED, will be DELETED\n"  if ($debug >= 2);
					delete $rh_refcodes->{$last_id};
				}
			}
		}
		elsif ($line =~ /^DE\s+Deleted\s+entry\./)
        {
            print STDERR "$last_id will be deleted\n" if ($debug >= 2);
			delete $rh_refcodes->{$last_id};
		}
    }
}

# Retrieving the files from the internet
sub __GetEnzymeDat
{
	my $self = shift;
	my $cmd = "wget -q -O - " .&FULLEC_DB_URL;
    my @a_ref = `$cmd`;
	carp "ERROR with $cmd - status=$?" if ($? ne '0');
	return @a_ref;
}

sub __GetClassTxt
{
	my $self = shift;
	my $cmd = "wget -q -O - " . &CLASS_DB_URL;
	my @a_ref = `$cmd`;
	carp "ERROR with $cmd - status=$?" if ($? ne '0');
	return @a_ref;
}

# Identify a common class from an array of codes
# Arg: an array of codes, ex. 2.1.1.3,2.1.1.36
# Return: a class (NOT VALIDATED) ex. 2.1.1.-
sub __Codes2Class
{
	my $self = shift;
	my @a_codes = @_;
	
	# Building a tree from the codes
	my $root = Tree::Simple->new('ROOT');
	foreach my $code (@a_codes)
	{
		my @a_nodes = split /\./,$code;
		my $t = $root;
		foreach my $n (@a_nodes)
		{
			# No child: create a new child
			if ( $t->getChildCount()==0)
			{
				$t = $t->addChild(Tree::Simple->new($n));
				$t = $t->getChild(0);
			}
			
			# One child: walk a new step
			else
			{
				# same as $n: just keep walking
				$t = $t->getChild(0);
				if ( $t->getNodeValue() eq $n )
				{
					next;
				}
				
				# '-': finished with this code
				elsif ( $t->getNodeValue() eq '-' )
				{
					last;
				}
				
				# any other value: replace by a '-' up to level 4
				else
				{
					do
					{
						$t -> setNodeValue('-');
						last if ( $t->getDepth() == 3 );
						if ( $t->getChildCount()==0 )
						{
							$t = $t -> addChild(Tree::Simple->new('-'));
						}
						else
						{
							$t = $t -> getChild(0);
						}
					}
					while ( &TRUE );
				}
			}
		}
	}
	
	# Build the class
	my @a_class=();
	my $t = $root;
	$t->traverse(sub {
		my ($t) = @_;
		push @a_class,$t->getNodeValue();
	});
	return join '.',@a_class;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;

