#!/usr/bin/perl

#
# $Id
#

# Copyright INRA

# Erika.Sallet@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Schiex@toulouse.inra.fr

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod 

=head1 NAME


=head1 SYNOPSIS

export EGNEP=/egnep/path
export EUGENEDIR=/eugene/c++/path
$EGNEP/bin/int/egn-euk.pl --indir $PWD/indir/ --outdir $PWD/outdir/ --workingdir $PWD/wordir --cfg $EGNEP/cfg/egn-euk.cfg

=head1 DESCRIPTION


=cut

BEGIN
{
    $ENV{LANG} = "C";

    use Cwd 'abs_path';
    use File::Basename;
    my ($dirprg) = dirname(abs_path($0));
    unshift(@INC, "$dirprg/../../lib");
    unshift(@INC, "$dirprg/../../corelib");

}

use strict;
use warnings;
use Carp;
use ParamParser;
use General;
use GeneralBioinfo;
use Logger;
use Cwd;
use List::Util  qw(first max maxstr min minstr reduce shuffle sum);
use threads;
use threads::shared;
use Data::Dumper;

# GLOBAL VAR
my $LOGGER_LEVEL_EXEC = 'DE';
my $o_logger;
my $logger_stamp;

my $DEBUG;
my $CLEAN;
my $FORCE;               # if 1 relaunch all computing even if .success file exists
my $INDEPENDENT_STRAND;  # if 1 annotate independently the two strands

my $TRNASCAN_EXT    = "trnascan";
my $RNAMMER_EXT     = "rnammer";
my $RFAMSCAN_EXT    = "rfamscan";
my $PRODIGAL_EXT    = "prodigal";
my $MAPPING_EXT     = "bed.specific.gz";
my $WIG_EXT         = "wig2expr";
my $EXPR_POS_EXT    = "position";
my $EXPR_REG_EXT    = "region";
my $TSTART_EXT      = "tstart";
my $TSTART_POS_EXT  = "tstart_position";
my $LTRHARVEST_EXT  = "ltrharvest";
my $REPET_EXT       = "repet_noexpressed_nosimprot";
my $ALLSEQFILE_NAME = "sequences";

my $DBDIR          = "db";
my $ROOT_ANNOT_DIR = "annotationV";
my $TRAINDIR       = "training";

my $MODE = "EUK";
my $VERSION;
my $DEBUG_DEFAULT = 0;
my $LINELEN       = 80;
my $PRINTINDENT   = 0;
my $DEFAULT_NCPUS = 8;
my $PROTECT_MARGIN=100;
my $MIN_LEN_SHORT_UNSPLICED = 10000000;

MAIN:
{
    my $o_param = New ParamParser(
                                  'GETOPTLONG',         \&Usage,
                                  'indir=s',            'outdir=s',
                                  'workingdir=s',       'cfg=s',
                                  'no_ncrna_detection', 'no_prodigal',
                                  'no_rfamscan',        'no_trnascan',
                                  'no_rnammer',         'no_repeat_masking',
                                  'debug=i',            'uniq=s', 'ncpus=i',
                                  'no_clean',           'no_red', 'force'
                                  );
    $o_param->SetBehaviour('ignore_space');
    $o_param->Update('ENV', "A");
    $o_param->SetUnlessDefined('ncpus',$DEFAULT_NCPUS);

    $o_param->AssertDirExists('EGNEP');
    $o_param->AssertDirExists('EUGENEDIR');
    $o_param->AssertFileExists('cfg');
    $o_param->AssertFullPath('cfg');
    $o_param->Update($o_param->Get('cfg'), 'A');
    $o_param->SetBehaviour('use_substitution_table');
    $o_param->SetSubstitution('%i', $o_param->Get('EGNEP'));
    $o_param->SetSubstitution('%e', $o_param->Get('EUGENEDIR'));
    $o_param->Update($o_param->Get('cfg'), 'U');
    $ENV{EGNPP} = $o_param->Get('EGNEP');    # Need for external program (ex get_BlastX.pl)
    my $cmdline = "$0 @ARGV\n";
    $VERSION = &GetVersion($o_param->Get('EGNEP'));
    $FORCE   = 0;
    $FORCE   = 1 if ($o_param->IsDefined('force'));
    &AssertParameters($o_param);

    ###################  Global parameters ##############################

    $DEBUG = $o_param->Get('debug');
	$INDEPENDENT_STRAND = $o_param->Get('independent_strand_annotation');
    my $rootworkingdir = $o_param->Get('workingdir');           # Working directory
	$rootworkingdir =~ s/\/$//;
	$o_param->Set('rootworkingdir', $rootworkingdir);           # Save the root working directory
	$o_param->Set('seq_fof', "$rootworkingdir/entries.fof");    # File of files of the sequences
	
	$o_param->SetUnlessDefined('output_prefix', $ALLSEQFILE_NAME);
	my $out_prefix = $o_param->Get('output_prefix');
	$out_prefix =~ s/\s+/_/g;
	$o_param->Set('output_prefix', $out_prefix);
	
	$o_param->SetUnlessDefined('chr_sorting_by_size', 0);
    my $allseqfile     = "$rootworkingdir/$ALLSEQFILE_NAME";    # multifasta of all sequences

    # Uniq code to generate temporary files/directory
    $o_param->SetUnlessDefined('uniq', time . "." . $$);        # Uniq value
    my $uniq = $o_param->Get('uniq');

    # Log file
    my $loggerfile = "$rootworkingdir/logger.$uniq.txt";
    $o_logger = New Logger($loggerfile, $LOGGER_LEVEL_EXEC, 'NOCOMPRESSION');
    $logger_stamp = $o_logger->CalcStamp();

    # Report file
    my $reportfile = $o_param->Get('outdir') . "/".$o_param->Get('output_prefix'). ".egnep_report.txt";
	unlink($reportfile) if (-f $reportfile);
    $o_param->Set('report', $reportfile);

    &PrintModeInfo($o_param, $loggerfile, $cmdline);

    ################## Create the working tree #####################

    my @a_seq;    # Array of the genomic sequences
    &CreateWorkSeqTree($o_param, $o_param->Get('genomedir'),
                       \@a_seq, $rootworkingdir, $ALLSEQFILE_NAME, $o_logger, $logger_stamp, $uniq);

    if ($MODE eq "EUK")
    {
        &EuGeneEuk($o_param, $rootworkingdir, $allseqfile, $o_logger, $logger_stamp, \@a_seq,$uniq);
    }
    else
    {
        &EuGeneProk($o_param, $rootworkingdir, $allseqfile, $ALLSEQFILE_NAME, $o_logger, $logger_stamp, $uniq, \@a_seq);
    }
}

sub GetVersion
{
    my ($EGNPP) = @_;

    my $version_file = "$EGNPP/VERSION";

    return "0" if (!-e $version_file);

    my $f_      = &General::GetStreamIn($version_file);
    my $version = <$f_>;
    chomp($version);
    $f_->close;

    return $version;
}

sub EuGeneEuk
{
    my ($o_param, $rootworkingdir, $allseqfile, $o_logger, $logger_stamp, $ra_seq,$uniq) = @_;

    my $paramDir = &GetParamDir($o_param);

    ##################################################################################
    # 1 BlastX
    my %h_database;    # BlastX databases
    &GetBlastXDb($o_param, \%h_database);
    &PrintSection("Protein database cleaning");
    &CleanProtDb($o_param, \%h_database);
    &PrintSection("Protein sequence similarity search");
    &BlastXAllSeq($o_param, $ra_seq, \%h_database, "prot_sim");
    ##################################################################################

    ##################################################################################
    # Find transcriptome and proteome to use for training
    my $trainestnb    = $o_param->Get('training_dbest_number');
    my $train_est_db  = $o_param->Get("est_" . $trainestnb . "_file");
    my $train_prot_db = $h_database{$o_param->Get('training_dbprot_number')}{file};
    ##################################################################################

    ##################################################################################
    # 2 Mapping of the transcriptomes
    &PrintSection("Transcriptome mapping");
    &MappingEstEuk($o_param, $allseqfile);
    ##################################################################################

    ##################################################################################
    # 3 Train IMM and search non canonical spliced sites
    &PrintSection("IMM model building");
    my $matrixfile = "$paramDir/eugene.mat";
    &BuildEGNMatrixEuk($o_param, $allseqfile, $train_est_db, $train_prot_db, $matrixfile, $o_logger, $logger_stamp);
    &PrintSection("Non canonical spliced site detection");
    &DetectNonCanSites($o_param, $o_param->Get('report'));
    ##################################################################################

    ##################################################################################
    # 4 ncRNA detection
    &PrintSection("ncRNA detection");
    &AddPrintIndent(2);
    &GetRfamScanAllSeq($o_param, $ra_seq);
    &GetRnammerAllSeq($o_param, $ra_seq);
    &GettRNAscanSEAllSeq($o_param, $ra_seq);
    &AddPrintIndent(-2);
    $o_param->Set('workingdir', $rootworkingdir);
    ##################################################################################

    ##################################################################################
    # 5 Add additional evidences
    if ($o_param->IsDefined('additional_list') && $o_param->Get('additional_list') ne "" )
    {
        &PrintSection("Complete with additional data");
	&LoadAdditionalEvidences($o_param);
    }
    ##################################################################################

    ##################################################################################
    # 6 Search specific repeat domains
    if (!$o_param->IsDefined('no_repeat_masking'))
    {
        &PrintSection("Species specific repeat domains building");
        if ($o_param->IsDefined('species_repeat_domains'))
        {
            print STDOUT "Use " . $o_param->Get('species_repeat_domains') . "\n";
        }
        else
        {
            my $repet_domain = &GetDbDir($o_param) . "/SpeciesRepeatDomain.fa";
            &SearchSpeciesRepeatDomains($o_param, $repet_domain, $allseqfile, $ra_seq, \%h_database, $matrixfile);
            $o_param->Set('species_repeat_domains', $repet_domain);
        }
        &StatSpeciesRepeatDomains($o_param);
    }
    ##################################################################################
   

    ##################################################################################
    # 7 Search repeat regions
    if (!$o_param->IsDefined('no_repeat_masking'))
    {
        &PrintSection("Repeat search");
        &SearchRepet($o_param, $allseqfile);
        &StatRepeatSearch($o_param);
    }
    ##################################################################################

    ##################################################################################
    # 8 EuGene annotation
    &PrintSection("Sequence annotation");
    my $annotv1Dir = &CreateAnnotTree($o_param, 1);
    my $egnv1Dir   = $annotv1Dir . "/raw_eugene";
    my $cfg_file; 
    my @a_res_gff3 = ();                              # Array of EuGene gff3 result files

    my $repeatmode = ($o_param->IsDefined('no_repeat_masking')) ? 0 : 1;
	if (!$INDEPENDENT_STRAND)
	{
		$cfg_file = "$paramDir/eugenev1.par";
		&BuildEGNEukParameterFile($o_param, $cfg_file, \%h_database, $repeatmode, 0);
		&LaunchAllEGN($o_param, $ra_seq, $cfg_file, $matrixfile, $egnv1Dir, "ANNOT", 0, "");
	}
	else
	{
		# strand forward
		$cfg_file = "$paramDir/eugenev1.FW.par";
		&BuildEGNEukParameterFile($o_param, $cfg_file, \%h_database, $repeatmode, 1);
		&LaunchAllEGN($o_param, $ra_seq, $cfg_file, $matrixfile, $egnv1Dir, "ANNOT", 0, "FW");
		
		# strand reverse
		$cfg_file = "$paramDir/eugenev1.RV.par";
		&BuildEGNEukParameterFile($o_param, $cfg_file, \%h_database, $repeatmode, -1);
		&LaunchAllEGN($o_param, $ra_seq, $cfg_file, $matrixfile, $egnv1Dir, "ANNOT", 0, "RV");

		# cat the two annotation files
		foreach my $seq (@$ra_seq)
		{     
			my ($seqpath, $seqname, $seqid, $seqlen) = &GetSeqInfo($seq);
			my $catcmd = "cat $egnv1Dir/FW$seqname.gff3 $egnv1Dir/RV$seqname.gff3 > $egnv1Dir/$seqname.unsort.gff3";
			&ExeCmd($catcmd, $o_logger, $logger_stamp);

			# sort 
			my $sortcmd = $o_param->Get('prg_biofileconverter');
			$sortcmd    .= " --kingdom eukaryote --input-ann-format GFF3 --output-ann-format GFF3 --input-ann-file  $egnv1Dir/$seqname.unsort.gff3 --output-ann-file $egnv1Dir/$seqname.gff3 > /dev/null 2>&1";
   
			&ExeCmd($sortcmd, $o_logger, $logger_stamp);

			&Gzip($o_logger, $logger_stamp, "$egnv1Dir/$seqname.unsort.gff3", "$egnv1Dir/FW$seqname.gff3","$egnv1Dir/RV$seqname.gff3");
		}
	}

	# Transfert ncRNA annotations
	&AddPrintIndent(2);
	my @a_tmp=();
	&Multithreading(\&SliceTransferAnnotncRNA,$o_param->Get('ncpus'), $ra_seq, \@a_tmp, $o_param, $egnv1Dir);
	&AddPrintIndent(-2);
	foreach my $seq (@$ra_seq)
    {
        my ($seqpath, $seqname, $seqid, $seqlen) = &GetSeqInfo($seq);
        push(@a_res_gff3, "$egnv1Dir/$seqname.transfert.gff3");
    }

	# consolidation
    my $outdir = $o_param->Get('outdir');    # Output directory
    &Consolidate($o_param, $outdir, \@a_res_gff3, $allseqfile, $o_logger, $logger_stamp);
    ##################################################################################

    ##################################################################################
    # 9 Cleaning work directories
    foreach my $seq (@$ra_seq)
    {
		my $workingdir = &GetWorkDir($seq, $o_param);

        # Compress all the temporary files
        my $cmd = "gzip -qr $workingdir/";
        &ExeCmd($cmd, $o_logger, $logger_stamp);
    }
    my $trainingroot = &GetTrainDir($o_param) . "/$ALLSEQFILE_NAME.training";
    my $cmd          = "gzip -f $trainingroot*";
    #&ExeCmd($cmd, $o_logger, $logger_stamp);
	
	$cmd = "rm -rf $outdir/full_eugene";
	&ExeCmd($cmd, $o_logger, $logger_stamp);
	
    ##################################################################################

    return;
}




sub LoadAdditionalEvidences
{
	my ($o_param) = @_;
	
	my $list  = $o_param->Get('additional_list');
	my @a_list = split(/\s+/, $list);
	
	foreach my $nb (@a_list)    # '2', '3'
	{
		my $additionalfile = $o_param->Get("additional_" . $nb . "_file");
		my $ext = &GetGFF3Ext($additionalfile);
		# split by sequence (and by strand if strand specific annotation)
		&SplitBySeq($o_param, $additionalfile, $ext);	
	}
	return;
}

=head2 procedure SplitBySeq
	
 Title        : SplitBySeq
 Usage        : 
 Procedure    : Split gff3 file by sequence. Save result in the directory of the corresponding sequence
 Prerequisite : prg_splitgff3 exists
 Args         : $o_param
                $gff3 input file to split
                $ext : extension to the new files
 Globals      : $INDEPENDENT_STRAND

=cut
sub SplitBySeq
{
	my ($o_param, $gff3, $ext) = @_;
	
	my ($path, $name) = &GetPathAndName($gff3);	
	my $step          = "Add additional file $name";
	my $firstfile     = &GetFirstSeqPath($o_param).".$ext.gff3";
	
    return if (&IsValidatedStep($firstfile, $step));
	
	&PrintStart($step);
	
	my $seq_fof = $o_param->Get('seq_fof'); 
	my $cmd = $o_param->Get('prg_splitgff3') . " --in $gff3 --fof $seq_fof --ext $ext";
	$cmd .= " --splitbystrand" if ($INDEPENDENT_STRAND);
	$cmd .= " && ". &GetValidateFileCmd($o_param, "$firstfile");
	&ExeCmd($cmd, $o_logger, $logger_stamp);
	
	&PrintStop($step);
	
	return;
}


sub SliceTransferAnnotncRNA
{
my($i,$ra_tasks,$ra_inputs,$sa_results,$o_param,$egnv1Dir) = @_;

	my $seq = $$ra_inputs[$i];
	my ($seqpath, $seqname, $seqid, $seqlen) = &GetSeqInfo($seq);
	&TransferAnnotncRNA($o_param, $seq, "$egnv1Dir/$seqname.gff3", "$egnv1Dir/$seqname.transfert.gff3");
}

=head2 procedure BlastXAllSeq
	
 Title        : BlastXAllSeq
 Usage        : 
 Procedure    : Search blastx command needed to be lauch for each sequence avd each database
                parallelize blastx against short sequences
                then call sequencially blastx against long sequences
                Results are gff3 files. For sequence $seq: $seq.blast$nb.gff3
 Prerequisite : prg_mapwithblast_limit has to be a key of o_param
 Args         : $ra_seq ref a an array of sequences
                $rh_database ref to a hash of proteome database:
                 key = prot database number or id, 
                 value=prot database path
                $step String to descrbe the task: just use for tmp file names
 Globals      : none

=cut

sub BlastXAllSeq
{
    my ($o_param, $ra_seq, $rh_database, $step) = @_;

    my $cmd;
    my %h_all_db;
    my $seqlenlimit = $o_param->Get('prg_mapwithblast_limit');
    

    $step =~ s/ +//g;
    my $rootworkdir = $o_param->Get('workingdir');
    my $blastdir    = "$rootworkdir/BLAST_$step\_" . $o_param->Get('uniq');
    mkdir($blastdir) if (!-d $blastdir);

    my $blastxcmd_sh         = "$blastdir/blastx.cmd.sh";
    my $blastxcmd_paraloop   = "$blastdir/blastx.cmd.paraloop";
    my $f_blastxcmd_sh       = &General::GetStreamOut($blastxcmd_sh);
    my $f_blastxcmd_paraloop = &General::GetStreamOut($blastxcmd_paraloop);

    foreach my $seq (@$ra_seq)
    {
        my ($seqpath, $seqname, $seqid, $seqlen) = &GetSeqInfo($seq);
		my $workingdir = &GetWorkDir($seq, $o_param);
        $o_param->Set('workingdir', $workingdir);

        foreach my $nb (sort keys %$rh_database)
        {
            my $db = $$rh_database{$nb}{file};
            my ($b, $dname) = &GetPathAndName($db);

            my $outfile = "$seq.blast$nb";
            my $step    = "$seqname BlastX $dname";

            next if (&IsValidatedStep("$outfile.gff3", $step));

            $h_all_db{$dname} = 1;
            $cmd = "export EGNPP=" . $o_param->Get('EGNEP') . ";";
           $cmd .= "export EGNEP=" . $o_param->Get('EGNEP') . ";";
            $cmd .= "cd $workingdir;";
            $cmd .=
                $o_param->Get('prg_get_blastx')
              . " --sequence $seq --cfg "
              . $o_param->Get('cfg')
				. " --db $db --outfile $outfile --workingdir $workingdir";
			$cmd .= " --protein_similarity_search_method " . $o_param->Get('protein_similarity_search_method');
            $cmd .= " --verbose " if ($DEBUG);

            # Mv $seq.blast$ext in $workingdir (we only keep the .gff3)
            $cmd .= ";mv $outfile $workingdir/";

			if ( exists($$rh_database{$nb}{pcs}) ) # actuellement la cle n'existe pas pour les repeats
			{
				# filter blastx results
				if ( -e "$db.seqlen" )
				{
					$cmd .= "; ".$o_param->Get('prg_blastx_filter')." --seqlen $db.seqlen --scov $$rh_database{$nb}{pcs} --in $outfile.gff3 ";
				}
				else
				{
					print STDERR "WARNING: unable to apply prg_blastx_filter, $db.seqlen is missing\n";
				}

				# split by strand
				if ($INDEPENDENT_STRAND)
				{
					$cmd .= " && ".&GetSplitByStrandCmd($o_param, "$outfile.gff3");
				}
			}
            my $validcmd = &GetValidateFileCmd($o_param, "$outfile.gff3");
            $cmd .= " && $validcmd";

            if ($seqlen >= $seqlenlimit)
            {
                print $f_blastxcmd_sh "$cmd\n";
            }
            else
            {
                print $f_blastxcmd_paraloop "$cmd\n";
            }
        }
    }
    $f_blastxcmd_sh->close;
    $f_blastxcmd_paraloop->close;
    $o_param->Set('workingdir', $rootworkdir);

    my $msg          = "BlastX " . join(" ", sort keys(%h_all_db));
    my $param_msg    = "  BLASTX PARAMETERS=" . $o_param->Get('blast_param');

	my $method          = $o_param->Get('protein_similarity_search_method');
    my $reduction_param = "NO_PROTEIN_DATABASE_REDUCTION";
	$reduction_param = "  UBLAST PARAMETERS=" . $o_param->Get('prg_usearch_ublast_param') if ($method eq "ublast_blastx");
	$reduction_param = "  DIAMOND PARAMETERS=" . $o_param->Get('prg_diamond_param')       if ($method eq "diamond_blastx");

    if (-s $blastxcmd_paraloop || -s $blastxcmd_sh)
    {
        &PrintStart($msg);
        &PrintMore($param_msg);
        &PrintMore($reduction_param);
    }

    # 1 launch blastx with paraloop
    &Paraloop($o_param, $blastxcmd_paraloop, $blastdir);

    # 2 launch blastx with shell
    if (!-z $blastxcmd_sh)
    {
        $cmd = $o_param->Get('prg_shell') . " $blastxcmd_sh";
        &ExeCmd($cmd, $o_logger, $logger_stamp);
    }

    &PrintStop($msg) if (-s $blastxcmd_paraloop || -s $blastxcmd_sh);

    return;
}

=head2 procedure Search Repet
	
 Title        : SearchRepet
 Usage        : 
 Function     : 1 create dabase of repeat : repbase+repeatdomain of species
                2 launch ltrharvest then mask LTR region of the genome
                3 blast the repeast database against the masked genome
                merge the resulted region with ltrharvest result (and mitehunter results @TODO)
                remove region which are expressed or with a protein similarity
                genereate a eugene formated file of repeat regions.
 Prerequisite : key 'species_repeat_domains' exists
 Args         : $allseqfile multifasta
 Globals      : none

=cut

sub SliceLTRAnalysis
{
	my($i,$ra_task,$ra_inputs,$ra_outputs,$o_param) = @_;
	
	my $seq = $$ra_inputs[$i];

	my $workingdir = &GetWorkDir($seq, $o_param);
	$o_param->Set('workingdir', $workingdir);

	# Launch LTRharvest
	my $ltrharvest_gff3 = &GetLTRHarvest($o_param, $seq, $workingdir);

	# Mask regions of the genomic sequence where LTRHarvest found repet
	my $maskedallseqfile = "$seq.masked";
	&MaskSequence($seq, $maskedallseqfile, $ltrharvest_gff3);
}

sub SearchRepet
{
    my ($o_param, $allseqfile) = @_;

    my $rootworkdir  = $o_param->Get('workingdir');
    my $uniq         = $o_param->Get('uniq');
    my $repet_domain = $o_param->Get('species_repeat_domains');

	# 0 Run REpeat Detector
	&Red($o_param, $allseqfile, "red", $rootworkdir) if (!$o_param->IsDefined('no_red'));
	
    # 1 Create a database of repet domain: RepBase + species repet domains
    my %h_repeat_db;
    my $repetdomain_and_repbase = &GetDbDir($o_param) . "/RepBase+SpeciesRepeatDomain.fa";
    &CreateDatabase($o_param, $repet_domain, $o_param->Get('repeat_sequence_db'), $repetdomain_and_repbase);
    $h_repeat_db{'rep'}{file} = $repetdomain_and_repbase;

    # 2 LTR masking of genomic sequences
    &PrintSection("LTR region search and masking");
    my @a_masked_seq;

	&AddPrintIndent(2);
    my $f_fof = $o_param->GetStreamIn('seq_fof');
	my @a_fof_raw = ();
    while (my $seq = <$f_fof>)
    {
        chomp($seq);
		push(@a_fof_raw,$seq);
        my $maskedallseqfile = "$seq.masked";
        push(@a_masked_seq, $maskedallseqfile); # lui je le laisserai trie l'ordre original pour suivre plus facilement l'avancee des calculs
	}
    $f_fof->close;

	my @a_fof = shuffle @a_fof_raw; # je melange les fichiers pour ne pas avoir les plus gros dans le meme block
	my @a_tmp=();
	&Multithreading(\&SliceLTRAnalysis,$o_param->Get('ncpus'),\@a_fof,\@a_tmp,$o_param);
	&AddPrintIndent(-2);
	
	# JG 20160425: avant de lancer le blastx je vais masker avec les resultats de red, ca ne sert a rien d'avoir des hits dans des regions qui seront masquees
	if (!$o_param->IsDefined('no_red'))
	{
		$f_fof = $o_param->GetStreamIn('seq_fof');
		while (my $seq = <$f_fof>)
		{
			chomp($seq);
			if ( -e "$seq.red.gff3" )
			{
#			print STDERR "je rejoute le masquage avec Red avant de faire le BX\n";
				my $maskedallseqfile = "$seq.masked";
				# en entree j'ai le fichier deja masque par ltrharvest
#			system("cp $maskedallseqfile $maskedallseqfile.ltrharvestonly");
				&MaskSequence($maskedallseqfile, $maskedallseqfile, "$seq.red.gff3");
			}
		}
		$f_fof->close;
	}

    $o_param->Set('workingdir', $rootworkdir);

    # 3 Blastx Repet sequences against masked genomic sequences
    &PrintSection("Repeat region similarity search");
    &BlastXAllSeq($o_param, \@a_masked_seq, \%h_repeat_db, "repeat");


    # 5 Build egn repeat file
    &PrintSection("Repeat region selection");
	&AddPrintIndent(2);
	&Multithreading(\&SliceBuildEGNRepetFile,$o_param->Get('ncpus'),\@a_fof,\@a_tmp,$o_param);
	&AddPrintIndent(-2);
    return;
}

sub SliceBuildEGNRepetFile
{
my($i,$ra_task,$ra_inputs,$ra_outputs,$o_param) = @_;

	my $seq = $$ra_inputs[$i];
    my $workingdir = &GetWorkDir($seq, $o_param);
	my @a_repetfiles;
	push(@a_repetfiles, "$seq.$LTRHARVEST_EXT.gff3");
	push(@a_repetfiles, "$seq.masked.blastrep.gff3");
	push(@a_repetfiles, "$seq.red.gff3") if (!$o_param->IsDefined('no_red'));

	#push(@a_repetfiles, $mitehunter_gff3); # MITEHunter?
	&BuildEGNRepetFile($o_param, $seq, \@a_repetfiles, $workingdir);
}

=head2 procedure BuildEGNRepetFile
	
 Title        : BuildEGNRepetFile
 Usage        : 
 Function     : build $seq.$REPET_EXT.gff3, a egn formated gff3 of the putative repeted regions 
                with no overlap with translated or transcripted regions
                               
 Prerequisite : 
 Args         : $seq
                $allrepet: gff3 of repeted regions of the sequence
 Globals      : none

=cut

sub BuildEGNRepetFile
{
    my ($o_param, $seq, $ra_repetfiles, $workdir) = @_;

    my ($seqpath, $seqname) = &GetPathAndName($seq);

    my $okrepetfile4egn = "$seq.$REPET_EXT.gff3";
    my $step            = "$seqname Extract repeated regions";

    return if (&IsValidatedStep($okrepetfile4egn, $step));

	my $prevdir = cwd();
	chdir($workdir);

	my %h_options=();

	$h_options{tmpdir}="$workdir";

    &PrintStart($step);

    my $allrepet = "$workdir/$seqname.allrepetregions.gff3";
    &MergeGFF3($o_param, $seq, $ra_repetfiles, $allrepet,"repeat_region",&FALSE,0,\%h_options);

    my @a_est_list    = split(/\s+/, $o_param->Get('est_list'));
    my @a_blastx_list = split(/\s+/, $o_param->Get('blastx_db_list'));

    # get est, blastx and ncRNA predictor result files
    # these regions has to be substracted to repet regions
    my @a_preservfiles=();
    foreach my $nb (@a_est_list)    # '2', '3'
    {
		if ($o_param->Get("est_$nb\_preserve"))
		{
			push(@a_preservfiles, "$seq.est$nb.gff3");
			push(@a_preservfiles, "$seq.est$nb.rm_unspliced_$MIN_LEN_SHORT_UNSPLICED.gff3") if ($o_param->Get("est_$nb\_remove_unspliced") == 2);
		}
    }
    foreach my $nb (@a_blastx_list)    # '2', '3'
    {
        push(@a_preservfiles, "$seq.blast$nb.gff3") if ($o_param->Get("blastx_db_$nb\_preserve"));
    }
    push(@a_preservfiles, "$seq.$TRNASCAN_EXT.gff3");
    push(@a_preservfiles, "$seq.$RFAMSCAN_EXT.gff3");
    push(@a_preservfiles, "$seq.$RNAMMER_EXT.gff3");

    my $allpreserv = "$workdir/$seqname.REPET.est_blastx_topreserv.gff3";
    &MergeGFF3($o_param, $seq, \@a_preservfiles, $allpreserv,"repeat_region",&FALSE,$PROTECT_MARGIN,\%h_options);

    my $okrepetfile = "$workdir/$seqname.REPET.est_blastx_preserved.gff3";
    &SubtractGFF3($o_param, $allrepet, $allpreserv, $okrepetfile);

    &SetTypeGFF3($okrepetfile, $okrepetfile4egn, "intergenic_region|intron", 2);
	
    &SplitByStrand($o_param, $okrepetfile4egn) if ($INDEPENDENT_STRAND);
	
    &PrintStop($step);

    unlink($allpreserv, $okrepetfile) if ($CLEAN);
    unlink($allrepet) if ($CLEAN);

    &ValidateStep($okrepetfile4egn);

	chdir($prevdir);

    return;
}

=head2 procedure SearchSpeciesRepeatDomains
	
 Title        : SearchSpeciesRepeatDomains
 Usage        : 
 Function     : 
                launch an eugene annotation then extract proteome, then
                search domains similar to repbase
 Prerequisite : 
 Args         : $allseqfile multifasta
 Globals      : none

=cut

sub SearchSpeciesRepeatDomains
{
    my ($o_param, $out, $allseqfile, $ra_seq, $rh_db_level, $matrixfile) = @_;

    my $cmd;
    my $step = "Search repeat domains specific to the genome";
    return if (&IsValidatedStep($out, $step));

    &PrintStart($step);
    &AddPrintIndent(2);

    # 1 Launch EuGene and extract proteome to get protein from the species
    my $annotv0Dir  = &CreateAnnotTree($o_param, 0);
    my $egnv0Dir    = $annotv0Dir . "/raw_eugene";
    my $cfg_v0_file = &GetParamDir($o_param) . "/eugenev0.par";
    &BuildEGNEukParameterFile($o_param, $cfg_v0_file, $rh_db_level, 0, 0);

    &LaunchAllEGN($o_param, $ra_seq, $cfg_v0_file, $matrixfile, $egnv0Dir, "REPEAT", 1, "");

    my @a_res_gff3;
    foreach my $seq (@$ra_seq)
    {
        my ($seqpath, $seqname) = &GetPathAndName($seq);

        #&LaunchEGN($o_param, $seq, $cfg_v0_file, $matrixfile, $egnv0Dir, $o_logger, $logger_stamp, 1);
        push(@a_res_gff3, "$egnv0Dir/$seqname.gff3");
    }
    &Consolidate($o_param, $annotv0Dir, \@a_res_gff3, $allseqfile, $o_logger, $logger_stamp);
	my $prefix = $o_param->Get('output_prefix');
    my $protv0 = "$annotv0Dir/$prefix\_prot.fna";

    # 2 Search putative repet domain in proteome V0

    &PrintStart("Search similarities with RepBase");
    my $repbase     = $o_param->Get('repeat_sequence_db');
    my $scov        = $o_param->Get('prg_extract_repet_domain_min_scov');
    my $info        = "";
    my $blast_param = "";

    if ($o_param->IsDefined('prg_extract_repet_domain_blastparam'))
    {
        $blast_param = $o_param->Get('prg_extract_repet_domain_blastparam');
        $info        = "  PARAMETERS BlastP: $blast_param";
        &PrintMore($info);
    }
    $info = "  FILTERS BlastP= subject alignment percentage > $scov %";
    &PrintMore($info);

    $cmd = $o_param->Get('prg_extract_repet_domain');
    $cmd .= " --in $protv0 --out $out --repbase $repbase --scov $scov";
    $cmd .= " --no-clean " if (!$CLEAN);
    $cmd .= " --blastparam $blast_param " if ($blast_param ne "");
    $cmd .= " > /dev/null 2>&1";
    &ExeCmd($cmd, $o_logger, $logger_stamp);
    &PrintStop("Search similarities with RepBase");
    &AddPrintIndent(-2);

    &ValidateStep($out);
    &PrintStop($step);

    return;
}

sub StatSpeciesRepeatDomains
{
    my ($o_param) = @_;

    my $repet_domain = $o_param->Get('species_repeat_domains');
    my $stat         = "";

    my ($gpath, $gname, $gid, $glen) = &GetSeqInfo($o_param->Get('workingdir') . "/$ALLSEQFILE_NAME");

    if (defined($repet_domain) && -e $repet_domain)
    {
        # number of domains
        my $species_repet_domain_nb = &CountSeq($repet_domain);
        my ($seqpath, $seqname, $seqid, $species_repet_domain_nt) = &GetSeqInfo($repet_domain);

        # percentage of the sequence covered by a repeat
        my $species_repet_domain_percent = sprintf("%0.1f", ($species_repet_domain_nt / $glen * 100));

        $stat .= "\n## " . $o_param->Get('organism') . " specific repeat domains\n";
        $stat .= "File=$repet_domain\n";
        $stat .=
          "Repeat domain number=$species_repet_domain_nb Repeat domain length=$species_repet_domain_nt nt ($species_repet_domain_percent% of genomic sequences)\n\n";
    }
	# &AppendFile($o_param->Get('report'), "$stat");

    return;
}

sub StatRepeatSearch
{
    my ($o_param) = @_;

    my $workdir = $o_param->Get('workingdir');
    my $report  = $o_param->Get('report');

    my ($seqpath, $seqname, $seqid, $seqlen) = &GetSeqInfo("$workdir/$ALLSEQFILE_NAME");

    # masked sequence length with ltrharvest
	 
    my $allltrharvest = "$workdir/$ALLSEQFILE_NAME.all.$LTRHARVEST_EXT.gff3.fa";
	if (! -e "$allltrharvest.gz")
	{
	    &ConcatFilesByExt($o_param, "$LTRHARVEST_EXT.gff3.fa.gz", $allltrharvest, $o_logger, $logger_stamp);
    	&Gzip($o_logger, $logger_stamp, $allltrharvest);
	}

    my ($ltrnb, $ltrnt) = &GetMultifastaInfo($o_param, "$allltrharvest.gz");
    my $percent = sprintf("%0.1f", $ltrnt / $seqlen * 100);
    #&AppendFile($report, "## LTR masking\nLTR region length=$ltrnt nt ($percent% of genomic sequences)\n");
	if (!$o_param->IsDefined('no_red') && -e "$workdir/Red/out/seq.merged.bed")
	{
		my $red_reg_nt      = &CountNtBed("$workdir/Red/out/seq.merged.bed"); 
		my $red_reg_percent = sprintf("%0.1f", ($red_reg_nt / $seqlen * 100));
		#&AppendFile($report, "## Red repeat predictions\nRed region length=$red_reg_nt nt ($red_reg_percent% of genomic sequences)\n");
	}
    my $allrepeat = "$workdir/$ALLSEQFILE_NAME.all.$REPET_EXT.gff3";
    &ConcatFilesByExt($o_param, "$REPET_EXT.gff3", $allrepeat, $o_logger, $logger_stamp);

    my $repeat_reg_nb = &CountNbFeatureGFF3($allrepeat, "intergenic_region")/2; # divide by 2 because there are two intergenic_feature features, one per strand 
    my $repeat_reg_nt = &CountFeatureNtGFF3($allrepeat, "intergenic_region")/2; # divide by 2 because there are two intergenic_feature features, one per strand 

    my $repeat_reg_percent = sprintf("%0.1f", ($repeat_reg_nt / $seqlen * 100));

    my $stat =
      "\n## Total repeat regions (after masking with Red + LTRHarvest + Species specific repeat search. After demasking \"protected\" expressed or translated regions)\n";
	$stat .= "Repeat region number = $repeat_reg_nb\n";
	$stat .= "Repeat region length = $repeat_reg_nt nt ($repeat_reg_percent% of genomic sequences)";

    &AppendFile($o_param->Get('report'), $stat);

    return;
}

sub ConcatFilesByExt
{
    my ($o_param, $ext, $out, $o_logger, $logger_stamp) = @_;

    &ExeCmd("rm -f $out", $o_logger, $logger_stamp);
    my $f_in = $o_param->GetStreamIn('seq_fof');
    while (my $l = <$f_in>)
    {
        chomp($l);
        #my ($d, $f) = $l =~ /(.+)\/(.+)/;
        my ($file) = "$l.$ext";
        if (-e $file)
        {
            my $cmd = ($ext =~ /.gz$/) ? "gzip -cd  " : "cat ";
            $cmd .= " $file >> $out";
            &ExeCmd($cmd, $o_logger, $logger_stamp);
        }
    }
    $f_in->close;
}

#######################################################################################
##    GFF3 tools
#######################################################################################

sub SetTypeGFF3
{
    my ($in, $out, $typelist, $both_strand) = @_;

    my @a_types = split(/\|/, $typelist);

    my $f_    = &General::GetStreamIn($in);
    my $f_out = &General::GetStreamOut($out);

    while (my $l = <$f_>)
    {
        chomp($l);

        foreach my $type (@a_types)
        {
            my @a_fields = split(/\t/, $l);
            $a_fields[6] = "+" if ($both_strand);
            $a_fields[2] = $type;
            $a_fields[8] =~ s/ID=/ID=$type:/;
            $a_fields[8] .= ";Ontology_term=SO:0000191" if ($type eq "intron");

            print $f_out join("\t", @a_fields) . "\n";

            if ($both_strand)
            {
                $a_fields[8] =~ s/ID=/ID=2_/;
                $a_fields[6] = "-";
                print $f_out join("\t", @a_fields) . "\n";
            }
        }
    }

    $f_->close;
    $f_out->close;

    return;

}

sub SubtractGFF3
{
    my ($o_param, $ori_gff3, $tosubtract_gff3, $res) = @_;

    my $cmd;

    $cmd = $o_param->Get('prg_subtractbed') . " -a $ori_gff3 -b $tosubtract_gff3 > $res";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    return;
}

# prerequisite
# gff3 contain only features of one sequence !

sub MergeGFF3
{
    my ($o_param, $seq, $ra_gff3, $res,$code,$stranded,$margin,$rh_options) = @_;

	my $bkdir = $$rh_options{tmpdir};
	$$rh_options{tmpdir} .= "/SORTTMP.$$.".time;
	mkdir $$rh_options{tmpdir};

    my ($seqpath, $seqname, $seqid, $seqlen) = &GetSeqInfo($seq);

    my $catgff3         = "$res.catandsorted";
	my $lof = join(',',@$ra_gff3);
	my $cmd = $o_param->Get('prg_mergesortgff3') . " --out $catgff3  --lof $lof --seqlen $seqlen --seqname $seqname --tmpdir $$rh_options{tmpdir} --margin $margin ";
	&ExeCmd($cmd, $o_logger, $logger_stamp);

    # Merge features
	if ( ! $stranded )
	{
		$cmd = $o_param->Get('prg_mergebed') . " -i $catgff3 > $res.bed";
		&ExeCmd($cmd, $o_logger, $logger_stamp);
	}
	else
	{
		if ( ! -z $catgff3 )
		{
			$cmd = $o_param->Get('prg_mergebed') . " -i $catgff3 -S + -c 7 -o distinct > $res.bed";
			&ExeCmd($cmd, $o_logger, $logger_stamp);
			$cmd = $o_param->Get('prg_mergebed') . " -i $catgff3 -S - -c 7 -o distinct >> $res.bed";
			&ExeCmd($cmd, $o_logger, $logger_stamp);
		}
		else
		{
			&ExeCmd("touch $res.bed", $o_logger, $logger_stamp);
		}
	}

    # Convert bed 2 gff3
    $cmd = $o_param->Get('prg_bed2gff3') . " --in  $res.bed --source eugene --code $code --out $res";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

	if ( $stranded )
	{
    	&ExeCmd("cp $res $res.bk", $o_logger, $logger_stamp);
		my $f_in = &GetStreamIn("$res.bk");
		my $f_out= &GetStreamOut("$res");
		while(my $l=<$f_in>)
		{
			chomp($l);
			my($strand)=$l=~/(\S+)$/;
			if ( $l!~/^#/ )
			{
				my @a_=split("\t",$l);
				$a_[6]=$strand;
				$l = join("\t",@a_);
			}
			print $f_out "$l;Ontology_term=SO:0000833\n";
		}
		$f_in->close;
		$f_out->close;
		unlink("$res.bk");
	}

    # remove temporary file
    unlink("$res.bed", $catgff3) if ($CLEAN && ! $stranded);

	$$rh_options{tmpdir}=$bkdir;

    return;

}

sub CountNbFeatureGFF3
{
    my ($gff3, $feat) = @_;

    my $f_in    = &General::GetStreamIn($gff3);
    my $gene_nb = 0;

    while (my $line = <$f_in>)
    {
        chomp $line;
        $gene_nb++ if ($line =~ /\t$feat\t/);
    }
    $f_in->close;

    return $gene_nb;
}

sub CountFeatureNtGFF3
{
    my ($gff3, $feat) = @_;

    my $f_in    = &General::GetStreamIn($gff3);
    my $feat_nt = 0;
    while (my $line = <$f_in>)
    {
        chomp $line;
        next if ($line !~ /\t$feat\t/);
        my @a_f = split(/\t/, $line);
        $feat_nt += ($a_f[4] - $a_f[3] + 1);
    }
    $f_in->close;

    return $feat_nt;

}

sub CountNtBed
{ 
	my ($bed) = @_;

    my $f_in    = &General::GetStreamIn($bed);
    my $feat_nt = 0;
    while (my $line = <$f_in>)
    {
        chomp $line;
        my @a_f = split(/\t/, $line);
        $feat_nt += ($a_f[2] - $a_f[1]);
    }
    $f_in->close;

    return $feat_nt;
}


sub GetAttNameGFF3
{
    my ($gff3, $feat, $rh_) = @_;

    my $f_in = &General::GetStreamIn($gff3);

    while (my $line = <$f_in>)
    {
        chomp $line;
        next if ($line !~ /\t$feat\t/);
        my ($name) = $line =~ /Name=([^;]+);?/;
        $$rh_{$name}++ if (defined($name) && $name ne "");
    }
    $f_in->close;

    return;
}

sub TransferAnnotncRNA
{
    my ($o_param, $seq, $gff3, $resgff3) = @_;

    my $ncrnafiles = "";
    $ncrnafiles .= " $seq.$TRNASCAN_EXT.gff3" if (-e "$seq.$TRNASCAN_EXT.gff3");
    $ncrnafiles .= " $seq.$RNAMMER_EXT.gff3"  if (-e "$seq.$RNAMMER_EXT.gff3");
    $ncrnafiles .= " $seq.$RFAMSCAN_EXT.gff3" if (-e "$seq.$RFAMSCAN_EXT.gff3");

	my $o_local_logger = ( $o_param->Get('ncpus') > 1 ) ? undef: $o_logger; # si je // ca va mettre le souk dans les logs

    # no ncrna
    if ($ncrnafiles eq "")
    {
        &ExeCmd("cp $gff3 $resgff3", $o_local_logger, $logger_stamp);
    }
    else
    {
        my $allncrna = $seq . ".allncrna.gff3";
        my $cmd      = "cat $ncrnafiles > $allncrna";
        &ExeCmd($cmd, $o_local_logger, $logger_stamp);

        $cmd = $o_param->Get('prg_transfer_gff3_attr');
        $cmd .=
          " --in $gff3 --ref $allncrna --out $resgff3 --intersectbed_type \"tRNA;rRNA;ncRNA\" --attrs \"subunit;rfam_acc;rfam_id;anti_codon;product\" ";
        $cmd .= $o_param->Get('prg_transfer_gff3_attr_param') if ($o_param->IsDefined('prg_transfer_gff3_attr_param'));
        $cmd .= " > /dev/null 2>&1";
        &ExeCmd($cmd, $o_local_logger, $logger_stamp);

        unlink($allncrna) if ($CLEAN);
    }

    return;
}

#######################################################################################
##    Training
#######################################################################################

=head2 procedure BuildEGNMatrixEuk
	
 Title        : BuildEGNMatrixEuk
 Usage        : 
 Function     : 
 Prerequisite : 
 Args         : $sequence
                $est        fasta   ref transcriptome 
                $prot       fasta   ref proteome
                $workingdir dir
                $matrixfile binary  result IMM matrix file
 Globals      : none

=cut

sub BuildEGNMatrixEuk
{
    my ($o_param, $sequence, $est, $prot, $matrixfile, $o_logger, $logger_stamp) = @_;

    my $step = "Build IMM models";
    return if (&IsValidatedStep($matrixfile, $step));

    &PrintStart($step);
    &AddPrintIndent(2);
    my $traindir = &GetTrainDir($o_param);

    my ($seqpath, $seqname) = $sequence =~ /(.+)\/(.+)/;
    my $training_blastxfile = "$traindir/$seqname.blasttraining";
    my $training_gmapfile   = "$traindir/$seqname.esttraining";
    my $trainingroot        = "$traindir/$seqname.training";
    my $cdsfile             = "$traindir/$seqname.training.CDS.fa";
    my $intronfile          = "$traindir/$seqname.training.INTRON.fa";

    # 1) Keep only long transcriptome sequences (len >= 3* hsp_training_length)
    my ($estpath, $estname) = $est =~ /(.+)\/(.+)/;
    my $minlen          = $o_param->Get('hsp_training_length') * 3;
    my $est_filter_file = &GetDbDir($o_param) . "/$estname.filterlen$minlen";
    &FilterFastaLen($o_param, $est, $est_filter_file, $minlen, $o_logger, $logger_stamp);

    # 2) blast ref proteome against ref transcriptome and compress output
    &BlastXBatch($o_param, $est_filter_file, $prot, $training_blastxfile, 1);
    $training_blastxfile .= ".gz";

    # 3) extract CDS region from the m8 blast results
    &BuildFastaCDSFromBlastXHits($training_blastxfile, $est_filter_file, $cdsfile, $o_param);

    # 4) Map ref transcriptome to genomic sequence (Use filters training_est_pcs|pci)
    $o_param->Set('training_est_file', $est_filter_file);    # need for Gmap
    &Gmap($o_param, $sequence, 'training_est', "", "", $training_gmapfile);
    if (!-s $training_gmapfile . ".gff3")
    {
        &PrintError(  "ERROR: no data to train eugene IMM "
                    . "(because no result for mapping of the reference transcriptome "
                    . "to the genomic sequence): choose an other reference transcriptome and launch again.\n");
    }

    # 5) extract introns sequences and non canonical splice sites
    &PrintStart("Build training datasets");
    &BuildTrainingDataSets($training_gmapfile . ".gff3", $sequence, $trainingroot, $intronfile, $o_param);
    &FastaTotrainIMM($intronfile);
    &PrintStop("Build training datasets");

    # 6) build matrix
    &BuildIMM($o_param, $cdsfile, $intronfile, $matrixfile, $o_logger, $logger_stamp);

    &ValidateStep($matrixfile);
    &AddPrintIndent(-2);
    &PrintStop($step);

    # 7) Clean or compress temporary files
    &Gzip($o_logger, $logger_stamp, $cdsfile, $intronfile);

    return;
}

=head2 procedure BuildIMM
	
 Title        : BuildIMM
 Usage        : 
 Procedure    : build imm matrix
 Args         : $cdsfile      cds sequences in egn_trainIMM format:
                              seq1name seq1
                              seq2name seq2
                $intronfile  intron sequence. same format than cds file
                $matrixfile  result binary matrix
 Prerequisite : 
 Globals      : none

=cut

sub BuildIMM
{
    my ($o_param, $cdsfile, $intronfile, $matrixfile, $o_logger, $logger_stamp) = @_;

    my $trainIMM = "export EUGENEDIR=" . $o_param->Get('EUGENEDIR') . "; " . $o_param->Get("prg_eugene_train_IMM");
    my $cmd      = "$trainIMM $cdsfile $intronfile > /dev/null 2>&1";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    $cmd = "mv $cdsfile.bin $matrixfile";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    return;

}

=head2 procedure BuildTrainingDataSets
	
 Title        : BuildTrainingDataSets
 Usage        : 
 Function     : extract intron sequences from gff3 file
 Args         : $gff3       gff3 file
                $seq        fasta/multifasta file of sequences
                $outintron  file Result fasta file
                $o_param 
 Prerequisite : 
 Globals      : none

=cut

sub BuildTrainingDataSets
{
    my ($gff3, $seq, $outprefix, $outintron, $o_param) = @_;

    my $cmd   = $o_param->Get('prg_build_training_dataset');
    my $param = $o_param->Get('prg_build_training_dataset_param');
    $cmd .=
      " --sequence $seq --gff3 $gff3 --outprefix $outprefix --parent_feature EST_match --child_feature match_part --intron_file $outintron $param > /dev/null 2>&1";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    return;
}

sub DetectNonCanSites
{
    my ($o_param, $report_file) = @_;

    my $f_ = IO::File->new(">>$report_file") or die "ERROR: Can't write in $report_file\n";
    print $f_ "\n## Splicing (canonical and non-canonical) sites detected in the training dataset\n";
	print $f_ "#Site\tDinucleotide\tCount\tCount/Total donor(or acceptor) site count*100\n";
    my $noncandon = "";
    my $noncanacc = "";

    # 1 search in the training datasets
    my $trainingroot = &GetTrainDir($o_param) . "/$ALLSEQFILE_NAME.training";
    &SearchNonCanSites($o_param, "$trainingroot.alt-acc", "$trainingroot.TP.acceptor", "acceptor", \$noncanacc, $f_);
    &SearchNonCanSites($o_param, "$trainingroot.alt-don", "$trainingroot.TP.donor",    "donor",    \$noncandon, $f_);
    $f_->close;

    # Read  noncansite in the cfg file
    my $f_model = $o_param->GetStreamIn('prg_eugene_euk_cfg_model');
    while (my $cfg_line = <$f_model>)
    {
		if ($cfg_line =~ /^EuGene.NonCanDon\s+(\S+)/)
		{
            my $cfgnoncandon = $1; chomp($cfgnoncandon);
			if ($noncandon eq "")
			{
				$noncandon = $cfgnoncandon;
			}
			else
			{
				$noncandon .= ",$cfgnoncandon";
			}
		}
		if ($cfg_line =~ /^EuGene.NonCanAcc\s+(\S+)/)
		{
			my $cfgnoncanacc = $1;chomp($cfgnoncanacc);
			if ($noncanacc eq "")
			{
				$noncanacc = $cfgnoncanacc;
			}
			else
			{
				$noncanacc .= ",$cfgnoncanacc";
			}
		}
    }
    $f_model->close;
  
    my %h_don;
    my %h_acc;
    foreach my $site (split(",", $noncandon))
    {
        $h_don{$site}++;
    }
    foreach my $site (split(",", $noncanacc))
    {
        $h_acc{$site}++;
    }

    # reinit variables
    $noncandon = "";
    $noncanacc = "";
    $noncandon = join(",", keys(%h_don)) if (scalar(keys(%h_don)) > 0);
    $noncanacc = join(",", keys(%h_acc)) if (scalar(keys(%h_acc)) > 0);

    $o_param->Set('noncandon', $noncandon);
    $o_param->Set('noncanacc', $noncanacc);

    return;
}

=head2 procedure SearchNonCanSites
	
 Title        : SearchNonCanSites
 Usage        : 
 Function     : keep a site as a non canonical splice site if it is present in the mapping 
                result more than  'noncansite_required_percent'. 
 Args         : $o_param 
                $noncansiteprefix: prefix of mutlifasta file of non canonical file extracted from the mapping
                $cansitefile multifasta of canonical splice sites
                $ra_noncansite array of found sites
 Prerequisite : 
 Globals      : none

=cut

sub SearchNonCanSites
{
    my ($o_param, $noncansiteprefix, $cansitefile, $type, $r_res, $f_) = @_;

    my $text                        = "";
    my $noncansite_required_percent = $o_param->Get('noncansite_required_percent');
    my $max_noncansite_candidate_nb = $o_param->Get('max_noncansite_candidate_nb');

    # count total site
    my $can_sites_nb = &CountDiffSeq($cansitefile);
    my $cansite = ($type eq "donor") ? "GT" : "AG";
    $text .= "Canonical $type\t$cansite\t$can_sites_nb\t\n";
	
	my $non_can_inc = 0;    # number of dinucleotide kept as non canonical
	my @a_noncanfiles = glob("$noncansiteprefix.??");
    foreach my $f (@a_noncanfiles)
    {
        chomp($f);

        # count noncanonical site (note the redondancy is possible, so call CountDiffSeq)
        my $nb_site         = &CountDiffSeq($f);
        my ($site)          = $f =~ /$noncansiteprefix\.(..)/;
        my $site_nb_percent = sprintf("%0.1f", $nb_site / $can_sites_nb * 100);
        if ($site_nb_percent >= $noncansite_required_percent)
        {
            print STDERR
              "$f : $nb_site sites / $can_sites_nb= $site_nb_percent% >= $noncansite_required_percent --> keep $site\n"
              if ($DEBUG);

            $$r_res .= "$site,";
            $non_can_inc++;
            $text .= "Non canonical $type\t$site\t$nb_site\t$site_nb_percent\%\n";
        }
        else
        {
            print STDERR
              "$f : $nb_site sites / $can_sites_nb= $site_nb_percent% < $noncansite_required_percent --> rejected $site\n"
              if ($DEBUG);
        }
    }

	print $f_ $text;

    if ($non_can_inc >= $max_noncansite_candidate_nb)
    {
		my $error_msg = "ERROR: the transcriptome mapping results show too many different non canonical $type sites ($non_can_inc > 'max_noncansite_candidate_nb'=$max_noncansite_candidate_nb). Check data/parameters used for transcriptome mapping. See files $noncansiteprefix*";
		print $f_ $error_msg."\n";
        &PrintError($error_msg);
    }
    
    $$r_res =~ s/,$//;
    
    return;
}

=head2 procedure BuildFastaCDSFromBlastXHits
	
 Title        : BuildFastaCDSFromBlastXHits
 Usage        : 
 Function     : extract potential cds part of transcript sequences
                filter blastx hits with :
                pci > hsp_training_pci,
                evalue > hsp_training_evalue, 
                aligment len > hsp_training_length
                len %3 = 0, 
                does not contain N or stop codon  
 Args         : $blastxfile m8 blastx result
                $seqfile    nucleic multifasta file (transcript file)
                $out        multifasta file of potential cds
                $o_param 
                
 Prerequisite : $seqfile is the query of the blastx results
                
 Globals      : none

=cut

sub BuildFastaCDSFromBlastXHits
{
    my ($blastxfile, $seqfile, $out, $o_param) = @_;

    my %h_seq;
    &FastaToHash($seqfile, \%h_seq);
    my %h_hit      = ();
    my $hsp_nb     = 0;
    my $hsp_pci    = $o_param->Get('hsp_training_pci');
    my $hsp_evalue = $o_param->Get('hsp_training_evalue');
    my $hsp_length = $o_param->Get('hsp_training_length');    # for blastx the length is in aa

    &PrintMore("BLASTX FILTERS= HSP_length > $hsp_length AA, identity percentage > $hsp_pci, e-value > $hsp_evalue");
    my $f_blast = &GetStreamIn($blastxfile);
    my $f_fasta = &GetStreamOut($out);

    my %h_deja = ();
    while (my $line = <$f_blast>)
    {
        next if ($line =~ /^#/);

        &ParseM8Line(\$line, 1, \%h_hit);
        next
          if (!(($h_hit{pci} >= $hsp_pci) && ($h_hit{"e-value"} <= $hsp_evalue) && ($h_hit{len} >= $hsp_length)))
          ;    # for blastx the length is in aa
        next if (defined($h_deja{$h_hit{target}}));
        $h_deja{$h_hit{target}} = 1;

        my ($start, $end) = &MinMax($h_hit{qstart}, $h_hit{qend});
        next if (($end - $start + 1) % 3 != 0);

        my $sseq = &ExtractSeq(\$h_seq{$h_hit{query}}{sequence}, $start, $end, $h_hit{strand});
        next if ($sseq =~ /N/);
        $sseq =~ s/\w\w\w$//;    # remove the last codon
        my $ok = &TRUE;
        while ($sseq =~ /(...)/g)
        {
            my $codon = $1;
            if ($codon =~ /TAA|TGA|TAG/)
            {
                $ok = &FALSE;
                last;
            }
        }
        next if (!$ok);

        $hsp_nb++;

        # replace the substring by N for next iteration
        substr($h_seq{$h_hit{query}}{sequence}, $start - 1, length($sseq), 'N' x length($sseq));

        $h_hit{query} =~ s/\|/-/g;
        print $f_fasta ">$h_hit{query}-$start,$end,$h_hit{strand}\n$sseq\n";
    }
    $f_blast->close;
    $f_fasta->close;

    &FastaTotrainIMM($out);

    my $hsp_min_nb = $o_param->Get('hsp_training_min_nb');
    if ($hsp_nb < $hsp_min_nb)
    {
        &PrintError(
            "ERROR: not enough HSP ($hsp_nb kept while $hsp_min_nb required) to build training CDS file. BlastX file=$blastxfile\n"
            );
        return 1;
    }

    return 0;

}

=head2 procedure FastaTotrainIMM
	
 Title        : FastaTotrainIMM
 Usage        : 
 Function     : transform fasta file to egn training format
                seq1name seq1
                seq2name seq2
                ...
 Args         : $in multifasta file
 Globals      : none

=cut

sub FastaTotrainIMM
{
    my ($in) = @_;

    my $ori = $in . ".tmp";
    my $cmd = "mv $in $ori";
    system($cmd);

    my $name  = "";
    my $seq   = "";
    my $f_in  = &GetStreamIn($ori);
    my $f_out = &GetStreamOut($in);
    while (my $line = <$f_in>)
    {
        chomp($line);
        if ($line =~ /^>/)
        {
            print $f_out "$name $seq\n" if ($name ne "");
            ($name) = $line =~ /^>(\S+)/;
            $seq = "";
        }
        else
        {
            $seq .= $line;
        }
    }
    print $f_out "$name $seq\n" if ($name ne "");

    $f_in->close;
    $f_out->close;

    unlink($ori) if ($CLEAN);

    return;
}

########################################################################
###  Transcriptome mapping functions
########################################################################

=head2 procedure GmapBuild

    Title        : GmapBuild
    Usage        : &GmapBuild($o_param, $seqfasta, $dir);
    Prerequisite : none
    Procedure    : Create gmap indexation directory (if not exists) 
                   and index genome sequence for gmap
    Args         : $o_param
                   $genome          Fasta sequence
                   $outdir          Directory of gmap indexation
    Globals      : none
=cut

sub GmapBuild
{
        my ($o_param, $genome, $outdir) = @_;

        my ($genomedir, $genomename) = $genome =~ /(.+)\/(.+)/; 

        mkdir($outdir) if (! -d  $outdir);

        # -d STRING  genome name
        # -D STRING  destination directory for installation
        my $cmd = $o_param->Get("prg_gmapbuild") . " -d $genomename -D $outdir $genome 2> $outdir/gmap_idx.$$.stde";
        print STDERR $cmd ."\n"; 
        system($cmd);

        return;
}


=head2 procedure MappingEstEuk
	
 Title        : MappingEstEuk
 Usage        : 
 Prerequisite : o_param object has the keys: 'est_list'
 Function     : Mapping transcriptome libraries  to genomic sequence $sequence
 Args         : $o_param 
                $sequence  fasta genomic sequence
                
 Globals      : none

=cut

sub MappingEstEuk
{
    my ($o_param, $sequence) = @_;

    my $cmd;
    my $est_list   = $o_param->Get('est_list');
    my $workdir    = $o_param->Get('workingdir');
    my $seq_fof    = $o_param->Get('seq_fof');
    
    my @a_est_list = split(/\s+/, $est_list);

    my $gmapbuild_dir = &GetDbDir($o_param);
	$gmapbuild_dir .= "/GMAP_INDEX";
    if ( ! &IsValidatedStep("$gmapbuild_dir", "Gmap build index") )
	{
		&GmapBuild($o_param, $sequence, $gmapbuild_dir);
        &ValidateStep("$gmapbuild_dir");
	}

    foreach my $nb (@a_est_list)    # '2', '3'
    {
        my $outext = "est$nb";
        &Gmap($o_param, $sequence, "est_$nb", $outext, $seq_fof, "", $gmapbuild_dir);
    }

    # Write mapping stats in report file
    my $info = "## Alignment of transcriptomes\n";
    $info .=
      "#Transcriptome_nb\tFile\tRaw_sequence_nb\tMapped_and_filtered_sequence_nb\tMapped_and_Filtered_sequence_percentage\n";

    # foreach est database
    foreach my $nb (@a_est_list)
    {
        my $db        = $o_param->Get("est_" . $nb . "_file");
        my $seq_nb    = &CountSeq($db);
		my $unspliced = $o_param->Get("est_" . $nb . "_remove_unspliced");

        # count the number of mapped seq (raw gmap results)
        my $rawgmap = "$workdir/$ALLSEQFILE_NAME.est$nb.gmap_raw.gff3.gz";
        my %h_raw_mapped_seq;
        &GetAttNameGFF3($rawgmap, "gene", \%h_raw_mapped_seq);
        my $gmapraw_seq_nb = scalar(keys(%h_raw_mapped_seq));

        # post traitement + comptage
        my %h_mapped_seq;

        my $f_fof = $o_param->GetStreamIn('seq_fof');
        while (my $seq = <$f_fof>)
        {
			chomp($seq);
            my $gff3           = "$seq.est$nb.gff3"; 
            my $unspliced_gff3 = "$seq.est$nb.rm_unspliced_$MIN_LEN_SHORT_UNSPLICED.gff3";
            if (  ! -e "$gff3.filt.done" ) # je ne peux pas utiliser un .success sur le gff3 car c'est utilise pour declencher le gmap
            {
				#JG, avant de compter je nettoie
				$cmd = $o_param->Get('prg_gmap_filter_patch')." --in $gff3 ";
				&ExeCmd($cmd, $o_logger, $logger_stamp);
				&ExeCmd("touch $gff3.filt.done", $o_logger, $logger_stamp);
				if ($INDEPENDENT_STRAND)
				{
					&SplitByStrand($o_param, $gff3);
					&SplitByStrand($o_param, $unspliced_gff3) if ($unspliced == 2);
				}
			}
            &GetAttNameGFF3($gff3,           "EST_match", \%h_mapped_seq);
            &GetAttNameGFF3($unspliced_gff3, "EST_match", \%h_mapped_seq) if ($unspliced == 2);
	    
        }
        $f_fof->close;
        
        my $mapped_seq_nb      = scalar(keys(%h_mapped_seq));
        my $mapped_seq_percent = sprintf("%0.1f", ($mapped_seq_nb / $seq_nb * 100));
        $info .= "$nb\t$db\t$seq_nb\t$mapped_seq_nb\t$mapped_seq_percent\n";
    }

    my $report_file = $o_param->Get('report');
    &AppendFile($report_file, $info);

    # Optional post filter based on intron abundancy
    my $intron_filter = $o_param->Get('gmap_intron_filter');
    if ($intron_filter == 1)
    {
	my $allestnb = 1000; # a ameliorer!!!! verifier que ca m'existe  pas
	
	my $f_fof = $o_param->GetStreamIn('seq_fof');
	while (my $seq = <$f_fof>)
    {
        chomp($seq);
        
		next if (&IsValidatedStep("$seq.est$allestnb.gff3", "Intron filter $seq"));
		
	    my @a_spliced_files;
	    my @a_unspliced_files;
	    my @a_to_remove;
	    
	    # 1 creation of two files, one with spliced est, one with unspliced est
	    foreach my $nb (@a_est_list)
	    {
			my $rmunspliced    = $o_param->Get("est_" . $nb . "_remove_unspliced");
			my $spliced_gff3   = "$seq.est$nb.gff3";
			my $unspliced_gff3 = "$seq.est$nb.rm_unspliced_$MIN_LEN_SHORT_UNSPLICED.gff3";

		# _remove_unspliced = 0 : the file contains spliced and unspliced hit:so split in two files
		# _remove_unspliced = 1 : je ne garde que les episses. je splitte quand meme pour supprimer les 
		# les non episses crees APRES le filtre sur les petits exons
		# _remove_unspliced = 2 : je garde episses et non episses. je splitte quand meme le premier fichier 
		# pour extraire les non episses crees APRES le filtre sur les petits exons
		
			$cmd = $o_param->Get('prg_splitspliced'). " --in $spliced_gff3 --ext est$nb\_SPLIT";
			&ExeCmd($cmd, $o_logger, $logger_stamp);
			push(@a_to_remove, "$seq.est$nb\_SPLIT.gff3", "$seq.est$nb\_SPLIT.rm_unspliced_$MIN_LEN_SHORT_UNSPLICED.gff3");
			push(@a_spliced_files, "$seq.est$nb\_SPLIT.gff3");
			if ($rmunspliced == 0)
			{
				push(@a_unspliced_files, "$seq.est$nb\_SPLIT.rm_unspliced_$MIN_LEN_SHORT_UNSPLICED.gff3");
			}
			if ($rmunspliced == 2)
			{		   
				push(@a_unspliced_files, $unspliced_gff3); # non episses d'origine
				push(@a_unspliced_files, "$seq.est$nb\_SPLIT.rm_unspliced_$MIN_LEN_SHORT_UNSPLICED.gff3"); # non episses cree apres le filtre sur petit exon
			}
	    }
	    
	    # 2 Get all the spliced ests then filter by intron incoherence
	    $cmd = "cat " . join(' ', @a_spliced_files) . " > $seq.est$allestnb.beforeintronfilter.gff3";
	    &ExeCmd($cmd, $o_logger, $logger_stamp);
	    $cmd = $o_param->Get('prg_gmap_intron_filter')." --in  $seq.est$allestnb.beforeintronfilter.gff3 --out $seq.est$allestnb.gff3"; 
	    &ExeCmd($cmd, $o_logger, $logger_stamp);
	    unlink("$seq.est$allestnb.beforeintronfilter.gff3");

	    # Get all the unspliced ests
	    $cmd = "cat " . join(' ', @a_unspliced_files) . " > $seq.est$allestnb.rm_unspliced_$MIN_LEN_SHORT_UNSPLICED.gff3";
	    &ExeCmd($cmd, $o_logger, $logger_stamp);

	    if ($INDEPENDENT_STRAND)
	    {
			&SplitByStrand($o_param, "$seq.est$allestnb.gff3");
			&SplitByStrand($o_param, "$seq.est$allestnb.rm_unspliced_$MIN_LEN_SHORT_UNSPLICED.gff3");
	    }
	    
        # remove the temporary files 
	    foreach my $f (@a_to_remove)
	    {
			unlink($f);
	    }
	    
	    &ValidateStep("$seq.est$allestnb.gff3");
	}
	
	$o_param->Set("est_" . $allestnb . "_file", "estall$allestnb");
	$o_param->Set("est_" . $allestnb . "_pcs", 0);
	$o_param->Set("est_" . $allestnb . "_pci", 0);
	$o_param->Set("est_" . $allestnb . "_remove_unspliced", 2);
	$o_param->Set("est_" . $allestnb . "_preserve", 1);
	$o_param->Set("est_" . $allestnb . "_training", 0);
	$o_param->Set('est_list', $allestnb);
    }
    
    return;
}

=head2 procedure Red
	
 Title        : Red
 Usage        : 
 Prerequisite : 
 Function     : Identification of repeats
 Args         : $o_param 
                $sequence      fasta     genomic sequence
                $outext        string    Extension of the result file. 
                $workingdir
                $uniq
 Globals      : none

=cut

sub Red
{
    my ($o_param, $sequence, $outext, $workingdir) = @_;

    my $step     = "REpeat Detection";
    my $firstseq = &GetFirstSeqPath($o_param);
    if ( &IsValidatedStep("$firstseq.$outext.gff3", $step))
    {
        return;
    }

	my $prevdir = cwd();

	#my $reddir = "$workingdir/Red.".time.".$$";
	my $reddir = "$workingdir/Red";
	mkdir $reddir if ( ! -d $reddir );
    &ExeCmd("ln -s $sequence $reddir/seq.fa", $o_logger, $logger_stamp) if ( ! -e "$reddir/seq.fa" );

	chdir($reddir);

    my $cmd = "cd $reddir ; mkdir -p out ; ".$o_param->Get('prg_red')." -gnm . -rpt out -msk out";
    $cmd .= " > /dev/null 2>&1 " if (!$DEBUG);
    &ExeCmd($cmd, $o_logger, $logger_stamp);

	my $f_in  = &GetStreamIn("$reddir/out/seq.rpt");
	my $f_out = &GetStreamOut("$reddir/out/seq.bed");
	while(my $l=<$f_in>)
	{
		chomp($l);
		$l=~s/^>//;
		my @a_ = split(' ',$l);
		print $f_out "$a_[0]\t$a_[$#a_-1]\t$a_[$#a_]\n";
	}
	$f_in->close;
	$f_out->close;

    &ExeCmd($o_param->Get('prg_mergebed')." -d ".$o_param->Get('prg_red_mergingdistance')." -i $reddir/out/seq.bed > $reddir/out/seq.merged.bed", $o_logger, $logger_stamp);

	$f_in  = &GetStreamIn("$reddir/out/seq.merged.bed");
	my $out = "$reddir/out/seq.merged.".$o_param->Get('prg_red_minlen').".red.gz";
	$f_out = &GetStreamOut($out);
	my %h_seqid=();
	while(my $l=<$f_in>)
	{
		chomp($l);
		my @a_ = split(' ',$l);
		next if (($a_[2]-$a_[1] < $o_param->Get('prg_red_minlen')));
		print $f_out "$l\n";
		$h_seqid{$a_[0]}=1;
	}
	$f_in->close;
	$f_out->close;

	if ( -e $out && ! -z $out )
	{
		&SplitBed($out, $outext, $o_param, $o_logger, $logger_stamp);
	}

	# ATTENTION, il faut aussi creer le fichier pour les autres
	my $f_fof = $o_param->GetStreamIn('seq_fof');
    while (my $seq = <$f_fof>)
    {
        chomp($seq);
		my($dir,$id)=$seq=~/(.+)\/(.+)/;
		if ( defined($h_seqid{$id}) )
		{
			my $seqworkdir = &GetWorkDir($seq, $o_param);
    		&ExeCmd($o_param->Get('prg_bed2gff3')." --in $seqworkdir/$id.$outext.$MAPPING_EXT --out $seq.$outext.gff3 --source red --code repeat_region", $o_logger, $logger_stamp);
		}
		else
		{
			&ExeCmd("touch $seq.$outext.gff3");
		}
	}

	&ValidateStep("$firstseq.$outext.gff3");

    &PrintStop($step);

	chdir($prevdir);

    return;
}


=head2 procedure Gmap
	
 Title        : Gmap
 Usage        : &Gmap();
 Prerequisite : o_param objects contains keys $param_prefix."_file", $param_prefix."_pcs"
                $param_prefix."_pci", $param_prefix."_remove_unspliced"
 Function     : Mapping the transcriptome library $param_prefix."_file" to genomic sequence $sequence
 Args         : $o_param 
                $sequence      fasta     genomic sequence
                $param_prefix  string    Example: "est_1"
                $outext        string    Extension of the result file. 
                                         Ex".est1" --> create seq1.est1.gff3, seq2.est1.gff3, etc 
                                         Use with split
                $splitfof      fof       If defined, split the result by sequence
                $outfile       file      Result file (user if splitdir is not defined)
 Globals      : none

=cut

sub Gmap
{
    my ($o_param, $sequence, $param_prefix, $outext, $splitfof, $outfile, $gmapbuild_dir) = @_;

    my $db  = $o_param->Get($param_prefix . "_file");
    my $pcs = $o_param->Get($param_prefix . "_pcs");
    my $pci = $o_param->Get($param_prefix . "_pci");
    my $cfg = $o_param->Get('cfg');
    my ($dbpath, $dbname) = &GetPathAndName($db);

    my $step     = "Gmap $dbname";
    my $firstseq = &GetFirstSeqPath($o_param);

    if (   &IsValidatedStep($outfile . ".gff3", $step)
        || &IsValidatedStep("$firstseq.$outext.gff3", $step))
    {
        return;
    }

    my $cmd = "export EGNPP=" . $o_param->Get('EGNEP') . ";";
    $cmd .= "export EGNEP=" . $o_param->Get('EGNEP') . ";";
    $cmd .= $o_param->Get('prg_get_est_gmap') . " --sequence $sequence --mrnadb $db --pcs $pcs --pci $pci --cfg $cfg";
	$cmd .= " --gmapbuild_dir $gmapbuild_dir " if ( defined($gmapbuild_dir) );
    $cmd .= "" if (!$CLEAN);
    $cmd .= " --outext $outext " if ($outext ne "");
    $cmd .= " --split $splitfof " if ($splitfof ne "");
    $cmd .= " --outfile $outfile " if ($outfile ne "");
    $cmd .= " --min_length_short_unspliced $MIN_LEN_SHORT_UNSPLICED " if ($o_param->Get($param_prefix . "_remove_unspliced") > 0);
    $cmd .= " > /dev/null 2>&1 " if (!$DEBUG);

    &PrintStart($step);
    my $param = "  PARAMETERS=" . $o_param->Get('prg_gmap_param');
    &PrintMore($param);
    $param = "  FILTERS=EST length percentage > $pcs, identity percentage > $pci";
    $param .= ", remove unspliced EST"            if ($o_param->Get($param_prefix . "_remove_unspliced") == 1);
	$param .= ", split spliced and unspliced EST" if ($o_param->Get($param_prefix . "_remove_unspliced") == 2);
    &PrintMore($param);

    &ExeCmd($cmd, $o_logger, $logger_stamp);

	

    if ($splitfof ne "")
    {
        &ValidateStep("$firstseq.$outext.gff3");
    }
    else
    {
        &ValidateStep($outfile . ".gff3");
    }
    &PrintStop($step);

    return;
}

sub GetSplitByStrandCmd
{
	my ($o_param, $f) = @_;
	
	my $cmd = $o_param->Get('prg_splitbystrand')." --gff3  $f"; 
	return $cmd;
}


sub SplitByStrand
{
	my ($o_param, $f) = @_;
	
	&ExeCmd(&GetSplitByStrandCmd($o_param, $f), $o_logger, $logger_stamp);
	
	return;
}


########################################################################
###  Prokaryote functions
########################################################################

=head2 procedure  EuGeneProk
	
 Title        : EuGeneProk
 Usage        : EuGene Prokaryote
 Function     : Launch eugene for Prokaryote genome
 Args         : $o_param 
 Globals      : none

=cut

sub EuGeneProk
{
    my ($o_param, $rootworkingdir, $allseqfile, $ALLSEQFILE_NAME, $o_logger, $logger_stamp, $uniq, $ra_seq) = @_;

    my $outdir = $o_param->Get('outdir');    # Output directory

    # Check and get the evidence input files
    my %h_files;
    &AssertProkEvidencePatternKeys($o_param, $o_param->Get('datadir'), \%h_files);

    # Check and get the BlastX protein databases
    my %h_database;                          # BlastX databases
    &AssertBlastX($o_param, \%h_database);

    # Load the template configuration files for EGN
    my %h_cfg_template;
    &GetTemplate($o_param, \%h_cfg_template);

    ################### Global compute ##############################

    &GetGlobalEvidence($o_param, \%h_files, $rootworkingdir, $allseqfile,
                       $ALLSEQFILE_NAME, $o_logger, $logger_stamp, $uniq);

    ################### Sequence specific compute ###################

    my @a_res_gff3;    # Array of EuGene gff3 result files
    my $annotastruct_cpt = -1;

    # for each fasta: create all evidences
    foreach my $seq (@$ra_seq)
    {
        printf STDOUT "ANALYSE sequence: $seq\n" if ($DEBUG);

        my ($seqpath, $seqname, $seqid, $seqlen) = &GetSeqInfo($seq);

        my $workingdir = $seqpath . "/work." . $uniq . "/";
        $o_param->Set('workingdir', $workingdir);
        my $matrixfile = "$seq.mat";
        my $cfg_file   = "$seq.eugene.par";

        my $annotastruct_content = "";
        $annotastruct_cpt = -1;    # Reinit the counter for this sequence

        # 1 Launch All BlastX
        foreach my $nb (keys %h_database)
        {
            &GetBlastX($o_param, $seq, $h_database{$nb}{file}, $nb, $workingdir, $o_logger, $logger_stamp);
        }

        # 2 Train EuGene
        &BuildEGNMatrix($o_param, $seq, $workingdir, $matrixfile, $o_logger, $logger_stamp);

        # 3 Manage all the kinds of RNAseq data
        foreach my $nb (sort keys(%h_files))
        {
            my $pat = "evidence_" . $nb;

            # Prepare the evidences for egn
            &GetEvidences($o_param, $seq, $seqid, $seqlen, $seqname, $workingdir, $pat, $h_files{$nb},
                          \$annotastruct_cpt, \%h_cfg_template, \$annotastruct_content,
                          $o_logger, $logger_stamp, $uniq);
        }

        # 4 ncRNA computing
        &GettRNAscanSE($o_param, $seq);
        &GetRnammer($o_param, $seq);
        &GetRfamScan($o_param, $seq);
        &CompleteCfg(\$annotastruct_content, \$annotastruct_cpt, $TRNASCAN_EXT,
                     $h_cfg_template{ncrna}, "$seq.$TRNASCAN_EXT.gff3");
        &CompleteCfg(\$annotastruct_content, \$annotastruct_cpt, $RFAMSCAN_EXT,
                     $h_cfg_template{ncrna}, "$seq.$RFAMSCAN_EXT.gff3");
        &CompleteCfg(\$annotastruct_content, \$annotastruct_cpt, $RNAMMER_EXT,
                     $h_cfg_template{ncrna}, "$seq.$RNAMMER_EXT.gff3");

        # 5 Launch Prodigal
        &GetProdigal($o_param, $seq, $seqlen, \$annotastruct_cpt, \%h_cfg_template, \$annotastruct_content, $o_logger,
                     $logger_stamp);

        # 6 Build the cfg file for egn
        &BuildEGNProkParameterFile($o_param, $cfg_file, \%h_database, $annotastruct_content, $annotastruct_cpt);

        # 7 Launch EuGene
        &LaunchEGN($o_param, $seq, $cfg_file, $matrixfile, $outdir, $o_logger, $logger_stamp, 0);

        # Compress all the temporary files
        my $cmd = "gzip -qr $workingdir/";
        &ExeCmd($cmd, $o_logger, $logger_stamp);

        printf STDOUT "ANALYSE sequence: $seq DONE.\n" if ($DEBUG);

        push(@a_res_gff3, "$outdir/$seqname.gff3");
    }

    # Compute statistics about annotation. Extract protein sequences.
    &Consolidate($o_param, $outdir, \@a_res_gff3, $allseqfile, $o_logger, $logger_stamp);
}

=head2 procedure PrintModeInfo
	
 Title        : PrintModeInfo
 Usage        : &PrintModeInfo($o_param, $file)
 Function     : Print to STDOUT and in the log file information about the mode
               (No prodigal, No ncrna)
 Args         : $o_param 
                $loggerfile
 Globals      : none

=cut

sub PrintModeInfo
{
    my ($o_param, $loggerfile, $cmd) = @_;

    printf STDOUT "%s\n", "#" x $LINELEN;
    printf STDOUT "%s\n", "#" x $LINELEN;
    print STDOUT "# $cmd";
    print STDOUT "# EuGene Pipeline $MODE - version $VERSION\n";
	print STDOUT "# Independent strand annotation\n" if ($INDEPENDENT_STRAND == 1);
    print STDOUT "# EUGENEDIR " . $o_param->Get('EUGENEDIR') . "\n";
    print STDOUT "# EGNEP " . $o_param->Get('EGNEP') . "\n";
    print STDOUT "# Log file $loggerfile\n";

    if ($o_param->IsDefined('no_ncrna_detection'))
    {
        printf STDOUT "# No ncRNA mode (tRNAscan-SE, rfam_scam and RNAmmer not launch)\n";
        $o_logger->Trace("No ncRNA mode (tRNAscan-SE, rfam_scam and RNAmmer not launch)\n",
                         $logger_stamp, $LOGGER_LEVEL_EXEC);
    }
    if ($o_param->IsDefined('no_prodigal'))
    {
        printf STDOUT "# No prodigal mode\n";
        $o_logger->Trace("No prodigal\n", $logger_stamp, $LOGGER_LEVEL_EXEC);
    }
    if ($o_param->IsDefined('no_rfamscan'))
    {
        printf STDOUT "# No rfam_scan mode\n";
        $o_logger->Trace("No rfam_scan\n", $logger_stamp, $LOGGER_LEVEL_EXEC);
    }
    if ($o_param->IsDefined('no_trnascan'))
    {
        printf STDOUT "# No trnascan-SE mode\n";
        $o_logger->Trace("No trnascan-SE\n", $logger_stamp, $LOGGER_LEVEL_EXEC);
    }
    if ($o_param->IsDefined('no_rnammer'))
    {
        printf STDOUT "# No Rnammer mode\n";
        $o_logger->Trace("No Rnammer\n", $logger_stamp, $LOGGER_LEVEL_EXEC);
    }

    printf STDOUT "%s\n", "#" x $LINELEN;
    printf STDOUT "%s\n", "#" x $LINELEN;
    print STDOUT "\n";

    return;
}

########################################################################
###  Assert functions
########################################################################

=head2 procedure AssertParameters

 Title        : AssertParameters
 Usage        : 
 Prerequisite : 
 Function     : assert tte required parameters
 Args         : $o_param 
 Globals      : none

=cut

sub AssertParameters
{
    my ($o_param) = @_;

    &AssertGeneralParameters($o_param);

    if ($MODE eq "EUK")
    {
        &AssertEukParameters($o_param);
    }
    else
    {
        &AssertProkParameters($o_param);
    }

    return;
}

=head2 procedure AssertGeneralParameters

 Title        : AssertGeneralParameters
 Usage        : 
 Prerequisite : 
 Function     : assert tte required parameters
 Args         : $o_param 
 Globals      : none

=cut

sub AssertGeneralParameters
{
    my ($o_param) = @_;

    $o_param->SetUnlessDefined('debug', $DEBUG_DEFAULT);
    $o_param->AssertInteger('debug');

    $CLEAN = 1;
    $CLEAN = 0 if $o_param->IsDefined('no_clean');

    $o_param->AssertDirExists('indir', 'outdir');
    $o_param->AssertFullPath('indir', 'outdir');
		
    $o_param->SetUnlessDefined('workingdir', $o_param->Get('outdir'));
    $o_param->AssertFullPath('workingdir', 'prg_get_blastx');
    $o_param->AssertDirExists('workingdir');

	foreach my $d ("outdir", "workingdir")
	{
		my $outdir = $o_param->Get($d);
		$outdir =~ s/\/$//;
		$o_param->Set($d, $outdir);
	}

    $o_param->AssertFileExists('prg_rnammer')
      if (   !$o_param->IsDefined('no_rnammer')
          && !$o_param->IsDefined('no_ncrna_detection'));

    $o_param->AssertFileExists('rfam_cm_file')
      if (   !$o_param->IsDefined('no_rfamscan')
          && !$o_param->IsDefined('no_ncrna_detection'));

    $o_param->AssertFileExists('prg_fasta2tree', 'prg_biofileconverter');

    $o_param->AssertDefined('prg_eugene', 'organism');
	

    return;
}

=head2 procedure AssertProkParameters

 Title        : AssertProkParameters
 Usage        : 
 Prerequisite : 
 Function     : assert tte required parameters
 Args         : $o_param 
 Globals      : none

=cut

sub AssertProkParameters
{
    my ($o_param) = @_;

    $o_param->SetUnlessDefined('genomedir', $o_param->Get('indir') . "/genome");
    $o_param->SetUnlessDefined('datadir',   $o_param->Get('indir') . "/data");
    $o_param->AssertDirExists('genomedir', 'datadir');

    $o_param->AssertFullPath('prg_egn-prok_train');
    $o_param->AssertFileExists('prg_eugene_prok_cfg_model', 'prg_eugene_cfg_low_gc', 'prg_eugene_cfg_high_gc');

    $o_param->AssertFileExists(
                               'prg_eugene_cfg_rnaseq_annotastruct', 'prg_eugene_cfg_ncrna_annotastruct',
                               'prg_eugene_cfg_blastx',              'prg_eugene_cfg_region_annotastruct',
                               'prg_eugene_cfg_tstart_annotastruct', 'prg_eugene_cfg_start_annotastruct',
                               'prg_shuffleseq',                     'prg_prodigal',
                               'prg_genome_statistics',              'prg_split_bed',
                               'prg_samtools',                       'prg_bam2fastq',
                               'prg_glint',                          'prg_bed_to_expr',
                               'prg_wig_to_expr',                    'prg_bed_filter',
                               'prg_tstart_detection'
                               );

    $o_param->AssertDefined(
                            'prg_glint_param',                      'prg_glint_small_param',
                            'prg_bed_to_expr_tstart_param',         'prg_bed_to_expr_tstart_scoremax',
                            'prg_bed_to_expr_tstart_scalingfactor', 'prg_bed_to_expr_tstart_min_coverage',
                            'prg_bed_to_expr_annot_param',          'prg_bed_to_expr_annot_scoremax',
                            'prg_bed_to_expr_annot_scalingfactor',  'prg_bed_to_expr_annot_min_coverage_percent',
                            'prg_bed_filter_param',                 'prg_bed_filter_pe_param',
                            'prg_bed_filter_small_pe_param',        'prg_tstart_detection_param',
                            'prg_tstart_detection_log_threshold',   'prg_tstart_detection_boostscore_factor',
                            'prg_wig_to_expr_param',                'prg_egn-prok_train_hsp_training_evalue'
                            );
    $o_param->AssertInteger(
                            'prg_glint_mate_dist',                    'prg_glint_mm_se',
                            'prg_glint_mm_pe',                        'prg_glint_mm_small_se',
                            'prg_glint_mm_small_pe',                  'prg_tstart_detection_derivative_threshold',
                            'prg_tstart_detection_tstart_ratio',      'prg_egn-prok_train_hsp_training_min_nb',
                            'prg_egn-prok_train_hsp_training_length', 'prg_egn-prok_train_hsp_training_pci',
                            'prg_prodigal_minlen'
                            );

    $o_param->AssertAllowedValue("prg_tstart_detection_boost_ig", "TRUE", "FALSE");

    return;
}

=head2 procedure AssertEukParameters

 Title        : AssertEukParameters
 Usage        : 
 Prerequisite : 
 Function     : assert the required parameters
 Args         : $o_param 
 Globals      : none

=cut

sub AssertEukParameters
{
    my ($o_param) = @_;

    $o_param->SetUnlessDefined('independent_strand_annotation', 0);
	$o_param->AssertAllowedValue("independent_strand_annotation", 0, 1);

    $o_param->AssertDirExists('indir');
    $o_param->Set('genomedir', $o_param->Get('indir'));

    $o_param->AssertFileExists('prg_eugene_euk_cfg_model');
	$o_param->AssertFileExists('prg_eugene_cfg_blastx', 'prg_eugene_cfg_est', 'prg_eugene_cfg_est_unspliced');

    $o_param->AssertFileExists('prg_eugene_cfg_ncrna_annotastruct', 'prg_eugene_cfg_repet_annotastruct');

    $o_param->AssertFileExists(
                               'prg_makeblastdb',      'prg_extract_seqlen',
                               'blast_cmd',            'prg_build_training_dataset',
                               'prg_eugene_train_IMM', 'prg_get_ltrharvest',
                               'prg_gt',               'prg_bed2gff3','prg_split_bed',
                               'prg_mergebed',         'prg_subtractbed', 'prg_mergesortgff3',
                               'prg_consolidate_gff3', 'prg_validate',
                               'prg_N50',              'prg_splitbystrand', 'prg_splitgff3', 'prg_splitspliced'
                               );

    # use by lipm_dbprot_remove_repbase.pl
    $o_param->AssertFileExists('prg_fasta_filter', 'blastp_cmd');
    $o_param->AssertFileExists('prg_fasta_splitter');
    $o_param->AssertDefined('prg_extract_repet_domain', 'prg_extract_repet_domain_min_scov',
                            'prg_get_est_gmap',         'prg_transfer_gff3_attr', 'prg_shell');

    $o_param->AssertDefined(
                            'prg_remove_repbase', 'prg_remove_repbase_blastparam',
                            'repeat_sequence_db', 'prg_gmapbuild',
                            'prg_gmap',           'prg_gmap_param', 'prg_gmap_filter_patch','prg_red','prg_red_minlen','prg_red_mergingdistance',
                            'prg_gmap_filter',    'blast_param', 'prg_gmap_intron_filter',
                            'prg_build_training_dataset_param'
                            );
    $o_param->AssertAllowedValue('gmap_intron_filter', 0, 1);

    $o_param->AssertInteger('hsp_training_min_nb', 'hsp_training_length', 'hsp_training_pci', 'blast_paraloop_chunk');
    $o_param->AssertDefined('hsp_training_evalue');

    $o_param->AssertDefined('prg_mapwithblast','protein_similarity_search_method');
    $o_param->AssertInteger('prg_mapwithblast_limit', 'prg_mapwithblast_win_size', 'prg_mapwithblast_win_overlap');
	$o_param->AssertAllowedValue('protein_similarity_search_method', "blastx", "ublast_blastx", "diamond_blastx");

    $o_param->AssertDefined('prg_usearch','prg_usearch_ublast_param') if ($o_param->Get('protein_similarity_search_method') eq "ublast_blastx");
	$o_param->AssertDefined('prg_diamond','prg_diamond_param')        if ($o_param->Get('protein_similarity_search_method') eq "diamond_blastx");

	
	# Assert wam directory exists
	$o_param->AssertDefined('wam_dataset');
	my $wam_dataset  = $o_param->Get('wam_dataset');
	my $wamdir = $o_param->Get('EUGENEDIR')."/models/WAM/$wam_dataset";
	$o_param->Set('wam_directory', $wamdir);
	$o_param->AssertDirExists('wam_directory');
	
    &AssertProteinDatabase($o_param->Get('repeat_sequence_db'));
    &AssertBlastXEuk($o_param);
    &AssertEstKeys($o_param);
	&AssertAdditionalKeys($o_param);

    if ($o_param->IsDefined('species_repeat_domains'))
    {
        $o_param->AssertFileExists('species_repeat_domains');
    }

    # si plus de 3 sites d'épissage non canoniques sont trouves, le programme sort.
    $o_param->SetUnlessDefined('max_noncansite_candidate_nb', 3);


    return;
}

=head2 function AssertBlastX

 Title        : AssertBlastX
 Usage        : 
 Prerequisite : blastx_db_list, training_db_file are defined in o_param
 Function     : assert the database exists and are formated. Fill $rh_db_list
 Args         : $o_param 
                $rh_db_list Ref to a hash: Key=nb, Value=path opf the database
 Globals      : none

=cut

sub AssertBlastX
{
    my ($o_param, $rh_db_list) = @_;

    $o_param->AssertDefined('blastx_db_list');

    my $list = $o_param->Get('blastx_db_list');    # "2 5 6"
    my @a_db_nb = split(/\s+/, $list);

    foreach my $nb (@a_db_nb)
    {
        my $db = $o_param->Get("blastx_db_$nb\_file");
        $db =~ s/\s+$//;

        &AssertProteinDatabase($db);
        $o_param->SetUnlessDefined("blastx_db_$nb\_weight", 1);
        $o_param->SetUnlessDefined("blastx_db_$nb\_pcs",0);
        $o_param->SetUnlessDefined("blastx_db_$nb\_activegaps",0);
        my $level = $o_param->Get("blastx_db_$nb\_weight");

        $$rh_db_list{$nb}{file}  = $db;
        $$rh_db_list{$nb}{weight}= $level;
        $$rh_db_list{$nb}{pcs}= $o_param->Get("blastx_db_$nb\_pcs");
        $$rh_db_list{$nb}{activegaps}= $o_param->Get("blastx_db_$nb\_activegaps");
		$$rh_db_list{$nb}{pcs}=~s/\s//g;
		$$rh_db_list{$nb}{activegaps}=~s/\s//g;
    }

    # assert protein database for training is defined
    &AssertProteinDatabase($o_param->Get('training_db_file'));

    return;
}

=head2 function AssertBlastXEuk

 Title        : AssertBlastXEuk
 Usage        : 
 Prerequisite : blastx_db_list, training_db_file are defined in o_param
 Function     : assert the database exists and are formated
 Args         : $o_param 
 Globals      : none

=cut

sub AssertBlastXEuk
{
    my ($o_param) = @_;

    my $training_db_found = 0;

    $o_param->AssertDefined('blastx_db_list');

    my $list = $o_param->Get('blastx_db_list');    # "2 5 6"
    my @a_db_nb = split(/\s+/, $list);

    foreach my $nb (@a_db_nb)
    {
        my $db = $o_param->Get("blastx_db_$nb\_file");
        $db =~ s/\s+$//;

        &AssertProteinDatabase($db);
        $o_param->AssertInteger("blastx_db_$nb\_preserve");
        $o_param->SetUnlessDefined("blastx_db_$nb\_weight",       1);
        $o_param->SetUnlessDefined("blastx_db_$nb\_remove_repet", 1);
        $o_param->AssertInteger("blastx_db_$nb\_remove_repet");
        $o_param->AssertInteger("blastx_db_$nb\_training");
        $o_param->AssertInteger("blastx_db_$nb\_activegaps");
        $o_param->AssertInteger("blastx_db_$nb\_pcs");
        $o_param->AssertAllowedValue("blastx_db_$nb\_training", 0, 1);

        if ($o_param->Get("blastx_db_$nb\_training") == 1)
        {
            if (!$training_db_found)
            {
                $o_param->Set('training_dbprot_number', $nb);
                $training_db_found = 1;
            }
            else
            {
                print STDERR "Warning: only one protein database is used for IMM training. (blastx_db_\$nb skipped).\n";
            }
        }
    }

    if ($training_db_found == 0)
    {
        Carp::croak "No blastx database for IMM training. Choose one with parameter 'blastx_bd_X_training=1'\n";
    }

    return;
}

=head2 function AssertProteinDatabase

 Title        : AssertProteinDatabase
 Usage        : 
 Prerequisite : 
 Function     : Assert the protein database  is formated. 
                If the database is a .pal, check the subdatabases are formated
 Args         : $db
 Globals      : none

=cut

sub AssertProteinDatabase
{
    my ($db) = @_;

    if (!-f $db . ".nin" && !-f $db . ".pin" && !-f $db . ".pal" && !-f $db . ".nal")
    {
        Carp::croak "The database \"$db\" does not exist or isn't indexed. "
          . "(Use 'makeblastdb' program to index).\n";
    }
    if (-e $db . ".pal")
    {
        my ($paldir, $file) = $db =~ /(.+)\/(.+)$/;
        my $f_ = &General::GetStreamIn($db . ".pal");
        while (my $line = <$f_>)
        {
            if ($line =~ /^DBLIST (.+)$/)
            {
                my @a_ff = split(/\s+/, $1);
                foreach my $elt (@a_ff)
                {
                    # if it is not an absolute path
                    $elt = "$paldir/$elt" if (dirname($elt) eq ".");
                    Carp::croak "The database \"$elt\" isn't indexed. " . "Use 'makeblastdb' program to format it.\n"
                      if (!-f "$elt.pin" && !-f $db . ".nin");
                }
                last;
            }
        }
    }

    return;
}

=head2 procedure CleanProtDb

 Title        : CleanProtDb
 Usage        : &CleanProtDb($o_param, \%h_database)
 Prerequisite : 
 Procedure    : remove repet in each protein database of the hashtable h_database. Index these new db.
                Replace the database in the hashtable.
 Args         : $o_param     ParamParser object
                $rh_database  Ref to a hash. Key: nb, value:protein database path
 Globals      : none

=cut

sub CleanProtDb
{
    my ($o_param, $rh_database) = @_;
    my $cmd;
    my $newdbdir = &GetDbDir($o_param);

    foreach my $nb (keys %$rh_database)
    {
        my $db = $$rh_database{$nb}{file};
        my ($dbdir, $dbname) = &GetPathAndName($db);
        my $localdb = $newdbdir . "/" . $dbname;

        #1  create a link in the db directory
        if (!-e $localdb && !-l $localdb)
        {
            $cmd = "ln -s $db $localdb";
            &ExeCmd($cmd, $o_logger, $logger_stamp);
        }

		# JG on doit pouvoir simplifier en testant $o_param->Get("blastx_db_$nb\_remove_repet") == 1 && -e $db && $db !~ /.pal$/ puis un seul else
        # not repeat cleaning: only create index
        if ($o_param->Get("blastx_db_$nb\_remove_repet") == 0)
        {
			if (!&IsValidatedStep($localdb, "Clean database $dbname"))
			{
				&MakeBlastDb($o_param, $localdb, $o_logger, $logger_stamp);
				&MakeReduceDbMethodIndex($o_param, $localdb, $o_logger, $logger_stamp);
				&ValidateStep($localdb);
			}
        }
        else
        {
            if (-e $db && $db !~ /.pal$/)
            {
                my $cleandb = $localdb . ".noTE";
                if (!&IsValidatedStep($cleandb, "Clean database $dbname"))
                {
                    &RemoveRepbase($o_param, $localdb, $cleandb);
                    &MakeBlastDb($o_param, $cleandb, $o_logger, $logger_stamp);	
		    		&MakeReduceDbMethodIndex($o_param, $cleandb, $o_logger, $logger_stamp);
                    &ValidateStep($cleandb);
                    &PrintDebug("Clean database $db --> $cleandb\n");
                }
                $localdb = $cleandb;
            }
            else
            {
                &MakeBlastDb($o_param, $localdb, $o_logger, $logger_stamp);
				&MakeReduceDbMethodIndex($o_param,  $localdb, $o_logger, $logger_stamp);
            }
        }
        $$rh_database{$nb}{file} = $localdb;
    }

    return;
}

=head2 procedure RemoveRepbase

 Title        : RemoveRepbase
 Usage        : 
 Prerequisite : Keys 'prg_remove_repbase'  'repeat_sequence_db' 'blastp_cmd' 
                'prg_fasta_filter' have to exist in $o_param  
 Function     : Remove repet sequence in the protein database $db
 Args         : $o_param     ParamParser object
                $db
                $cleandb
 Globals      : none

=cut

sub RemoveRepbase
{
    my ($o_param, $db, $cleandb) = @_;

    &PrintStart("Clean database $db");
    my $pg_remove_repbase = $o_param->Get('prg_remove_repbase');
    my $rep_db            = $o_param->Get('repeat_sequence_db');
    my $blastparam        = $o_param->Get('prg_remove_repbase_blastparam');
    my $blastpprg         = $o_param->Get('blastp_cmd');
    my $prg_fasta_filter  = $o_param->Get('prg_fasta_filter');

    &PrintMore("  PARAMETERS: $blastparam") if ($blastparam ne "");

    my $param = "";
    $param .= " --blastparam $blastparam " if ($blastparam ne "");
    my $cmd = "$pg_remove_repbase --in $db --out $cleandb --repbase $rep_db $param ";
    $cmd .= " --blastpprg $blastpprg --fastafilterprg $prg_fasta_filter ";
    $cmd .= "> /dev/null 2>&1" if (!$DEBUG);

    &ExeCmd($cmd, $o_logger, $logger_stamp);

    &PrintStop("Clean database $db");

    return;
}

=head2 procedure AssertProkEvidencePatternKeys

 Title        : AssertProkEvidencePatternKeys
 Usage        : 
 Prerequisite : -
 Function     : Assert 'evidence_list' is defined. According to this value,  
                assert all the attending keys are defined in o_param. 
                Fill $rh_files with 
                the evidence numbers and the corresponding files
 Args         : $o_param     ParamParser object
                $indir       directory of input data
                $rh_files    Ref to a hask Key: evidence type number, 
                             value: ref to an array of files
 Globals      : none

=cut

sub AssertProkEvidencePatternKeys
{
    my ($o_param, $indir, $rh_files) = @_;

    $o_param->AssertDefined('evidence_list');
    my $evidence_list = $o_param->Get('evidence_list');
    my @a_evidence_list = split(/\s+/, $evidence_list);

    # Assert the keys
    foreach my $nb (@a_evidence_list)    # '2', '3'
    {
        my $root       = "evidence_" . $nb;
        my $typekey    = $root . "_type";
        my $patternkey = $root . "_pat";
        my $formatkey  = $root . "_format";

        $o_param->AssertAllowedValue($formatkey, 'fastq', 'bed', 'bedpe', 'fasta', 'wig', 'sam', 'bam', 'pair');
        $o_param->AssertDefined($patternkey, $formatkey);

        my $format = $o_param->Get($formatkey);

        if ($format =~ /fast[qa]/)
        {
            $o_param->AssertDefined($typekey);
            $o_param->AssertAllowedValue($typekey, 'ose', 'ope');
        }

        if ($format eq "bed")
        {
            $o_param->Set($typekey, 'ose');
        }
        if ($format eq "bedpe")
        {
            $o_param->Set($typekey, 'ope');
        }

        # Need to know the orientation for all format except wig
        if (   $format eq "bam"
            || $format eq "sam"
            || $format =~ /fast[qa]/
            || $format =~ /^bed/)
        {
            $o_param->AssertAllowedValue($root . "_small",        0, 1);
            $o_param->AssertAllowedValue($root . "_oriented_end", 1, 2);
        }

        if ($format eq "wig")
        {
            $o_param->AssertAllowedValue($root . "_strand", 'forward', 'reverse');
        }
        if ($format eq "pair")
        {
            my $ndf_filename = $o_param->Get($root . "_ndf");
            my $ndf_file     = $indir . "/" . $ndf_filename;
            $o_param->Set($root . "_ndf", $ndf_file);
            $o_param->AssertFileExists($root . "_ndf");
            $o_param->AssertDefined('prg_glint_tiling_param');
        }

        my $pattern = $o_param->Get($patternkey);
        $o_param->Set('pattern_path', "$indir/$pattern");
        my $ra_files = $o_param->GetListOfFiles('pattern_path');

        if (scalar(@$ra_files) == 0)
        {
            print "No file $indir/$pattern found for RNAseq data number $nb --> IGNORED\n";
            next;
        }

        if ($format =~ /fast[aq]/ && $o_param->Get($typekey) =~ /ope/)
        {
            # Check the number of files is modulo2
            if ((scalar(@$ra_files) % 2) != 0)
            {
                Carp::croak "ERROR looking for paired-end files: "
                  . $o_param->Get('pattern_path') . ". "
                  . scalar(@$ra_files)
                  . " files found (not a pair number). Two files are attending per library.\n";
            }

            my @a_gp;
            &GroupPair($ra_files, \@a_gp);
            @$ra_files = @a_gp;
        }
        $o_param->Delete('pattern_path');

        $$rh_files{$nb} = $ra_files;
    }

    return;
}

=head2 procedure AssertEstKeys

 Title        : AssertEstKeys
 Usage        : 
 Prerequisite : -
 Function     : Assert 'est_list' is defined. According to this value,  
                assert all the attending keys are defined in o_param. 
                Fill $rh_files with 
                the evidence numbers and the corresponding files
 Args         : $o_param     ParamParser object
                $rh_files    Ref to a hask Key: evidence type number, 
                             value: ref to an array of files
 Globals      : none

=cut

sub AssertEstKeys
{
    my ($o_param) = @_;

    my $found_training = 0;
    $o_param->AssertDefined('est_list');
    my $transcriptome_list = $o_param->Get('est_list');
    my @a_transcriptome_list = split(/\s+/, $transcriptome_list);

    # Assert the transcriptome file exist
    foreach my $nb (@a_transcriptome_list)    # '2', '3'
    {
        $o_param->AssertFileExists("est_" . $nb . "_file");
        $o_param->AssertInteger("est_" . $nb . "_pcs");
        $o_param->AssertInteger("est_" . $nb . "_pci");
        $o_param->AssertInteger("est_" . $nb . "_preserve");
        $o_param->AssertInteger("est_" . $nb . "_training");
        if ($o_param->Get("est_" . $nb . "_training") == 1)
        {
            if (!$found_training)
            {
                $o_param->Set('training_dbest_number', $nb);
                $found_training = 1;
            }
            else
            {
                print STDERR "Warning: only one transcriptome file is used for IMM training. (est_" . $nb
                  . " skipped).\n";
            }
        }

        $o_param->AssertAllowedValue("est_" . $nb . "_remove_unspliced", 0, 1, 2, "TRUE", "FALSE");
		$o_param->Set("est_" . $nb . "_remove_unspliced", 0) if ($o_param->Get("est_" . $nb . "_remove_unspliced") eq "FALSE");
		$o_param->Set("est_" . $nb . "_remove_unspliced", 1) if ($o_param->Get("est_" . $nb . "_remove_unspliced") eq "TRUE");
    }

    if ($found_training == 0)
    {
        Carp::croak "No transcriptome selected for IMM training. Choose one with parameter 'est_X_training=1'\n";
    }

    $o_param->AssertInteger('training_est_pcs', 'training_est_pci');

    $o_param->AssertAllowedValue("training_est_remove_unspliced", 0, 1, 2, "TRUE", "FALSE");
	$o_param->Set("training_est_remove_unspliced", 0) if (	$o_param->Get("training_est_remove_unspliced") eq "FALSE");
	$o_param->Set("training_est_remove_unspliced", 1) if (	$o_param->Get("training_est_remove_unspliced") eq "TRUE");

    return;
}

sub AssertAdditionalKeys
{
	my ($o_param) = @_;
	
	if ( $o_param->IsDefined('additional_list'))
	{
		my $list   = $o_param->Get('additional_list');
		my @a_list = split(/\s+/, $list);
		
		# Assert the file exist
		foreach my $nb (@a_list)    # '2', '3'
		{
			$o_param->AssertFileExists("additional_" . $nb . "_file");
			my $ext = "";
			$ext = &GetGFF3Ext($o_param->Get("additional_" . $nb . "_file"));
			if (!defined ($ext) || $ext eq "")
			{
				Carp::croak "additional_${nb}_file must end with .\$ext.gff3\n";	
			}
			$o_param->AssertFileExists("additional_" . $nb . "_cfg_template");
		}
	}
	return;
}

=head2 procedure GroupPair

 Title        : GroupPair
 Usage        : 
 Prerequisite : $ra_file array of existing files
 Procedure    : concatenate paired files and fill the array $ra_gp 
 Args         : $ra_file Ref to an array of files
                $ra_gp: Ref to an array to fill with pair of files.
 Globals      : none

=cut

sub GroupPair
{
    my ($ra_file, $ra_gp) = @_;

    my $file1 = "";    # First file
    foreach my $file (@$ra_file)
    {
        if ($file1 ne "")
        {
            # Conatenate the two files
            push(@$ra_gp, "$file1,$file");
            $file1 = "";
        }
        else
        {
            $file1 = $file;
        }
    }

    return;
}

=head2 procedure BuildEGNMatrix

 Title        : BuildEGNMatrix
 Usage        : 
 Prerequisite : prg_egn-prok_train has to be defined
 Function     : Build the eugene matrix file, using the blastx result
 Args         : $o_param               ParamParser object
                $seq                   fasta file path
                $workingdir            working directory         
                $matrixfile            Result matrix file path
 
 Globals      : none

=cut

sub BuildEGNMatrix
{
    my ($o_param, $seq, $workingdir, $matrixfile, $o_logger, $logger_stamp) = @_;

    if ($DEBUG && (-e $matrixfile))
    {
        print STDOUT "DEBUG MODE: $matrixfile used.\n";
        return;
    }

    my $training_blastxfile;
    my $training_db_file = $o_param->Get('training_db_file');

    my ($seqpath, $seqname) = $seq =~ /(.+)\/(.+)/;

    # Check if the blastX against this protein database was previously done
    my $prev_blastx = &SearchBlastXResult($o_param, $workingdir, $seqname, $training_db_file);
    if ($prev_blastx ne "")
    {
        print STDOUT "Prev blastX result found : $prev_blastx\n" if ($DEBUG);
        $training_blastxfile = $prev_blastx;
    }
    else
    {
        print STDERR "No Prev BlastX result found for  $training_db_file\n" if ($DEBUG);

        # compute the BlastX
        &GetBlastX($o_param, $seq, $training_db_file, "mat", $workingdir, $o_logger, $logger_stamp);
        $training_blastxfile = "$workingdir/$seqname.blastmat";
    }

    if (&IsOKTrainingDatasets($training_blastxfile, $seq, $o_param))
    {
        my $prg_train = $o_param->Get('prg_egn-prok_train');
        my $cmd_train =
            "$prg_train --cfg "
          . $o_param->Get('cfg')
          . " --infile $seq --workingdir $workingdir "
          . "--blastxfile $training_blastxfile --matrixfile $matrixfile";
        &ExeCmd($cmd_train, $o_logger, $logger_stamp);

        $o_param->SetUnlessDefined('reference_matrix', $matrixfile);
    }
    else
    {
        if ($o_param->IsDefined('reference_matrix'))
        {
            &PrintInfo("Not enough blastx hsp to build the matrix of $seq ==> Use a previous one.\n");

            my $cmd = "cp " . $o_param->Get('reference_matrix') . " $matrixfile";
            &ExeCmd($cmd, $o_logger, $logger_stamp);
        }
        else
        {
            # If the blastx file is not empty
            if (-s $training_blastxfile)
            {
                &PrintInfo("WARNING: build the matrix for $seq but not enough blastx hsp (no previous matrix found)\n");

                my $prg_train = $o_param->Get('prg_egn-prok_train');
                my $cmd_train =
                    "$prg_train --cfg "
                  . $o_param->Get('cfg')
                  . " --infile $seq --workingdir $workingdir "
                  . "--blastxfile $training_blastxfile --matrixfile $matrixfile";
                &ExeCmd($cmd_train, $o_logger, $logger_stamp);
            }
            else
            {
                my $info =
                    "ERROR: impossible to train EuGene on $seq. "
                  . "BlastX results are mandatory but file $training_blastxfile is empty. "
                  . " You should try an other training protein database 'training_db_file'\n";
                $o_logger->Trace("$info", $logger_stamp, $LOGGER_LEVEL_EXEC);
                Carp::croak($info);
            }
        }
    }
    &PrintStop("BUILD IMM MODELS");
    return;
}

=head2 function IsOKTrainingDatasets

 Title        : IsOKTrainingDatasets
 Usage        : 
 Prerequisite : -
 Function     : Count the number of HSP after filtering (len, pci, evalue)
                Return 1 if its upper than the threshold
 Args         : $blastxfile M8 blastx result file
                $seqfile fasta file
                $o_param
 Globals      : none

=cut

sub IsOKTrainingDatasets
{
    my ($blastxfile, $seqfile, $o_param) = @_;

    my %h_seq;
    &FastaToHash($seqfile, \%h_seq);
    my %h_hit      = ();
    my $hsp_nb     = 0;
    my $hsp_pci    = $o_param->Get('prg_egn-prok_train_hsp_training_pci');
    my $hsp_evalue = $o_param->Get('prg_egn-prok_train_hsp_training_evalue');
    my $hsp_length = $o_param->Get('prg_egn-prok_train_hsp_training_length');

    my $f_blast = &GetStreamIn($blastxfile);
    my %h_deja  = ();
    while (my $line = <$f_blast>)
    {
        next if ($line =~ /^#/);

        &ParseM8Line(\$line, 1, \%h_hit);
        next
          if (!(($h_hit{pci} >= $hsp_pci) && ($h_hit{"e-value"} <= $hsp_evalue) && ($h_hit{len} >= $hsp_length)))
          ;    # for blastx the length is in aa
        next if (defined($h_deja{$h_hit{target}}));
        $h_deja{$h_hit{target}} = 1;

        my ($start, $end) = &MinMax($h_hit{qstart}, $h_hit{qend});
        next if (($end - $start + 1) % 3 != 0);
        my ($query) = $h_hit{query};
        my $sseq = &ExtractSeq(\$h_seq{$h_hit{query}}{sequence}, $start, $end, $h_hit{strand});
        next if ($sseq =~ /N/);
        $sseq =~ s/\w\w\w$//;    # j'enleve le dernier codon
        my $ok = &TRUE;
        while ($sseq =~ /(...)/g)
        {
            my $codon = $1;
            if ($codon =~ /TAA|TGA|TAG/)
            {
                $ok = &FALSE;
                last;
            }
        }
        $hsp_nb++ if ($ok);
    }
    $f_blast->close;

    return 1 if ($hsp_nb >= $o_param->Get('prg_egn-prok_train_hsp_training_min_nb'));
    return 0;
}

=head2 procedure BuildEGNProkParameterFile

 Title        : BuildEGNProkParameterFile
 Usage        : 
 Prerequisite : $o_param has the following keys: 'organism' 'prg_eugene_cfg_model' 
                'prg_eugene_cfg_gc' 'prg_eugene_cfg_blastx' and 'gc_percent'
 Function     : Create the complete eugene configuration file
 Args         : $o_param                       ParamParser object
                $cfg_file                      File to create
                $rh_db_level           hashref Ref to a hash : key=blastx number, 
                                               value=level of confidence of the blastx result
                $annotastruct_content  string  Annotastruct instances description
                $annotastruct_cpt      int     Index of the last AnnotaStruct instance        
 Globals      : none

=cut

sub BuildEGNProkParameterFile
{
    my ($o_param, $cfg_file, $rh_db_level, $annotastruct_content, $annotastruct_cpt) = @_;

    # Result file
    my $f_cfg = &General::GetStreamOut($cfg_file);

    # Copy the conf model file
    my $f_model = $o_param->GetStreamIn('prg_eugene_prok_cfg_model');
    while (my $cfg_line = <$f_model>)
    {
        $cfg_line = "EuGene.organism " . $o_param->Get('organism') . "\n" if ($cfg_line =~ /^EuGene.organism/);
        print $f_cfg $cfg_line;
    }
    $f_model->close;

    # add to the cfg file the part depending on the GC% of the organism
    my $cfg_gc_model =
      ($o_param->Get('gc_percent') eq "high")
      ? $o_param->Get('prg_eugene_cfg_high_gc')
      : $o_param->Get('prg_eugene_cfg_low_gc');
    my $f_gc_model = &General::GetStreamIn($cfg_gc_model);
    while (my $cfg_line = <$f_gc_model>)
    {
        print $f_cfg $cfg_line;
    }
    $f_gc_model->close;

    # Add the instances of AnnotaStruct
    print $f_cfg $annotastruct_content;
    print $f_cfg "Sensor.AnnotaStruct.use " . ($annotastruct_cpt + 1) . "\n##\n";

    # Create the blastX instance
    if ($o_param->Get('blastx_db_list') ne "")
    {
        my $blastx_line;
        my $f_blastx = $o_param->GetStreamIn('prg_eugene_cfg_blastx');
        while (my $line = <$f_blastx>)
        {
            $line =~ s/\s+$//;
            $line =~ s/\s+/\t/g;
            $line .= "\n";
            $blastx_line .= $line;
        }
        $f_blastx->close;
        my $levels='';
		my $activegaps='';
        foreach my $nb (sort keys %$rh_db_level)
        {
            my $val = $$rh_db_level{$nb}{weight};
            $blastx_line =~ s/BlastX\.level$nb\*\t0\.0/BlastX\.level$nb\*\t$val/;
            $levels .= $nb;
			if ( $$rh_db_level{$nb}{activegaps} )
			{
				$activegaps.=$nb;
			}
        }
        $blastx_line =~ s/BlastX\.levels\t\d/BlastX\.levels\t$levels/;
		if ( $activegaps ne '' )
		{
        	$blastx_line =~ s/BlastX\.activegaps\s+NONE/BlastX\.activegaps\t$activegaps/;
		}
        print $f_cfg $blastx_line;
        print $f_cfg "Sensor.BlastX.use	1";
    }
    else
    {
        print $f_cfg "Sensor.BlastX.use	0";
    }

    $f_cfg->close;

    return;
}

=head2 procedure BuildEGNEukParameterFile

 Title        : BuildEGNEukParameterFile
 Usage        : 
 Prerequisite : $o_param has the following keys: 'organism' 'prg_eugene_cfg_model' 
                'prg_eugene_cfg_gc' 'prg_eugene_cfg_blastx' and 'gc_percent'
 Function     : Create the complete eugene configuration file
 Args         : $o_param                       ParamParser object
                $cfg_file                      File to create
                $rh_db_level           hashref Ref to a hash : key=blastx number, 
                                               value=level of confidence of the blastx result   
                $repeat                        If 1, add an annotastruct instance to mask repeat regions
                $strand             integer    0 both strand, -1 reverse, 1 forward
 Globals      : none

=cut

sub BuildEGNEukParameterFile
{
    my ($o_param, $cfg_file, $rh_db_level, $repet, $strand) = @_;

	my $strandext = "";
	$strandext    = ".FW" if ($strand == 1);
	$strandext    = ".RV" if ($strand == -1);

    # cfg file
    my $f_cfg = &General::GetStreamOut($cfg_file);

    #### 1 Copy the conf model file (without noncansplice sites)
    my $f_model = $o_param->GetStreamIn('prg_eugene_euk_cfg_model');
    while (my $cfg_line = <$f_model>)
    {
		# change the organism name
        $cfg_line = "EuGene.organism " . $o_param->Get('organism') . "\n" if ($cfg_line =~ /^EuGene.organism/);
		
		# change the WAM model to use
		if ($cfg_line =~ /^(SignalWAM\.species\S+)\s/)
		{
			my $attribute   = $1;
			my $wam_dataset = $o_param->Get('wam_dataset');
			$cfg_line       = "$attribute $wam_dataset\n";
		}
		# changes according to the selected strand
		$cfg_line = "MarkovConst.affectedStrand $strand\n" if ($strand != 0  && $cfg_line =~ /^MarkovConst.affectedStrand/); 
		$cfg_line = "Output.initid 1000000\n"              if ($strand == -1 && $cfg_line =~ /^Output.initid/);

		# remove the uncanonical splice site information
        next if ($cfg_line =~ /^EuGene.NonCanDon/ || $cfg_line =~ /^EuGene.NonCanAcc/);
        print $f_cfg $cfg_line;
    }
    $f_model->close;

    # 2.3 If found, write in the configuration file the non canonical spliced sites
    my $noncanacc = $o_param->Get('noncanacc');
    my $noncandon = $o_param->Get('noncandon');
    print $f_cfg "##\nEuGene.NonCanDon $noncandon\n##\n" if ($noncandon ne "");
    print $f_cfg "##\nEuGene.NonCanAcc $noncanacc\n##\n" if ($noncanacc ne "");


    # Create the blastX instance
    if ($o_param->Get('blastx_db_list') ne "")
    {
        my $blastx_line;
        my $f_blastx = $o_param->GetStreamIn('prg_eugene_cfg_blastx');
        while (my $line = <$f_blastx>)
        {
            $line =~ s/\s+$//;
            $line =~ s/\s+/\t/g;
            $line .= "\n";
            $blastx_line .= $line;
        }
        $f_blastx->close;
        my $levels='';
		my $activegaps='';
        foreach my $nb (sort keys %$rh_db_level)
        {
            my $val = $$rh_db_level{$nb}{weight};
            $blastx_line =~ s/BlastX\.level$nb\*\t0\.0/BlastX\.level$nb\*\t$val/;
            $levels .= $nb;
			if ( $$rh_db_level{$nb}{activegaps} )
			{
				$activegaps.=$nb;
			}
        }
        $blastx_line =~ s/BlastX\.levels\t\d/BlastX\.levels\t$levels/;
		if ( $activegaps ne '' )
		{
        	$blastx_line =~ s/BlastX\.activegaps\s+NONE/BlastX\.activegaps\t$activegaps/;
		}
        print $f_cfg $blastx_line;
        print $f_cfg "Sensor.BlastX.use	1\n##\n";
    }
    else
    {
        print $f_cfg "Sensor.BlastX.use	0\n##\n";
    }

    # Add the Est instances
    my @a_est_list = split(/\s+/, $o_param->Get('est_list'));
    my $est_content;
    my $est_cpt = -1;

    my $est_template           = &ReadFile($o_param->Get('prg_eugene_cfg_est'));    
	my $est_template_unspliced = &ReadFile($o_param->Get('prg_eugene_cfg_est_unspliced'));
    foreach my $nb (@a_est_list)    # '2', '3'
    {
        &CompleteCfg(\$est_content, \$est_cpt, ".est$nb$strandext", $est_template);
		
		if ($o_param->Get("est_".$nb. "_remove_unspliced") == 2)
		{
			&CompleteCfg(\$est_content, \$est_cpt, ".est$nb.rm_unspliced_$MIN_LEN_SHORT_UNSPLICED$strandext", $est_template_unspliced);
		}
    }
    print $f_cfg $est_content;
    print $f_cfg "Sensor.Est.use " . ($est_cpt + 1) . "\n##\n";

    # Add the annotastruct instances
    my $annot_content = "";
    my $annot_cpt     = -1;

    # the ncrna instances
    my $template_ncrna = &ReadFile($o_param->Get('prg_eugene_cfg_ncrna_annotastruct'));
    &CompleteCfg(\$annot_content, \$annot_cpt, "$TRNASCAN_EXT$strandext", $template_ncrna)
      if (!$o_param->IsDefined('no_ncrna_detection') && !$o_param->IsDefined('no_trnascan'));
    &CompleteCfg(\$annot_content, \$annot_cpt, "$RFAMSCAN_EXT$strandext", $template_ncrna)
      if (!$o_param->IsDefined('no_ncrna_detection') && !$o_param->IsDefined('no_rfamscan'));
    &CompleteCfg(\$annot_content, \$annot_cpt, "$RNAMMER_EXT$strandext", $template_ncrna)
      if (!$o_param->IsDefined('no_ncrna_detection') && !$o_param->IsDefined('no_rnammer'));

    if ($repet)
    {
        # the repeat instance
        my $template_repet = &ReadFile($o_param->Get('prg_eugene_cfg_repet_annotastruct'));
        &CompleteCfg(\$annot_content, \$annot_cpt, "$REPET_EXT$strandext", $template_repet);
    }

	if ($o_param->IsDefined('additional_list') && $o_param->Get('additional_list') ne "")
    { 
		my @a_add_list = split(/\s+/, $o_param->Get('additional_list'));
		foreach my $nb (@a_add_list)    # '2', '3'
		{
			my $template = &ReadFile($o_param->Get("additional_" . $nb . "_cfg_template"));
			my $file     = $o_param->Get("additional_" . $nb . "_file");
			my $ext      = &GetGFF3Ext($file);
			&CompleteCfg(\$annot_content, \$annot_cpt, $ext, $template);
		}
	}

    print $f_cfg $annot_content;
    print $f_cfg "Sensor.AnnotaStruct.use " . ($annot_cpt + 1) . "\n##\n";

    $f_cfg->close;

    return;
}


sub GetGFF3Ext
{
	my ($gff3file) = @_;
	
	my $ext = "";
	($ext) = $gff3file =~ /\.([^\.]+)\.gff3/i;
	
	return $ext;
}


=head2 procedure GetTemplate

 Title        : GetTemplate
 Procedure     :fill $rh_template
 Args         : $o_param   
                $rh_template key=position/region/ncrna/tstart/blastx, 
                             value=template of the cfg part for this plugin
 Globals      : none

=cut

sub GetTemplate
{
    my ($o_param, $rh_template) = @_;

    # configuration file part for blastX and AnnotaStruct
    $$rh_template{position} = &ReadFile($o_param->Get('prg_eugene_cfg_rnaseq_annotastruct'));
    $$rh_template{region}   = &ReadFile($o_param->Get('prg_eugene_cfg_region_annotastruct'));
    $$rh_template{ncrna}    = &ReadFile($o_param->Get('prg_eugene_cfg_ncrna_annotastruct'));
    $$rh_template{tstart}   = &ReadFile($o_param->Get('prg_eugene_cfg_tstart_annotastruct'));
    $$rh_template{start}    = &ReadFile($o_param->Get('prg_eugene_cfg_start_annotastruct'));

    return;
}

=head2 procedure GetGlobalEvidence

 Title        : GetGlobalEvidence
 Usage        : 
 Prerequisite : 
 Function     : Launch the compute which have to be done together on all the input sequences 
                * Case FASTQ: mapping and mapping filtering
                * Case wig: 
                Then split the results in the appropriate directory  
 Args         : $o_param 
                $rh_evidence_type ref to a hash  key=evidence nb, value=ref to an array of files
                $workingdir       dir path       Output directory
                $seqfile          file           Input multifasta sequence file
                $rootname         String         Prefix of the output files 
 Globals      : none

=cut

sub GetGlobalEvidence
{

    my ($o_param, $rh_evidence_type, $workingdir, $seqfile, $rootname, $o_logger, $logger_stamp, $uniq) = @_;

	my $prevdir = cwd();
    chdir($workingdir);

    # For each type of evidence
    foreach my $nb (sort keys(%$rh_evidence_type))
    {
        my $prefix      = "evidence_" . $nb;
        my $orientation = $o_param->Get($prefix . "_type");
        my $format      = $o_param->Get($prefix . "_format");
        my $small       = $o_param->Get($prefix . "_small");

        &PreProcessing($prefix, $seqfile, $workingdir, $o_param, $o_logger, $logger_stamp);

        # for each file of evidence
        foreach my $evidence (@{$$rh_evidence_type{$nb}})
        {
            my $evidence_name = &GetEvidenceName($evidence);

            my $tomap_fastq;
            my $tomap_format;

            # Conversion SAM|BAM to FASTQ
            if ($format =~ /[sb]am/)
            {
                my $bam_file;
                if ($format eq "sam")
                {
                    $bam_file = "$workingdir/$rootname.$evidence_name.bam";
                    &Sam2Bam($evidence, $bam_file, $o_param, $o_logger, $logger_stamp);
                }
                else
                {
                    $bam_file = $evidence;
                }
                my $fastq_file_root = "$workingdir/$rootname.$evidence_name";
                ($tomap_fastq, $orientation) =
                  &Bam2Fastq($bam_file, $fastq_file_root, $o_param, $o_logger, $logger_stamp);
                $tomap_format = "fastq";
                &SetBamOrientation($prefix, $orientation, $o_param);
            }

            if ($format =~ /fast[aq]/)
            {
                $tomap_fastq  = $evidence;
                $tomap_format = $format;
            }

            # MAPPING
            my $mapping_bed_file;
            if ($format =~ /fast[aq]/ || $format =~ /[sb]am/)
            {
                $mapping_bed_file = "$workingdir/$rootname.$evidence_name.bed";
                &Mapping($seqfile, $tomap_fastq, $tomap_format, $orientation, $mapping_bed_file,
                         $small,   $o_param,     $o_logger,     $logger_stamp);
                $mapping_bed_file .= ".gz" if (-e "$mapping_bed_file.gz");
            }
            elsif ($format eq "bed" || $format eq "bedpe")
            {
                $mapping_bed_file = $evidence;
            }

            # FILTERING OF MAPPING RESULT
            if ($format =~ /fast[aq]/ || $format eq "bed" || $format eq "bedpe" || $format =~ /[sb]am/)
            {
                my $filter_bed_file = "$workingdir/$rootname.$evidence_name.$MAPPING_EXT";
                &MappingFiltering($mapping_bed_file, $orientation, $small, $filter_bed_file, $prefix, $o_param,
                                  $o_logger, $logger_stamp);

                # Split the bed result by sequence
                &SplitBed($filter_bed_file, $evidence,  $o_param, $o_logger, $logger_stamp);
            }

            # Tiling array
            if ($format eq "pair")
            {
                &SplitPair2Expr($evidence, $seqfile, $workingdir, $prefix,
                                $rootname, $o_param, $o_logger, $logger_stamp);
            }

            # wig file
            if ($format eq "wig")
            {
                &SplitWig2Expr($o_param, $evidence, $o_param->Get($prefix . "_strand"),
                               $workingdir, $rootname, $o_logger, $logger_stamp);
            }
        }
    }
	chdir($prevdir);

    return;
}

=head2 procedure PreProcessing

 Title        : PreProcessing
 Procedure    : Do primary computing needed for all files of the evidence type
                If format = pair, save the coordinates of the sequences included in the ndf file
 Prerequisite : None
 Args         : $prefix     Root of the key of the current evidence type. Ex evidence_2 
                $seqfile    Multifasta file of the genomic sequènces
                $workingdir Temp directory
                $o_param
                $o_logger
                $logger_stamp
 Globals      : none

=cut

sub PreProcessing
{
    my ($prefix, $seqfile, $workingdir, $o_param, $o_logger, $logger_stamp) = @_;

    if ($o_param->Get($prefix . "_format") eq "pair")
    {
        &LoadNdfSequences($prefix, $seqfile, $workingdir, $o_param, $o_logger, $logger_stamp);
    }

    return;
}

=head2 procedure SetBamOrientation

 Title        : SetBamOrientation
 Procedure     :Check that all the sam/bam files have the same orientation. 
                Save the orientation in the o_param object
 Prerequisite : 
 Args         : $prefix
                $orientation
                $o_param
 Globals      : none

=cut

sub SetBamOrientation
{
    my ($prefix, $orientation, $o_param) = @_;

    if (   $o_param->IsDefined($prefix . "_type")
        && $o_param->Get($prefix . "_type") ne $orientation)
    {
        Carp::croak(  "Problem with the bam data $prefix. All the files "
                    . "have to be paired-end or single end, not both.\n "
                    . "You should create two evidence families to fix that.\n");
    }
    $o_param->Set($prefix . "_type", $orientation);

    return;
}

=head2 procedure Sam2Bam

 Title        : Sam2Bam
 Procedure     :Convert sam file to bam
 Prerequisite : prg_samtools has to be defined. gzip has to be in the path
 Args         : $sam_file Sam file
                $bam_file result bam file
                $o_param
                $o_logger
                $logger_stamp
 Globals      : none

=cut

sub Sam2Bam
{
    my ($sam_file, $bam_file, $o_param, $o_logger, $logger_stamp) = @_;

    if ($DEBUG && (-e $bam_file))
    {
        print STDERR "DEBUG MODE: $bam_file already computed.\n";
        return;
    }
    my $in_exe_path = &GetCompressPath($sam_file);    # "/bin/gzip" for example
    if ($in_exe_path ne "")
    {
        $in_exe_path .= " -cd ";                      # Ex "/bin/gzip -cd "
    }
    else
    {
        $in_exe_path = "cat ";
    }

    # -S entry=sam; -b to convert in bam
    my $cmd = "$in_exe_path $sam_file | " . $o_param->Get('prg_samtools') . " view -b -S - > $bam_file ";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    return;

}

=head2 function Bam2Fastq

 Title        : Bam2Fastq
 Procedure     :Convert bam file to fastq and return the fatsq result files(s) 
                and the type (ope or ose)
 Prerequisite : prg_bam2fastq has to be defined. gzip is in the path
 Args         : $bam_file        Input bam file
                $fastq_file_root Root of the result fastq file(s)
                $o_param
                $o_logger
                $logger_stamp
 Globals      : none

=cut

sub Bam2Fastq
{
    my ($bam_file, $fastq_file_root, $o_param, $o_logger, $logger_stamp) = @_;

    my $fastqfile;      # name of the result fastq file(s)
    my $orientation;    # ope or ose

    if ($DEBUG)
    {
        if ((-e "$fastq_file_root\_1.fastq.gz") && (-e "$fastq_file_root\_2.fastq.gz"))
        {
            print STDERR
              "DEBUG MODE: $fastq_file_root\_1.fastq.gz and $fastq_file_root\_2.fastq.gz already computed.\n";
            $fastqfile   = "$fastq_file_root\_1.fastq.gz,$fastq_file_root\_2.fastq.gz";
            $orientation = "ope";
            return ($fastqfile, $orientation);
        }
        elsif (-e "$fastq_file_root.fastq.gz")
        {
            print STDERR "DEBUG MODE: $fastq_file_root.fastq.gz already computed\n";
            $fastqfile   = "$fastq_file_root.fastq.gz";
            $orientation = "ose";
            return ($fastqfile, $orientation);
        }
    }

    my $cmd = $o_param->Get('prg_bam2fastq') . " --output $fastq_file_root#.fastq $bam_file > /dev/null 2>&1";

    &ExeCmd($cmd, $o_logger, $logger_stamp);

    if (-e "$fastq_file_root\_1.fastq" && "$fastq_file_root\_2.fastq")
    {
        $orientation = "ope";
        $cmd         = "gzip -f $fastq_file_root\_1.fastq  $fastq_file_root\_2.fastq";
        $fastqfile   = "$fastq_file_root\_1.fastq.gz,$fastq_file_root\_2.fastq.gz";
    }
    else
    {
        $orientation = "ose";
        $cmd         = "gzip -f $fastq_file_root.fastq";
        $fastqfile   = "$fastq_file_root.fastq.gz";
    }

    # Zip files
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    #print "Return $fastqfile, $orientation\n";
    return ($fastqfile, $orientation);
}

=head2 procedure SplitWig2Expr

 Title        : SplitWig2Expr
 Procedure    : Read a wig file. Convert a wig file in GFF3 files
                One containing expression depth with filtering an other without (used to search tstart)
                Split thes files per sequence and save them in workingdir
 Prerequisite : prg_wig_to_expr , prg_wig_to_expr_annot_param
                prg_bed_to_expr_annot_scoremax  prg_bed_to_expr_annot_scalingfactor
                prg_bed_to_expr_annot_min_coverage_percent has to be defined
 Args         : $o_param   
                $wigfile
                $strand
                $workingdir
                $rootname
 Globals      : none

=cut

sub SplitWig2Expr
{
    my ($o_param, $wigfile, $strand, $workingdir, $rootname, $o_logger, $logger_stamp) = @_;

    my $ra_1 = &Wig2Expr4EGN($o_param, $wigfile, $strand, $workingdir, $rootname, $o_logger, $logger_stamp);
    my $ra_2 = &Wig2Expr4tStart($o_param, $wigfile, $strand, $workingdir, $rootname, $o_logger, $logger_stamp);

    my @a_created_files = (@$ra_1, @$ra_2);

    return \@a_created_files;
}

=head2 procedure Wig2Expr4EGN

 Title        : Wig2Expr4EGN
 Procedure    : Convert a wig file in expression files (.region and .position gff3 files). 
                Split by seq and move it in the appropriated sequence directory
 Prerequisite : prg_wig_to_expr , prg_wig_to_expr_annot_param
                prg_bed_to_expr_annot_scoremax  prg_bed_to_expr_annot_scalingfactor
                prg_bed_to_expr_annot_min_coverage_percent has to be defined
 Args         : $o_param   
                $wigfile
                $strand
                $workingdir
                $rootname
 Globals      : none

=cut

sub Wig2Expr4EGN
{
    my ($o_param, $wigfile, $strand, $workingdir, $rootname, $o_logger, $logger_stamp) = @_;

    my $evidence_name  = &GetEvidenceName($wigfile);
    my $tmpoutfilename = "$rootname.$evidence_name.$WIG_EXT";
    my $tmpoutfile     = "$workingdir/$tmpoutfilename";

    my $cde;

    $cde = $o_param->Get('prg_wig_to_expr') . " " . $o_param->Get('prg_wig_to_expr_annot_param');
    $cde .= " --wigfile $wigfile --outprefix $tmpoutfile --strand $strand";
    $cde .= " --tag WIG --split ";
    $cde .= " --scoremax " . $o_param->Get('prg_bed_to_expr_annot_scoremax');
    $cde .= " --scalingfactor " . $o_param->Get('prg_bed_to_expr_annot_scalingfactor');
    $cde .= " --min_coverage_percent " . $o_param->Get('prg_bed_to_expr_annot_min_coverage_percent');

    &ExeCmd($cde, $o_logger, $logger_stamp);

    # For each sequence, copy region.gff3 and position.gff3 file
    my $ra_created_files;

    my @a_ = glob("$workingdir/*.$tmpoutfilename.$EXPR_POS_EXT.gff3");
    foreach my $file (@a_)
    {
        my ($seqname) = $file =~ /$workingdir\/(.+)\.$tmpoutfilename/;
        my $seqdir = &GetSeqDir($seqname, $o_param);
        my $reg_file = "$workingdir/$seqname.$tmpoutfilename.$EXPR_REG_EXT.gff3";

        # If the sequence directory exists, copy in
        if (-d $seqdir)
        {
            # Copy position file
            $cde = "mv $file $seqdir/$seqname.$evidence_name.$EXPR_POS_EXT.gff3";
            &ExeCmd($cde, $o_logger, $logger_stamp);
            push(@$ra_created_files, "$seqdir/$seqname.$evidence_name.$EXPR_POS_EXT.gff3");

            # Copy region file
            $cde = "mv $reg_file $seqdir/$seqname.$evidence_name.$EXPR_REG_EXT.gff3";
            &ExeCmd($cde, $o_logger, $logger_stamp);
            push(@$ra_created_files, "$seqdir/$seqname.$evidence_name.$EXPR_REG_EXT.gff3");

        }
        else
        {
            &PrintInfo(  "Warning: $wigfile contains data from $seqname but this sequence "
                       . "is not in the input/genome directory. Data ignored\n");
            &ExeCmd("gzip -f $file $reg_file", $o_logger, $logger_stamp);
        }
    }

    return $ra_created_files;
}

=head2 procedure Wig2Expr4tStart

 Title        : Wig2Expr4tStart
 Procedure     :Convert wig file in expression file for tstart detection. 
                Split by seq and move it in the working sequence directory
 Args         : $o_param   
                $wigfile
                $strand
                $workingdir
                $rootname
 Globals      : none

=cut

sub Wig2Expr4tStart
{
    my ($o_param, $wigfile, $strand, $workingdir, $rootname, $o_logger, $logger_stamp) = @_;

    my @a_created_files;

    my $evidence_name = &GetEvidenceName($wigfile);
    my $outfilename   = "$rootname.$evidence_name.$TSTART_POS_EXT";
    my $outfile       = "$workingdir/$outfilename";

    my $cde;
    $cde = $o_param->Get('prg_wig_to_expr') . " " . $o_param->Get('prg_wig_to_expr_annot_param');
    $cde .= " --wigfile $wigfile --outprefix $outfile --strand $strand";
    $cde .= " --tag WIG --split ";
    $cde .= " --scoremax " . $o_param->Get('prg_bed_to_expr_tstart_scoremax');
    $cde .= " --scalingfactor " . $o_param->Get('prg_bed_to_expr_tstart_scalingfactor');
    $cde .= " --min_coverage " . $o_param->Get('prg_bed_to_expr_tstart_min_coverage');

    &ExeCmd($cde, $o_logger, $logger_stamp);

    my @a_ = glob("$workingdir/*.$outfilename.position.gff3");
    foreach my $file (@a_)
    {
        my ($seqname) = $file =~ /$workingdir\/(.+)\.$outfilename.position.gff3/;
        my $reg_file = "$workingdir/$seqname.$outfilename.region.gff3";
        my $seqdir = &GetWorkSeqDir($seqname, $o_param);
        if (-d $seqdir)
        {
            $cde = "mv $file $seqdir/$seqname.$evidence_name.$TSTART_POS_EXT.gff3";
            &ExeCmd($cde, $o_logger, $logger_stamp);
            push(@a_created_files, "$seqdir/$seqname.$evidence_name.$TSTART_POS_EXT.gff3");
            unlink($reg_file);
        }
        else
        {
            &PrintInfo(  "Warning: $wigfile contains data from $seqname but this "
                       . "sequence is not in the input/genome directory. Data ignored\n");
            &ExeCmd("gzip $file $reg_file", $o_logger, $logger_stamp);
        }
    }

    return \@a_created_files;
}

=head2 procedure SplitBed

 Title        : SplitBed
 Procedure     :Split bed file per sequence and move files in sequence specific directories
 Args         : $bed_file Bed file  
                $evidence 
 Globals      : none

=cut

sub SplitBed
{
    my ($bed_file, $evidence, $o_param, $o_logger, $logger_stamp) = @_;

    my ($workingdir, $bedname) = $bed_file =~ /(.+)\/(.+)/;
    my $prg_split_bed = $o_param->Get('prg_split_bed');

    # split the results and move it in the sequence directory
	my $prevdir = cwd();
    chdir($workingdir);

    my $cmd = "$prg_split_bed --in $bed_file";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    my @a_ = glob("$workingdir/*.$bedname");
    foreach my $file (@a_)
    {
        my ($seqname) = $file =~ /$workingdir\/(.+)\.$bedname/;
		my $seqd = &GetWorkSeqDir($seqname, $o_param);
		my $cmd  = "mv $file $seqd/" . &GetMappingFileName($seqname, $evidence);
        #my $cmd = "mv $file $seqdir/$seqname/work.$uniq/" . &GetMappingFileName($seqname, $evidence);
        &ExeCmd($cmd, $o_logger, $logger_stamp);
    }
    chdir($prevdir);

    return;
}

=head2 procedure GetEvidences

 Title        : GetEvidences
 Usage        : 
 Prerequisite : $prefix . "_pat" $prefix . "_type"  $prefix . "_format" $prefix . "_small" exists
 Function     : Manage one type of RNAseq data
 Args         : $o_param 
                $seqfile file     Input fasta sequence file
                $seqid  string    Name of the sequence
                $seqlen integer   Length of the sequence  
                $seqname string   Name of the sequence
                $workdir dir path Output directory
                $prefix           Prefix of the type of evidence ex: "evidence_2"
                $annotastruct_cpt Use to count the number of rnaseq data computed. 
                                  This value is update by this procedure.
                $annotastruct_cpt,
                $rh_template
                $cfg_content 
 Globals      : none

=cut

sub GetEvidences
{
    my ($o_param, $seqfile, $seqid, $seqlen, $seqname, $workdir, $prefix, $ra_evidences_files,
        $r_annotastruct_cpt, $rh_template, $r_cfg_content, $o_logger, $logger_stamp, $uniq)
      = @_;

    my $orientation = $o_param->Get($prefix . "_type");
    my $format      = $o_param->Get($prefix . "_format");

    &PrintInfo("Load $prefix RNAseq data\n");

    my $data_type = "";
    $data_type = "MAPPED" if (   $format =~ /fast[aq]/
                              || $format eq "bed"
                              || $format eq "bedpe"
                              || $format =~ /[sb]am/);

    foreach my $evidence (@$ra_evidences_files)
    {
        &PrintInfo("File $evidence..\n");

        # Name of this evidence
        my $evidence_name      = &GetEvidenceName($evidence);
        my $expr_position_file = "$seqfile.$evidence_name.$EXPR_POS_EXT.gff3";
        my $expr_region_file   = "$seqfile.$evidence_name.$EXPR_REG_EXT.gff3";
        my $tstart_file_gff3   = "$seqfile.$evidence_name.$TSTART_EXT.gff3";

        if (!$DEBUG
            || ($DEBUG && !(-e $expr_position_file && -e $expr_region_file && -e $tstart_file_gff3)))
        {
            my $expr_4tstart_file = "$workdir/$seqname.$evidence_name.$TSTART_POS_EXT.gff3";
            if ($data_type eq "MAPPED")
            {
                # 0 Get the already compute mapping file
                my $filtering_bed_file = "$workdir/" . &GetMappingFileName($seqname, $evidence);
                if (-e $filtering_bed_file)
                {
                    # 1 Build the expression files to integrate in EuGene
                    &Bed2Expr4EGN($filtering_bed_file, $orientation, $expr_position_file,
                                  $expr_region_file, $o_param, $o_logger, $logger_stamp, $uniq);

                    # 2 Build the expression files to detect start of transcription
                    &Bed2Expr4tStart($filtering_bed_file, $orientation, $expr_4tstart_file,
                                     $o_param, $o_logger, $logger_stamp, $uniq);
                }
                else
                {
                    &PrintInfo("No mapping result for $seqfile $evidence_name\n");
                }
            }
            if (-e $expr_4tstart_file)
            {
                # 4 Detect start of transcription
                &TStartDetection(
                                 $expr_4tstart_file, $tstart_file_gff3, $seqid,         $seqlen,
                                 $seqname,           $workdir,          $evidence_name, $o_param,
                                 $o_logger,          $logger_stamp
                                 );
                &Gzip($o_logger, $logger_stamp, $expr_4tstart_file);
            }
            else
            {
                &PrintInfo("No expression result for $seqfile $evidence_name\n");
            }
        }
        else
        {
            print STDERR "DEBUG MODE: use region, position and tstart files for $evidence_name.\n";
        }

        &CompleteCfg($r_cfg_content, $r_annotastruct_cpt, "$evidence_name.$EXPR_POS_EXT",
                     $$rh_template{position}, $expr_position_file)
          if (-e $expr_position_file);
        &CompleteCfg($r_cfg_content, $r_annotastruct_cpt, "$evidence_name.$EXPR_REG_EXT",
                     $$rh_template{region}, $expr_region_file)
          if (-e $expr_region_file);
        &CompleteCfg($r_cfg_content, $r_annotastruct_cpt, "$evidence_name.$TSTART_EXT",
                     $$rh_template{tstart}, $tstart_file_gff3)
          if (-e $tstart_file_gff3);

    }

    return;
}

=head2 procedure CompleteCfg

 Title        : CompleteCfg
 Usage        : 
 Prerequisite : 
 Function     : Complete $cfg_content string. If $evidence_file exists, 
                add an instance of a plugin. substitute CPT by $cpt 
                and EXT by $ext in the template and concatenate this string to cfg_content
 Args         : $r_cfg_content   String ref  Parameter file part
                $r_cpt           Ref to int  Cpt to create new instance of the plugin
                $ext             string      Extension of the instance
                $template        string      Template of a plugin instance
 Globals      : none

=cut

sub CompleteCfg
{
    my ($r_cfg_content, $r_cpt, $ext, $template, $evidence_file) = @_;

    if (!defined($evidence_file) || $evidence_file eq "" || ($evidence_file ne "" && -e $evidence_file))
    {
        $$r_cpt++;
        my $content = $template;
        $content =~ s/CPT/$$r_cpt/g;
        $content =~ s/EXT/$ext/g;
        $content .= "##\n";

        $$r_cfg_content .= $content;
    }

    return;
}

=head2 procedure TStartDetection

 Title        : TStartDetection
 Usage        : 
 Prerequisite : prg_tstart_detection and prg_tstart_detection_param prg_tstart_detection_derivative_threshold
                prg_tstart_detection_log_threshold prg_tstart_detection_boostscore_factor
                prg_tstart_detection_boost_ig, prg_tstart_detection_tstart_ratio  are defined in $o_param
 Procedure     : 
 Args         : $infile input fasta file
                $seqid  name of the sequence
                $seqlen Len of the sequence
                $outfile 
                $format input format gff3 or wig
                $o_param 
 Globals      : none

=cut

sub TStartDetection
{
    my ($infile, $outfile, $seqid, $seqlen, $seqname, $workdir, $evidence_name, $o_param, $o_logger, $logger_stamp) =
      @_;

    # Test to catch case where no
    if (-e $infile)
    {
        my $tmpfile = "$workdir/$seqname.$evidence_name.$TSTART_EXT.tmp";

        my $cmd;
        $cmd = $o_param->Get('prg_tstart_detection') . " " . $o_param->Get('prg_tstart_detection_param');
        $cmd .= " --derivative_threshold " . $o_param->Get('prg_tstart_detection_derivative_threshold');
        $cmd .= " --log_threshold " . $o_param->Get('prg_tstart_detection_log_threshold');
        $cmd .= " --boostscore_factor " . $o_param->Get('prg_tstart_detection_boostscore_factor');
        $cmd .= " --tstart_ratio " . $o_param->Get('prg_tstart_detection_tstart_ratio');
        $cmd .= " --boost_ig " if ($o_param->Get('prg_tstart_detection_boost_ig') eq "TRUE");
        $cmd .= " --in $infile --out $tmpfile --seq_len $seqlen --seq_name \"$seqid\" 2> /dev/null";

        &ExeCmd($cmd, $o_logger, $logger_stamp);

        $cmd = "mv $tmpfile.gff3 $outfile";
        &ExeCmd($cmd, $o_logger, $logger_stamp);
    }
    else
    {
        # No expression file for this sequence, for this evidence
        # Possible if evidence is a wig file which contain no data about the sequence
        &PrintInfo("INFO TStartDetection: $seqid $evidence_name No data found in $evidence_name for $seqid.\n");
    }

    return;
}

=head2 procedure GettRNAscanSE

 Title        : GettRNAscanSE
 Usage        : 
 Prerequisite : prg_get_trnascanse has to be defined
 Function     : Launch trnascanse. Result file is seq.trnascanse.gff3
 Args         : $o_param  ParamParser object
                $seq      fasta file path
 Globals      : none

=cut

sub GettRNAscanSE
{
    my ($o_param, $seq) = @_;

    my ($path, $name) = &GetPathAndName($seq);
    my $step          = "$name tRNAscanSE";
    my $trnascan_file = "$seq.$TRNASCAN_EXT.gff3";

    my $cmd = &GettRNAscanSECmd($o_param, $seq, 0);
    return if ($cmd eq "");

    &PrintStart($step);
    &ExeCmd($cmd, $o_logger, $logger_stamp);
    &ValidateStep($trnascan_file);
    &PrintStop($step);

    return;
}

=head2 procedure GettRNAscanSEAllSeq

 Title        : GettRNAscanSEAllSeq
 Usage        : 
 Prerequisite :  has to exist
 Function     : Launch trnascanSE for each sequence
 Args         : $o_param  ParamParser object
                $ra_seq   Ref to an array of sequence paths
 Globals      : none

=cut	

sub GettRNAscanSEAllSeq
{
    my ($o_param, $ra_seq) = @_;

    my $cmd;

    my $rootworkdir = $o_param->Get('workingdir');
    my $trnadir     = $rootworkdir . "/TRNASCANSE_" . $o_param->Get('uniq');
    mkdir($trnadir);
    my $cmd_paraloop   = "$trnadir/trnascanse.cmd.paraloop";
    my $f_cmd_paraloop = &General::GetStreamOut($cmd_paraloop);

    foreach my $seq (@$ra_seq)
    {
		my $workingdir = &GetWorkDir($seq, $o_param);
        $o_param->Set('workingdir', $workingdir);

        $cmd = &GettRNAscanSECmd($o_param, $seq, 1);
        next if ($cmd eq "");
        print $f_cmd_paraloop "$cmd\n";
    }
    $f_cmd_paraloop->close;
    $o_param->Set('workingdir', $rootworkdir);

    if (-s $cmd_paraloop)
    {
        my $msg = "tRNA search";
        &PrintStart($msg);
        &Paraloop($o_param, $cmd_paraloop, $trnadir);
        &PrintStop($msg);
    }

    return;
}

sub GettRNAscanSECmd
{
    my ($o_param, $seq, $validate) = @_;

    my ($path, $name) = &GetPathAndName($seq);
    my $step          = "$name tRNAscanSE";
    my $trnascan_file = "$seq.$TRNASCAN_EXT.gff3";

    return ""
      if (   $o_param->IsDefined('no_ncrna_detection')
          || $o_param->IsDefined('no_trnascan')
          || &IsValidatedStep($trnascan_file, $step));

    my $cmd =
        $o_param->Get('prg_get_trnascanse')
      . " --sequence $seq --cfg "
      . $o_param->Get('cfg')
      . " --outfile $trnascan_file --workingdir "
      . $o_param->Get('workingdir');
    $cmd .= " --kingdom eukaryote "  if ($MODE eq "EUK");
    $cmd .= " --kingdom prokaryote " if ($MODE eq "PROK");

	# split by strand
	if ($INDEPENDENT_STRAND)
	{
		$cmd .= " && ".&GetSplitByStrandCmd($o_param, $trnascan_file);
	}
	
	
    if ($validate == 1)
    {
        my $validcmd = &GetValidateFileCmd($o_param, $trnascan_file);
        $cmd .= " && $validcmd";
    }

    return $cmd;
}

=head2 procedure GetRfamScan

 Title        : GetRfamScan
 Usage        : 
 Prerequisite : prg_get_rfamscan has to exist
 Function     : Launch rfamscan. Result file is seq.rfamscan.gff3
 Args         : $o_param  ParamParser object
                $seq      fasta file path
 Globals      : none

=cut

sub GetRfamScan
{
    my ($o_param, $seq) = @_;

    my ($path, $name) = &GetPathAndName($seq);
    my $step = "$name Rfam";

    my $cmd = &GetRfamScanCmd($o_param, $seq, 0);
    return if ($cmd eq "");

    &PrintStart($step);
    &ExeCmd($cmd, $o_logger, $logger_stamp);
    &PrintStop($step);
    &ValidateStep("$seq.$RFAMSCAN_EXT.gff3");

    return;
}

=head2 procedure GetRfamScanAllSeq

 Title        : GetRfamScanAllSeq
 Usage        : 
 Prerequisite : prg_get_rfamscan has to exist
 Function     : Launch rfamscan for each sequence
 Args         : $o_param  ParamParser object
                $ra_seq   Ref to an array of sequence paths
 Globals      : none

=cut	

sub GetRfamScanAllSeq
{
    my ($o_param, $ra_seq) = @_;

    my $cmd;

    my $rootworkdir = $o_param->Get('workingdir');
    my $rfamdir     = $rootworkdir . "/RFAM_" . $o_param->Get('uniq');
    mkdir($rfamdir);
    my $cmd_paraloop   = "$rfamdir/rfamscan.cmd.paraloop";
    my $f_cmd_paraloop = &General::GetStreamOut($cmd_paraloop);

    foreach my $seq (@$ra_seq)
    {
        my $workingdir = &GetWorkDir($seq, $o_param);
        $o_param->Set('workingdir', $workingdir);

        $cmd = &GetRfamScanCmd($o_param, $seq, 1);
        next if ($cmd eq "");
        print $f_cmd_paraloop "$cmd\n";
    }
    $f_cmd_paraloop->close;
    $o_param->Set('workingdir', $rootworkdir);

    if (-s $cmd_paraloop)
    {
        my $msg = "Rfam similarity search";
        &PrintStart($msg);
        &Paraloop($o_param, $cmd_paraloop, $rfamdir);
        &PrintStop($msg);
    }

    return;
}

=head2 procedure GetRfamScanCmd

 Title        : GetRfamScanCmd
 Usage        : 
 Prerequisite : prg_get_rfamscan has to exist
 Function     : Get rfamscan command line
 Args         : $o_param  ParamParser object
                $seq      fasta file path
                $validate If 1 create a success file at the end of the execution
 Return       : string rfam_scan command              
 Globals      : none

=cut

sub GetRfamScanCmd
{
    my ($o_param, $seq, $validate) = @_;

    my ($path, $name) = &GetPathAndName($seq);
    my $step      = "$name Rfam";
    my $rfam_file = "$seq.$RFAMSCAN_EXT.gff3";

    return ""
      if (   $o_param->IsDefined('no_ncrna_detection')
          || $o_param->IsDefined('no_rfamscan')
          || &IsValidatedStep($rfam_file, $step));

    my $cmd = "export EGNPP=" . $o_param->Get('EGNEP') . ";";
    $cmd .=
        $o_param->Get('prg_get_rfam')
      . " --fasta_file $seq "
      . " --output_file $rfam_file --working_dir "
      . $o_param->Get('workingdir')
      . " --cm_file "
      . $o_param->Get('rfam_cm_file');
    $cmd .= " --verbose " if ($DEBUG);
	
    # split by strand
	if ($INDEPENDENT_STRAND)
	{
		$cmd .= " && ".&GetSplitByStrandCmd($o_param, $rfam_file);
	}
	
    if ($validate == 1)
    {
        my $validcmd = &GetValidateFileCmd($o_param, $rfam_file);
        $cmd .= " && $validcmd";
    }

    return $cmd;
}

=head2 procedure GetRnammer

 Title        : GetRnammer
 Usage        : 
 Prerequisite : prg_get_Rnammer has to be defined
 Function     : Launch rnammer. Result file is seq.rnammer.gff3
 Args         : $o_param  ParamParser object
                $seq      fasta file path
 Globals      : none

=cut

sub GetRnammer
{
    my ($o_param, $seq) = @_;

    my ($path, $name) = &GetPathAndName($seq);
    my $step = "$name Rnammer";

    my $rnammer_file = "$seq.$RNAMMER_EXT.gff3";

    return if (&IsValidatedStep($rnammer_file, $step));

    &PrintStart($step);
    my $cmd = GetRnammerCmd($o_param, $seq, 0);
    return if ($cmd eq "");

    &ExeCmd($cmd, $o_logger, $logger_stamp);

    &ValidateStep($rnammer_file);
    &PrintStop($step);

    return;
}

sub GetRnammerCmd
{
    my ($o_param, $seq, $validate) = @_;

    my ($path, $name) = &GetPathAndName($seq);
    my $step         = "$name Rnammer";
    my $rnammer_file = "$seq.$RNAMMER_EXT.gff3";

    return ""
      if (   $o_param->IsDefined('no_ncrna_detection')
          || $o_param->IsDefined('no_rnammer')
          || &IsValidatedStep($rnammer_file, $step));

    my $cmd =
        $o_param->Get('prg_get_rnammer')
      . " --sequence $seq --cfg "
      . $o_param->Get('cfg')
      . " --outfile $rnammer_file --workingdir "
      . $o_param->Get('workingdir');
    $cmd .= " --kingdom eukaryote "  if ($MODE eq "EUK");
    $cmd .= " --kingdom prokaryote " if ($MODE eq "PROK");

	# split by strand
	if ($INDEPENDENT_STRAND)
	{
		$cmd .= " && ".&GetSplitByStrandCmd($o_param, $rnammer_file);
	}
	
    if ($validate == 1)
    {
        my $validcmd = &GetValidateFileCmd($o_param, $rnammer_file);
        $cmd .= " && $validcmd";
    }

    return $cmd;
}

=head2 procedure GetRnammerAllSeq

 Title        : GetRnammerAllSeq
 Usage        : 
 Prerequisite :  has to exist
 Function     : Launch rnammer for each sequence
 Args         : $o_param  ParamParser object
                $ra_seq   Ref to an array of sequence paths
 Globals      : none

=cut	

sub GetRnammerAllSeq
{
    my ($o_param, $ra_seq) = @_;

    my $cmd;

    my $rootworkdir = $o_param->Get('workingdir');
    my $rnammerdir  = $rootworkdir . "/RNAMMER_" . $o_param->Get('uniq');
    mkdir($rnammerdir);

    my $cmd_paraloop   = "$rnammerdir/rnammer.cmd.paraloop";
    my $f_cmd_paraloop = &General::GetStreamOut($cmd_paraloop);

    foreach my $seq (@$ra_seq)
    {
        my $workingdir = &GetWorkDir($seq, $o_param);
        $o_param->Set('workingdir', $workingdir);

        $cmd = &GetRnammerCmd($o_param, $seq, 1);
        next if ($cmd eq "");
        print $f_cmd_paraloop "$cmd\n";
    }
    $f_cmd_paraloop->close;
    $o_param->Set('workingdir', $rootworkdir);

    if (-s $cmd_paraloop)
    {
        my $msg = "Rnammer search";
        &PrintStart($msg);
        &Paraloop($o_param, $cmd_paraloop, $rnammerdir);
        &PrintStop($msg);
    }

    return;
}

=head2 procedure GetProdigal

 Title        : GetProdigal
 Usage        : 
 Prerequisite :
 Function     : Launch prodigal. Result file is seq.prodigal.gff3
 Args         : $o_param  ParamParser object
                $seq      fasta file path
                $seqlen    integer    Sequence length
                $r_annotastruct_cpt
                $rh_cfg_template
                $r_cfg_content
 Globals      : none

=cut

sub GetProdigal
{
    my ($o_param, $seq, $seqlen, $r_annotastruct_cpt, $rh_cfg_template, $r_cfg_content, $o_logger, $logger_stamp) = @_;

    return if ($o_param->IsDefined('no_prodigal'));

    my $cmd;
    my $prg_prodigal    = $o_param->Get('prg_prodigal');
    my $prodigal_file   = "$seq.$PRODIGAL_EXT.gff3";
    my $prodigal_minlen = $o_param->Get('prg_prodigal_minlen');

    if ($seqlen >= $prodigal_minlen)
    {
        $cmd = "$prg_prodigal -f gff -i $seq -o $seq.$PRODIGAL_EXT.gff > /dev/null 2>&1";
        &ExeCmd($cmd, $o_logger, $logger_stamp);

        # Create a reference training set for prodigal
        if (!$o_param->IsDefined('reference_prodigal_train'))
        {
            $cmd = $o_param->Get('prg_prodigal') . " -f gff -i $seq -t $seq.$PRODIGAL_EXT.train > /dev/null 2>&1";
            &ExeCmd($cmd, $o_logger, $logger_stamp);
            $o_param->Set('reference_prodigal_train', "$seq.$PRODIGAL_EXT.train");
        }
    }
    else
    {
        if ($o_param->IsDefined('reference_prodigal_train'))
        {
            my $trainset = $o_param->Get('reference_prodigal_train');
            &PrintInfo(
                "Sequence length smaller than prodigal minlen ($prodigal_minlen): use the prodigal training set $trainset\n"
                );

            $cmd = "$prg_prodigal -f gff -i $seq -o $seq.$PRODIGAL_EXT.gff -t $trainset > /dev/null 2>&1";
            &ExeCmd($cmd, $o_logger, $logger_stamp);
        }
        else
        {
            &PrintInfo(  "Sequence size smaller than prodigal minlen ($prodigal_minlen): "
                       . "no prodigal training set available so use -p meta option\n");

            $cmd = "$prg_prodigal -f gff -i $seq -o $seq.$PRODIGAL_EXT.gff -p meta > /dev/null 2>&1";
            &ExeCmd($cmd, $o_logger, $logger_stamp);
        }
    }

    # Create a eugene formated GFF3 file (Add the SO ontology to the CDS lines)
    my $f_in  = &General::GetStreamIn("$seq.$PRODIGAL_EXT.gff");
    my $f_out = &General::GetStreamOut($prodigal_file);
    while (my $l = <$f_in>)
    {
        if ($l !~ /^#/ && $l =~ /\tCDS\t/)
        {
            chomp($l);
            $l .= ";Ontology_term=SO:0005845\n";
        }
        print $f_out $l;
    }
    $f_out->close;
    $f_in->close;

    $o_logger->Trace("Create $prodigal_file\n", $logger_stamp, $LOGGER_LEVEL_EXEC);

    &CompleteCfg($r_cfg_content, $r_annotastruct_cpt, $PRODIGAL_EXT, $$rh_cfg_template{start}, $prodigal_file);

    return;
}

sub GetLTRHarvest
{
    my ($o_param, $seq, $workdir) = @_;

    my ($path, $name) = &GetPathAndName($seq);

    my $ltrharvest_file = "$seq.$LTRHARVEST_EXT.gff3";
    my $step            = "$name LTRHarvest";

    return $ltrharvest_file if (&IsValidatedStep($ltrharvest_file, $step));

    &PrintStart($step);

    my $cmd =
        $o_param->Get('prg_get_ltrharvest')
      . " --sequence $seq "
      . " --cfg "
      . $o_param->Get('cfg')
      . " --outfile $ltrharvest_file --workingdir $workdir";
    $cmd .= " --debug " if ($DEBUG);
	my $o_local_logger = ( $o_param->Get('ncpus') > 1 ) ? undef: $o_logger;
    &ExeCmd($cmd, $o_local_logger, $logger_stamp);

    &Gzip($o_local_logger, $logger_stamp, $ltrharvest_file . ".fa");

    &PrintStop($step);
    &ValidateStep($ltrharvest_file);
    return $ltrharvest_file;
}

=head2 procedure LoadNdfSequences

 Title        : LoadNdfSequences
 Procedure    : Extract from the ndf file the sequences. Map them to the genomic sequences
                Save the coordinate in data structure.
 Prerequisite : None
 Args         : $prefix     Root of the key of the current evidence type. Ex evidence_2 
                $seqfile    Multifasta file of the genomic sequènces
                $workingdir Temp directory
                $o_param
                $o_logger
                $logger_stamp
 Globals      : none

=cut

sub LoadNdfSequences
{
    my ($prefix, $seqfile, $workingdir, $o_param, $o_logger, $logger_stamp) = @_;

    my $ndf_file = $o_param->Get($prefix . "_ndf");
    my ($ndf_path, $ndf_name) = $ndf_file =~ /(.+)\/(.+)/;
    &PrintInfo("Extract data from $ndf_file\n");

    # 1) Extract from ndf file the probes and map them to the genomc loci
    my $fasta_file = $workingdir . "/" . $ndf_name . ".fna.gz";
    my $glintres   = $workingdir . "/" . $ndf_name . ".fna.glint";
    &Ndf2Fasta($ndf_file, $fasta_file);

    my $cmd;
    if (!$DEBUG || ($DEBUG && !-e $glintres . ".gz"))
    {
        my $cmd =
            $o_param->Get('prg_glint')
          . " map $seqfile $fasta_file --input-format fasta -o $glintres "
          . $o_param->Get('prg_glint_tiling_param');
        &ExeCmd($cmd, $o_logger, $logger_stamp);
    }
    $glintres .= ".gz";

    print STDERR "Load $ndf_file...\n" if ($DEBUG);

    #Parse results

    my %h_chr;
    my %h_start;
    my %h_end;
    my %h_strand;
    my %h_seqid;

    my %h_m8_res;
    my $f_map_res = &General::GetStreamIn($glintres);
    while (my $m8_line = <$f_map_res>)
    {
        chomp($m8_line);
        &GeneralBioinfo::ParseM8Line(\$m8_line, 1, \%h_m8_res);

        my $seqid = $h_m8_res{query};

        # If already view, remove the probe because not specific
        if (defined($h_seqid{$seqid}))
        {
            delete($h_chr{$seqid});
            delete($h_start{$seqid});
            delete($h_end{$seqid});
            delete($h_strand{$seqid});
            next;
        }
        $h_seqid{$seqid}  = $seqid;
        $h_chr{$seqid}    = $h_m8_res{target};
        $h_start{$seqid}  = $h_m8_res{tstart};
        $h_end{$seqid}    = $h_m8_res{tend};
        $h_strand{$seqid} = $h_m8_res{strand};
    }

    $o_param->Set("$prefix\_chrs",    \%h_chr);
    $o_param->Set("$prefix\_starts",  \%h_start);
    $o_param->Set("$prefix\_ends",    \%h_end);
    $o_param->Set("$prefix\_strands", \%h_strand);

    return;
}

=head2 procedure SplitPair2Expr

 Title        : SplitPair2Expr
 Procedure    : Convert pair file in wig, then convert wig to expr files and 
                split them per sequence. Save results in the sequence directories
 Prerequisite : 
 Args         : $pair_file     Pair file from nimblescan
                $seqfile       multifasta of all the genomic sequences
                $workingdir    Temporary directory where create the temp files
                $evidence_key  Exe: evidence_5. Used to get sup parameters
                $rootname      Root name of the temporary files
 Globals      : none

=cut

sub SplitPair2Expr
{
    my ($pair_file, $seqfile, $workingdir, $evidence_key, $rootname, $o_param, $o_logger, $logger_stamp) = @_;

    my $evidence_name = &GetEvidenceName($pair_file);
    my ($pair_path, $pair_name) = $pair_file =~ /(.+)\/(.+)/;

    if ($DEBUG)
    {
        my $globseqdir = $o_param->Get('rootworkingdir');
        my @a_pos      = glob("$globseqdir/*/*/*$evidence_name.$EXPR_POS_EXT.gff3");
        my @a_reg      = glob("$globseqdir/*/*/*$evidence_name.$EXPR_REG_EXT.gff3");
        my @a_tstart   = glob("$globseqdir/*/*/*$evidence_name.$TSTART_EXT.gff3");
        if (scalar(@a_pos) > 0 && scalar(@a_reg) > 0 && scalar(@a_tstart) > 0)
        {
            my $text = "DEBUG MODE: use already computed file for tiling pair evidence $evidence_name:\n";
            $text .= join(' ', @a_pos) . " " . join(' ', @a_reg) . " " . join(' ', @a_tstart) . "\n";
            &PrintInfo($text);

            return;
        }
    }

    my $wig_file_plus  = "$workingdir/$evidence_name\_FWD.$WIG_EXT.gz";
    my $wig_file_minus = "$workingdir/$evidence_name\_REV.$WIG_EXT.gz";
    &Pair2Wig($pair_file, $wig_file_plus, $wig_file_minus, $evidence_key, $o_param);

    # Convert wig file (forward strand) to expression files
    my $ra_created_files =
      &SplitWig2Expr($o_param, $wig_file_plus, "forward", $workingdir, $rootname, $o_logger, $logger_stamp);

    # Convert wig file (reverse strand) to expression files
    &SplitWig2Expr($o_param, $wig_file_minus, "reverse", $workingdir, $rootname, $o_logger, $logger_stamp);

    # Concatenate forward and reverse expression files
    foreach my $forward_f (@$ra_created_files)
    {
        my $reverse_f = $forward_f;
        my $cat_f     = $forward_f;
        $reverse_f =~ s/_FWD\.([^\.]+)\.gff3/_REV\.$1\.gff3/;
        $cat_f =~ s/_FWD\.([^\.]+)\.gff3/\.$1\.gff3/;

        my $cmd = "cat $forward_f $reverse_f > $cat_f ";
        &ExeCmd($cmd, $o_logger, $logger_stamp);
        unlink($forward_f, $reverse_f);
    }

    return;
}

=head2 procedure Pair2Wig

 Title        : Pair2Wig
 Procedure    : Convert pair file in wig. create two wig files, one per strand
 Prerequisite : 
 Args         : $pair_file      Pair file from nimblescan
                $wig_file_plus  wig file for strand +
                $wig_file_minus wig file for strand -
                $prefix
 Globals      : none

=cut

sub Pair2Wig
{
    my ($pair_file, $wig_file_plus, $wig_file_minus, $prefix, $o_param) = @_;

    if ($DEBUG && (-e $wig_file_plus) && (-e $wig_file_minus))
    {
        print STDERR "DEBUG MODE: $wig_file_plus and $wig_file_minus already computed.\n";
        return;
    }

    my $rh_chr    = $o_param->Get("$prefix\_chrs");
    my $rh_start  = $o_param->Get("$prefix\_starts");
    my $rh_end    = $o_param->Get("$prefix\_ends");
    my $rh_strand = $o_param->Get("$prefix\_strands");

    my %h_probe_intensity;    # key   = "chr"_"strand",
                              # value = array of intensity per position
    my %h_probe_nb;           # key   = "chr"_"strand",
                              # value = array containing the number of probe matchs per position

    my $key;

    # Parse  pair files
    my $f_pair = &General::GetStreamIn($pair_file);
    while (my $line = <$f_pair>)
    {
        chomp($line);
        my @a_line = split(/\t/, $line);
        my $probid = $a_line[3];
        my $signal = $a_line[9];

        if (defined($$rh_chr{$probid}))
        {
            $key = $$rh_chr{$probid} . "_" . $$rh_strand{$probid};

            $h_probe_intensity{$key} = [] if (!defined $h_probe_intensity{$key});
            $h_probe_nb{$key}        = [] if (!defined $h_probe_nb{$key});
            my $ra_intensity = $h_probe_intensity{$key};
            my $ra_nb        = $h_probe_nb{$key};
            my $starti       = $$rh_start{$probid} - 1;
            my $endi         = $$rh_end{$probid} - 1;

            for (my $i = $starti; $i <= $endi; $i++)
            {
                $$ra_intensity[$i] += $signal;
                $$ra_nb[$i] += 1;
            }
        }
    }

    print STDERR "end of read pair file\n" if ($DEBUG);

    # Save in wig files
    # create one wig file per strand
    my $f_wig_plus  = &General::GetStreamOut($wig_file_plus);
    my $f_wig_minus = &General::GetStreamOut($wig_file_minus);

    my $f_current;
    foreach my $seq (keys %h_probe_intensity)
    {
        my ($chr, $strand) = $seq =~ /(.+)_([-\+])$/;
        $f_current = ($strand eq "+") ? $f_wig_plus : $f_wig_minus;

        print $f_current "track type=wiggle_0 name=\"tiling $chr $strand\"\n";
        print $f_current "fixedStep chrom=$chr start=1 step=1 span=1\n";

        my $ra_intensity = $h_probe_intensity{$seq};
        my $ra_nb        = $h_probe_nb{$seq};
        my $l            = scalar(@$ra_intensity);

        #print STDERR "length array ! $l\n";
        for (my $i = 0; $i < $l; $i++)
        {
            if (defined($$ra_intensity[$i]))
            {
                my $mean_intensity = $$ra_intensity[$i] / $$ra_nb[$i];
                printf $f_current "%.2f\n", $mean_intensity;
            }
            else
            {
                print $f_current "0\n";
            }
        }
    }
    $f_wig_plus->close;
    $f_wig_minus->close;

    return;
}

=head2 procedure Ndf2Fasta

 Title        : Ndf2Fasta
 Procedure    : Convert ndf to multifasta
 Prerequisite : 
 Args         : $ndf    Ndf file generated by nimblescan
                $fasta  Result fasta file
 Globals      : none

=cut

sub Ndf2Fasta
{
    my ($ndf, $fasta) = @_;

    my $f_      = &General::GetStreamIn($ndf);
    my $f_fasta = &General::GetStreamOut($fasta);
    my @a_line;
    while (my $line = <$f_>)
    {
        chomp($line);
        next if ($line =~ /^PROBE_DESIGN_ID/);
        @a_line = split(/\t/, $line);
        if ($a_line[1] eq "BLOCK1")
        {
            print $f_fasta ">$a_line[12]\n$a_line[5]\n";
        }
    }

    $f_->close;
    $f_fasta->close;

    return;
}

=head2 procedure Consolidate
	
 Title        : Consolidate
 Usage        : 
 Prerequisite : prg_biofileconverter and prg_genome_statistics in o_param correspond to existing programs.
 Function     : Compute protein sequences file, and statistics about structural annotation  
 Args         : $o_param 
                $outdir      Output directory
                $ra_res_gff3 Ref to an array of gff3 files
                $allseq_file Mutifastafile of genomic sequences in working directory
 Globals      : none

=cut

sub Consolidate
{
    my ($o_param, $outdir, $ra_res_gff3, $allseq_file, $o_logger, $logger_stamp) = @_;

    my $cmd = "";

    my ($allseq_dir) = $allseq_file =~ /(.+)\/[^\/]+$/; 
	$allseq_dir =~ s/\/$//;
	my $allseq_name  = $o_param->Get('output_prefix');
	
    &PrintStart("Annotation consolidation");

    # copy the multifasta file in the result directory if workdir and outdir are different
    if ($allseq_dir ne $outdir)
    {
		$cmd .= "rm -f $outdir/$allseq_name;" if (-f "$outdir/$allseq_name");
        $cmd .= "cp -f $allseq_file $outdir/$allseq_name";
        &ExeCmd($cmd, $o_logger, $logger_stamp);

        # zip the multifasta file in the working dir
        #&Gzip($o_logger, $logger_stamp, $allseq_file);
    }

    # Cat all the gff3 results in a uniq file in the 'full_eugene' directory
    my $annotdir = "$outdir/full_eugene";
    &ExeCmd("mkdir $annotdir", $o_logger, $logger_stamp) if (!-d "$annotdir");
    &ExeCmd("rm -f $annotdir/$allseq_name.gff3", $o_logger, $logger_stamp);
    foreach my $file (@$ra_res_gff3)
    {
        &ExeCmd("cat $file >> $annotdir/$allseq_name.gff3", $o_logger, $logger_stamp);
    }
    my $genenb = &CountNbFeatureGFF3("$annotdir/$allseq_name.gff3", "gene");
    my $lengenenb = length($genenb) + 1;

    # Rename genes --> create $outdir/sequences.gff3
    $cmd = $o_param->Get('prg_consolidate_gff3');
    $cmd .= " --dirin $annotdir --dirout $outdir --len_gene_number $lengenenb";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    # Compute protein fasta file
    $cmd = $o_param->Get('prg_biofileconverter');
    $cmd .= " --kingdom bacteria " if ($MODE eq "PROK");
    $cmd .=
        " --input-ann-file $outdir/$allseq_name.gff3 --input-seq-file $outdir/$allseq_name --all-contigs "
      . "--output-ann-file $outdir/$allseq_name.fna --input-ann-format GFF3  "
      . "--output-ann-format FASTA  > /dev/null 2>&1";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    # Compute basic statistics
    $cmd = $o_param->Get('prg_genome_statistics');
    $cmd .= " --kingdom prokaryote " if ($MODE eq "PROK");
    $cmd .=
      " --in  $outdir/$allseq_name.gff3 --out  $outdir/$allseq_name --scaffolds  $outdir/$allseq_name  > /dev/null 2>&1";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    &Gzip($o_logger, $logger_stamp, "$annotdir/$allseq_name.gff3");

    &PrintStop("Annotation consolidation");

    return;
}

=head2 procedure CreateWorkSeqTree

 Title        : CreateWorkSeqTree
 Usage        : 
 Prerequisite : keys 'prg_fasta2tree' in o_param is an existing program
 Procedure    : Search fasta files in $indir, Create a ordered tree of files
                in the working directory. (from the longer sequences to the smaller)
                Compute global GC percent
 Args         : $o_param     Param object
                $indir       Input directory containing fasta sequences
                $ra_seq      Ref to an array  The method fills it with the path of the FASTA files
                $workingdir  directory where create the ordered fiel tree                
                $root        Name of the concatenation of fasta files
 Globals      : none

=cut

sub CreateWorkSeqTree
{
    my ($o_param, $indir, $ra_seq, $workingdir, $root, $o_logger, $logger_stamp, $uniq) = @_;

    &PrintStart("Create tree");

    my $f_all = &General::GetStreamOut("$workingdir/$root.tmp");
    my $files;

    # Create a multifasta file with the fasta sequences
    foreach my $f (glob("$indir/*"))
    {
        if (&IsFasta($f))
        {
            &PrintDebug("Load fasta file $f\n");

            my $f_in = &General::GetStreamIn($f);
            while (my $l = <$f_in>)
            {
                print $f_all $l;
            }
            $files .= "$f ";
        }
        else
        {
            printf STDERR "WARNING: $f ignored. Dont look like a FASTA file\n";
            $o_logger->Trace("WARNING: $f ignored. Dont look like a FASTA file\n");
        }
    }
    $f_all->close;

    &Fasta2Tree($o_param, "$workingdir/$root.tmp", $workingdir, $o_logger, $logger_stamp);

    # Remove temporary full fasta file
    my $cmd = "rm $workingdir/$root.tmp";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    # For each sequence, create a workingdirectory
    # And save it in a full sorted FASTA file
    my $f_sorted_all = &General::GetStreamOut("$workingdir/$root");
    my $f_fof        = &General::GetStreamIn("$workingdir/entries.fof");
	my $prevdir      = cwd();

    while (my $f = <$f_fof>)
    {
        chomp($f);
        push(@$ra_seq, $f);
        my ($seqpath, $seqname) = $f =~ /(.+)\/(.+)/;

        # Create the workingdir for this sequence
        chdir($seqpath);
        mkdir("work.$uniq");
        my $f_in = &General::GetStreamIn($f);
        while (my $l = <$f_in>)
        {
            print $f_sorted_all $l;
        }
    }
    $f_sorted_all->close;
	chdir($prevdir);

    print STDERR "Cat $files in $workingdir/$root\n" if ($DEBUG);
    $o_logger->Trace("Cat $files in $workingdir/$root\n", $logger_stamp, $LOGGER_LEVEL_EXEC);

    &ComputeGCPercent($o_param, "$workingdir/$root");

    &PrintStop("Create tree");

    return;
}

#################################################################################
# Fasta manipulation
#################################################################################

=head2 function IsFasta

 Title        : IsFasta
 Usage        : IsFasta($file)
 Prerequisite : none
 Function     : return true if its the file looks like a fasta. Just assert the first line start by '>'
 Returns      : ture/false
 Args         : $seq  a file
 Globals      : none

=cut

sub IsFasta
{
    my ($seq) = @_;

    my $f_seq = &General::GetStreamIn($seq);
    my $line  = <$f_seq>;
    return 0 if (defined($line) && $line !~ /^>\S+/);
    return 1;
}

=head2 procedure FilterFastaLen

 Title        : FilterFastaLen
 Usage        : 
 Prerequisite : 
 Procedure    : keep only sequence with a length > $minlen
 Args         : $in     fasta file  input
                $out    fasta file  output
                $minlen integer     Minimum length
 Globals      : none

=cut

sub FilterFastaLen
{
    my ($o_param, $in, $out, $minlen, $o_logger, $logger_stamp) = @_;

    my $cmd = $o_param->Get('prg_fasta_filter');
    $cmd .= " --in $in --out $out --min_length $minlen";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    return;
}

=head2 function GetFirstSeqName
	
 Title        : GetFirstSeqName
 Usage        : my $name = &GetFirstSeqName($fasta)
 Prerequisite : 
 Function     : Extract the seq name from the first fasta header
 Args         : $fasta   Fasta file
 Return       : first seq name
 Globals      : none

=cut

sub GetFirstSeqName
{
    my ($fasta) = @_;
    my $f_in    = &GetStreamIn($fasta);
    my $l       = <$f_in>;
    my ($name)  = $l =~ />(\S+)/;
    return $name;
}



sub GetFirstSeqPath
{
	my ($o_param) = @_;

	my $f_in    = $o_param->GetStreamIn('seq_fof');
    my $l       = <$f_in>;
	chomp($l);

    return $l;
}



=head2 function ComputeGCPercent

 Title        : ComputeGCPercent
 Usage        : &ComputeGCPercent($seqfile)
 Function     : Compute the gc percent (nb of G|C) vs length of the sequence except 'N'
                Set variable gc_percent to "high" or "low" 
 Args         : $o_param
                $seqfile Multifasta file
 Return       : gc_percent
 Globals      : none

=cut

sub ComputeGCPercent
{
    my ($o_param, $seqfile) = @_;

    my $gccount     = 0;
    my $lenwithoutN = 0;
    my $f_          = &General::GetStreamIn($seqfile);
    while (my $line = <$f_>)
    {
        chomp($line);
        next if $line =~ /^>/;
        ($gccount) += $line =~ tr/GCgc/GCgc/;
        $line =~ s/N//gi;    # remove N
        $lenwithoutN += length($line);
    }
    $f_->close;

    my $gc_percent = $gccount / $lenwithoutN * 100;
    print STDOUT "GC% : $gc_percent\n" if ($DEBUG);
    if ($gc_percent >= 50)
    {
        $o_param->Set('gc_percent', "high");
    }
    else
    {
        $o_param->Set('gc_percent', "low");
    }
    return $gc_percent;
}

=head2 function Fasta2Tree
	
 Title        : Fasta2Tree
 Usage        : 
 Procedure    : generate a tree from multifasta file
                0001/seq2/seq1
                     seq2/seq2
 Args         : $o_param
                $mutifasta 
                $dir Directory where create tree
 Globals      : none

=cut

sub Fasta2Tree
{
    my ($o_param, $multifasta, $dir, $o_logger, $logger_stamp) = @_;

    if (-d "$dir/0001")
    {
        &PrintDebug("Warning: tree $dir/0001 already exists.\n");
        return;
    }

    my $cmd = $o_param->Get('prg_fasta2tree');
    $cmd .= " --sequence $multifasta --outdir $dir --type dna  --max_subdirectories 1000000";
	$cmd .= " --sort_by_len" if ($o_param->Get('chr_sorting_by_size') == 1);
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    return;
}

# count the number of different ID in the multifasta file
sub CountDiffSeq
{
    my ($seq) = @_;

    my %h_seq;
    &GeneralBioinfo::FastaToHash($seq, \%h_seq, 1);
    my $seq_nb = scalar(keys(%h_seq));
    return $seq_nb;
}

# count the number of sequences in the multifastafile
sub CountSeq
{
    my ($seq) = @_;

    my $seq_nb = 0;
    my $f_in   = &General::GetStreamIn($seq);
    while (my $line = <$f_in>)
    {
        chomp($line);
        $seq_nb++ if ($line =~ /^>/);
    }
    $f_in->close;

    return $seq_nb;
}

sub GetMultifastaInfo
{
    my ($o_param, $fasta) = @_;

    my $cmd     = $o_param->Get('prg_N50') . " " . $fasta;
    my $res     = `$cmd`;
    my ($nbreg) = $res =~ /^NUM\t(\d+)/;
    my ($ntnb)  = $res =~ /BP\t(\d+)/;

    #print "nb de regions $nbreg : nt = $ntnb nt\n";

    return ($nbreg, $ntnb);
}

########################################################################################
########################################################################################

=head2 function Mapping

 Title        : Mapping
 Usage        : 
 Prerequisite : $o_param has the keys: prg_glint_mate_dist, prg_glint, 
                prg_glint_mm_small_pe, prg_glint_mm_pe, prg_glint_small_param, prg_glint_param
 Function     : Launch glint. Adapt parameters of the type of data (ose/ope, small)
 Args         : $seq  a fasta file
                $evidence fastq file
                $format  fast or fasta
                $orientation ope/ose
                $outfile
                $small 1 if it is small rna, 0 else
                $o_param
 Globals      : none

=cut

sub Mapping
{
    my ($seq, $evidence, $format, $orientation, $outfile, $small, $o_param, $o_logger, $logger_stamp) = @_;

    if ($DEBUG && (-e "$outfile.gz"))
    {
        print STDERR "DEBUG MODE: use $outfile.gz.\n";
        return;
    }

    my $cmd;
    my $mm = 0;

    if ($orientation =~ /ope/)
    {
        #Carp::croak("An even number of filenames is expected for paired read files\n") if (($#$ra_lof+1) %2 != 0 );
        my $matedist = " --mate-maxdist " . $o_param->Get('prg_glint_mate_dist');
        $cmd = $o_param->Get('prg_glint') . " mappe $matedist ";

        # lorsque les sequences recherches sont tres courtes, je veux 0 mm,
        # sinon 2 reste accceptable car c'est le meilleur
        # et que je vais jouer sur la paire pour retrouver la specificite
        $mm = ($small) ? $o_param->Get('prg_glint_mm_small_pe') : $o_param->Get('prg_glint_mm_pe');
    }
    else
    {
        $cmd = $o_param->Get('prg_glint') . " map ";

        # lorsque les sequences recherches sont tres courtes,
        # je veux 0 mm, sinon 1 reste accceptable car c'est le meilleur
        $mm = ($small) ? $o_param->Get('prg_glint_mm_small_se') : $o_param->Get('prg_glint_mm_se');
    }
    my $glint_params = " --mmis $mm --input-format $format ";
    if ($small)
    {
        $glint_params .= $o_param->Get('prg_glint_small_param') . " ";
    }
    else
    {
        $glint_params .= $o_param->Get('prg_glint_param') . " ";
    }

    #my $outfile = ( $orientation =~ /pe/ ) ? "$outdir/$ref.vs.$outprefix.bedpe" : "$outdir/$ref.vs.$outprefix.bed";

    # 1. mapping
    $cmd .= " $seq  $evidence $glint_params -o $outfile 2> /dev/null";

    print STDERR "MAPPING... $cmd\n" if ($DEBUG);
    &ExeCmd($cmd, $o_logger, $logger_stamp);
    print STDERR "Mapping done.\n" if ($DEBUG);

    return;

}

=head2 procedure MappingFiltering

 Title        : MappingFiltering
 Usage        : MappingFiltering
 Prerequisite : prg_bed_filter and prg_bed_filter_param have to be defined in o_param
 Function     : 
 Returns      : none
 Args         : $mapping_file file    bed/bedpe file
                $orientation  string  ose/ope
                $small        integer [0 or 1] 
                $outfile      file    bed result file
                $pat          string  pattern, for instance 'evidence_1'
                $o_param
               
 Globals      : none

=cut

sub MappingFiltering
{
    my ($mapping_file, $orientation, $small, $outfile, $pat, $o_param, $o_logger, $logger_stamp) = @_;

    if ($DEBUG && (-e $outfile))
    {
        print STDERR "DEBUG MODE: use $outfile file.\n";
        return;
    }

    my $cde;
    my $prg_filter       = $o_param->Get('prg_bed_filter');
    my $prg_filter_param = $o_param->Get('prg_bed_filter_param');

    if ($orientation =~ /ope/)
    {
        $prg_filter .= " --bedpe";
        if ($small)
        {
            $prg_filter .= " " . $o_param->Get('prg_bed_filter_pe_small_param');
        }
        else
        {
            $prg_filter .= " " . $o_param->Get('prg_bed_filter_pe_param');
        }
    }
    $prg_filter .= " --oriented_end " . $o_param->Get($pat . "_oriented_end") . " ";    # evidence_x_oriented_end

    $cde = "$prg_filter --in $mapping_file --out $outfile $prg_filter_param";

    print STDERR "FILTERING... $cde\n" if ($DEBUG);
    &ExeCmd($cde, $o_logger, $logger_stamp);
    print STDERR "Filtering done.\n" if ($DEBUG);

    return;
}

=head2 procedure Bed2Expr4EGN

 Title        : Bed2Expr4EGN
 Usage        : Bed2Expr4EGN()
 Prerequisite : prg_bed_to_expr prg_bed_to_expr_annot_param workingdir prg_bed_to_expr_annot_scoremax
                prg_bed_to_expr_annot_min_coverage_percent prg_bed_to_expr_annot_scalingfactor
 Function     : Convert bed[pe] file to expression files for the annotation (position.gff3 and region.gff3) 
 Returns      : none
 Args         : $filtering_bed_file   file    bed/bedpe file
                $orientation  string  ose/ope
                $outpositionfile      file      
                $outregionfile        file 
                $o_param
               
 Globals      : none

=cut

sub Bed2Expr4EGN
{
    my ($filtering_bed_file, $orientation, $outpositionfile, $outregionfile, $o_param, $o_logger, $logger_stamp, $uniq)
      = @_;

    my $cde;

    my $out = $o_param->Get('workingdir') . "/expr_gff3." . $uniq;
    my $tag = "glint";                                               #TODO: mettre un flag dependant de la banque

    $cde = $o_param->Get('prg_bed_to_expr') . " " . $o_param->Get('prg_bed_to_expr_annot_param');

    $cde .= " --tag $tag ";
    $cde .= " --scoremax " . $o_param->Get('prg_bed_to_expr_annot_scoremax');
    $cde .= " --scalingfactor " . $o_param->Get('prg_bed_to_expr_annot_scalingfactor');
    $cde .= " --min_coverage_percent " . $o_param->Get('prg_bed_to_expr_annot_min_coverage_percent');
    $cde .= " --interval --bedpe " if ($orientation =~ /ope/);
    $cde .= " --bedfile $filtering_bed_file --outprefix $out";

    &ExeCmd($cde, $o_logger, $logger_stamp);

    # Rename files *position.gff3 and *region.gff3
    $cde = "mv $out.position.gff3 $outpositionfile";
    &ExeCmd($cde, $o_logger, $logger_stamp);

    $cde = "mv $out.region.gff3 $outregionfile";
    &ExeCmd($cde, $o_logger, $logger_stamp);
    unlink("$out.non_intergenic.gff3");

    return;
}

=head2 procedure Bed2Expr4tStart

 Title        : Bed2Expr4tStart
 Usage        : Bed2Expr4tStart()
 Prerequisite : prg_bed_to_expr prg_bed_to_expr_annot_param workingdir 
                prg_bed_to_expr_tstart_scoremax prg_bed_to_expr_tstart_scalingfactor
                prg_bed_to_expr_tstart_min_coverage,
                have to be keys of o_param
 Function     : Convert bed[pe] file to expression files for the tStart detection (position.gff3) 
 Returns      : none
 Args         : $bed_file             file    bed/bedpe file
                $orientation          string  ose/ope
                $outpositionfile      file      
                $o_param
               
 Globals      : none

=cut

sub Bed2Expr4tStart
{
    my ($bed_file, $orientation, $outpositionfile, $o_param, $o_logger, $logger_stamp, $uniq) = @_;

    my $cde;

    my $out = $o_param->Get('workingdir') . "/expr_tStart_gff3." . $uniq;
    my $tag = "glint";                                                      #TODO: mettre un flag dependant de la banque

    $cde = $o_param->Get('prg_bed_to_expr') . " " . $o_param->Get('prg_bed_to_expr_tstart_param');

    $cde .= " --tag $tag ";
    $cde .= " --scoremax " . $o_param->Get('prg_bed_to_expr_tstart_scoremax');
    $cde .= " --scalingfactor " . $o_param->Get('prg_bed_to_expr_tstart_scalingfactor');
    $cde .= " --min_coverage " . $o_param->Get('prg_bed_to_expr_tstart_min_coverage');
    $cde .= " --interval --bedpe " if ($orientation =~ /ope/);
    $cde .= " --bedfile $bed_file --outprefix $out";

    &ExeCmd($cde, $o_logger, $logger_stamp);

    # Rename file *position.gff3
    $cde = "mv $out.position.gff3 $outpositionfile";
    &ExeCmd($cde, $o_logger, $logger_stamp);

    unlink("$out.region.gff3", "$out.non_intergenic.gff3");

    return;
}

sub LaunchEGN
{
    my ($o_param, $seq, $cfg, $matrix, $outdir, $o_logger, $logger_stamp, $validate) = @_;

    my ($path, $seqname) = &GetPathAndName($seq);
    my $step = "$seqname EuGene";

    return if ($validate == 1 && &IsValidatedStep("$outdir/$seqname.gff3", $step));

    &PrintStart($step);
    my $cmd = $o_param->Get('prg_eugene');
    $cmd .=
      " -A $cfg -m $matrix -pg -O $outdir/ $seq > $outdir/$seqname.eugene.stdout 2> $outdir/$seqname.eugene.stderr";

    &ExeCmd($cmd, $o_logger, $logger_stamp);
    &ValidateStep("$outdir/$seqname.gff3") if ($validate == 1);
    &PrintStop($step);

    return;
}

sub GetEuGeneCmd
{
    my ($o_param, $seq, $cfg, $matrix, $outdir, $prefixoutfile, $validate) = @_;

    my ($path, $seqname) = &GetPathAndName($seq);
    my $step        = "$prefixoutfile$seqname EuGene";
    my $eugene_file = "$outdir/$prefixoutfile$seqname.gff3";

    return "" if ($validate == 1 && &IsValidatedStep($eugene_file, $step));

    my $cmd = "export EUGENEDIR=" . $o_param->Get('EUGENEDIR') . "; " . $o_param->Get('prg_eugene');
    $cmd .=
      " -A $cfg -m $matrix -pg -O  $outdir/$prefixoutfile $seq > $outdir/$prefixoutfile$seqname.eugene.stdout 2> $outdir/$prefixoutfile$seqname.eugene.stderr";

    if ($validate == 1)
    {
        my $validcmd = &GetValidateFileCmd($o_param, $eugene_file);
        $cmd .= " && $validcmd";
    }
    return $cmd;
}


# validate: if 1, create a .success file
# prefix: prefix of the eugene result gff3 files
sub LaunchAllEGN
{
    my ($o_param, $ra_seq, $cfg, $matrix, $outdir, $step, $validate, $prefix) = @_;

    my $cmd;

    my $rootworkdir = $o_param->Get('workingdir');

    my $eugenedir   = $outdir . "/EGN_$step\_" . $o_param->Get('uniq').$prefix;
    mkdir($eugenedir);
	
    my $cmd_paraloop   = "$eugenedir/eugene.cmd.paraloop";
    my $f_cmd_paraloop = &General::GetStreamOut($cmd_paraloop);
	
    foreach my $seq (@$ra_seq)
    {
        my $workingdir = &GetWorkDir($seq, $o_param);
        $o_param->Set('workingdir', $workingdir);
		
        $cmd = &GetEuGeneCmd($o_param, $seq, $cfg, $matrix, $outdir, $prefix, $validate);
        next if ($cmd eq "");
        print $f_cmd_paraloop "$cmd\n";
    }
    $f_cmd_paraloop->close;

    $o_param->Set('workingdir', $rootworkdir);

    if (-s $cmd_paraloop)
    {
        my $msg = "EuGene annotation";
        &PrintStart($msg);
        &Paraloop($o_param, $cmd_paraloop, $eugenedir);
        &PrintStop($msg);
    }

    return;

}

=head2 function GetEvidenceName

 Title        : GetEvidenceName
 Usage        : 
 Prerequisite : none 
 Function     : Get the name of the evidence
 Args         : $file Path of a evidence file (fastq for instance)
 Globals      : none

=cut

sub GetEvidenceName
{
    my ($file) = @_;

    my ($evidence_name) = $file =~ /[^;]+\/([^;]+)/;
	if ( !defined($evidence_name) )
	{
		return($file);
	}
    $evidence_name =~ s/\.(lzma|xz|gz|zip|fna)$//;    # Remove extension
    $evidence_name =~ s/\.[^.]+$//;                   # Remove the type of the file (fastq, wig, ...)

    return $evidence_name;
}

sub ExeCmd
{
    my ($cmd, $o_logger, $logger_stamp) = @_;

    if ($DEBUG)
    {
        print STDERR "---------------------------------------------------------------------\n";
        print STDERR $cmd . "\n";
        print STDERR "---------------------------------------------------------------------\n";
    }
	if ( defined($o_logger) ) # lorsque je vais //er je vais envoyer un o_logger non defini
	{
    	$o_logger->Trace("$cmd\n", $logger_stamp, $LOGGER_LEVEL_EXEC);
	}
    system($cmd) && Carp::croak "ERROR >$cmd<\n";

    return;
}

=head2 procedure Paraloop

 Title        : Paraloop
 Usage        : 
 Prerequisite : 'prg_paraloop', 'workingdir' has to be keys of $o_param
 Function     : Launch commands with paraloop program
 Args         : $o_param      ParamParser object
                $cmd_paraloop File of command lines
                $paraloopdir  Directory where launch paraloop. 
                              (optional)
 Globals      : none

=cut

sub Paraloop
{
    my ($o_param, $cmd_paraloop, $paraloopdir) = @_;

    my $currentdir = cwd();

    if ($paraloopdir eq "")
    {
        $paraloopdir = $o_param->Get('workingdir') . "/PARALOOP_" . $o_param->Get('uniq');
    }

    mkdir($paraloopdir) if (!-d $paraloopdir);
    chdir($paraloopdir);

    # create a tmp
    my $cmd = $o_param->Get('prg_paraloop')
      . " --program=Shell --input $cmd_paraloop --output $paraloopdir/paraloop.output --clean > /dev/null 2>&1";
	my $wc = `wc -l $cmd_paraloop`;	my @a_wc=split(" ", $wc);
	if ( $a_wc[0] <= 1 )
	{
		$cmd =~ s/\-\-interleaved//;
	}
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    chdir($currentdir);

    return;
}

sub GetMappingFileName
{
    my ($seqname, $evidencefile) = @_;
    return "$seqname." . &GetEvidenceName($evidencefile) . "." . $MAPPING_EXT;
}

# seq : full path of an input sequence file
# Return the working directory of the sequence
sub GetWorkDir
{
	my ($seq, $o_param) = @_;
	
	my $uniq = $o_param->Get('uniq');
	$seq =~ s/\.masked$//;
	my ($seqdir, $seqname) = $seq =~ /(.+)\/(.+)$/;

	return "$seqdir/work.$uniq/";
}

# seqname: sequence name
# Return the working directory of the sequence
sub GetWorkSeqDir
{
    my ($seqname, $o_param) = @_;
	
    $seqname =~ s/\.masked$//;
	
	my $seqpath = &GetSeqPath($seqname, $o_param);
	return &GetWorkDir($seqpath, $o_param);
}

sub GetSeqDir
{
    my ($seqname, $o_param) = @_;
	
	my $seqpath = &GetSeqPath($seqname, $o_param);
	my ($dir,$name) = $seqpath =~ /(.+\/)([^\/]+)$/;
    return $dir;
}


sub GetSeqPath
{
	my ($seqname, $o_param) = @_;
	
	my $seq   = "";
	my $f_fof = $o_param->GetStreamIn('seq_fof');
	
	while (my $l = <$f_fof>)
	{
		chomp($l);
		if ($l =~ /\/$seqname\/$seqname$/)
		{
            $seq = $l;
			last;
		}
	}
	$f_fof->close;
	
    return $seq;
}


sub GetDbDir
{
    my ($o_param) = @_;

    my $rootworkingdir = $o_param->Get('workingdir');
    my $newdbdir       = "$rootworkingdir/$DBDIR";
    if (!-d $newdbdir)
    {
        mkdir $newdbdir;
    }
    return $newdbdir;
}

# if not exists create the annotation tree
# dir/raw_eugene
# dir/full_eugene
sub CreateAnnotTree
{
    my ($o_param, $nb) = @_;

    my $annotdir = $o_param->Get('workingdir') . "/$ROOT_ANNOT_DIR$nb/";
    if (!-d $annotdir)
    {
        mkdir $annotdir;
    }
    mkdir "$annotdir/raw_eugene"  if (!-d "$annotdir/raw_eugene");
    mkdir "$annotdir/full_eugene" if (!-d "$annotdir/full_eugene");

    return $annotdir;
}

sub GetParamDir
{
    my ($o_param) = @_;

    my $paramdir = $o_param->Get('workingdir') . "/egn_param/";
    if (!-d $paramdir)
    {
        mkdir $paramdir;
    }
    return $paramdir;
}

sub GetTrainDir
{
    my ($o_param) = @_;

    my $dir = $o_param->Get('workingdir') . "/$TRAINDIR/";
    if (!-d $dir)
    {
        mkdir $dir;
    }
    return $dir;
}

=head2 function MaskSequence
	
 Title        : MaskSequence
 Usage        : &MaskSequence($seqin, $seqout, $gff3)
 Procedure    : regions of gff3 file are masked by 'N'
 Args         : $seqin fastafile
                $seqout masked result fasta file
                $gff3 gff3 file of region of seqin
 Globals      : none

=cut

sub MaskSequence
{
    my ($seqin, $seqout, $gff3) = @_;

    my ($p,  $n)  = &GetPathAndName($seqin);
    my ($pg, $ng) = &GetPathAndName($gff3);

    my $step = "$n masked with $ng";
    &PrintStart($step);

    # load sequences
    my %h_seq;
    &FastaToHash($seqin, \%h_seq);

    my %h_gff3line = ();

    my $f_gff3 = &GetStreamIn($gff3);
    while (my $line = <$f_gff3>)
    {
        chomp($line);

        #print "$line\n";
        next if ($line =~ /^#/);
        &ParseGFF3Line(\$line, 2, \%h_gff3line);

        #print $h_gff3line{start}. " len: ". $h_gff3line{len}."\n";
        substr($h_seq{$h_gff3line{seqid}}{sequence}, $h_gff3line{start} - 1, $h_gff3line{len}, 'N' x $h_gff3line{len});
    }
    $f_gff3->close;
    &HashToFasta($seqout, \%h_seq);

    &PrintStop($step);

    return;
}

=head2 function GetSeqInfo
	
 Title        : GetSeqInfo
 Usage        : ($p, $n, id, $len) = &GetSeqInfo($o_param)
 Function     : Return the seq directory, the seq file name, 
                the seqid and the seq len
 Args         : $seq : Path of a fasta file
 Globals      : none

=cut

sub GetSeqInfo
{
    my ($seq) = @_;

    my ($seqpath, $seqname) = $seq =~ /(.+)\/(.+)/;

    # Get name and length of the sequence
    my %h_seq;
    &FastaToHash($seq, \%h_seq, 1);
    my $seqid;
    my $seqlen = 0;
    foreach my $k (keys %h_seq)
    {
        $seqid .= $k;
        $seqlen += $h_seq{$k}{len};
    }

    #my @a_keys = keys(%h_seq);
    #my $seqid  = $a_keys[0];
    #my $seqlen = $h_seq{$seqid}{len};

    return ($seqpath, $seqname, $seqid, $seqlen);
}

sub GetPathAndName
{
    my ($f) = @_;

    my ($path, $name) = $f =~ /(.+)\/(.+)/;

    return ($path, $name);
}

=head2 function ReadFile

 Title        : ReadFile
 Function     : Return the content of the file
 Args         : $file 
 Globals      : none

=cut

sub ReadFile
{
    my ($file) = @_;

    my $content = "";
    my $f_      = &General::GetStreamIn($file);
    while (my $line = <$f_>)
    {
        $content .= $line;
    }
    $f_->close;

    return $content;
}

sub Gzip
{
    my ($o_logger, $logger_stamp, @a_files) = @_;

    my $cmd = "gzip -f ";
    foreach my $file (@a_files)
    {
        $cmd .= " $file ";
    }

    &ExeCmd($cmd, $o_logger, $logger_stamp);

    return;

}

sub CatSeq
{
    my ($ra_seq, $newseq) = @_;

    my $f_fullseq = &General::GetStreamOut($newseq);
    foreach my $seq (@$ra_seq)
    {
        my $f_ = &General::GetStreamIn($seq);
        while (my $line = <$f_>)
        {
            print $f_fullseq $line;
        }
        $f_->close;
    }
    $f_fullseq->close;

    return;
}

##################################################################################
##                                   BLAST FUNCTIONS
##################################################################################

=head2 function GetBlastXDb

 Title        : GetBlastXDb
 Usage        : 
 Prerequisite : blastx_db_list is an existing key in o_param
 Function     : Fill $rh_db_list
 Args         : $o_param 
                $rh_db_list  Ref to a hash: Key=nb, Value=path of the database
                $rh_db_level Ref to a hash: Key=nb, Value=weignth for this database
 Globals      : none

 JG: l'init de file et weight etait fait 2 fois, ici et dans Assert..., je fais pareil pour pcs et activegaps

=cut

sub GetBlastXDb
{
    my ($o_param, $rh_db_list) = @_;

    my $list = $o_param->Get('blastx_db_list');    # "2 5 6"
	$list =~ s/^\s+//;
    my @a_db_nb = split(/\s+/, $list);

    foreach my $nb (@a_db_nb)
    {
        my $db = $o_param->Get("blastx_db_$nb\_file");
        $db =~ s/\s+$//;
        my $level = $o_param->Get("blastx_db_$nb\_weight");

        $$rh_db_list{$nb}{file}  = $db;
        $$rh_db_list{$nb}{weight}= $level;
        $$rh_db_list{$nb}{pcs}= $o_param->Get("blastx_db_$nb\_pcs");
        $$rh_db_list{$nb}{activegaps}= $o_param->Get("blastx_db_$nb\_activegaps");
		$$rh_db_list{$nb}{pcs}=~s/\s//g;
		$$rh_db_list{$nb}{activegaps}=~s/\s//g;
    }

    return;
}

=head2 procedure MakeBlastDb

 Title        : MakeBlastDb
 Function     : Index protein fasta file with makeblastdb
 Args         : $o_param
                $db Protein fasta file
 Prerequisite : $db has to be a prot fasta file
 Globals      : none

=cut

sub MakeBlastDb
{
    my ($o_param, $db, $o_logger, $logger_stamp) = @_;

    my ($dbpath, $dbname) = &GetPathAndName($db);

    #&PrintStart("Index $dbname");

    my $pg  = $o_param->Get('prg_makeblastdb');
    my $cmd = "$pg -dbtype prot -in $db -parse_seqids > /dev/null 2>&1";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

	$cmd = $o_param->Get('prg_extract_seqlen')." --in $db --out $db.seqlen ";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    #&PrintStop("Index $dbname");
    return;
}


sub MakeReduceDbMethodIndex
{
    my ($o_param, $db, $o_logger, $logger_stamp) = @_;

	my $method = $o_param->Get('protein_similarity_search_method');
	if ($method eq "diamond_blastx")
	{
		&MakeDiamondDb($o_param, $db, $o_logger, $logger_stamp);
	}
	elsif ($method eq "ublast_blastx")
	{
		&MakeUBlastDb($o_param, $db, $o_logger, $logger_stamp);
	}

    return;
}

=head2 procedure MakeUBlastDb

 Title        : MakeUBlastDb
 Function     : Index protein fasta file with usearch 
 Args         : $o_param
                $db Protein fasta file
 Prerequisite : $db has to be a prot fasta file
 Globals      : none

=cut

sub MakeUBlastDb
{
    my ($o_param, $db, $o_logger, $logger_stamp) = @_;

    # format for ublast
    if (!-e $db . ".udb")
    {
        my $usearch_cmd = $o_param->Get('prg_usearch');
        my $cmd         = $usearch_cmd . " -makeudb_ublast $db -output $db.udb";
        $cmd .= " > /dev/null 2>&1 " if (!$DEBUG);
        &ExeCmd($cmd, $o_logger, $logger_stamp);
    }

    return;
}


=head2 procedure MakeDiamondDb

 Title        : MakeDiamondDb
 Function     : Index protein fasta file with diamond
 Args         : $o_param
                $db Protein fasta file
 Prerequisite : $db has to be a prot fasta file
 Globals      : none

=cut

sub MakeDiamondDb
{
    my ($o_param, $db, $o_logger, $logger_stamp) = @_;

    # format for diamond
    if (!-e $db.".dmnd")
    {
	my $cmd = $o_param->Get('prg_diamond'). " makedb --in $db -d $db";
	$cmd .= " > /dev/null 2>&1" if (!$DEBUG); 
	&ExeCmd($cmd, $o_logger, $logger_stamp);
    }
    return;
}

=head2 function GetBlastPlusCmd

 Title        : GetBlastPlusCmd
 Usage        : 
 Prerequisite : blast_cmd and blast_param keys exists in o_param
 Function     : Return blast+ command
 Args         : $o_param  ParamParser object
                $seq      fasta file path
                $db       protein fasta file path
                $out_file
 Globals      : none

=cut

sub GetBlastPlusCmd
{
    my ($o_param, $seq, $db, $out_file) = @_;

    my $blast_cmd   = $o_param->Get('blast_cmd');
    my $blast_param = $o_param->Get('blast_param');

    # Blastplus cmd
    my $cmd = "$blast_cmd -db $db -query $seq -out $out_file $blast_param";

    return $cmd;
}

=head2 procedure GetBlastX

 Title        : GetBlastX
 Usage        : 
 Prerequisite : prg_get_blastx has to be defined
 Function     : Launch blastx. Result file is seq.blast$ext
 Args         : $o_param  ParamParser object
                $seq      fasta file path
                $db       protein fasta file path
                $ext      Extension of the blastx result file
                $workingdir working directory
 Globals      : none

=cut

sub GetBlastX
{
    my ($o_param, $seq, $db, $ext, $workingdir, $o_logger, $logger_stamp) = @_;

    my ($a, $sname) = &GetPathAndName($seq);
    my ($b, $dname) = &GetPathAndName($db);

    my $step    = "$sname BlastX $dname";
    my $outfile = "$seq.blast$ext";

    return if (&IsValidatedStep("$outfile.gff3", $step));

    my $param_msg = "  PARAMETERS=" . $o_param->Get('blast_param') . "\n";

    &PrintStart($step);
    &PrintMore($param_msg);

    my $cmd =
        $o_param->Get('prg_get_blastx')
      . " --sequence $seq --cfg "
      . $o_param->Get('cfg')
		. " --db $db --outfile $outfile --workingdir $workingdir";
	$cmd .= " --protein_similarity_search_method " . $o_param->Get('protein_similarity_search_method');
    $cmd .= " --verbose " if ($DEBUG);
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    # Mv $seq.blast$ext in $workingdir (we only keep the .gff3)
    $cmd = "mv $outfile $workingdir/";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    &PrintStop($step);

    &ValidateStep("$outfile.gff3");

    return;
}

# TODO A transformer en programme externe?????
# cas dutilisation: banque proteique contre bc de sequences nucleiques (ex contre transcriptome)
# OBSOLETE: not used
sub GetBlastXParaloop
{
    my ($o_param, $seq, $db, $out, $o_logger, $logger_stamp, $compress) = @_;

    my ($a, $sname) = &GetPathAndName($seq);
    my ($b, $dname) = &GetPathAndName($db);

    my $tmpdir = &GetTrainDir($o_param) . "/tmp_paraloop";
    mkdir $tmpdir if (!-d $tmpdir);

    my $step      = "BlastX $sname $dname";
    my $param_msg = "  PARAMETERS=" . $o_param->Get('blast_param') . "\n";
    return if (&IsValidatedStep($out, $step));

    &PrintStart($step);
    &PrintMore($param_msg);

    my $cfg   = $out . ".paraloop.cfg";
    my $f_cfg = &GetStreamOut($cfg);
    print $f_cfg "PARALOOP_Blast_origin=blast_plus\n";
    print $f_cfg "PARALOOP_Blast_path=" . $o_param->Get('blast_cmd') . "\n";
    print $f_cfg "PARALOOP_Blast_params=" . $o_param->Get('blast_param') . "\n";
    print $f_cfg "PARALOOP_Blast_chunk=" . $o_param->Get('blast_paraloop_chunk') . "\n";
    print $f_cfg "PARALOOP_tmp=$tmpdir\n";
    $f_cfg->close;

    my $cmd = $o_param->Get('prg_paraloop');
    $cmd .= " --program Blast --input $seq --db $db --cfg $cfg --output $out";
    $cmd .= " > /dev/null 2>&1";

    &ExeCmd($cmd, $o_logger, $logger_stamp);

    &PrintStop($step);
    &ValidateStep($out);

    &Gzip($o_logger, $logger_stamp, $out) if ($compress);

    return;

}

=head2 procedure BlastXBatch

 Title        : BlastXBatch
 Usage        : 
 Prerequisite : prg_fasta_splitter, blast_cmd and blast_param have to be
                defined in $o_param
 Function     : use to blast proteome against multifasta file of short
                sequences (trasncriptome for instance)
 Args         : $o_param  ParamParser object
                $seq      multifasta file path
                $db       protein fasta file path
                $out      blast outfile (m8 file)
                $compress If true gzip the result file
 Globals      : none

=cut

sub BlastXBatch
{
    my ($o_param, $seq, $db, $out, $compress) = @_;

    my $suffix = "training";

    my ($a,     $sname) = &GetPathAndName($seq);
    my ($b,     $dname) = &GetPathAndName($db);
    my ($opath, $oname) = &GetPathAndName($out);

    my $step      = "BlastX $sname $dname";
    my $param_msg = "  PARAMETERS=" . $o_param->Get('blast_param') . "\n";

    return if (&IsValidatedStep($out, $step));

    &PrintStart($step);
    &PrintMore($param_msg);

    # Create a dedicated directory
    my $tmpdir = $opath . "/BLASTBATCH_" . $o_param->Get('uniq');
    mkdir $tmpdir if (!-d $tmpdir);

    # split multifasta file in slices of equivalent slices in nt
    my $cmd = $o_param->Get('prg_fasta_splitter');
    $cmd .= " --in $seq --outprefix $tmpdir/seq --maxlength_per_slice 10000000";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    # create a paraloop command file
    my $paraloop_cmd = "$tmpdir/paraloop.cmd";
    my $f_para       = &General::GetStreamOut($paraloop_cmd);

    my @a_cat;
    my @a_files = glob("$tmpdir/seq*");
    foreach my $subseq (@a_files)
    {
        chomp($subseq);
        next if ($subseq !~ /seq\d+$/);
        my $tmpout = $subseq . ".blast$suffix";
        my $cmdline = &GetBlastPlusCmd($o_param, $subseq, $db, $tmpout);
        print $f_para "$cmdline\n" if ($cmdline ne "");
        push(@a_cat, $tmpout);
    }
    $f_para->close;

    # Launch paraloop
    &Paraloop($o_param, $paraloop_cmd, $tmpdir);

    # concat intermediar result files
    $cmd = "cat " . join(' ', @a_cat) . " > $out";
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    # remove intermediar result files
    $cmd = "rm " . join(' ', @a_cat);
    &ExeCmd($cmd, $o_logger, $logger_stamp);

    &PrintStop($step);
    &ValidateStep($out);

    &Gzip($o_logger, $logger_stamp, $out) if ($compress);

    return;
}

=head2 function SearchBlastXResult

 Title        : SearchBlastXResult
 Usage        : 
 Prerequisite : blastx_db_list and training_db_file are defined in o_param
 Function     : Search if a previous BlastX result
 Args         : $o_param 
                $workingdir
                $seqname
                $training_db_file
              
 Globals      : none

=cut

sub SearchBlastXResult
{
    my ($o_param, $workingdir, $seqname, $training_db_file) = @_;

    my $list = $o_param->Get('blastx_db_list');    # "2 5 6"
    my @a_db_nb = split(/\s+/, $list);

    foreach my $nb (@a_db_nb)
    {
        if ($o_param->Get("blastx_db_$nb\_file") eq $training_db_file)
        {
            if (-f $workingdir . "/" . $seqname . ".blast$nb")
            {
                return $workingdir . "/" . $seqname . ".blast$nb";
            }
        }
    }

    return "";
}

sub CreateDatabase
{
    my ($o_param, $f1, $f2, $res) = @_;

    my $step = "Create repeat database (RepBase+SpeciesRepeatDom)";
    return if (&IsValidatedStep($res, $step));

    &PrintStart($step);
    my $cmd = "cat $f1 $f2 > $res";
    &ExeCmd($cmd, $o_logger, $logger_stamp);
    &MakeBlastDb($o_param, $res, $o_logger, $logger_stamp);
    &MakeReduceDbMethodIndex($o_param, $res, $o_logger, $logger_stamp);

    &PrintStop($step);
    &ValidateStep($res);

    return;
}

########################################################################
###  Tools to print information
########################################################################

sub PrintInfo
{
    my ($text) = @_;

    $text .= "\n" if ($text !~ /\n$/);
    print STDOUT "$text";
    $o_logger->Trace("$text", $logger_stamp, $LOGGER_LEVEL_EXEC);

    return;
}

sub PrintDebug
{
    my ($text) = @_;

    $text .= "\n" if ($text !~ /\n$/);
    print STDERR "DEBUG: $text" if ($DEBUG);
    $o_logger->Trace("$text", $logger_stamp, $LOGGER_LEVEL_EXEC);

    return;
}

sub PrintError
{
    my ($msg) = @_;

    print STDERR $msg;
    $o_logger->Trace("$msg", $logger_stamp, $LOGGER_LEVEL_EXEC);
    exit;
}

=head2 function IsValidatedStep

 Title        : IsValidatedStep
 Usage        :  if (&IsValidatedStep($f, $msg))
 Prerequisite : 
 Function     : return 1 if the file exists and the file $f.succeed exists 
                (it means the step which create the program has finished properly)
 Args         : $file
              
 Globals      : none

=cut

sub IsValidatedStep
{
    my ($file, $msg) = @_;

    if ($FORCE)
    {
        unlink($file . ".success") if (-e $file . ".success");
    }

    if ((-e $file || -e "$file.gz") && -e $file . ".success")
    {
        my $textlen = length($msg);
        my $len     = $LINELEN - $textlen - $PRINTINDENT * 2;
		$len = 1 if ($len <= 0);
        printf STDOUT "%s%s%savailable\n", ' ' x ($PRINTINDENT * 2), $msg, "." x $len;
        return 1;
    }
    return 0;
}

=head2 function ValidateStep

 Title        : ValidateStep
 Usage        : ValidateStep($f)
 Prerequisite : 
 Function     : create file $f.succeed
 Args         : $file
              
 Globals      : none

=cut

sub ValidateStep
{
    my ($file) = @_;

    if (-e $file)
    {
        &General::GetStreamOut($file . ".success");
    }
    else
    {
        Carp::croak("ERROR: file >$file< should exist.\n");
    }
    return;
}

sub ValidateFile
{
    my ($o_param, $file, $noexe) = @_;
    my $cmd = &GetValidateFileCmd($o_param, $file);
    &ExeCmd($cmd, $o_logger, $logger_stamp);
    return;
}

sub GetValidateFileCmd
{
    my ($o_param, $file) = @_;
    my $cmd = $o_param->Get('prg_validate') . " --file $file";
    return $cmd;
}

sub PrintStart
{
    my ($task) = @_;

    my $textlen = length($task);
    my $len     = $LINELEN - $textlen - $PRINTINDENT * 2;
	$len = 1 if ($len <= 0);	
    printf STDOUT "%s%s%sstarted\n", ' ' x ($PRINTINDENT * 2), $task, "." x $len;
    return;
}

sub PrintStop
{
    my ($task) = @_;

    my $textlen = length($task);
    my $len     = $LINELEN - $textlen - $PRINTINDENT * 2;		
	$len = 1 if ($len <= 0);
    printf STDOUT "%s%s%sdone\n", ' ' x ($PRINTINDENT * 2), $task, "." x $len;

    return;
}

sub PrintMore
{
    my ($txt) = @_;

    chomp($txt);
    printf STDOUT "%s%s\n", ' ' x ($PRINTINDENT * 2), $txt;

    return;
}

sub PrintSection
{
    my ($name) = @_;

    my $char    = "#";
    my $textlen = length($name);
    my $l1      = ($LINELEN - $textlen - ($PRINTINDENT * 2)) / 2 - 2;
    my $l2      = $l1;
    $l2 += 1 if ($textlen % 2 == 1);
    printf STDOUT "%s%s  %s  %s\n", ' ' x ($PRINTINDENT * 2), $char x $l1, $name, $char x $l2;
}

sub AddPrintIndent
{
    my ($val) = @_;
    $PRINTINDENT += $val;

    return;
}

sub AppendFile
{
    my ($file, $txt) = @_;

    my $f_ = IO::File->new(">>$file") or die "ERROR: Can't write in $file\n";
    print $f_ $txt;
    $f_->close;
    return;
}

sub Usage
{
    print STDERR<<END
$0
[Mandatory]
    --indir      dirname        Directory containing genomic sequence fasta files: *.[fna|fa|fasta]* 
    --outdir     dirname        Existing directory where will be stored the annotations
    --cfg        filename       Configuration file

[Optional]
    --workingdir dirname        [Default outdir] directory where will be run the computations

[Advanced optional]
    --no_ncrna_detection        Flag If activate, dont compute ncrna program detection (tRNAscan-SE, rfam_scam and Rnammer)
    --no_rfamscan               Flag If activate, don't launch rfam_scan
    --no_trnascan               Flag If activate, don't launch tRNAscan-SE
    --no_rnammer                Flag If activate, don't launch Rnammer
    --no_repeat_masking         Flag
    --no_red                    Flag
    --help
    --debug      integer        [Default 0] 0: don't display message, 
                                            1: display command line 
    --force                     Flag If activate, relaunch all steps, even if already done
    --no_clean
END
}

sub UsageProk
{
    print STDERR<<END
$0
[Mandatory]
    --indir      dirname        Directory containing two directories:
                                  - genome/*.[fna|fa|fasta]* : genomic sequence fasta files
                                  - data/*                   : data files(fastq/fasta/deb/bedpe/) + (wig/bam)
    --outdir     dirname        Existing directory where will be stored the EuGene predictions
    --cfg        filename       Configuration file

[Optional]
    --workingdir dirname        [Default outdir] directory where will be run the computations

[Advanced optional]
    --no_ncrna_detection        Flag If activate, dont compute ncrna program detection (tRNAscan-SE, rfam_scam and Rnammer)
    --no_rfamscan               Flag If activate, don't launch rfam_scan
    --no_trnascan               Flag If activate, don't launch tRNAscan-SE
    --no_rnammer                Flag If activate, don't launch Rnammer
    --help
	--ncpus      integer        [$DEFAULT_NCPUS] set the number of threads for the main job
    --debug      integer        [Default 0] 0: don't display message, 
                                            1: display command line 
    --force                     Flag If activate, relaunch all steps, even if already done
    --no_clean
END
}
