#!/usr/bin/perl

=pod 

=head1 NAME

ReduceBlast 

=head1 SYNOPSIS

    targeted_blast_search.pl --query sequence.fasta --db database.fasta --output output.blast.m8 --cfg blast.cfg
    [--query_loc 100000-300000]


=head1 DESCRIPTION

Blastx can be a time consuming task. The goal of this program is to 
first reduce the target protein database using quick program ublast (usearch suite), 
and then launch blastx against the reduced database.
Usearch  works well with short query sequences. You can use the option query_loc=start-end to extract a query subsequence 
The option --slice aloow to split the query sequence in sequences of 15000nt before launch ublast. 
CAREFUL: usearch(64bits) sometimes raise memory errors using that query multifasta file.
To fix this issue, if usearch execution failed, classic blastx+ research is performed.

=cut

BEGIN
{
    $ENV{LANG} = "C";
}

use FindBin;
use lib "$FindBin::Bin/../../lib/";
use lib "$FindBin::Bin/../../corelib/"; 

use strict; 
use warnings;
use ParamParser;
use Carp;
use General;
use GeneralBioinfo;


my $VERBOSE;


MAIN:
{
    my $o_param = New ParamParser('GETOPTLONG', "query=s", "db=s", "output=s", 
		"cfg=s", "query_loc=s", "verbose", "noclean", "slice_len=i", "db_length=i", "prokaryote", "protein_similarity_search_method=s");
    $o_param->SetUsage(my $usage=sub { &Usage(); } );
    $o_param->AssertFileExists('cfg');
	$o_param->SetBehaviour('ignore_space');
    $o_param->Update($o_param->Get('cfg'),'A');
	
	$o_param->SetSubstitution('%i', $ENV{'EGNEP'});
	$o_param->SetBehaviour('use_substitution_table');
	

	$VERBOSE = ($o_param->IsDefined('verbose')) ? 1 : 0;
	
    # Check the parameters
    $o_param->AssertFileExists('query');	
    $o_param->AssertDefined('db', 'output');
    $o_param->AssertFullPath('output');
    $o_param->AssertDefined('blast_cmd', 'blast_param');
	$o_param->AssertAllowedValue('protein_similarity_search_method', "blastx", "ublast_blastx", "diamond_blastx");
	
    $o_param->AssertDefined('prg_blastdbcmd', 'prg_makeblastdb');
    
    if ($o_param->Get('protein_similarity_search_method') eq "diamond_blastx")
	{
		$o_param->SetUnlessDefined('slice_len', 1000);
	}
	
    if ($o_param->IsDefined('slice_len'))
    {
		$o_param->AssertInteger('slice_len');
		$o_param->AssertDefined('prg_fasta2overlappingwins');
		my $slice_len = $o_param->Get('slice_len');
		my $step      = sprintf ("%d", $slice_len*90/100);
		$o_param->Set('prg_fasta2overlappingwins_param', "--maxlen $slice_len  --step $step");
	}
	
    my $cmd;
    my $query  = $o_param->Get('query');
    my $db     = $o_param->Get('db');
	my $method = $o_param->Get('protein_similarity_search_method');
	
    $o_param->SetUnlessDefined('db_length', &GetDbSize($db));
    $o_param->AssertInteger('db_length');
    my $db_length = $o_param->Get('db_length');
    
    # Check the database was indexed for blastx+
    if (!-f $db.".pin" && ! -f "$db.00.pin")  
    {
		croak "The database \"$db\" is not indexed. "
			. "(Index with the command :'\$EGNEP/bin/ext/ncbi-blast/bin/makeblastdb -in $db -dbtype prot|nucl  -parse_seqids').\n";
    }
    
    my $output     = $o_param->Get('output');
    my $tmp_prefix = "$output.".time.".$$";
    $o_param->Set('tmp_prefix', $tmp_prefix);
 
    # 0 extractseq [option]
    if ($o_param->IsDefined('query_loc'))
    {
	my $loc = $o_param->Get('query_loc') ;
	my ($start, $end) = ( $loc  =~ /(\d+)-(\d+)/);
	if ($start > $end)
	{
	    croak "query_loc (>$loc<) is not well-formed. Should be start-stop where start < stop.\n";  
	}
	my $newquery = "$tmp_prefix.query.fa";
	&ExtractSubSeq($query, $newquery, $o_param->Get('query_loc'));
	$query = $newquery;		
    }
    
    my $formatted_query;
    if ($o_param->IsDefined('slice_len'))
    {
	# Split input sequence
	&Fasta2Wins($o_param, $query, "$tmp_prefix.query.slices.fa"); 
	$formatted_query = "$tmp_prefix.query.slices.fa";
    }
    else
    {
	$formatted_query = $query;
    }	

	if ($method eq "blastx")
	{
		# Launch blastx+ against the full db
		&ComputeBlast($o_param, $query, $db, $output, $db_length,0);
    }
    else
    {
		##### Database reduction
		
		# Launch diamond (or ublast) first
		my $ublastout = "$tmp_prefix.$method.m8";
		my $ublastres ;
		if ($method eq "ublast_blastx")
		{
			$ublastres = &UBlast($o_param, $formatted_query, $db, $ublastout);
		}
		elsif ($method eq "diamond_blastx")
		{
			$ublastres = &Diamond($o_param, $formatted_query, $db, $ublastout);
		} 
		else
		{
		    # default mode (should not append) 
			$ublastres = &Diamond($o_param, $formatted_query, $db, $ublastout);
		}
		
		if ($ublastres != 0)
		{
	    	print STDERR "$method failed with sequence >$formatted_query< against the database >$db<: launch classical Blastx+\n";
	   		# Launch blastx+ against the full db
	    	&ComputeBlast($o_param, $query, $db, $output, $db_length,0);
		}
		else
		{
	    	# If Diamond (or Ublast) result is not empty
	    	if (-s $ublastout)
	    	{
			my $protein_id_file    = "$tmp_prefix.$method.prot.id";
			my $protein_fasta_file = "$tmp_prefix.$method.prot.fa";
			# Extract proteins and format the database for blastx+
			&GetSubjectIDsFromM8($ublastout, $protein_id_file);
			&BuildBlastDb($o_param, $db, $protein_id_file, $protein_fasta_file);
		
			# Launch blastx+ against this new db
			&ComputeBlast($o_param, $query, $protein_fasta_file, $output, $db_length, 1);
		
			if (!$o_param->IsDefined('noclean'))
			{
		    	unlink($protein_id_file, $protein_fasta_file);
		    	# remove index files for the reduced protein database
		    	my ($protdir, $protname) = ( $protein_fasta_file =~ /(.+)\/(.+)/ );
		    	$cmd = "find $protdir -name \"$protname.p*\" -exec rm \'{}\' \\;";
		    	print STDERR "$cmd\n" if ($VERBOSE);
		    	system($cmd);
			}
	    	}
	    	else # diamond (or ublast) empty
	    	{	
				print STDERR "No $method result with sequence >$formatted_query< against >$db<.\n" if ($VERBOSE);
				# No hit: create an empty output file
				my $f_out = &GetStreamOut($output);
				$f_out->close();
	    	}

		##### Database reduction end
		}
    }

    # Coordinate recompute
    if (-s $output && $o_param->IsDefined('query_loc'))
    {
	&UpdateCoordinate($o_param, $output, $o_param->Get('query_loc'));
    }

    # Clean
    &Clean($o_param);
    
}



sub Clean
{
    my ($o_param) = @_;
    
    my $tmp_prefix = $o_param->Get('tmp_prefix');
    return if ($o_param->IsDefined('noclean'));

    my $method = $o_param->Get('protein_similarity_search_method');
    unlink("$tmp_prefix.query.slices.fa") if (-e "$tmp_prefix.query.slices.fa");
    unlink("$tmp_prefix.$method.m8") if (-e "$tmp_prefix.$method.m8");
    unlink("$tmp_prefix.query.fa") if ($o_param->IsDefined('query_loc') && -e "$tmp_prefix.query.fa");
    
    
    my $protein_id_file    = "$tmp_prefix.$method.prot.id";
    my $protein_fasta_file = "$tmp_prefix.$method.prot.fa";
    unlink($protein_id_file) if (-e $protein_id_file);
    
    if (-e $protein_fasta_file)
    {
	unlink($protein_fasta_file) if (-e $protein_fasta_file);
	# remove index files for the reduced protein database
	my ($protdir, $protname) = ( $protein_fasta_file =~ /(.+)\/(.+)/ );
	my $cmd = "find $protdir -name \"$protname.p*\" -exec rm \'{}\' \\;";
	print STDERR "$cmd\n" if ($VERBOSE);
	system($cmd);
    }
    
    return;
}




sub UpdateCoordinate
{
	my ($o_param, $m8file, $query_loc) = @_;

	my $cmd;
	my ($start, $end) = ($query_loc  =~ /(\d+)-(\d+)/);
	my $oldm8file =  "$m8file.oldcoordinate";
	
	$cmd = "mv $m8file $oldm8file";
	system($cmd);

	my $f_in  = &General::GetStreamIn($oldm8file);
	my $f_out = &General::GetStreamOut($m8file);
	while (my $l = <$f_in>)
	{
		chomp($l);
		my @a_f = split("\t", $l);
		$a_f[6] = $a_f[6] + $start - 1;
		$a_f[7] = $a_f[7] + $start - 1;
		print $f_out join("\t", @a_f)."\n";
	}
	$f_in->close;
	$f_out->close;

	if (!$o_param->IsDefined('noclean'))
    {
		$cmd = "rm $oldm8file";
		system($cmd);
	}

	return;
}


sub ExtractSubSeq
{
	my ($query, $newquery, $query_loc) = @_;
	
	croak "Error: query_loc format should be 'start-end'.\n" if ($query_loc !~ /\d+-\d+/ );
	
	my ($start, $end) = ($query_loc  =~ /(\d+)-(\d+)/);
	croak "Error in query_loc values: start ($start) should be lower than end ($end).\n" if ($start > $end);
	
	my %h_seq;
	&GeneralBioinfo::FastaToHash($query,\%h_seq);
	my @a_keys = keys %h_seq;
	my $seqid  = $a_keys[0];

	croak "Error in query_loc values: end ($end) upper than sequence length (".$h_seq{$seqid}{len}.")\n" if ($end > $h_seq{$seqid}{len});
	my $subseq    = &GeneralBioinfo::ExtractSeq(\$h_seq{$seqid}{sequence}, $start, $end);

	my $f_ = &General::GetStreamOut($newquery);
	print $f_ ">".$seqid."\n$subseq\n";
	$f_->close;

	return;
}


=head2 procedure Fasta2Wins

 Title        : Fasta2Wins
 Usage        : 
 Prerequisite : prg_fasta2overlappingwins prg_fasta2overlappingwins_param are keys of o_param object
 Procedure    : convert $fasta file in multifasta file of sequences of 15000nt 
 Returns      : None
 Args         : $o_param
                $fasta      Fasta file
                $win_fasta  Output mutlifasta file
 Globals      : none

=cut
sub Fasta2Wins
{
	my ($o_param, $fasta, $win_fasta) = @_;
	
	my $cmd = $o_param->Get('prg_fasta2overlappingwins');
	$cmd .= " --in $fasta --out $win_fasta ". $o_param->Get('prg_fasta2overlappingwins_param');
	$cmd .= " > /dev/null 2>&1" if (!$VERBOSE);

	print STDERR "$cmd\n" if ($VERBOSE);
	system($cmd);

	return;
}


=head2 procedure UBlast

 Title        : UBlast
 Usage        : 
 Prerequisite : 
 Procedure    : Index database if not done, then launch ublast
 Returns      : None
 Args         : $o_param
                $query
                $db
                $ublastout
 Globals      : none

=cut
sub UBlast
{
	my ($o_param, $query, $db, $ublastout) = @_;
	
	my $cmd;
	
	my $usearch_cmd          = $o_param->Get('prg_usearch');
    my $usearch_ublast_param = $o_param->Get('prg_usearch_ublast_param');
    
    # format for ublast
    if (!-e $db.".udb")
    {
		$cmd = $usearch_cmd. " -makeudb_ublast $db -output $db.udb";
		$cmd .= " > /dev/null 2>&1" if (!$VERBOSE);
		print STDERR "$cmd\n" if ($VERBOSE);
		system($cmd);
	}
	# launch ublast
	$cmd = $usearch_cmd. " -ublast $query -db $db.udb -blast6out $ublastout $usearch_ublast_param ";
	$cmd .= " > /dev/null 2>&1" if (!$VERBOSE);
	print STDERR "$cmd\n" if ($VERBOSE);
    my $res = system($cmd);
    
    return $res;
}


sub DiamondBlastP
{
	my ($o_param, $query, $db, $diamondout) = @_;
	
	my $cmd;
	
	my $diamond_cmd   = $o_param->Get('prg_diamond');
    my $diamond_param = $o_param->Get('prg_diamond_param');
    
    # format for diamond
    if (!-e $db.".dmnd")
    {
		$cmd = $diamond_cmd. " makedb --in $db -d $db";
		$cmd .= " > /dev/null 2>&1" if (!$VERBOSE);
		print STDERR "$cmd\n" if ($VERBOSE);
		system($cmd);
	}
	
	# 0 build a gff3 of one region
	my $querygff3   = $query.".gff3";
	my $queryorf    = $query.".orf.gff3";
	my $queryorfpep = $query.".orf.pep.fa";
	&SeqToGff3($query, $querygff3);
	
	# 1 Launch orfinder
	my $minorf = 120;
	$minorf    = 40 if ($o_param->IsDefined('prokaryote'));
	
	$cmd = $o_param->Get('prg_gt') . " orffinder -allorfs -types region -min $minorf -matchdescstart -seqfile $query $querygff3 > $queryorf";
	print STDERR "$cmd\n" if ($VERBOSE);
	system($cmd);
	
	# 2 convert in proteic sequence
	$cmd = $o_param->Get('prg_gt') . " extractfeat -type reading_frame -translate -seqfile $query --matchdescstart ";
	#$cmd .= "-gcode 11 " if ($o_param->IsDefined('prokaryote'));
	$cmd .= "$queryorf > $queryorfpep";
	print STDERR "$cmd\n" if ($VERBOSE);
	system($cmd);
	
	# launch diamond blastP
	$cmd = $diamond_cmd. " blastp -d $db -q $queryorfpep -o $diamondout --outfmt 6 $diamond_param ";
	$cmd .= " > /dev/null 2>&1" if (!$VERBOSE);
	print STDERR "$cmd\n" if ($VERBOSE);
    my $res = system($cmd);

	unlink($querygff3, $queryorf, $queryorfpep);
	unlink($query.".des", $query.".esq",$query.".md5",$query.".ois", $query.".sds");
	
    return $res;
}



sub SeqToGff3
{
	my ($seq, $gff3) = @_;
	
	my %h_seq;
	my $cpt = 1;
	my $f_out = &General::GetStreamOut($gff3);
	print $f_out "##gff-version 3\n";
	&GeneralBioinfo::FastaToHash($seq, \%h_seq, 0);
    foreach my $name (keys(%h_seq))
    {
		print $f_out "##sequence-region $name 1 $h_seq{$name}{len}\n";
		print $f_out "$name\t.\tregion\t1\t$h_seq{$name}{len}\t.\t.\t.\tID=$name.$cpt\n";
		$cpt++;
    }
	$f_out->close;
	
	return;		
}

sub Diamond
{
   my ($o_param, $query, $db, $diamondout) = @_;
	
	my $cmd;
	
	my $diamond_cmd   = $o_param->Get('prg_diamond');
    my $diamond_param = $o_param->Get('prg_diamond_param');
    
    # format for diamond
    if (!-e $db.".dmnd")
    {
		$cmd = $diamond_cmd. " makedb --in $db -d $db";
		$cmd .= " > /dev/null 2>&1" if (!$VERBOSE);
		print STDERR "$cmd\n" if ($VERBOSE);
		system($cmd);
	}
	# launch diamond
    $cmd = $diamond_cmd. " blastx -d $db -q $query -o $diamondout --outfmt 6 $diamond_param ";
    $cmd .= " --query-gencode 11 " if ($o_param->IsDefined('prokaryote'));
	$cmd .= " > /dev/null 2>&1" if (!$VERBOSE);
	print STDERR "$cmd\n" if ($VERBOSE);
    my $res = system($cmd);
    
    return $res;
}

=head2 procedure GetSubjectIDsFromM8

 Title        : GetSubjectIDsFromM8
 Usage        : 
 Prerequisite : 
 Procedure    : Extract from a m8 file the subject id in column2
 Returns      : None
 Args         : $m8 file
                $subject_id_file Output file, containibng disctinct subject ID 
                (column2 of m8 file)
 Globals      : none

=cut
sub GetSubjectIDsFromM8
{
	my ($m8, $subject_id_file) = @_;
    
    my %h_id;
    my $f_       = &General::GetStreamIn($m8);
	my $f_protid = &General::GetStreamOut($subject_id_file);
    foreach my $l (<$f_>)
    {
		chomp($l);
		my @a_lines = split(/\t/, $l);
		my $protid  = $a_lines[1];
		if (!exists $h_id{$protid})
		{
			print $f_protid "$protid\n";
			$h_id{$protid} = 1;
		}
	}
    $f_->close;
    $f_protid->close;
    
    return;
}
    
    
=head2 procedure BuildBlastDb

 Title        : BuildBlastDb
 Usage        : 
 Prerequisite : prg_blastdbcmd and prg_makeblastdb are keys of o_param and existing programs
 Procedure    : Extract proteins from id list and build a blast+ database
 Returns      : None
 Args         : $o_param
                $db                  multifasta proteic
                $protein_id_file     file of protein IDs
                $protein_fasta_file  output proteic multifasta 
 Globals      : none

=cut
sub BuildBlastDb
{
	my ($o_param, $db, $protein_id_file, $protein_fasta_file) = @_;
    
    my $cmd;
    

    $cmd = $o_param->Get('prg_blastdbcmd') . " -db $db -dbtype prot -entry_batch $protein_id_file -out $protein_fasta_file";
	$cmd .= " > /dev/null 2>&1" if (!$VERBOSE);
    print STDERR "$cmd\n" if ($VERBOSE);
    system($cmd);
    
    # 3 formatdb for blast+
    $cmd = $o_param->Get('prg_makeblastdb') . " -in $protein_fasta_file -dbtype prot  -parse_seqids";
	$cmd .= " > /dev/null 2>&1" if (!$VERBOSE);
    print STDERR "$cmd\n" if ($VERBOSE);
    system($cmd);
    
	return;
}


    
=head2 procedure ComputeBlast

 Title        : ComputeBlast
 Usage        : &ComputeBlast($o_param, $seq, $db, $out_file);
 Prerequisite : $db is a file which was indexed by formatdb, 
                $blast_cmd program exists.
 Procedure    : Blast $seq against the database $db. The result file is $out_file 
 Returns      : None
 Args         : $o_param
                $seq            Query sequence
                $db             Database
                $out_file       Blast output file
                $db_length      Full database length
 Globals      : none

=cut

sub ComputeBlast
{
    my ($o_param, $seq, $db, $out_file, $db_length, $noseg) = @_;
	
    my $blast_cmd      = $o_param->Get('blast_cmd');
    my $blast_param    = $o_param->Get('blast_param');    
    if ($noseg == 1)
    {
	# No Filter query sequence with SEG 
	$blast_param =~ s/-seg +(yes|no)/ /;
	$blast_param .= " -seg no ";
    }
    # Execute Blast
    my $cmd          = "$blast_cmd -db $db -query $seq -out $out_file $blast_param";

    $cmd .= " -dbsize $db_length " if (defined($db_length) && $db_length > 0);

    #$cmd .= " > /dev/null 2>&1" if (!$VERBOSE);
    print $cmd if ($VERBOSE);
    my $system_out   = system($cmd);
    my $blast_output = $system_out >> 8;
    if ($blast_output != 0)
    {
	my $signal_number = $system_out & 127;
	unlink($out_file) if (-e $out_file); # remove output file if created
	&Clean($o_param);
    	croak "Blast execution failed. Exit value: $blast_output, signal number: $signal_number.\n";
    }
	
    return;
}



sub GetDbSize
{
    my ($db) = @_;

    my 	%h_seq;    
    &GeneralBioinfo::FastaToHash($db, \%h_seq, 1);
    my $len = 0;
    foreach my $s (keys(%h_seq))
    {
	$len += $h_seq{$s}{len};
    }
    return $len;
}

=item Usage

	$o_param->Usage();
	Print the usage of the program

=cut

sub Usage
{
print STDERR<<END
Usage:$0	

[Mandatory]
    --query 	filename        Sequence file path
    --db        filename        Database path
    --output	filename        Blast output file (tabular format)
    --cfg       filename        Full path of the configuration file
	--protein_similarity_search_method 	[blastx,ublast_blastx,diamond_blastx]
[Optional]	
    --prokaryote                Adapt parameters for prokaryotic genomes
    --query_loc integer-integer Location on the query sequence in 1-based offsets 
                                (Format: start-stop)
    --slice_len  integer        If defined, split the query sequence in slices of length slice_len. 
                                NOTE: in 'diamond_blastx' mode, if this parameter is not 
                                defined, the value is automatically set to 1000.
                                CAREFUL: usearch64 sometimes raises memory errors
                                with that kind of multifasta as query input.
    --db_length integer         Database length (Value computed if not defined). 
                                BlastX option '-dbsize'
    --verbose   
    --noclean   
	--help
END
}

